
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#include "boid.h"


/* *************************************************************
************************************************************* */
int boid::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'b':
         if (!strcmp(token, "boid_range")) {
            get_token(infile, token);
            nb_min_radius = atof(token);
            get_token(infile, token);
            nb_max_radius = atof(token);
            return 1;
         }

         break;

      case 'g':
         if (!strcmp(token, "gravity")) {
            gflag = 1;
            get_token(infile, token);
            gravity_vel = atof(token);
            get_token(infile, token);
            gravity[0] = atof(token);
            get_token(infile, token);
            gravity[1] = atof(token);
            get_token(infile, token);
            gravity[2] = atof(token);

            normalize3(gravity);
            return 1;
         }

         break;

      case 'l':
         if (!strcmp(token, "leader_range")) {
            get_token(infile, token);
            nl_min_radius = atof(token);
            get_token(infile, token);
            nl_max_radius = atof(token);
            return 1;
         }

         break;

      case 'm':
         if (!strcmp(token, "max_turn_radius")) {
            get_token(infile, token);
            max_turn_rad = deg2rad(atof(token));
            return 1;
         }

         break;

      case 'r':
         if (!strcmp(token, "red_alert")) {
            get_token(infile, token);
            ra_radius = atof(token);
            return 1;
         }

         break;

      case 'v':
         if (!strcmp(token, "velocity")) {
            get_token(infile, token);
            velocity = atof(token);
            return 1;
         }

         if (!strcmp(token, "velocity_range")) {
            get_token(infile, token);
            minspeed = atof(token);
            get_token(infile, token);
            maxspeed = atof(token);
            return 1;
         }

         break;

      case 'y':
         if (!strcmp(token, "yellow_alert")) {
            get_token(infile, token);
            ya_radius = atof(token);
            return 1;
         }

         break;

      default:
         break;
   }

   return quark::parse(infile, token);
}


/* *************************************************************
************************************************************* */
int boid::query_collision(quark *drunk) {

   switch (drunk->bflag) {      // check for yellow alert

      case BNONE:
           return NO_COLLISION;

      case BPOINT:
         if (!query_point_sphere_intersect(drunk->oldcenter, oldcenter, ya_radius))
            return NO_COLLISION;
         if (query_point_sphere_intersect(drunk->oldcenter, oldcenter, ra_radius))
            return RED_ALERT;

         return YELLOW_ALERT;

      case BSPHERE:
         if (!query_sphere_sphere_intersect(drunk->oldcenter, drunk->bsphere, oldcenter, ya_radius))
            return NO_COLLISION;
         if (query_sphere_sphere_intersect(drunk->oldcenter, drunk->bsphere, oldcenter, ra_radius))
            return RED_ALERT;

         return YELLOW_ALERT;

      case BRECT:
         if (!query_sphere_rect_intersect(oldcenter, ya_radius, drunk->bnormal))
            return NO_COLLISION;
         if (query_sphere_rect_intersect(oldcenter, ra_radius, drunk->bnormal))
            return RED_ALERT;

         return YELLOW_ALERT;

      case BPLANE:
         if (!query_sphere_plane_intersect(oldcenter, ya_radius, drunk->bnormal[0]))
            return NO_COLLISION;
         if (query_sphere_plane_intersect(oldcenter, ra_radius, drunk->bnormal[0]))
            return RED_ALERT;

         return YELLOW_ALERT;
   }

   return NO_COLLISION;
}


/* *************************************************************
************************************************************* */
void boid::calc_reflect(quark *drunk, float *dir, float *ref) {

   float n[3];

   switch (drunk->bflag) {              // calc reflect normal

      case BPOINT:
         if (!point_sphere_reflectn3vf(drunk->oldcenter, oldcenter, n)) {
            n[0] = -dir[0];
            n[1] = -dir[1];
            n[2] = -dir[2];
            normalize3(n);
         }

         break;

      case BSPHERE:
         if (!sphere_sphere_reflectn3vf(oldcenter, drunk->oldcenter, n)) {
            n[0] = -dir[0];
            n[1] = -dir[1];
            n[2] = -dir[2];
            normalize3(n);
         }

         break;

      case BPLANE:
         sphere_plane_reflectn3vf(drunk->bnormal[0], n);
         break;

      case BRECT:
         sphere_rect_reflectn3vf(oldcenter, drunk->bvertex, drunk->bnormal, n);
         break;

      default:
         break;
   }

   if (dotproduct3(n, dir) < -CORRECT) {
      ref[0] = -dir[0];
      ref[1] = -dir[1];
      ref[2] = -dir[2];
      reflect(n,ref);
   }

   else
      copyarray3(ref, dir);
}


/* *************************************************************
************************************************************* */
void boid::calc_avoid_leader_boid(float *avoid, quark *qtr, int *aflag, quark **nl,
                      float *nld, quark **nb, float *nbd, quark *parent) {

   linktype *ptr;
   float    reflect[3];
   float    temp;

   if (qtr != parent && qtr != this) {
      switch (query_collision(qtr)) {
         case NO_COLLISION:
            break;

         case YELLOW_ALERT:
            if (*aflag == RED_ALERT)
               break;

            if (*aflag == NO_COLLISION)
               *aflag = YELLOW_ALERT;

            calc_reflect(qtr, direction, reflect);
            addarray3(avoid, reflect);

            break;

         case RED_ALERT:
            calc_reflect(qtr, direction, reflect);
            *aflag = RED_ALERT;

            if (*aflag == YELLOW_ALERT) {
               copyarray3(avoid, reflect);
            }

            else
               addarray3(avoid, reflect);

            break;
      }

      if (*aflag != RED_ALERT)
         if (qtr->whatami() == BOIDBAIT) {
            if ((temp = distance3(qtr->oldcenter, oldcenter)) < *nbd || *nbd < 0) {
               *nb = qtr;
               *nbd = temp;
            }

            if (temp < *nld || *nld < 0) {
               *nl = qtr;
               *nld = temp;
            }

         }

         else
            if (qtr->whatami() == BOID)
               if ((temp = distance3(qtr->oldcenter, oldcenter)) < *nbd || *nbd < 0) {
                  *nb = qtr;
                  *nbd = temp;
               }

   }

   for (ptr = qtr->edge; ptr != (linktype *)NULL && ptr->link != parent; ptr=ptr->next)
      calc_avoid_leader_boid(avoid, ptr->link, aflag, nl, nld, nb, nbd, qtr);
}


/* *************************************************************
    NOTE: boid class is independent on external xforms....
          and this object is not actually drawn - its an
          animation device, not a 3d object (its child is though :) )

// FUTURE:
//6) note: for animals - if going down, glide
//                        if going up, flap

************************************************************* */
void boid::whereami(int frameno, molecule **mptr, int *obcount, int wflag,
			quark *actors, quark *parent, vector4f *mx) {

   linktype *ptr;
   atom     *atr;
   int      aflag = NO_COLLISION;                       // collision type;
   float    avoid[3] = {0,0,0};

   quark    *nearest_leader, *nearest_boid;
   float    nl_dist, nb_dist;

   float    weight;
   float    dir[3];
   float    temp;

   int      i, j;
   float    xmx[4][4], r[4][4];      // local xform mx

   if (!firstflag) {
//printf("oc %f %f %f\n", oldcenter[0], oldcenter[1], oldcenter[2]);

      copyarray3(dir, direction);          // backup old dir

      nl_dist = nb_dist = -1.0;
      nearest_leader = nearest_boid = (quark *)NULL;

      for (atr = (atom *)actors; atr != (atom *)NULL; atr = atr->next)
         if (atr->ob != (quark *)NULL)
            calc_avoid_leader_boid(avoid, atr->ob, &aflag, &nearest_leader,
                                   &nl_dist, &nearest_boid, &nb_dist, (quark *)NULL);

      if (aflag != NO_COLLISION)
         normalize3(avoid);

      if (aflag == RED_ALERT) {                            // collision
         copyarray3(direction, avoid);
         velocity = maxspeed;
      }

      else {
         weight = 1.0;

         if (aflag == YELLOW_ALERT) {                      // near collision
            copyarray3(direction, avoid);
            temp = minspeed + (maxspeed-minspeed) * ya_weight;
            smultarray3(direction, temp);
            weight -= ya_weight;
         }

         else
            direction[0] = direction[1] = direction[2] = 0;

         if (weight > 0) {
            if (nearest_leader != (quark *)NULL)           // nearest flock leader
               if (nl_dist > nl_max_radius) {
                  copyarray3(avoid, nearest_leader->oldcenter);

                  subarray3(avoid, oldcenter);
                  temp = normalize3(avoid);
                  temp = (temp - nl_max_radius)/nl_max_radius;

                  if (temp > weight)
                     temp = weight;

                  weight -= temp;
                  temp = minspeed + (maxspeed-minspeed)*temp;
                  smultarray3(avoid, temp);
                  addarray3(direction, avoid);
               }

               else
                  if (nl_dist < nl_min_radius) {
                     copyarray3(avoid, oldcenter);
                     subarray3(avoid, nearest_leader->oldcenter);
                     temp = normalize3(avoid);
                     temp = (nl_min_radius - temp)/nl_min_radius;

                     if (temp > weight)
                        temp = weight;

                     weight -= temp;
                     temp = minspeed + (maxspeed-minspeed)*temp;
                     smultarray3(avoid, temp);

                     addarray3(direction, avoid);

                  }

            if (weight > 0) {
               if (nearest_boid != (quark *)NULL)          // nearest boid
                  if (nb_dist > nb_max_radius) {
                     copyarray3(avoid, nearest_boid->oldcenter);
                     subarray3(avoid, oldcenter);
                     temp = normalize3(avoid);
                     temp = (temp - nb_max_radius)/nb_max_radius;

                     if (temp > weight)
                        temp = weight;

                     weight -= temp;
                     temp = minspeed + (maxspeed-minspeed)*temp;
                     smultarray3(avoid, temp);
                     addarray3(direction, avoid);
                  }

                  else
                     if (nb_dist < nb_min_radius) {
                        copyarray3(avoid, oldcenter);
                        subarray3(avoid, nearest_boid->oldcenter);
                        temp = normalize3(avoid);
                        temp = (nb_min_radius - temp)/nb_min_radius;

                        if (temp > weight)
                           temp = weight;

                        weight -= temp;
                        temp = minspeed + (maxspeed-minspeed)*temp;
                        smultarray3(avoid, temp);
                        addarray3(direction, avoid);
                     }

               if (weight > 0) {
                  if (gflag) {                     // gravity
                     temp = maxspeed * weight;

                     if (temp >= gravity_vel)
                        weight = (temp - gravity_vel)/maxspeed;
                     else {
                        copyarray3(avoid, gravity);
                        temp = gravity_vel - temp;
                        smultarray3(avoid, temp);
                        addarray3(direction, avoid);
                        weight = 0;
                     }

                  }

                  if (weight > 0) {                // part of previous path
                     temp = velocity * weight;
                     copyarray3(avoid, dir);
                     smultarray3(avoid, temp);
                     addarray3(direction, avoid);
                     weight = 0;
                  }

               }

            }

         }

      }

      velocity = normalize3(direction);

      if (velocity < minspeed)
         velocity = minspeed;
      else
         if (velocity > maxspeed)
            velocity = maxspeed;

      temp = ACOS(dotproduct3(direction, dir));

      if (temp > max_turn_rad && aflag != RED_ALERT)
         temp = max_turn_rad;

      if (temp > CORRECT) {
         init_mx(i,j,r);
         copyarray3(r[0], dir);
         if (abs(dotproduct3(dir, direction)) > 1.0-CORRECT) {   // 0 or 2PI turn
            copyarray3(r[1], vup);
            xproduct(r[2], r[0], r[1]);
            normalize3(r[2]);
         }

         else {
            xproduct(r[2], dir, direction);
            normalize3(r[2]);
            xproduct(r[1], r[2], r[0]);
            normalize3(r[1]);
         }

         transpose(r);
         init_mx(i,j,xmx);
         xmx[0][0] = xmx[1][1] = COS(temp);
         xmx[0][1] = -(xmx[1][0] = SIN(temp));
         matmatmulto(r, xmx);
         direction[0] = 1;
         direction[1] = direction[2] = 0;
         matvecmulto(xmx, direction); 

//printf("%f %f %f\n", dir[0], dir[1], dir[2]);
//printf("%f %f %f\n", direction[0], direction[1], direction[2]);
      }

      copyarray3(avoid, direction);
      smultarray3(avoid, velocity);
      addarray3(center, avoid);            // new location!!!!

      copyarray3(r[2], direction);

      if (abs(dotproduct3(dir, direction)) > 1.0-CORRECT) {   // 0 or 2PI turn
         copyarray3(r[1], vup);
         xproduct(r[0], r[1], r[2]);
         normalize3(r[0]);
      }

      else {
         xproduct(r[0], r[2], dir);                // nZ x oZ cause oZ in wrong dir
         normalize3(r[0]);
         xproduct(r[1], r[2], r[0]);
         normalize3(r[1]);
//         copyarray3(vup, r[1]);
      }

   }

   else {                       // first frame
      copyarray3(direction, initxform[2]);
      normalize3(direction);
      copyarray3(vup, initxform[1]);
      normalize3(vup);

// where am i ???? not initxform...

      center[0] = initxform[0][3];
      center[1] = initxform[1][3];
      center[2] = initxform[2][3];

      copyarray3(r[2], direction);
      copyarray3(r[1], vup);

      copyarray3(r[0], initxform[0]);
      normalize3(r[0]);

      inversemx(mx, old_base);
   }

//printf("*** %f %f %f | %f\n", direction[0], direction[1], direction[2], velocity);

   copymx3x3(xmx, r);
   xmx[0][3] = center[0];
   xmx[1][3] = center[1];
   xmx[2][3] = center[2];
   xmx[3][0] = xmx[3][1] = xmx[3][2] = 0;
   xmx[3][3] = 1;

   matmatmulto(mx, old_base);

   matmatmulto(old_base, xmx);
   center[0] = xmx[0][3];
   center[1] = xmx[1][3];
   center[2] = xmx[2][3];
   center[3] = 1.0;
   copyarray3(direction, xmx[2]);
   normalize3(direction);
   copyarray3(vup, xmx[1]);
   normalize3(vup);

   if (gflag)
      matvecmultv(xmx, gravity);

   for (ptr=edge; ptr!=(linktype *)NULL; ptr=ptr->next)  // apply xform to
      if (ptr->link != parent)                           //    children
         ptr->link->whereami(frameno, mptr, obcount, wflag, actors, this, xmx);

   inversemx(mx, old_base);
}

