
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#include "flight.h"
#define CATCHUP 30
#define ICATCHUP 0.03333333333

/* ********************************************************
   This function reads in the data for this child
******************************************************** */
int flight::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'a':

         if (strlen(token) < 6)
            break;

         switch (token[5]) {
            case 'x':
               if (!strcmp(token, "axis_x")) {
                  get_token(infile, token);
                  old_vx[0] = atof(token);
                  get_token(infile, token);
                  old_vx[1] = atof(token);
                  get_token(infile, token);
                  old_vx[2] = atof(token);
                  normalize3(old_vx);
                  return 1;
               }

               break;

            case 'y':
               if (!strcmp(token, "axis_y")) {
                  get_token(infile, token);
                  old_vy[0] = atof(token);
                  get_token(infile, token);
                  old_vy[1] = atof(token);
                  get_token(infile, token);
                  old_vy[2] = atof(token);
                  normalize3(old_vy);
                  return 1;
               }

               break;

            case 'z':
               if (!strcmp(token, "axis_z")) {
                  get_token(infile, token);
                  old_vz[0] = atof(token);
                  get_token(infile, token);
                  old_vz[1] = atof(token);
                  get_token(infile, token);
                  old_vz[2] = atof(token);
                  normalize3(old_vz);
                  return 1;
               }

               break;

            default:
               break;
         }

      default:
         break;
   }

   return quark::parse(infile, token);
}


/* *************************************************************
   place object based on local and parent xforms

1) calc new location
2) vpn = new direction = new loc - old loc 
3) if new direction about same as old direction, use old vup, vright
4) else calc new vup/vright based upon plane created by new/old direction vectors
5) 2 possible vups/right -> vup, smallest rotation...


want to rotate around "z" until y[1] is positive, and x[1] = 0, as long as
z != { 0, +-1, 0 }
1) figure out positive or neg rotation around z, and the rotation
2) 1) A = old matrix, B = rotate mx
3) new matrix C = A*B


************************************************************* */
void flight::whereami(int frameno, molecule **mptr, int *obcount, int wflag,
			quark *actors, quark *parent, vector4f *mx) {

   float    xmx[4][4];    		  // local xform mx
   linktype *ptr;                         // pointer
   vector3f new_vz, new_vy, new_vx, t_u, t_v;
   float temp, temp2;
   int defflag = 1;		// flag for no sig. change in direction
   int i, j;
   float one = 1 - CORRECT4;

   calc_initmx(mx, xmx);

   center[0] = xmx[0][3];
   center[1] = xmx[1][3];
   center[2] = xmx[2][3];
   center[3] = 1;

   copyarray3(new_vz, center);
   subarray3(new_vz, oldcenter);

   temp = normalize3(new_vz);
   temp2 = abs(dotproduct3(new_vz, old_vz));

   if (!firstflag) {
      if (temp > CORRECT4)
//         if (temp2 < 0.9961) {		// over 5 deg change in direction
         if (temp2 < one) {
            defflag = 0;

            xproduct(new_vx, new_vz, old_vz);
            normalize3(new_vx);
            xproduct(new_vy, new_vz, new_vx);
//printf("pitch %d\n", frameno);

            if (dotproduct3(new_vx, old_vx) < 0) {
               new_vx[0] = -new_vx[0];
               new_vx[1] = -new_vx[1];
               new_vx[2] = -new_vx[2];

               new_vy[0] = -new_vy[0];
               new_vy[1] = -new_vy[1];
               new_vy[2] = -new_vy[2];
            }

         }

/*
         else
            if (temp2 < one) {		// 5 deg or less change in direction
printf("yaw %d\n", frameno);
               defflag = 0;
               xproduct(new_vx, old_vy, new_vz);
               normalize3(new_vx);
               xproduct(new_vy, new_vz, new_vx);
            }
*/

      if (defflag) {
         if (fraction) {		// rotate toward "normal" orientation
//printf("roll %d\n", frameno);
            fraction--;

            vecmatmultv(angle, old_vx);
            vecmatmultv(angle, old_vy);
            vecmatmultv(angle, old_vz);
         }

      }

      else {
         if (abs(new_vz[1]) < 1-CORRECT && (new_vy[1] < 0 || abs(new_vx[1]) > CORRECT)) {	// calc rotation back to "normal"

            if (new_vy[1] >= one)
               fraction = 0;		// no rotate - should not get here
            else {
               if (new_vy[1] <= -one) {
                  fraction = CATCHUP;
                  temp = PI;
               }

               else {
                  copyarray3(t_v, new_vx);
                  t_v[1] = 0;
                  normalize3(t_v);

                  xproduct(t_u, new_vz, t_v);	// "y" axis that we want to rotate to...
                  normalize3(t_u);

                  temp = dotproduct3(t_u, new_vy);	// cos between y axis we have and the one we want

                  if (temp < one) {		// at the orientation we want
                     fraction = CATCHUP;
                     temp = ACOS(temp);

                     xproduct(t_v, t_u, new_vy);
                     if (dotproduct3(new_vx, t_v) < 0)
                        temp = -temp;
                  }

                  else
                     fraction = 0;
               }

               if (fraction) {
                  temp *= ICATCHUP;
                  init_mx(i, j, angle);
                  angle[0][0] = angle[1][1] = COS(temp);
                  angle[0][1] = -(angle[1][0] = SIN(temp));
               }

            }

         }

         else
            fraction = 0;

         copyarray3(old_vx, new_vx);
         copyarray3(old_vy, new_vy);
         copyarray3(old_vz, new_vz);      
      }

   }

   temp = magnitude3(xmx[0]);

   temp2 = magnitude3(xmx[1]);
   if (temp2 < temp)
      temp = temp2;

   temp2 = magnitude3(xmx[2]);
   if (temp2 < temp)
      temp = temp2;

   copyarray3(xmx[0], old_vx);
   smultarray3(xmx[0], temp);
   copyarray3(xmx[1], old_vy);
   smultarray3(xmx[1], temp);
   copyarray3(xmx[2], old_vz);
   smultarray3(xmx[2], temp);

   for (ptr=edge; ptr!=(linktype *)NULL; ptr=ptr->next)  // apply xform to
      if (ptr->link != parent)                           //    children
         ptr->link->whereami(frameno, mptr, obcount, wflag, actors, this, xmx);
}

