
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */



#include "convert.h"

int reverse_normal = 0;

void lw_shade::write_data(FILE *outfile) {

   fprintf(outfile, "KA %f %f %f\n", shade[0]/255.0, shade[1]/255.0, shade[2]/255.0);
   fprintf(outfile, "KP %f %f %f\n", (shade[0]*diff)/(255.0*256.0), (shade[1]*diff)/(255.0*256.0), (shade[2]*diff)/(255.0*256.0));
   fprintf(outfile, "KS %f %f %f\n", (shade[0]*spec)/(255.0*256.0), (shade[1]*spec)/(255.0*256.0), (shade[2]*spec)/(255.0*256.0));
   fprintf(outfile, "SPECN %f\n", specn);

   if (flag_lum)
      fprintf(outfile, "LUM %f %f %f\n\n", (shade[0]*lum)/(255.0*256.0), (shade[1]*lum)/(255.0*256.0), (shade[2]*lum)/(255.0*256.0));
   else
      fprintf(outfile, "LUM 0 0 0\n\n");
}


void lw_shade::write_notes(FILE *outfile) {

   if (trans > 0.0)
      fprintf(outfile, "   TRANSPARENCY - %f\n", trans);
}


list *list::get_node() {

   list *x;

   if (back != (list *)NULL) {
      x = back->get_node();
      if (x==back)
         back = (list *)NULL;
      return x;
   }

   if (next != (list *)NULL) {
      x = next->get_node();
      if (x==next)
         next = (list *)NULL;
      return x;
   }

   return this;
}


list *list::find_id(int x) {

   list *y;

   if (id == x)
      return this;

   if (back != (list *)NULL) {
      y = back->find_id(x);
      if (y != (list *)NULL)
         return y;
   }

   if (next != (list *)NULL) {
      y = next->find_id(x);
      if (y != (list *)NULL)
         return y;
   }

   return (list *)NULL;
}


void list::insert_id(list *x) {

   if (x->id == id)
      delete x;
   else if (x->id < id)
      if (back == (list *)NULL)
         back = x;
      else
         back->insert_id(x);
   else
      if (next == (list *)NULL)
         next = x;
      else
         next->insert_id(x);
}


int vertexlist::insert(float *point, int *vertexcount) {

   int flag;

   if (abs(point[0] - pt[0]) < CORRECT)
      if (abs(point[1] - pt[1]) < CORRECT)
         if (abs(point[2] - pt[2]) < CORRECT)
            return id;
         else
            flag = (point[2] < pt[2]);
      else
         flag = (point[1] < pt[1]);
   else
      flag = (point[0] < pt[0]);

   if (flag) {
      if (back == (list *)NULL) {
         back = new vertexlist;
         (*vertexcount)++;
         back->id = *vertexcount;
         copyarray3(((vertexlist *)back)->pt, point);
         return back->id;
      }

      return ((vertexlist *)back)->insert(point, vertexcount);
   }

   if (next == (list *)NULL) {
      next = new vertexlist;
      (*vertexcount)++;
      next->id = *vertexcount;
      copyarray3(((vertexlist *)next)->pt, point);
      return next->id;
   }

   return ((vertexlist *)next)->insert(point, vertexcount);
}


void vertexlist::write_data(FILE *outfile) {

   if (back != (list *)NULL)
      back->write_data(outfile);

   fprintf(outfile, "%f %f %f\n", pt[0], pt[1], pt[2]);

   if (next != (list *)NULL)
      next->write_data(outfile);
}


void objectlist::adjust_totals(int *polycount, int *edgecount) {

   if (back != (list *)NULL)
      ((objectlist *)back)->adjust_totals(polycount, edgecount);

   if (lw != (lw_shade *)NULL)
      if (lw->flag_dbl_side) {
         (*polycount)++;
         (*edgecount) += polynum;
         other_id = *polycount;
         lw->count++;
      }

   if (next != (list *)NULL)
      ((objectlist *)next)->adjust_totals(polycount, edgecount);
}


void objectlist::write_dbl_side(FILE *outfile) {

   int i;

   if (back != (list *)NULL)
      ((objectlist *)back)->write_dbl_side(outfile);

   if (lw != (lw_shade *)NULL)
      if (lw->flag_dbl_side) {
         fprintf(outfile, "%d ", polynum);

         if (!rnflag)
            for (i=polynum-1; i>-1; i--)
               fprintf(outfile, "%d ", vptr[i]);
         else
            for (i=0; i<polynum; i++)
               fprintf(outfile, "%d ", vptr[i]);

         fprintf(outfile, "\n");
      }

   if (next != (list *)NULL)
      ((objectlist *)next)->write_dbl_side(outfile);
}


void objectlist::write_ilm(FILE *outfile, int base) {

   if (back != (list *)NULL)
      ((objectlist *)back)->write_ilm(outfile, base);

   if (lw != (lw_shade *)NULL && base != lw->id) {
      fprintf(outfile, "# %d\n", id);

      lw->write_data(outfile);

      if (lw->flag_dbl_side) {
         fprintf(outfile, "# %d\n", other_id);
         lw->write_data(outfile);
      }

   }

   if (next != (list *)NULL)
      ((objectlist *)next)->write_ilm(outfile, base);
}


void objectlist::write_notes(FILE *outfile) {

   if (back != (list *)NULL)
      ((objectlist *)back)->write_notes(outfile);

   if (lw != (lw_shade *)NULL)
      if (lw->trans > 0.0) {
         fprintf(outfile, "# %d\n", id);
         lw->write_notes(outfile);

         if (lw->flag_dbl_side) {
            fprintf(outfile, "# %d\n", other_id);
            lw->write_notes(outfile);
         }

      }

   if (next != (list *)NULL)
      ((objectlist *)next)->write_notes(outfile);
}


void objectlist::write_data(FILE *outfile) {

   int i;

   if (back != (list *)NULL)
      back->write_data(outfile);

   fprintf(outfile, "%d ", polynum);

   if (!rnflag)
      for (i=0; i<polynum; i++)
         fprintf(outfile, "%d ", vptr[i]);
   else
      for (i=polynum-1; i>-1; i--)
         fprintf(outfile, "%d ", vptr[i]);

   fprintf(outfile, "\n");

   if (next != (list *)NULL)
      next->write_data(outfile);
}


int converter::read_ply(char *filename) {

   int i, j;

   objectlist *otr;

   vector3f point;

   FILE *infile;

   if (!(infile=fopen(filename, "r"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   polycount = 0;
   vertexcount = 0;
   edgecount = 0;

   printf("Reading...\n");

   while(1) {
      if (fscanf(infile, "%ld", &i) == EOF)
         break;

      polycount++;
      edgecount += i;

      otr = new objectlist;
      otr->id = polycount;
      otr->rnflag = reverse_normal;

      if (ob == (objectlist *)NULL)
         ob = otr;
      else
         ob->insert_id(otr);

      otr->vptr = new int[i];
      otr->polynum = i;

      for (j=0; j<i; j++) {

         fscanf(infile, "%f %f %f", &point[0], &point[1], &point[2]);

         if (head == (vertexlist *)NULL) {
            head = new vertexlist;
            copyarray3(head->pt, point);
            head->id = vertexcount = otr->vptr[i-1] = 1;
         }

         else
            otr->vptr[i-(1+j)] = head->insert(point, &vertexcount);
      }

   }

   fclose(infile);
   return 1;
}


/* ***********************************************************************

Ok this is the BYU format (well at least this is what I think is the
        BYU format.)

Line 1 has the number of parts, number of vertices, number of polygons and
        number of elements in the connectivity array.

Line 2 has the start polygon number and end polygon number for part 1
Line 3 has the start polygon number and end polygon number for part 2
.
.
.

Line 18 has the x y z coordinates for vertex 1 and vertex 2
Line 19 has the x y z coordinates for vertex 3 and vertex 4
.
.
.
Line 12715 has the connectivity array for the first polygon
        example:
                1 2 3 -4
        vertex 1 to vertex 2 to vertex 3 to vertex 4 and back to vertex 1

        a negative number show the last vertex of a polygon

        source to read in BYU format
        written by Kalvin Quinkert kalvin@uswest.com
        sometime between January 1992 and May 1992
        modified by Alex Harper March 1995
*********************************************************************** */

int converter::read_byu(char *filename) {

   int i, j, k;
   int partcount, temp1, temp2;
   objectlist *otr;
   intlist *ltr, *lltr;
   FILE *infile;
   vector3f point;
   int *index;
   int vcount;

   if (!(infile=fopen(filename, "r"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   printf("Reading...\n");

   fscanf(infile, "%ld %ld %ld %ld", &partcount, &vcount, &polycount, &edgecount);

   index = new int[vcount+1];

   for (i=0;i<partcount;i++)
      fscanf(infile, "%ld %ld\n", &temp1, &temp2);

   vertexcount = 0;

   for (i=0; i < vcount; i++) {
      fscanf(infile, "%f %f %f", &point[0], &point[1], &point[2]);

      if (head == (vertexlist *)NULL) {

         head = new vertexlist;
         copyarray3(head->pt, point);
         head->id = vertexcount = index[1] = 1;
      }

      else
         index[i+1] = head->insert(point, &vertexcount);
   }

                                        // reverse vertex list
   ltr = (intlist *)NULL;

   for(i=0; i<polycount; i++) {
      j = 0;

      while (1) {
         j++;
         lltr = new intlist;
         fscanf(infile, "%ld", &lltr->x);
         lltr->next = ltr;
         ltr = lltr;

         if (lltr->x < 0) {
            lltr->x = -lltr->x;
            break;
         }

      }

      otr = new objectlist;
      otr->id = i;
      otr->rnflag = reverse_normal;

      if (ob == (list *)NULL)
         ob = otr;
      else
         ob->insert_id(otr);

      otr->vptr = new int[j];
      otr->polynum   = j;

      for (k=j-1; k>-1; k--) {
         otr->vptr[k] = index[ltr->x];
         lltr = ltr;
         ltr = ltr->next;
         delete lltr;
      }

   }

   delete index;
   return 1;
}


/* ***********************************************************************
  LightWave 3D objects are stored as IFF files with a FORM type of
  LWOB.  Currently, a FORM LWOB must contain a PNTS chunk, a SRFS
  chunk, and a POLS chunk (in that order).  These may be followed by
  zero or more SURF chunks.  LightWave 2.0 also has the ability to
  save and load surface descriptions as FORM LWOB files that contain
  only a SURF chunk.
*********************************************************************** */

int converter::read_lw(char *filename) {

   FILE *infile;
   char buffer[5];
   int  bytesize;		// 32 bit int
   int  i;
   int *index;

   if (!(infile=fopen(filename, "r"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   printf("Reading...\n");

   for (i=0; i<4; i++)
      buffer[i] = getuchar(infile);
   buffer[4] = 0;

   if (strcmp(buffer, "FORM")) {
      printf("%s is not in Lightwave format... Aborting...\n", filename);
      fclose(infile);
      exit(1);
   }

   bytesize = getuint(infile);

   for (i=0; i<4; i++)
      buffer[i] = getuchar(infile);
   buffer[4] = 0;
   bytesize -= 4;

   if (strcmp(buffer, "LWOB")) {
      printf("%s is not in Lightwave format... Aborting...\n", filename);
      fclose(infile);
      exit(1);
   }

   for (i=0; i<4; i++)
      buffer[i] = getuchar(infile);
   buffer[4] = 0;
   bytesize -= 4;

   if (!strcmp(buffer, "SURF")) {
      printf("Only a SURF chunk...  Aborting...\n");
      fclose(infile);
      exit(1);
   }

   if (!strcmp(buffer, "PNTS")) {

      read_pnts(infile, &index, &bytesize);

      for (i=0; i<4; i++)
         buffer[i] = getuchar(infile);
      buffer[4] = 0;
      bytesize -= 4;

      if (strcmp(buffer, "SRFS")) {
         printf("Invalid SRFS chunk -> %s... Aborting...\n", buffer);
         fclose(infile);
         exit(1);
      }

      read_srfs(infile, &bytesize);

      for (i=0; i<4; i++)
         buffer[i] = getuchar(infile);
      buffer[4] = 0;
      bytesize -= 4;

      if (strcmp(buffer, "POLS")) {
         printf("Invalid POLS chunk -> %s... Aborting...\n", buffer);
         fclose(infile);
         exit(1);
      }

      read_pols(infile, index, &bytesize);

      while (bytesize > 0) {

         for (i=0; i<4; i++)
            buffer[i] = getuchar(infile);
         buffer[4] = 0;
         bytesize -= 4;

         if (strcmp(buffer, "SURF")) {
            printf("Invalid SURF chunk -> %s... Aborting...\n", buffer);
            fclose(infile);
            exit(1);
         }

         read_surf(infile, &bytesize);
      }

      fclose(infile);
      delete index;
      return 1;
   }

   printf("%s is not in Lightwave format... Aborting...\n", filename);
   fclose(infile);
   exit(1);
   return 0;
}


// negate z to get "pixcon" coord system
int converter::read_pnts(FILE *infile, int **index, int *bytesize) {

    int i, j;
    vector3f point;

   printf("PNTS...\n");

   j = getuint(infile);
   (*bytesize) -= 4 + j;

   if (*bytesize < 0) {
      printf("Invalid bytesize... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   j /= 12;

   *index = new int[j];

   for (i=0; i<j; i++) {
      *((unsigned int *)(&point[0])) = getuint(infile);
      *((unsigned int *)(&point[1])) = getuint(infile);
      *((unsigned int *)(&point[2])) = getuint(infile);

      if (head == (vertexlist *)NULL) {
         head = new vertexlist;
         copyarray3(head->pt, point);
         head->id = (*index)[0] = vertexcount = 1;
      }

      else
         (*index)[i] = head->insert(point, &vertexcount);
   }

   return 1;
}


int converter::read_srfs(FILE *infile, int *bytesize) {

   int j;
   int len;
   charlist *buffer, *head = (charlist *)NULL;

   printf("SRFS...\n");

   j = getuint(infile);
   (*bytesize) -= 4 + j;

   if (*bytesize < 0) {
      printf("Invalid SRFS - bad chunk size... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   for (scount=0; j > 0; scount++) {

      buffer = new charlist;

      len = getstring(infile, buffer->x, MAXSTRLEN);

      if (len < 1) {
         printf("Invalid string size (%d:%d)... Aborting...\n", polycount, len);
         fclose(infile);
         exit(1);
      }

      j -= len;

      if ((len & 0x01)) {		// if (odd # of bytes, read 1 more)
         getuchar(infile);
         j--;
      }

      buffer->next = head;
      head = buffer;
   }

   if (j < 0) {
      printf("SRFS - chunk size != listed chunk size... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   surface_list = new lw_shade[scount];

   for (j=scount-1; j > -1; j--) {
      surface_list[j].id = j;
      strcpy(surface_list[j].name, head->x);
      buffer = head;
      head = head->next;
      delete buffer;
   }

   return 1;
}


int converter::read_pols(FILE *infile, int *index, int *bytesize) {

   int chunk_size;
   int i, j, k;
   objectlist *otr;

   printf("POLS...\n");

   chunk_size = getuint(infile);
   (*bytesize) -= 4 + chunk_size;

   if (*bytesize < 0) {
      printf("POLS: Invalid bytesize... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   edgecount = 0;
   polycount = 0;

   while (chunk_size > 0) {

      otr = new objectlist;
      otr->rnflag = reverse_normal;

      polycount++;
      otr->id = polycount;
      otr->polynum   = getushort(infile);
      otr->vptr = new int[otr->polynum];
      edgecount += otr->polynum;

      chunk_size -= (1 + otr->polynum + 1) << 1;

      for (i=0; i<otr->polynum; i++)
         otr->vptr[i] = index[getushort(infile)];

      k = (short)getushort(infile);

      if (otr->polynum > 2) {
         otr->lw = &surface_list[abs(k)-1];
         otr->lw->count++;

         if (ob == (objectlist *)NULL)
            ob = otr;
         else
            ob->insert_id(otr);
      }

      else {
         polycount--;
         edgecount -= otr->polynum;
         delete otr;
      }

      if (k < 0) {		// detail polygons...

         k = getushort(infile);
         chunk_size -= 2;

         for (j=0; j<k; j++) {
            otr = new objectlist;
            otr->rnflag = reverse_normal;

            polycount++;
            otr->id = polycount;
            otr->polynum = getushort(infile);
            otr->vptr = new int[otr->polynum];
            edgecount += otr->polynum;

            chunk_size -= (1 + otr->polynum + 1) << 1;

            for (i=0; i<otr->polynum; i++)
               otr->vptr[i] = index[getushort(infile)];
     
            otr->lw = &surface_list[getushort(infile)];

            if (otr->polynum > 2) {
               otr->lw->count++;
               ob->insert_id(otr);
            }

            else {
               polycount--;
               edgecount -= otr->polynum;
               delete otr;
            }

         }

      }

   }

   if (chunk_size != 0) {
      printf("Data read != chunksize (%d)...  Aborting...\n", chunk_size);
      fclose(infile);
      exit(1);
   }

   return 1;
}


int converter::read_surf(FILE *infile, int *bytesize) {

   int chunk_size, sub_chunk_size;
   char buffer[MAXSTRLEN];
   int i;
   int temp;
   int ptr;

   int texflag = MASK_NULL;

   printf("SURF...\n");

   chunk_size = getuint(infile);
   (*bytesize) -= 4 + chunk_size;

   if (*bytesize < 0) {
      printf("Invalid bytesize... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   temp = getstring(infile, buffer, MAXSTRLEN);

   if (temp < 1) {
      printf("Underestimated string size (%d:%d)... Aborting...\n", temp, MAXSTRLEN);
      fclose(infile);
      exit(1);
   }

   chunk_size -= temp;

   if ((temp & 0x01)) {		// if (odd # of bytes, read 1 more)
      getuchar(infile);
      chunk_size--;
   }

   for (ptr=0; ptr<scount; ptr++)
      if (!strcmp(surface_list[ptr].name, buffer))
         break;

   if (ptr == scount) {
      printf("Invalid polygon reference... Aborting...\n");
      fclose(infile);
      exit(1);
   }

   while (chunk_size > 0) {

      for (i=0; i<4; i++)
         buffer[i] = getuchar(infile);
      buffer[4] = 0;
      sub_chunk_size = getushort(infile);
      chunk_size -= 6 + sub_chunk_size;

	// read 4 bytes - rgb {extra}  surface color
      if (!strcmp(buffer, "COLR")) {

         printf("COLR...\n");

         if (sub_chunk_size != 4) {
            printf("Invalid COLR chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         surface_list[ptr].shade[0] = getuchar(infile);
         surface_list[ptr].shade[1] = getuchar(infile);
         surface_list[ptr].shade[2] = getuchar(infile);

         getuchar(infile);
      }

      else if (!strcmp(buffer, "FLAG")) {

         printf("FLAG...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid FLAG chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         temp = getushort(infile);

         if (!surface_list[ptr].flag_lum) {
            surface_list[ptr].flag_lum = temp & 0x01;
            surface_list[ptr].lum = 256;
         }

         surface_list[ptr].flag_dbl_side = (unsigned char)((temp >> 8) & 0x01);
      }

      else if (!strcmp(buffer, "LUMI")) {
         
         printf("LUMI...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid LUMI chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         surface_list[ptr].lum = getushort(infile);
      }

      else if (!strcmp(buffer, "DIFF")) {
         
         printf("DIFF...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid DIFF chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         surface_list[ptr].diff = getushort(infile);
      }

      else if (!strcmp(buffer, "SPEC")) {

         printf("SPEC...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid SPEC chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         surface_list[ptr].spec = getushort(infile);
      }

      else if (!strcmp(buffer, "GLOS")) {

         printf("GLOS...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid GLOS chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         i = getushort(infile);

         switch(i) {
            case 1024:
               surface_list[ptr].specn = 7.0;
               break;
            case 256:
               surface_list[ptr].specn = 5.0;
               break;
           case 64:
               surface_list[ptr].specn = 3.0;
               break;
           case 16:
               surface_list[ptr].specn = 1.0;
               break;
           default:
               printf("default specn %d\n", i);
               surface_list[ptr].specn - 1.0;
         }

      }

      else if (!strcmp(buffer, "TRAN")) {
         printf("TRAN...\n");

         if (sub_chunk_size != 2) {
            printf("Invalid TRAN chunk size... Aborting...\n");
            fclose(infile);
            exit(1);
         }

         i = getushort(infile);

         if (!i)
            surface_list[ptr].trans = -1.0;
         else
            surface_list[ptr].trans = i/256.0;
      }

      else if (!strcmp(buffer, "DTEX")) {
         printf("DTEX...\n");

         texflag = MASK_DTEX;
         surface_list[ptr].texture.init(MASK_DTEX);
         
         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);

//printf("diffuse texture type %s\n", buffer);
      }

      else if (!strcmp(buffer, "STEX")) {
         printf("STEX...\n");

         texflag = MASK_STEX;
         if (surface_list[ptr].texture.id != MASK_DTEX)
            surface_list[ptr].texture.init(MASK_STEX);

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);

//printf("specular texture type %s\n", buffer);
      }

      else if (!strcmp(buffer, "CTEX")) {
         printf("CTEX...\n");

         texflag = MASK_CTEX;

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);
      }

      else if (!strcmp(buffer, "RTEX")) {
         printf("RTEX...\n");

         texflag = MASK_RTEX;

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);
      }

      else if (!strcmp(buffer, "TTEX")) {
         printf("TTEX...\n");

         texflag = MASK_TTEX;

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);
      }

      else if (!strcmp(buffer, "BTEX")) {
         printf("BTEX...\n");

         texflag = MASK_BTEX;

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);
      }

      else if (!strcmp(buffer, "LTEX")) {
         printf("LTEX...\n");

         texflag = MASK_LTEX;

         i = getstring(infile, buffer, MAXSTRLEN);

         if (i < 1) {
            printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
            fclose(infile);
            exit(1);
         }
 
         if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
            getuchar(infile);
      }

      else if (!strcmp(buffer, "TIMG")) {
         printf("TIMG...\n");

         if (texflag == surface_list[ptr].texture.id) {
            i = surface_list[ptr].texture.read_texname(infile);

            if (i < 1) {
               printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
               printf("Invalid bytesize (%d:%d)... Aborting...\n", i, sub_chunk_size);
               fclose(infile);
               exit(1);
            }

            if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
               getuchar(infile);
         }

         else {
            i = getstring(infile, buffer, MAXSTRLEN);

            if (i < 1) {
               printf("Underestimated string size (%d:%d)... Aborting...\n", i, MAXSTRLEN);
               printf("Invalid bytesize (%d:%d)... Aborting...\n", i, sub_chunk_size);
               fclose(infile);
               exit(1);
            }

            if ((i & 0x01)) 		// if (odd # of bytes, read 1 more)
               getuchar(infile);
         }

//printf("image filename %s\n", surface_list[ptr].texture.texname);
      }

      else if (!strcmp(buffer, "TCTR")) {
         printf("TCTR...\n");

         if (texflag == surface_list[ptr].texture.id) {
         }

         else {
         }

getfloat(infile);         
getfloat(infile);         
getfloat(infile);         
//         printf("%f %f %f\n", getfloat(infile), getfloat(infile), getfloat(infile));
//3 4-byte floats - x y z of texture center
      }

      else if (!strcmp(buffer, "TSIZ")) {
         printf("TSIZ...\n");
getfloat(infile);         
getfloat(infile);         
getfloat(infile);         
//         printf("%f %f %f\n", getfloat(infile), getfloat(infile), getfloat(infile));
//3 4-byte floats - x y z of texture size
      }



/*
     else if (!strcmp(buffer, "TVAL")) {
//short - texture value for current texture
     }
*/



/*

        else if (!strcmp(sub_chunk, "TFLG")) {
short - 7 bits


xaxis
yaxis
z axis
world coord
negative image
pixel blending
antialiasing

        }












      else if (!strcmp(sub_chunk, "REFL")) {
short
0 - 256
0 - 100 percent reflection
IGNORE!!!!
if not set -> default to zero

        }

        else if (!strcmp(sub_chunk, "RIMG")) {
filename even bytes
image to be used as reflection map...
if reflect flag set, and no RIMG, then use sky/ground or
	raytracing as reflect map....
IGNORE
        }

        else if (!strcmp(sub_chunk, "RSAN")) {
4 byte float - angle of reflection map seam (deg)
IGRNORE
        }

        else if (!strcmp(sub_chunk, "RIND")) {
4 byte float - refractive index - ratio of speed of light in a vacum
		to speed of light in material
IGGNORE

        }

        else if (!strcmp(sub_chunk, "EDGE")) {
4 byte float - edge transparency
IGNORE
        }

        else if (!strcmp(sub_chunk, "SMAN")) {
4 byte float - max angle (deg) between 2 adj polygons that can
be smooth shaded
IGNORE
        }


        else if (!strcmp(sub_chunk, "TFAL")) {
3 4-byte floats - x y z of texture falloff
        }

        else if (!strcmp(sub_chunk, "TVEL")) {
3 4-byte floats - x y z of texture velocity
        }

        else if (!strcmp(sub_chunk, "TCLR")) {
4 bytes - rgb-null - color for texture - color modifier
        }



        else if (!strcmp(sub_chunk, "TAMP")) {
4-byte float - amplitude of current bump texture
        }

        else if (!strcmp(sub_chunk, "TFRQ")) {
short - number of noise freq/wave sources for current texture
        }




        else if (!strcmp(sub_chunk, "TSP0")) {
4-byte float - represent button 0 that controls special params 
		(Contrast/Turbulence/Wavelength/etc)
        }

        else if (!strcmp(sub_chunk, "TSP1")) {
4-byte float - represent button 0 that controls special params 
		(Contrast/Turbulence/Wavelength/etc)
        }

        else if (!strcmp(sub_chunk, "TSP2")) {
4-byte float - represent button 0 that controls special params 
		(Contrast/Turbulence/Wavelength/etc)
        }
*/


      else {
         printf("Skipping %s chunk...\n", buffer);
         for (i=0; i<sub_chunk_size; i++)
            getuchar(infile);
      }

   }

   return 1;
}


int converter::write_spg(char *filename, int sflag) {

   list *qtr, *ptr;
   FILE *outfile;

   if (!vertexcount || !polycount || !edgecount || head == (list *)NULL || ob == (list *)NULL) {
      printf("NULL Object... Aborting...\n");
      exit(1);
   }

   if (sflag) {
      qtr = head;
      head = (vertexlist *)NULL;
      head = (vertexlist *)qtr->get_node();

      if (head != qtr)			// sort on id not number
         while (qtr != (list *)NULL) {
            ptr = qtr->get_node();
            if (ptr == qtr)
               qtr = (list *)NULL;
            head->insert_id(ptr);
         }

   }

   if (!(outfile=fopen(filename, "w"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   printf("Writing...\n");

   ob->adjust_totals(&polycount, &edgecount);

   fprintf(outfile, "%d %d %d\n", vertexcount, polycount, edgecount);

   head->write_data(outfile);
   ob->write_data(outfile);
   ob->write_dbl_side(outfile);

   fclose(outfile);
   return 1;
}


int converter::write_ilm(char *filename) {

   FILE *outfile;
   int base;
   int i;

   if (!(outfile=fopen(filename, "w"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   printf("Writing...\n");

   base = 0;

   for (i=1; i<scount; i++)
      if (surface_list[base].count < surface_list[i].count)
         base = i;

   surface_list[base].write_data(outfile);

   fprintf(outfile, "no_sides %d\n\n", polycount-surface_list[base].count);
   ob->write_ilm(outfile, base);

   fclose(outfile);
   return 1;
}


int converter::write_notes(char *filename) {

   FILE *outfile;

   if (!(outfile=fopen(filename, "w"))) {
      printf("Cannot open %s.  Aborting...\n", filename);
      exit(1);
   }

   printf("Writing...\n");

   ob->write_notes(outfile);

   fclose(outfile);
   return 1;
}


/*
int converter::read_symbol2d(FILE *infile, char *buffer) {

   int i, j;

   list *qtr;
   float point[3];
   float3dlist *fhead, *ptr;

   polycount = 0;
   vertexcount = 0;
   edgecount = 0;

   printf("Reading 2D...\n");

   fscanf(infile, "%s", buffer);

   while(strcmp(buffer, "3DFACET") ) {

      fhead = (float3dlist *)NULL;
      ptr = 
      


      polycount++;
      edgecount += i;

      qtr = new list;
      qtr->next = ob;
      ob = qtr;

      ob->vptr = new int[i];
      ob->polynum   = i;

      for (j=0; j<i; j++) {

         fscanf(infile, "%f %f %f", &point[0], &point[1], &point[2]);

         if (head == (list *)NULL) {

            head = new list;
            copyarray3(head->pt, point);
            head->id = vertexcount = ob->vptr[i-1] = 1;
         }

         else
            ob->vptr[i-(1+j)] = head->insert(point, &vertexcount);
      }

   }

   fclose(infile);
   return 1;
}


void extract_symbol(char *filename) {

   converter *x;
   FILE *infile;
   char buffer[1000], buffer2[256], name[256];

   if (!(infile = fopen(argv[1], "r"))) {
      printf("Could not access %s... Aborting...\n", argv[1]);
      exit(0);
   }

   fscanf(infile, "%s", buffer);
   if (strcmp(buffer, "EXECUTE")) {
      printf("%s is not a symbol file... Aborting...\n", argv[1]);
      fclose(infile);
      exit(0);
   }

   do {
      if (fscanf(infile, "%s", buffer) == EOF)
         break;
   } while (strcmp(buffer, "SYMBOL"));

   while (1) {
      if (!strcmp(buffer, "SYMBOL")) {
         fscanf(infile, "%s", name);
         fscanf(infile, "%s", buffer);
      }

      else
         if (!strcmp(buffer2, "END"))
            break;
         else
            if (!strcmp(buffer2, "2DFACET")) {
               x = new converter;
               x->read_symbol2d(infile, buffer);
               sprintf(buffer2, "%s.2d", name);
               x->write_spg(buffer2, 1);
               delete x;
            }

            else
               if (!strcmp(buffer2, "3DFACET")) {
                  x = new converter;
                  x->read_symbol3d(infile, buffer);
                  sprintf(buffer2, "%s.2d", name);
                  x->write_spg(buffer2, 1);
                  delete x;
               }

               else
                  fscanf(infile, "%s", buffer);
   }

   fclose(infile);
}
*/


void main(int argc, char **argv) {

   converter x;
   char buffer[MAXSTRLEN];

   if (argc < 4) {
      printf("usage: convert <source (ply/byu/lw)> <fromfilename> <tofilename> <-r>\n");
      exit(1);
   }

   if (argc == 5)
      reverse_normal = 1;

   if (!strcmp(argv[1], "ply")) {
      x.read_ply(argv[2]);

      strcpy(buffer, argv[3]);
      strcat(buffer, ".spg");
      x.write_spg(buffer, 1);

      return;
   }

   if (!strcmp(argv[1], "byu")) {
      x.read_byu(argv[2]);

      strcpy(buffer, argv[3]);
      strcat(buffer, ".spg");
      x.write_spg(buffer, 1);

      return;
   }

   if (!strcmp(argv[1], "lw")) {
      x.read_lw(argv[2]);

      strcpy(buffer, argv[3]);
      strcat(buffer, ".spg");
      x.write_spg(buffer, 1);

      strcpy(buffer, argv[3]);
      strcat(buffer, ".ilm");
      x.write_ilm(buffer);

      strcpy(buffer, argv[3]);
      strcat(buffer, ".nts");
      x.write_notes(buffer);

      return;
   }

/*
  if (!strcmp(argv[1], "symbol")) {
      extract_symbol(argv[2]);
      return;
   }
*/

//   printf("usage: convert <source (ply/byu/symbol)> <fromfilename> <tofilename>\n");
   printf("usage: convert <source (ply/byu)> <fromfilename> <tofilename>\n");

   return;
}



