
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

#if !defined __CONVERT_H
#define __CONVERT_H

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAXSTRLEN 256

#include "matrix.h"
#include "io.h"

#define MASK_NULL	0
#define MASK_CTEX	1
#define MASK_DTEX	2
#define MASK_STEX	4
#define MASK_RTEX	8
#define MASK_TTEX	16
#define MASK_BTEX	32
			// dunno what LTEX is for... (light texture???)
#define MASK_LTEX	64

typedef char string[MAXSTRLEN];

class textype {

   public:
      char texname[MAXSTRLEN];
      int id;

      void init(int mask) {

         id = mask;
         texname[0] = 0;
      }

      textype() {
         init(MASK_NULL);
      }

      int read_texname(FILE *infile) {

         char buffer[MAXSTRLEN];
         int  serial;
         int count;

         count = getstring(infile, texname, MAXSTRLEN);

         if (count < 1 || sscanf(texname, "%s %d", buffer, &serial) < 2)
            return count;

         sprintf(texname, "%s.%d", buffer, serial);
         return count;
      }

};


class lw_shade {

   public:
      string name;
      int id;
      int count;

      char flag_dbl_side, flag_lum;
      float lum; 	// 0..256 (0..1)

      float trans;

      float shade[3];
      float diff, spec;	// 0..256
      float specn;	// 16, 64, 256, 1024;

      textype texture;

      lw_shade() {
         shade[0] = shade[1] = shade[2] = 0;
         flag_dbl_side = flag_lum = 0;
         diff = spec = 0;
         specn = 1;
         trans = -1;
      }

      void write_data(FILE *outfile);
      void write_notes(FILE *outfile);
};


class charlist {

   public:
      string x;
      charlist *next;
};


class intlist {

   public:
      int x;
      intlist *next;
};


class float3dlist {

   public:
      vector3f pt;
      float3dlist *next;
};


class list {

   public:
      int id;
      list *back, *next;

      list() { back = next = (list *)NULL; }

      ~list() {

         if (back != (list *)NULL)
            delete back;
         if (next != (list *)NULL)
            delete next;
      }

      virtual void write_data(FILE *outfile) = 0;
      void insert_id(list *x);
      list *find_id(int x);

      list *get_node();
};


class vertexlist : public list {

   public:
      vector3f pt;

      void write_data(FILE *outfile);
      int  insert(float *point, int *vertexcount);
};


class objectlist : public list {

   public:
      int polynum;
      int  *vptr;
      lw_shade *lw;
      int other_id;
      int rnflag;

      objectlist() {
         vptr = (int *)NULL;
         lw = (lw_shade *)NULL;
         rnflag = 0;
      }

      ~objectlist() {

         if (vptr != (int *)NULL)
            delete vptr;
      }

      void adjust_totals(int *polycount, int *edgecount);
      void write_ilm(FILE *outfile, int base);
      void write_dbl_side(FILE *outfile);

      void write_data(FILE *outfile);
      void write_notes(FILE *outfile);
};


class converter {

   public:
      int polycount;
      int vertexcount;
      int edgecount;

      lw_shade *surface_list;
      int scount;

      vertexlist *head;                      // vertex list
      objectlist *ob;                        // polygon list

      converter() {
         head = (vertexlist *)NULL;
         ob = (objectlist *)NULL;
         surface_list = (lw_shade *)NULL;
      }

      ~converter() {
         if (head != (vertexlist *)NULL)
            delete head;
         if (ob != (objectlist *)NULL)
            delete ob;
         if (surface_list != (lw_shade *)NULL)
            delete surface_list;
      }

      int read_ply(char *filename);

      int read_byu(char *filename);

      int read_lw(char *filename);
      int read_pnts(FILE *infile, int **index, int *bytesize);
      int read_srfs(FILE *infile, int *bytesize);
      int read_pols(FILE *infile, int *index, int *bytesize);
      int read_surf(FILE *infile, int *bytesize);

//      int read_symbol2d(FILE *infile, char *buffer);
//      int read_symbol3d(FILE *infile, char *buffer);

      int write_spg(char *filename, int sflag);
      int write_ilm(char *filename);
      int write_notes(char *filename);
};

#endif
