
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#include "base.h"

pc::pc() {

   int i, j;

   base_color = &gshade;
   shaderlist = (shaderlisttype *)NULL;
   mctype = WFBW;
   mcinfo = CINULL;
   center[0] = center[1] = center[2] = 0;
   center[3] = 1;
   size = 1;
   init_mx(i, j, rotate);
   sflag = 0;
   lob = (shadelist *)NULL;
   frame = 0;

   texname2[0] = texname3[0] = 0;

   next = (pc *)NULL;
}


int pc::read_ilum(char *filename, shadelist **itr) {

   for (lob=*itr; lob!=(shadelist *)NULL; lob=lob->next)
      if (!strcmp(filename, lob->shadename))
         return 0;

   lob = new shadelist;

   lob->read_data(filename);
   lob->next = *itr;
   *itr = lob;

   return 1;
}


/* *************************************************************
   this function is the catch-all render object function
************************************************************* */
void pc::beamscan_children(light *lparm, subslave grandchild[]) {

   begin_scan();
   datacopy();
   beamscan(lparm, grandchild);
   end_scan();
}


/* *************************************************************
   this function is the catch-all render object function
************************************************************* */
void pc::scan_children(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) {

   begin_scan();
   datacopy();
   scan(cparm, lparm, zbuff, grandchild, shadptr);
   end_scan();
}


/* *************************************************************
   this function is the catch-all render object function
************************************************************* */
void pc::scan_children(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

   begin_scan();
   datacopy();
   scan(cparm, lparm, grandchild, shadptr);
   end_scan();
}


void pc::calc_color(float *surface, float fscale, float *normal, light *lmain, light *lparm, eye *parm, texcolortype *texcolor,
                    int flagtr, int flag3t, int flagl, float *rcolor, float *acolor) {

   shaderparamtype sparam;
   int k;
   light *ltr;

   if (flag3t) {
      copyarray4(sparam.in, surface);
      copyarray4(sparam.pt, surface);
      copyarray3(sparam.normal, normal);

      matvecmulto(iworld, sparam.in);

      smultarray3(sparam.in, fscale);

      sparam.out[0] = sparam.out[1] = sparam.out[2] = 1;
      sparam.add[0] = sparam.add[1] = sparam.add[2] = 0;

      for (k=0; k<shaderlist->shade->scount; k++)
         if (shaderlist->shade->s[k] != (shader *)NULL && shaderlist->shade->stype[k] == SURFACE)
            shaderlist->shade->s[k]->query_data((void *)(&sparam));

      copyarray3(surface, sparam.pt);
      copyarray3(normal, sparam.normal);

      multarray3(texcolor->color.ka, sparam.out);
      addarray3(texcolor->color.ka, sparam.add);

      multarray3(texcolor->color.kp, sparam.out);
      multarray3(texcolor->color.ks, sparam.out);

      if (flagl)
         multarray3(texcolor->color.lum, sparam.out);

      if (lmain != (light *)NULL)
         lmain->set_ambient(texcolor->color.ka, texcolor->ambient);
   }

   rcolor[0] = rcolor[1] = rcolor[2] = 0;

   if (lmain != (light *)NULL) {
      lmain->intensity(surface, normal, rcolor, parm, &texcolor->color, id);
      copyarray3(acolor, texcolor->ambient);
   }

   else
      acolor[0] = acolor[1] = acolor[2] = 0;

   if (flagl)
      addarray3(acolor, texcolor->color.lum);

   if (!flagtr)
      for (ltr=lparm; ltr!=(light *)NULL; ltr=ltr->next)
         if (ltr->ltype==BEAM_POINT || ltr->ltype==BEAM_FAR)
            ltr->intensity(surface, normal, acolor, parm, &texcolor->color, id);

   if (flag3t) {
      sparam.out[0] = sparam.out[1] = sparam.out[2] = 1;
      sparam.add[0] = sparam.add[1] = sparam.add[2] = 0;

      for (k=0; k<shaderlist->shade->scount; k++)
         if (shaderlist->shade->s[k] != (shader *)NULL && shaderlist->shade->stype[k] == ATMOSPHERE)
            shaderlist->shade->s[k]->query_data((void *)(&sparam));

      multarray3(rcolor, sparam.out);
      multarray3(acolor, sparam.out);
      addarray3(acolor, sparam.add);
   }

}


void pc::calc_color(float *surface, float fscale, float *normal, light *lmain, light *lparm, eye *parm, texcolortype *texcolor,
                float *bcolor, int flagtr, int flag2t, int flag3t, int flagl, float *rcolor, float *acolor) {

   shaderparamtype sparam;
   int k;
   light *ltr;

   if (flag3t) {
      copyarray4(sparam.in, surface);
      copyarray4(sparam.pt, surface);
      copyarray3(sparam.normal, normal);

      matvecmulto(iworld, sparam.in);

      smultarray3(sparam.in, fscale);

      sparam.out[0] = sparam.out[1] = sparam.out[2] = 1;
      sparam.add[0] = sparam.add[1] = sparam.add[2] = 0;

      for (k=0; k<shaderlist->shade->scount; k++)
         if (shaderlist->shade->s[k] != (shader *)NULL && shaderlist->shade->stype[k] == SURFACE)
            shaderlist->shade->s[k]->query_data((void *)(&sparam));

      copyarray3(surface, sparam.pt);
      copyarray3(normal, sparam.normal);

      multarray3(texcolor->color.ka, sparam.out);
      addarray3(texcolor->color.ka, sparam.add);

      multarray3(texcolor->color.kp, sparam.out);
      multarray3(texcolor->color.ks, sparam.out);

      if (flagl)
         multarray3(texcolor->color.lum, sparam.out);

      if (lmain != (light *)NULL)
         lmain->set_ambient(texcolor->color.ka, texcolor->ambient);
   }

   if (lmain != (light *)NULL) {
      copyarray3(rcolor, bcolor);

      if (flag2t) {
         multarray3(rcolor, texcolor->color.kp);
      }

      else
         if (flag3t)
            multarray3(rcolor, sparam.out);

      copyarray3(acolor, texcolor->ambient);
   }

   else {
       rcolor[0] = rcolor[1] = rcolor[2] = 0;
       acolor[0] = acolor[1] = acolor[2] = 0;
   }

   if (flagl)
      addarray3(acolor, texcolor->color.lum);

   if (!flagtr)
      for (ltr=lparm; ltr!=(light *)NULL; ltr=ltr->next)
         if (ltr->ltype==BEAM_POINT || ltr->ltype==BEAM_FAR)
            ltr->intensity(surface, normal, acolor, parm, &texcolor->color, id);

   if (flag3t) {
      sparam.out[0] = sparam.out[1] = sparam.out[2] = 1;
      sparam.add[0] = sparam.add[1] = sparam.add[2] = 0;

      for (k=0; k<shaderlist->shade->scount; k++)
         if (shaderlist->shade->s[k] != (shader *)NULL && shaderlist->shade->stype[k] == ATMOSPHERE)
            shaderlist->shade->s[k]->query_data((void *)(&sparam));

      multarray3(rcolor, sparam.out);
      multarray3(acolor, sparam.out);
      addarray3(acolor, sparam.add);
   }

}


int pc::parse(FILE *infile, char *token) {

   vector4f tmx[4];

   switch (token[0]) {

      case '2':
         if (!strcmp(token, "2d_texture")) {
            get_token(infile, texname2);
            mcinfo |= CITEXTURE;
            return 1;
         }

         return 0;

      case '3':
         if (!strcmp(token, "3d_texture")) {
            get_token(infile, texname3);
            mcinfo |= CI3DTEX;
            return 1;
         }

         return 0;

      case 'b':
         if (!strcmp(token, "biplane")) {
            mcinfo |= CIBITMAP;
            return 1;
         }

         return 0;

      case 'f':
         if (!strcmp(token, "frame")) {
            get_token(infile, token);
            frame = atoi(token) - 1;
            return 1;
         }

         return 0;

      case 'l':
         if (!strcmp(token, "luminate")) {
            mcinfo |= CILUM;
            return 1;
         }

         return 0;

      case 'r':

         if (!strcmp(token, "raycast")) {
            mcinfo |= CIRAYCAST;
            return 1;
         }

         if (strlen(token) < 8)
            return 0;

         switch(token[7]) {

            case 'x':
               if (!strcmp(token, "rotate_x")) {
                  get_token(infile, token);
                  rotate_mx_x(rotate, deg2rad(atof(token)));
                  return 1;
               }

               return 0;

            case 'y':
               if (!strcmp(token, "rotate_y")) {
                  get_token(infile, token);
                  rotate_mx_y(rotate, deg2rad(atof(token)));
                  return 1;
               }

               return 0;

            case 'z':
               if (!strcmp(token, "rotate_z")) {
                  get_token(infile, token);
                  rotate_mx_z(rotate, deg2rad(atof(token)));
                  return 1;
               }

               return 0;

            default:
               break;
         }

         return 0;

      case 's':
         if (!strcmp(token, "scale")) {
            get_token(infile, token);
            size = atof(token);
            return 1;
         }

         if (!strcmp(token, "shadow")) {
            sflag = 1;
            get_token(infile, token);
            splane[0] = atof(token);
            get_token(infile, token);
            splane[1] = atof(token);
            get_token(infile, token);
            splane[2] = atof(token);
            get_token(infile, token);
            splane[3] = atof(token);

            return 1;
         }

         if (!strcmp(token, "shade")) {
            get_token(infile, token);
            lower_case(token);

            switch (token[0]) {
               case 'a':
                  if (!strcmp(token, "alex")) {
                     get_token(infile, colorname);
                     mctype = ALEX;
                     return 1;
                  }

                  return 0;

               case 'b':
                  if (!strcmp(token, "bw")) {
                     mctype = BW;
                     return 1;
                  }

                  return 0;

               case 'c':
                  if (!strcmp(token, "constant")) {
                     get_token(infile, colorname);
                     mctype = CONSTANT;
                     return 1;
                  }

                  return 0;

               case 'd':
                  if (!strcmp(token, "dot")) {
                     mctype = DOT;
                     return 1;
                  }

                  return 0;

               case 'g':			// line/particle called before this
                  if (!strcmp(token, "gouraud")) {
                     get_token(infile, colorname);
                     mctype = GOURAUD;
                     return 1;
                  }

                  return 0;

               case 'i':
                  if (!strcmp(token, "irradiant")) {
                     get_token(infile, colorname);
                     mctype = IRRADIANT;
                     mcinfo |= CILUM;
                     return 1;
                  }

                  return 0;

               case 'p':
                  if (!strcmp(token, "phong")) {
                     get_token(infile, colorname);
                     mctype = PHONG;
                     return 1;
                  }

                  if (strlen(token) < 2)
                     return 0;

                  switch (token[1]) {

                     case 'a':
                        if (!strcmp(token, "palex")) {
                           get_token(infile, colorname);
                           mctype = PALEX;
                           return 1;
                        }

                        return 0;

                     case 'b':
                        if (!strcmp(token, "pbw")) {
                           mctype = PBW;
                           return 1;
                        }

                        return 0;

                     case 'c':
                        if (!strcmp(token, "pconstant")) {
                           get_token(infile, colorname);
                           mctype = PCONSTANT;
                           return 1;
                        }

                        return 0;

                     case 'd':
                        if (!strcmp(token, "pdot")) {
                           mctype = PDOT;
                           return 1;
                        }

                        return 0;

                     case 'g':
                        if (!strcmp(token, "pgouraud")) {
                           get_token(infile, colorname);
                           mctype = PGOURAUD;
                           return 1;
                        }

                        return 0;

                     case 'p':
                        if (!strcmp(token, "pphong")) {
                           get_token(infile, colorname);
                           mctype = PPHONG;
                           return 1;
                        }

                        return 0;

                     case 'w':
                        if (!strcmp(token, "pwfbw")) {
                           mctype = PWFBW;
                           return 1;
                        }

                        return 0;

                     default:
                        break;
                  }

                  return 0;

               case 'r':
                  if (!strcmp(token, "rsolid")) {
                     get_token(infile, colorname);
                     mctype = RSOLID;
                     return 1;
                  }

                  return 0;

               case 'w':
                  if (!strcmp(token, "wfbw")) {
                     mctype = WFBW;
                     return 1;
                  }

                  return 0;

               default:
                  break;
            }

            return 0;
         }

         return 0;
         
      case 't':
         if (strlen(token) < 6)
            return 0;

         switch (token[5]) {

            case 't':
               if (!strcmp(token, "tr_matrix")) {
                  get_token(infile, token);
                  tmx[0][0] = atof(token);
                  get_token(infile, token);
                  tmx[0][1] = atof(token);
                  get_token(infile, token);
                  tmx[0][2] = atof(token);
                  get_token(infile, token);
                  tmx[0][3] = atof(token);

                  get_token(infile, token);
                  tmx[1][0] = atof(token);
                  get_token(infile, token);
                  tmx[1][1] = atof(token);
                  get_token(infile, token);
                  tmx[1][2] = atof(token);
                  get_token(infile, token);
                  tmx[1][3] = atof(token);

                  get_token(infile, token);
                  tmx[2][0] = atof(token);
                  get_token(infile, token);
                  tmx[2][1] = atof(token);
                  get_token(infile, token);
                  tmx[2][2] = atof(token);
                  get_token(infile, token);
                  tmx[2][3] = atof(token);

                  get_token(infile, token);
                  tmx[3][0] = atof(token);
                  get_token(infile, token);
                  tmx[3][1] = atof(token);
                  get_token(infile, token);
                  tmx[3][2] = atof(token);
                  get_token(infile, token);
                  tmx[3][3] = atof(token);

                  matmatmulto(tmx, rotate);

                  return 1;
               }

               return 0;

            case 'l':
               if (!strcmp(token, "translate")) {
                  get_token(infile, token);
                  rotate[0][3] += atof(token);
                  get_token(infile, token);
                  rotate[1][3] += atof(token);
                  get_token(infile, token);
                  rotate[2][3] += atof(token);

                  return 1;
               }

               return 0;
         
            case 'p':
               if (!strcmp(token, "transparent")) {
                  mcinfo |= CITRANSPARENT;
                  return 1;
               }

               return 0;

            default:
               break;
         }

         return 0;

      default:
         break;
   }

   return 0;
}


void pc::preprocess(listtype *mainlist, shader *sparm) {

   if (((mctype > BW && mctype <= PHONG) || mctype > PBW) && query_whatami() != PARTICLE && query_whatami() != LINE) {

      read_ilum(colorname, &mainlist->itr);

      if (mcinfo & CITEXTURE)
         read_tex(texname2, filename, &mainlist->tdtr, &mainlist->texptr);
      else if (mcinfo & CI3DTEX)
         read_3dtex(texname3, filename, sparm, &mainlist->shltype);
   }

}
