
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#if !defined (__BASE_H)
#define __BASE_H

#include "light.h"
#include "zbuffer.h"
#include "shader.h"
#include "texpoly.h"
#include "texsbfast.h"

#define CINULL		0
#define CITEXTURE	1
#define CILUM		2
#define CIRAYCAST	4
#define CI3DTEX		8
#define CITRANSPARENT	16
#define CIBITMAP	32

typedef enum colorenum {
   SHADOW       =  -1,

   DOT		=  0,
   WFBW         =  1,
   BW           =  2,
   CONSTANT     =  3,
   ALEX         =  4,
   GOURAUD      =  5,
   PHONG        =  6,

   PDOT		= 40,
   PWFBW        = 41,
   PBW          = 42,
   PCONSTANT    = 43,
   PALEX        = 44,
   PGOURAUD     = 45,
   PPHONG       = 46,

   RSOLID       = 91,
   IRRADIANT    = 92,
} colortype;


#define MINDIM  5
#define MINBBOX 1.5


class pc;
class subslave;
class slave;
class master;

typedef struct grandchildstruct {
   eye      *cam;
   light    *lit;
   zbuffer  *zbuff;
   edgetype **e;
   pc       *source;
   int      id;
   int      *doneflag;
   volatile int done;
} grandchildtype;


class datatype {

   public:
      union {
         struct {                 // polygonal data
            int                  countvertex;           // vertex count
            int                  countedge;             // # of edges
            int                  countobject;           // # of polygons
            vector4f              *mvertex;
            int                  *object;               // polygon/vertex pointer list
            int                  *polynum;              // # of verticies in a polygon
            int                  **edgeptr;
         } pdata;

         struct {                 // surf. pat. data structure
            vector4f             *data;
            int                 vrow, vcol;             // x, y count
         } sdata;

      };

      char		dataname[MAXSTRLEN];   // data file for object
      datatype	*next;                 // ptr to next in list
      vector4f bbox[2];
};


class pc;
typedef pc * pcptrtype;

class listtype {                                // main pointer to objects

   public:

      datatype     *polydtr;                       // ptr to polygon objects
      datatype     *spatdtr;                       // ptr to patch objects
      camera       **camdtr;
      light        **lightdtr;
      pc           **header;
      shadelist    *itr;
      pc           **raddtr;
      shaderlisttype *shltype;
      texpolygon   *texptr;
      texbase      *tdtr;
      texsbfast    *sbptr;

      listtype() {
         polydtr  = (datatype *)NULL;
         spatdtr  = (datatype *)NULL;
         itr      = (shadelist *)NULL;
         shltype  = (shaderlisttype *)NULL;
         texptr   = (texpolygon *)NULL;
         tdtr     = (texbase *)NULL;
         sbptr    = (texsbfast *)NULL;

         header   = (pc **)NULL;
         camdtr   = (camera **)NULL;
         lightdtr = (light **)NULL;
         raddtr   = (pc **)NULL;
      }

      void init(int x) {

         int i;

         if (header != (pc **)NULL)
            delete header;
         header = new pcptrtype[x];

         if (camdtr != (camera **)NULL)
            delete camdtr;
         camdtr = new cameraptrtype[x];

         if (lightdtr != (light **)NULL)
            delete lightdtr;
         lightdtr = new lightptrtype[x];

         if (raddtr != (pc **)NULL)
            delete raddtr;
         raddtr = new pcptrtype[x];

         for (i=0; i<x; i++) {
            header[i] = (pc *)NULL;
            camdtr[i] = (camera *)NULL;
            lightdtr[i] = (light *)NULL; 
            raddtr[i] = (pc *)NULL;
         }

      }

};


/* *************************************************************
   This is the base class for objects to be rendered by the system
************************************************************* */
class pc {

   protected:
      vector4f bbox[2];
      vector4f box[8];

      virtual int  bound_box(eye *parm) = 0;
      virtual void transform(eye *parm) = 0;
              void calc_color(float *surface, float fscale, float *normal, light *lmain, light *lparm, eye *parm, texcolortype *texcolor,
                              int flagtr, int flag3t, int flagl, float *rcolor, float *acolor);
              void calc_color(float *surface, float fscale, float *normal, light *lmain, light *lparm, eye *parm, texcolortype *texcolor,
                              float *bcolor, int flagtr, int flag2t, int flag3t, int flagl, float *rcolor, float *acolor);

   public:
      int	frame;
      shadetype	*base_color;

      shaderlisttype *shaderlist;

      colortype	mctype;             // coloring
      unsigned int	mcinfo;
      float	center[4];          // object center in wc
      float	size;               // size change
      float	rotate[4][4];       // rotation matrix
      float	world[4][4];
      float	iworld[4][4];
      int	id;                 // object id #

      char	sflag;
      float	splane[4];          // shadow plane

      shadelist	*lob;

      char	texname2[MAXSTRLEN], texname3[MAXSTRLEN], colorname[MAXSTRLEN], filename[MAXSTRLEN];

      pc	*next;              // ptr to next object

      virtual void begin_scan() {}
      virtual void end_scan()   {}
      virtual void colorize(grandchildtype *child) = 0;
      virtual void datacopy() = 0;
      virtual void beamscan(light *lparm, subslave grandchild[]) = 0;
      virtual void scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) = 0;
      virtual void scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) = 0;
      virtual int read_data(char filename[], datatype **dptr) = 0;
      virtual int read_data(FILE *infile) = 0;
      virtual int  read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) = 0;
      virtual int  read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) = 0;
      virtual rendertype query_whatami() = 0;

      pc();

      virtual int  parse(FILE *infile, char *token);
      virtual void preprocess(listtype *mainlist, shader *sparm);

      int  read_ilum(char *filename, shadelist **itr); 

      void beamscan_children(light *lparm, subslave grandchild[]);

      void scan_children(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr);
      void scan_children(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr);
};

typedef pc ** pppc;

#include "control.h"


#endif
