
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
************************************************************* */

#include <stdio.h>
#include "matrix.h"


/* ***********************************************************************
                normal orders front, back left, top, right, bottom
                polygon orders : 0-3 front, 4-7 back, 0-4-5-1 left, etc
*********************************************************************** */
int line_rect_intersect(float *pt, float *v, vector4f *p, vector4f *n, float *flag) {

   vector4f ptemp[4];
   float t;
   float intersect[2];
   int ptr = 0;

   if (line_poly_intersect(pt, v, n[0], p, 4, &t)) {          // front face
      intersect[0] = t;
      ptr++;
   }

   if (line_poly_intersect(pt, v, n[1], &p[4], 4, &t)) {      // back face
      intersect[ptr] = t;
      ptr++;

      if (ptr == 2) {
         if (intersect[0]*intersect[1] > CORRECT)       // outside the box
            if (intersect[0] > 0) {                     // face to box
               if (intersect[0] < intersect[1])
                  *flag = intersect[0];
               else
                  *flag = intersect[1];
               return 1;
            }

                                                        // inside box or face away from box
         if (intersect[0] > intersect[1])
            *flag = intersect[0];
         else
            *flag = intersect[1];
         return 1;
      }

   }
                                                // left
   copyarray4(ptemp[0], p[0]);
   copyarray4(ptemp[1], p[4]);
   copyarray4(ptemp[2], p[5]);
   copyarray4(ptemp[3], p[1]);

   if (line_poly_intersect(pt, v, n[2], ptemp, 4, &t)) {
      intersect[ptr] = t;
      ptr++;

      if (ptr == 2) {
         if (intersect[0]*intersect[1] > CORRECT)       // outside the box
            if (intersect[0] > 0) {                     // face to box
               if (intersect[0] < intersect[1])
                  *flag = intersect[0];
               else
                  *flag = intersect[1];
               return 1;
            }

         if (intersect[0] > intersect[1])               // inside box or face aw         if (intersect[0] > intersect[1])
            *flag = intersect[0];
         else
            *flag = intersect[1];
         return 1;
      }

   }

   copyarray4(ptemp[0], p[1]);
   copyarray4(ptemp[1], p[5]);
   copyarray4(ptemp[2], p[6]);
   copyarray4(ptemp[3], p[2]);

   if (line_poly_intersect(pt, v, n[3], ptemp, 4, &t)) {
      intersect[ptr] = t;
      ptr++;

      if (ptr == 2) {
         if (intersect[0]*intersect[1] > CORRECT)       // outside the box
            if (intersect[0] > 0) {                     // face to box
               if (intersect[0] < intersect[1])
                  *flag = intersect[0];
               else
                  *flag = intersect[1];
               return 1;
            }

         if (intersect[0] > intersect[1])               // inside box or face aw         if (intersect[0] > intersect[1])
            *flag = intersect[0];
         else
            *flag = intersect[1];
         return 1;
      }

   }

   copyarray4(ptemp[0], p[2]);
   copyarray4(ptemp[1], p[6]);
   copyarray4(ptemp[2], p[7]);
   copyarray4(ptemp[3], p[3]);

   if (line_poly_intersect(pt, v, n[4], ptemp, 4, &t)) {
      intersect[ptr] = t;
      ptr++;

      if (ptr == 2) {
         if (intersect[0]*intersect[1] > CORRECT)       // outside the box
            if (intersect[0] > 0) {                     // face to box
               if (intersect[0] < intersect[1])
                  *flag = intersect[0];
               else
                  *flag = intersect[1];
               return 1;
            }

         if (intersect[0] > intersect[1])               // inside box or face aw         if (intersect[0] > intersect[1])
            *flag = intersect[0];
         else
            *flag = intersect[1];
         return 1;
      }

   }

   copyarray4(ptemp[0], p[3]);
   copyarray4(ptemp[1], p[7]);
   copyarray4(ptemp[2], p[4]);
   copyarray4(ptemp[3], p[0]);

   if (line_poly_intersect(pt, v, n[5], ptemp, 4, &t)) {
      intersect[ptr] = t;
      ptr++;

      if (ptr == 2) {
         if (intersect[0]*intersect[1] > CORRECT)       // outside the box
            if (intersect[0] > 0) {                     // face to box
               if (intersect[0] < intersect[1])
                  *flag = intersect[0];
               else
                  *flag = intersect[1];
               return 1;
            }

         if (intersect[0] > intersect[1])                // inside box or face aw         if (intersect[0] > intersect[1])
            *flag = intersect[0];
         else
            *flag = intersect[1];
         return 1;
      }

   }

   if (ptr == 1) {                              // should never happen unless I screwed the pooch
      *flag = intersect[0];
      printf("FUCKUP in LINE_RECT!!!\n");
      return 1;
   }

   return 0;
}


/* **********************************************************************
   This function calculates the distance between an objects center, and
   a plane.
*********************************************************************** */
int line_plane_intersect(float *plane, float *pt, float *vector, float *t) {

   float x;

   x = dotproduct3(plane, vector);

   if (not_zero(x)) {
      *t = -(plane[3] + dotproduct3(pt, plane))/x;
      return 1;
   }

   *t = 0;
   return 0;
}


/* **********************************************************************
        sphere -
        pt     - point on line
        vector - line vector
	NOTE:	if inside sphere, return positive t
*********************************************************************** */
int line_sphere_intersect(float radius, float *sphere, float *pt, float *vector, float *t) {

   float a, b, c, check, t1, t2;

   b = c = pt[0]-sphere[0];
   b *= vector[0];
   c *= c;

   a = pt[1]-sphere[1];
   b += (a*vector[1]);
   c += SQR(a);

   a = pt[2]-sphere[2];
   b += (a*vector[2]);
   c += SQR(a) - SQR(radius);

   b += b;

   a = SQR(vector[0]) + SQR(vector[1]) + SQR(vector[2]);

   check = SQR(b) - 4*a*c;

   if (check < 0) {
      *t = 0;
      return 0;
   }

   check = sqrt(check);
   a += a;

   t1 = (-b - check)/a;
   t2 = (-b + check)/a;

   if (t1 < 0) {
      if (t2 < 0) {
         if (t1 < t2)
            *t = t2;
         else
            *t = t1;

         return 1;
      }

      *t = t2;            // signs are different - we are inside
      return -1;
   }

   if (t2 < 0) {     // signs are different - we are inside
      *t = t1;
      return -1;
   }

   if (t1 < t2)
      *t = t1;
   else
      *t = t2;

   return 1;
}


/* **********************************************************************
*********************************************************************** */
int line_ellipse_intersect(float *scale, float *pt, float *vector, float *t) {

   float a, b, c, check, t1, t2;
   float mscale[3];

   mscale[0] = SQR(scale[0]);
   mscale[1] = SQR(scale[1]);
   mscale[2] = SQR(scale[2]);

   b = c = pt[0];
   b *= vector[0]/mscale[0];
   c *= c/mscale[0];

   a = pt[1];
   b += (a*vector[1])/mscale[1];
   c += SQR(a)/mscale[1];

   a = pt[2];
   b += (a*vector[2])/mscale[2];
   c += SQR(a)/mscale[2] - 1;

   b += b;

   a = SQR(vector[0])/mscale[0] + SQR(vector[1])/mscale[1] + SQR(vector[2])/mscale[2];

   check = SQR(b) - 4*a*c;

   if (check < 0) {
      *t = 0;
      return 0;
   }

   check = sqrt(check);
   a += a;

   t1 = (-b - check)/a;
   t2 = (-b + check)/a;

   if (t1 < 0) {
      if (t2 < 0) {
         if (t1 < t2)
            *t = t2;
         else
            *t = t1;

         return 1;
      }

      *t = t2;            // signs are different - we are inside
      return -1;
   }

   if (t2 < 0) {          // signs are different - we are inside
      *t = t1;
      return -1;
   }

   if (t1 < t2)
      *t = t1;
   else
      *t = t2;

   return 1;
}


/* **********************************************************************
*********************************************************************** */
int line_poly_intersect(float *pt, float *vector, float *plane, vector4f *vertex, int count, float *point) {

   float t;
   int  i;
   float v0[3], v1[3], v2[3];
   float angle;

   if (!(line_plane_intersect(plane, pt, vector, &t)))
      return 0;

   point[0] = pt[0] + t*vector[0];
   point[1] = pt[1] + t*vector[1];
   point[2] = pt[2] + t*vector[2];

   copyarray3(v2, vertex[0]);
   subarray3(v2, point);
   normalize3(v2);

   copyarray3(v0, v2);

   for (i=1, angle = 0; i<count; i++) {
      copyarray3(v1, v2);
      copyarray3(v2, vertex[i]);
      subarray3(v2, point);
      normalize3(v2);

      t = dotproduct3(v1, v2);
      if (t < 1-CORRECT)
         angle += ACOS(t);
   }

   copyarray3(v1, v0);

   t = dotproduct3(v1, v2);
   if (t < 1-CORRECT)
      angle += ACOS(t);

   if (angle < TWOPI2D-CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
*********************************************************************** */
int query_point_poly_intersect(float *pt, float *normal, int count, vector4f *vertex) {

   float angle;
   int  i;
   float v0[3], v1[3], v2[3];

   if (abs(dotproduct3(pt, normal) + normal[3]) > CORRECT)
      return 0;

   copyarray3(v2, vertex[0]);
   subarray3(v2, pt);
   normalize3(v2);

   copyarray3(v0, v2);

   for (i=1, angle=0; i<count; i++) {
      copyarray3(v1, v2);
      copyarray3(v2, vertex[i]);
      subarray3(v2, pt);
      normalize3(v2);
      angle += ACOS(dotproduct3(v1, v2));
   }

   copyarray3(v1, v0);
   angle += ACOS(dotproduct3(v1, v2));

   if (angle < TWOPI2D-CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
********************************************************************** */
int query_point_plane_intersect(float *pt, float *n) {

   float t;

   line_plane_intersect(n, pt, n, &t);

   if (t < -CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
********************************************************************** */
int query_point_sphere_intersect(float *pt, float *center, float r) {

   float sum[3];

   copyarray3(sum, pt);
   subarray3(sum, center);

   if (magnitude3(sum) - r < CORRECT)
      return 1;

   return 0;
}


/* ***********************************************************************
*********************************************************************** */
int query_point_rect_intersect(float *pt, vector4f *n) {

   float t;

   line_plane_intersect(n[0], pt, n[0], &t);
   if (t < -CORRECT)
      return 0;

   line_plane_intersect(n[1], pt, n[1], &t);
   if (t < -CORRECT)
      return 0;

   line_plane_intersect(n[2], pt, n[2], &t);
   if (t < -CORRECT)
      return 0;

   line_plane_intersect(n[3], pt, n[3], &t);
   if (t < -CORRECT)
      return 0;

   line_plane_intersect(n[4], pt, n[4], &t);
   if (t < -CORRECT)
      return 0;

   line_plane_intersect(n[5], pt, n[5], &t);
   if (t < -CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
********************************************************************** */
int query_sphere_sphere_intersect(float *c1, float r1, float *c2, float r2) {

   float sum[3];

   copyarray3(sum, c1);
   subarray3(sum, c2);

   if (magnitude3(sum) - (r1+r2) < CORRECT)
      return 1;

   return 0;
}


/* ***********************************************************************
*********************************************************************** */
int query_sphere_plane_intersect(float *c, float r, float *n) {

   float t;

   line_plane_intersect(n, c, n, &t);

   if (t + r > -CORRECT)
      return 1;

   return 0;
}


/* ***********************************************************************
*********************************************************************** */
int query_sphere_rect_intersect(float *c, float r, vector4f *n) {

   float t;

   line_plane_intersect(n[0], c, n[0], &t);
   if (abs(t) < CORRECT)
      return 1;
   if (t + r <= -CORRECT)
      return 0;

   line_plane_intersect(n[1], c, n[1], &t);
   if (abs(t) < CORRECT)
      return 1;
   if (t + r <= -CORRECT)
      return 0;

   line_plane_intersect(n[2], c, n[2], &t);
   if (abs(t) < CORRECT)
      return 1;
   if (t + r <= -CORRECT)
      return 0;

   line_plane_intersect(n[3], c, n[3], &t);
   if (abs(t) < CORRECT)
      return 1;
   if (t + r <= -CORRECT)
      return 0;

   line_plane_intersect(n[4], c, n[4], &t);
   if (abs(t) < CORRECT)
      return 1;
   if (t + r <= -CORRECT)
      return 0;

   line_plane_intersect(n[5], c, n[5], &t);
   if (t + r <= -CORRECT)
      return 0;

   return 1;
}

/* **********************************************************************
   Theory - if a plane intersects a rectangle, then it must intersect at
   LEAST 3 edges, therefore we can stop if we find one intersection.
   order of vertices - 1-4 one side, 5-8 opposite side where 1,5 match,
   2,6 match, etc. - 12 sets of compares

   new theory - if a plane intersects a rectangle, it must intersect with
   at least1 of the 4 cross-rectangle diagonals -
   only need to check 3 (2 creates a plane, 3 creates a space). (1,7 2,8 3,5)
********************************************************************** */
int query_plane_rect_intersect(float *n, vector4f *p) {

   float v[3];
   float t;

   copyarray3(v, p[6]);
   subarray3(v, p[0]);

   if (line_plane_intersect(n, p[0], v, &t)) {

      if (t < 0 || t > magnitude3(v))
         return 0;
      return 1;
   }

   copyarray3(v, p[7]);
   subarray3(v, p[1]);

   if (line_plane_intersect(n, p[1], v, &t)) {

      if (t < 0 || t > magnitude3(v))
         return 0;
      return 1;
   }

   copyarray3(v, p[4]);
   subarray3(v, p[2]);

   if (line_plane_intersect(n, p[2], v, &t)) {

      if (t < 0 || t > magnitude3(v))
         return 0;
      return 1;
   }

   return 0;  // theoretically, the program should never get here!!!
}


/* **********************************************************************
Cube Theory - if 2 cubes intersect each other, then at least one corner
of one cube is within another cube

Rectangle Theory 1: if 2 rectangles intersect, then there exists at
least 1 plane from each rectangle st the intersection of the planes
intersects both polygons that represent those planes.

Rectangle Theory 2: if bounding spheres that bound the rectangles
do not intersect, the rectangles do not intersect.

Rectangle Theory 3: if a corner of a rectangle is w/in the other,
there is an intersection

Rectangle Theory 4: if 2 rectangles intersect, then the edge of one
must intersect the edge of another unless one completely encloses the
other.
********************************************************************** */
int query_rect_rect(vector4f *n1, vector4f *p1, vector4f *n2, vector4f *p2) {

   float bsr[2];
   float center[2][4];
   float tr, dist;
   float t[4];
   int i;

   copyarray4(center[0], p1[0]);             // if bs's dont intersect,
   addarray3(center[0], p1[6]);              //   rect's dont intersect
   smultarray3(center[0], 0.5);

   copyarray3(t, center[0]);
   subarray3(t, p1[0]);
   bsr[0] = magnitude3(t);

   copyarray4(center[1], p2[0]);
   addarray3(center[1], p2[6]);
   smultarray3(center[1], 0.5);

   copyarray3(t, center[1]);
   subarray3(t, p2[0]);
   bsr[1] = magnitude3(t);

   if (!query_sphere_rect_intersect(center[0], bsr[0], n2))
      return 0;
                                        // now see if a vertex(pt) of one
                                        // intersects the other
   for (i=0; i<8; i++) {
      if (query_point_rect_intersect(p1[i], n2))
         return 1;
      if (query_point_rect_intersect(p2[i], n1))
         return 1;
   }
                                        // now check to see if edge of one
                                        // intersects the other
   for (i=0; i<3; i++) {
      copyarray3(t, p1[i+1]);
      subarray3(t, p1[i]);
      tr = magnitude3(t);

      if (line_rect_intersect(p1[i], t, p2, n2, &dist))
         if (dist <= tr)
            return 1;

      copyarray3(t, p2[i+1]);
      subarray3(t, p2[i]);
      tr = magnitude3(t);

      if (line_rect_intersect(p2[i], t, p1, n1, &dist))
         if (dist <= tr)
            return 1;

      copyarray3(t, p1[i+5]);
      subarray3(t, p1[i+4]);
      tr = magnitude3(t);

      if (line_rect_intersect(p1[i+4], t, p2, n2, &dist))
         if (dist <= tr)
            return 1;

      copyarray3(t, p2[i+5]);
      subarray3(t, p2[i+4]);
      tr = magnitude3(t);

      if (line_rect_intersect(p2[i+4], t, p1, n1, &dist))
         if (dist <= tr)
            return 1;
   }

   copyarray3(t, p1[0]);
   subarray3(t, p1[3]);
   tr = magnitude3(t);

   if (line_rect_intersect(p1[3], t, p2, n2, &dist))
      if (dist <= tr)
         return 1;

   copyarray3(t, p2[0]);
   subarray3(t, p2[3]);
   tr = magnitude3(t);

   if (line_rect_intersect(p2[3], t, p1, n1, &dist))
      if (dist <= tr)
         return 1;

   copyarray3(t, p1[4]);
   subarray3(t, p1[7]);
   tr = magnitude3(t);

   if (line_rect_intersect(p1[7], t, p2, n2, &dist))
      if (dist <= tr)
         return 1;

   copyarray3(t, p2[4]);
   subarray3(t, p2[7]);
   tr = magnitude3(t);

   if (line_rect_intersect(p2[7], t, p1, n1, &dist))
      if (dist <= tr)
         return 1;

   for (i=0; i<4; i++) {
      copyarray3(t, p1[i+4]);
      subarray3(t, p1[i]);
      tr = magnitude3(t);

      if (line_rect_intersect(p1[i], t, p2, n2, &dist))
         if (dist <= tr)
            return 1;

      copyarray3(t, p2[i+4]);
      subarray3(t, p2[i]);
      tr = magnitude3(t);

      if (line_rect_intersect(p2[i], t, p1, n1, &dist))
         if (dist <= tr)
            return 1;
   }

   return 0;
}


/* **********************************************************************
*********************************************************************** */
int line2d_line2d_intersect(float *v1, float *v2, float *pt) {

   float temp[3];

   if (abs(v1[0]) < CORRECT) {
      if (abs(v1[1]) < CORRECT || abs(v2[0]) < CORRECT)
         return 0;

      pt[1] = -v1[2]/v1[1];
      pt[0] = -(v2[1]*pt[1] + v2[2])/v2[0];

      return 1;
   }

   if (abs(v1[1]) < CORRECT) {
      if (abs(v1[0]) < CORRECT || abs(v2[1]) < CORRECT)
         return 0;

      pt[0] = -v1[2]/v1[0];
      pt[1] = -(v2[0]*pt[0] + v2[2])/v2[1];
      return 1;
   }

   if (abs(v2[0]) < CORRECT) {
      if (abs(v2[1]) < CORRECT || abs(v1[0]) < CORRECT)
         return 0;

      pt[1] = -v2[2]/v2[1];
      pt[0] = -(v1[1]*pt[1] + v1[2])/v1[0];
      return 1;
   }

   if (abs(v2[1]) < CORRECT) {
      if (abs(v2[0]) < CORRECT || abs(v1[1]) < CORRECT)
         return 0;

      pt[0] = -v2[2]/v2[0];
      pt[1] = -(v1[0]*pt[0] + v1[2])/v1[1];
      return 1;
   }

   temp[0] = v1[0] - (v2[0]*v1[1])/v2[1];

   if (abs(temp[0]) < CORRECT)
      return 0;

   temp[2] = v1[2] - (v2[2]*v1[1])/v2[1];

   pt[0] = -temp[2]/temp[0];
   pt[1] = -(v1[0]*pt[0] + v1[2])/v1[1];

   if (abs(v2[0]*pt[0] + v2[1]*pt[1] + v2[2]) > CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
   This function reflects an vector about a normal to a plane
*********************************************************************** */
void reflect(float normal[], float vector[]) {

   float mag;
   int    i;

   mag = abs(dotproduct3(normal, vector));

   for (i=0; i<3; i++)
      vector[i] += double(mag*normal[i]);

   normalize3(vector);
}


/* **********************************************************************
********************************************************************** */
int point_sphere_reflectn3vf(float *pt, float *center, float *normal) {

   copyarray3(normal, pt);
   subarray3(normal, center);

   if (normalize3(normal) < CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
********************************************************************** */
int sphere_sphere_reflectn3vf(float *c1, float *c2, float *normal) {

   copyarray3(normal, c1);
   subarray3(normal, c2);

   if (normalize3(normal) < CORRECT)
      return 0;

   return 1;
}


/* **********************************************************************
********************************************************************** */
int sphere_plane_reflectn3vf(float *n, float *normal) {

   copyarray3(normal, n);
   normalize3(normal);
   return 1;
}


/* **********************************************************************
          normal orders front, back left, top, right, bottom
          polygon orders : 0-3 front, 4-7 back, 0-4-5-1 left, etc
********************************************************************** */
int sphere_rect_reflectn3vf(float *c, vector4f *p, vector4f *n, float *normal) {

   float t, dt;
   float ct = -1;
   float pt[3];
   vector4f ptemp[4];

   line_plane_intersect(n[0], c, n[0], &t); // front face
   dt = abs(t);

   pt[0] = c[0] + t*n[0][0];
   pt[1] = c[1] + t*n[0][1];
   pt[2] = c[2] + t*n[0][2];

   if (query_point_poly_intersect(pt, n[0], 4, p)) { // if point in polygon
      ct = dt;
      copyarray3(normal, n[0]);
   }

   line_plane_intersect(n[1], c, n[1], &t); // back face
   dt = abs(t);

   if (ct < 0 || dt < ct) {
      pt[0] = c[0] + t*n[1][0];
      pt[1] = c[1] + t*n[1][1];
      pt[2] = c[2] + t*n[1][2];

      if (query_point_poly_intersect(pt, n[1], 4, &p[4])) { // if point in polygon
         ct = dt;
         copyarray3(normal, n[1]);
      }

   }

   line_plane_intersect(n[2], c, n[2], &t); // back face
   dt = abs(t);

   if (ct < 0 || dt < ct) {
      pt[0] = c[0] + t*n[2][0];
      pt[1] = c[1] + t*n[2][1];
      pt[2] = c[2] + t*n[2][2];

      copyarray4(ptemp[0], p[0]);
      copyarray4(ptemp[1], p[4]);
      copyarray4(ptemp[2], p[5]);
      copyarray4(ptemp[3], p[1]);

      if (query_point_poly_intersect(pt, n[2], 4, ptemp)) { // if point in polygon
         ct = dt;
         copyarray3(normal, n[2]);
      }

   }

   line_plane_intersect(n[3], c, n[3], &t); // back face
   dt = abs(t);

   if (ct < 0 || dt < ct) {
      pt[0] = c[0] + t*n[3][0];
      pt[1] = c[1] + t*n[3][1];
      pt[2] = c[2] + t*n[3][2];

      copyarray4(ptemp[0], p[1]);
      copyarray4(ptemp[1], p[5]);
      copyarray4(ptemp[2], p[6]);
      copyarray4(ptemp[3], p[2]);

      if (query_point_poly_intersect(pt, n[3], 4, ptemp)) { // if point in polygon
         ct = dt;
         copyarray3(normal, n[3]);
      }

   }

   line_plane_intersect(n[4], c, n[4], &t); // back face
   dt = abs(t);

   if (ct < 0 || dt < ct) {
      pt[0] = c[0] + t*n[4][0];
      pt[1] = c[1] + t*n[4][1];
      pt[2] = c[2] + t*n[4][2];

      copyarray4(ptemp[0], p[2]);
      copyarray4(ptemp[1], p[6]);
      copyarray4(ptemp[2], p[7]);
      copyarray4(ptemp[3], p[3]);

      if (query_point_poly_intersect(pt, n[4], 4, ptemp)) { // if point in polygon
         ct = dt;
         copyarray3(normal, n[4]);
      }

   }

   line_plane_intersect(n[5], c, n[5], &t); // back face
   dt = abs(t);

   if (ct < 0 || dt < ct) {
      pt[0] = c[0] + t*n[5][0];
      pt[1] = c[1] + t*n[5][1];
      pt[2] = c[2] + t*n[5][2];

      copyarray4(ptemp[0], p[3]);
      copyarray4(ptemp[1], p[7]);
      copyarray4(ptemp[2], p[4]);
      copyarray4(ptemp[3], p[0]);

      if (query_point_poly_intersect(pt, n[5], 4, ptemp)) { // if point in polygon
         ct = dt;
         copyarray3(normal, n[5]);
      }

   }

   if (ct < 0)          // If here, then cant determine (most cases sphere is nowhere near rectangloid
      return 0;

   return 1;
}


/*
	NOTE:	if inside, return positive
*/

/* **********************************************************************
********************************************************************** */
int line_cylinder_intersect(vector4f pt, vector4f v, float length, float radius, float *t, int *side) {

   float a, b, c, check, t1, t2;
   float topt, bottomt;
   int tflag, bflag, t1flag, t2flag, iflag;
   vector4f plane;
   float rsquare;

   rsquare = SQR(radius);

   plane[0] = plane[2] = 0;
   plane[1] = 1;
   plane[3] = length;
   bflag = line_plane_intersect(plane, pt, v, &bottomt);

   if (bflag) {
      a = pt[0] + bottomt*v[0];
      b = pt[2] + bottomt*v[2];

      if (SQR(a) + SQR(b) > rsquare)
         bflag = 0;
   }

   plane[3] = -length;
   tflag = line_plane_intersect(plane, pt, v, &topt);

   if (tflag) {
      a = pt[0] + topt*v[0];
      b = pt[2] + topt*v[2];

      if (SQR(a) + SQR(b) > rsquare)
         tflag = 0;
   }

   if (tflag && bflag) {
      *t = topt;
      *side = CYL_TOP;

      if (topt < 0) {
         if (bottomt < 0) {
            if (topt < bottomt) {
               *t = bottomt;
               *side = CYL_BOTTOM;
            }

            return 1;
         }
 
         *t = bottomt; 		// signs are different
         *side = CYL_BOTTOM;
         return -1;
      }

      if (bottomt > 0) {
         if (topt > bottomt) {
            *t = bottomt;
            *side = CYL_BOTTOM;
         }

         return 1;
      }

      else			// signs are different
         return -1; 
   }

   b = c = pt[0];
   b *= v[0];
   c *= c;

   b += (pt[2]*v[2]);
   c += SQR(pt[2]) - SQR(radius);

   b += b;
   a = SQR(v[0]) + SQR(v[2]);

   check = SQR(b) - 4*a*c;

   if (check < 0) {
      *t = 0;
      return 0;
   }

   check = sqrt(check);
   a += a;

   t1 = (-b - check)/a;
   t2 = (-b + check)/a;

   a = pt[1] + t1*v[1];
   t1flag = (a < length && a > -length);

   a = pt[1] + t2*v[1];
   t2flag = (a < length && a > -length);

   if (!(t1flag || t2flag)) {
      *t = 0;
      return 0;
   }

   iflag = -1;

   *side = CYL_SIDE;

   if (t1flag) {
      iflag = (t1 > 0);
      *t = t1;
   }

   if (t2flag) {
      if (iflag > -1) {
         if ((t2>0) == iflag) {
            if (*t < 0) {
               if (t2 > *t)
                  *t = t2;
            }

            else
               if (t2 < *t)
                  *t = t2;

            return 1;
         }

         if (t2 > *t)	// signs are different
            *t = t2;

         return -1;
      }

      iflag = (t2 > 0);
      *t = t2;
   }

   if (tflag) {
      if ((topt>0) == iflag) {
         if (*t < 0) {
            if (topt > *t) {
               *side = CYL_TOP;
               *t = topt;
            }

         }

         else
            if (topt < *t) {
               *side = CYL_TOP;
               *t = topt;
            }

         return 1;
      }

      if (topt > *t) {	// different signs
         *side = CYL_TOP;
         *t = topt;
      }

      return -1;
   }

   if (bflag) {
      if ((bottomt>0) == iflag) {
         if (*t < 0) {
            if (bottomt > *t) {
               *side = CYL_BOTTOM;
               *t = bottomt;
            }

         }

         else
            if (bottomt < *t) {
               *side = CYL_BOTTOM;
               *t = bottomt;
            }

         return 1;
      }

      if (bottomt > *t) {		// different signs
         *side = CYL_BOTTOM;
         *t = bottomt;
      }

      return -1;
   }

   *t = 0;
   return 0;
}


/* **********************************************************************
   0 - top
   1 - bottom
********************************************************************** */
int line_cone_intersect(vector4f pt, vector4f v, float length, float *radius, float *t, int *side) {

   float a, b, c, d, e, f, check, t1, t2;
   float topt, bottomt;
   int tflag, bflag, t1flag, t2flag, iflag;
   vector4f plane;

   plane[0] = plane[2] = 0;
   plane[1] = 1;
   plane[3] = length;
   bflag = line_plane_intersect(plane, pt, v, &bottomt);

   if (bflag) {
      a = pt[0] + bottomt*v[0];
      b = pt[2] + bottomt*v[2];

      if (SQR(a) + SQR(b) > SQR(radius[1]))
         bflag = 0;
   }

   plane[3] = -length;
   tflag = line_plane_intersect(plane, pt, v, &topt);

   if (tflag) {
      a = pt[0] + topt*v[0];
      b = pt[2] + topt*v[2];

      if (SQR(a) + SQR(b) > (radius[0]))
         tflag = 0;
   }

   if (tflag && bflag) {
      *t = topt;
      *side = CYL_TOP;

      if (topt < 0) {
         if (bottomt < 0) {
            if (topt < bottomt) {
               *t = bottomt;
               *side = CYL_BOTTOM;
            }

            return 1;
         }

         *t = bottomt;          // signs are different
         *side = CYL_BOTTOM;
         return -1;
      }

      if (bottomt > 0) {
         if (topt > bottomt) {
            *t = bottomt;
            *side = CYL_BOTTOM;
         }

         return 1;
      }

      else                      // signs are different
         return -1;
   }

   d = (radius[0] - radius[1])/(2*length);
   e = 0.5*(radius[1] + radius[0]);
   f = SQR(d);

   check = d*e;
   c = f*pt[1];
   b = pt[0]*v[0] + pt[2]*v[2] - v[1]*(c + check);
   b += b; 
   c = SQR(pt[0]) + SQR(pt[2]) - (pt[1]*(c + check + check) + SQR(e));
   a = SQR(v[0]) + SQR(v[2]) - SQR(v[1])*f;

   check = SQR(b) - 4*a*c;

   if (check < 0) {
      *t = 0;
      return 0;
   }

   check = sqrt(check);
   a += a;

   t1 = (-b - check)/a;
   t2 = (-b + check)/a;

   a = pt[1] + t1*v[1];
   t1flag = (a < length && a > -length);

   a = pt[1] + t2*v[1];
   t2flag = (a < length && a > -length);

   if (!(t1flag || t2flag)) {
      *t = 0;
      return 0;
   }

   iflag = -1;

   *side = CYL_SIDE;

   if (t1flag) {
      iflag = (t1 > 0);
      *t = t1;
   }

   if (t2flag) {
      if (iflag > -1) {
         if ((t2>0) == iflag) {
            if (*t < 0) {
               if (t2 > *t)
                  *t = t2;
            }

            else
               if (t2 < *t)
                  *t = t2;

            return 1;
         }

         if (t2 > *t)   // signs are different
            *t = t2;

         return -1;
      }

      iflag = (t2 > 0);
      *t = t2;
   }

   if (tflag) {
      if ((topt>0) == iflag) {
         if (*t < 0) {
            if (topt > *t) {
               *side = CYL_TOP;
               *t = topt;
            }

         }

         else
            if (topt < *t) {
               *side = CYL_TOP;
               *t = topt;
            }

         return 1;
      }

      if (topt > *t) {  // different signs
         *side = CYL_TOP;
         *t = topt;
      }

      return -1;
   }

   if (bflag) {
      if ((bottomt>0) == iflag) {
         if (*t < 0) {
            if (bottomt > *t) {
               *side = CYL_BOTTOM;
               *t = bottomt;
            }

         }

         else
            if (bottomt < *t) {
               *side = CYL_BOTTOM;
               *t = bottomt;
            }

         return 1;
      }

      if (bottomt > *t) {               // different signs
         *side = CYL_BOTTOM;
         *t = bottomt;
      }

      return -1;
   }

   *t = 0;
   return 0;
}


