
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#include "cone.h"
#include <string.h>

#define COUNTVERTEX   20
		// COUNTVERTEX/2
#define COUNTSIDE     10
		// countside + 2
#define COUNTOBJECT   12
		// 1.0/countside
#define ICOUNTSIDE    0.1
		// countside*4
#define COUNTSIDEEDGE 40
		// countsideedge + 2*countside
#define COUNTEDGE     60

/* *************************************************************
************************************************************* */
int cone::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'l':
         if (!strcmp(token, "length")) {
            get_token(infile, token);
            length = atof(token) * 0.5;

            return 1;
         }

         break;

      case 't':
         if (!strcmp(token, "top_radius")) {
            get_token(infile, token);
            radius[0] = atof(token);

            return 1;
         }

         break;

      case 'b':
         if (!strcmp(token, "bottom_radius")) {
            get_token(infile, token);
            radius[1] = atof(token);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


/* *************************************************************
	assumption: no shear, circle base and apex
************************************************************* */
float cone::calc_area() {

   float r[2];
   float l, m;

   if (radius[0] < radius[1]) {
      copyarray2(r, radius)
   }

   else {
      r[0] = radius[1];
      r[1] = radius[0];
   }

   if (r[0] < CORRECT)
      return cone_area(r[1], length+length);

   l = length + length;
   m = r[1]*l/(r[1] - r[0]);

   return cone_area(r[1], m) - cone_area(r[0], m-l);
}


/* *************************************************************
************************************************************* */
void cone::pg(eye *parm, light *lparm, zbuffer *zbuff, int inc) {

   int  i, j;
   vector4f surface, pzero;
   vector4f temp;
   vector4f normal;
   vector4f pt, pt2;
   float t;
   int   side;
   int   flag;
   float rcolor[3], acolor[3];
   light *lmain;
   char  flag2t = mcinfo & CITEXTURE;
   char  flag3t = mcinfo & CI3DTEX;
   char  flagl  = mcinfo & CILUM;
   char  flagt  = flag2t || flag3t;
   char  flagtr = mcinfo & CITRANSPARENT;

   texcolortype texcolor;
   shaderparamtype sparam;

   float piinverse = 1.0/PI;
   float ratio;
   float tx, ty;
   float di, dj, wi, wj, wt;

   float area;

   if (flag2t)
      area = sqrt((tob->query_x()*tob->query_y())/(calc_area()*SQR(parm->query_scale())));

   ratio = 1.0/(length + length);

   normal[3] = 0;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype != POINT && lmain->ltype != FAR; lmain=lmain->next);

   texcolor = *base_color;

   if (lmain != (light *)NULL && !flagt)
      lmain->set_ambient(texcolor.color.ka, texcolor.ambient);

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;

   matvecmulto(iworld, pzero);

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   parm->parascreen2map(surface);

   dj = 1.0/parm->query_scale();
   di = dj*MAXSUBPROC;

   wt = surface[0];
   wi = surface[1];

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {
         surface[0] = wj;
         surface[1] = wi;
         surface[2] = -1;

         copyarray3(pt, surface);
         matvecmultv(iworld, pt);

         flag = line_cone_intersect(pzero, pt, length, radius, &t, &side);

         if (flag && (-t>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front)) {
            surface[0] *= t;
            surface[1] *= t;
            surface[2] = -t;

            switch (side) {

               case CYL_TOP:
                  normal[0] = normal[2] = 0;
                  if (flag < 0)
                     normal[1] = -1;
                  else
                     normal[1] = 1;

                  break;

               case CYL_BOTTOM:
                  normal[0] = normal[2] = 0;
                  if (flag < 0)
                     normal[1] = 1;
                  else
                     normal[1] = -1;
              
                  break;

               default:			// CYL_SIDE
                  if (abs(radius[0] - radius[1]) < CORRECT) {
                     if (flag < 0) {
                        normal[0] = wcenter[0] - surface[0];
                        normal[2] = wcenter[2] - surface[2];
                     }

                     else {
                        normal[0] = surface[0] - wcenter[0];
                       normal[2] = surface[2] - wcenter[2];
                     }

                     normal[1] = 0;
                     break;
                  }

                  pt2[0] = pzero[0] + pt[0]*t;		// x component
                  pt2[1] = pzero[2] + pt[2]*t;		// z component
                  normalize2(pt2);

                  temp[0] = pt2[0]*radius[1];	// bottom
                  temp[1] = -length;
                  temp[2] = pt2[1]*radius[1];

                  pt2[2] = pt2[1]*radius[0];	// top
                  pt2[1] = length;
                  pt2[0] = pt2[0]*radius[0];

                  subarray3(temp, pt2);		// "z" axis
//                  normalize3(temp);

                  pt2[0] = temp[2];
                  pt2[1] = 0;
                  pt2[2] = -temp[0];
//                  normalize3(pt2);		// "x" axis

                  if (radius[0] < radius[1])
                     if (flag < 0)
                        xproduct(normal, pt2, temp);	// -y = X x Z
                     else
                        xproduct(normal, temp, pt2);	// y = Z x X
                  else
                     if (flag < 0)
                        xproduct(normal, temp, pt2);	// -y = X x Z
                     else
                        xproduct(normal, pt2, temp);	// y = Z x X

                  break;
            }

            matvecmultv(world, normal);
            normalize3(normal);

            if (flagt)
               texcolor = *base_color;

            if (flag2t) {
               smultarray3(pt, t);
               addarray3(pt, pzero);

               switch(side) {

                  case CYL_TOP:
                     ty = 1;
                     break;

                  case CYL_BOTTOM:
                     ty = 0;
                     break;

                  default:
                     ty = (pt[1] + length)*ratio;
                     break;
               }

               normalize3(pt);

               temp[0] = pt[0];
               temp[1] = pt[2];
               normalize2(temp);
               if (temp[0] >= 0.0)
                  tx = ACOS(temp[1])*piinverse*0.5;
               else
                  tx = 1-ACOS(temp[1])*piinverse*0.5;

               query_texel(tx, ty, &texcolor, area*(-surface[2]));
               if (lmain != (light *)NULL)
                  lmain->set_ambient(texcolor.color.ka, texcolor.ambient);
            }

            if (flag3t)
               calc_color(surface, shaderlist->scale, normal, lmain, lparm, parm, &texcolor, flagtr, 1, flagl, rcolor, acolor);
            else
               calc_color(surface, 1, normal, lmain, lparm, parm, &texcolor, flagtr, 0, flagl, rcolor, acolor);

            if (flagtr) {
               if (zbuff->zbuff[i][j] > front) {
                  zbuff->zbuff[i][j] = -999999999;
                  copyarray3(zbuff->ybuff[i][j], rcolor);
                  copyarray3(zbuff->ambientbuff[i][j], acolor);
               }

               else {
                  if (rcolor[0] > zbuff->ybuff[i][j][0])
                     zbuff->ybuff[i][j][0] = rcolor[0];
                  if (rcolor[1] > zbuff->ybuff[i][j][1])
                     zbuff->ybuff[i][j][1] = rcolor[1];
                  if (rcolor[2] > zbuff->ybuff[i][j][2])
                     zbuff->ybuff[i][j][2] = rcolor[2];

                  if (acolor[0] > zbuff->ambientbuff[i][j][0])
                     zbuff->ambientbuff[i][j][0] = acolor[0];
                  if (acolor[1] > zbuff->ambientbuff[i][j][1])
                     zbuff->ambientbuff[i][j][1] = acolor[1];
                  if (acolor[2] > zbuff->ambientbuff[i][j][2])
                     zbuff->ambientbuff[i][j][2] = acolor[2];
               }

            }

            else {
               zbuff->zbuff[i][j] = surface[2];
               copyarray3(zbuff->ybuff[i][j], rcolor);
               copyarray3(zbuff->ambientbuff[i][j], acolor);
            }

         }

      }

}


void cone::lt(light *lparm, int inc) {

   int i, j;
   float surface[4], pzero[4];
   float t;
   int   side;
   int   flag;
   float wi, wj, wt, di, dj;

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   copyarray3(pzero, surface);

   lparm->parascreen2map(surface);

   pzero[0] += 1;
   pzero[1] += MAXSUBPROC;
   pzero[3] = 0;
   lparm->parascreen2map(pzero);

   dj = pzero[0] - surface[0];
   di = pzero[1] - surface[1];
   wt = surface[0];
   wi = surface[1];

   pzero[0] = pzero[1] = 0;

   if (lparm->ltype == BEAM_POINT) {
      pzero[2] = 0;
      pzero[3]   = 1;
      matvecmulto(iworld, pzero);
      surface[3] = 0;
   }

   else {
      pzero[2] = -1;
      pzero[3]   = 0;
      matvecmultv(iworld, pzero);
      surface[3] = 1;
   }

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {

         surface[0] = wj;
         surface[1] = wi;

         if (lparm->ltype == BEAM_POINT) {
            surface[2] = -1;
            matvecmultv(iworld, surface);
            flag = line_cone_intersect(pzero, surface, length, radius, &t, &side);
         }

         else {
            surface[2] = 0;
            matvecmulto(iworld, surface);
            flag = line_cone_intersect(surface, pzero, length, radius, &t, &side);
         }

         if (flag && (-t>lparm->lbuff.lbuff->zbuff.pdata[i][j] || lparm->lbuff.lbuff->zbuff.pdata[i][j] > front)) {
            lparm->lbuff.lbuff->zbuff.pdata[i][j] = -t;
            lparm->lbuff.lbuff->idbuff.pdata[i][j] = id;
         }

      }

}


void cone::transform(eye *parm) {

   float width;
   int i, j;

   copymx4x4(world, rotate);

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   matmatmulto(parm->transform, world);

   wcenter[0] = world[0][3];
   wcenter[1] = world[1][3];
   wcenter[2] = world[2][3];
   wcenter[3] = 1;

   length *= size;
   radius[0] *= size;
   radius[1] *= size;

   inversemx(world, iworld);

   if (radius[0] > radius[1])
      width = radius[0];
   else
      width = radius[1];

   bbox[0][0] = -width;
   bbox[0][1] = -length;
   bbox[0][2] = -width;
   bbox[0][3] = 1;

   bbox[1][0] = width;
   bbox[1][1] = length;
   bbox[1][2] = width;
   bbox[1][3] = 1;

   make_box(bbox, box);

   matvecmulto(world, box[0]);
   copyarray3(bbox[0], box[0]);
   copyarray3(bbox[1], box[0]);

   for (i=1; i<8; i++) {
      matvecmulto(world, box[i]);

      for (j=0; j<3; j++)
         if (box[i][j] < bbox[0][j])
            bbox[0][j] = box[i][j];
         else
            if (box[i][j] > bbox[1][j])
               bbox[1][j] = box[i][j];
   }

}


void cone::geo2poly(texpolygon **ptob, shaderlisttype *slist) {

   int i, j, k, l;
   float fi, fd;
   float x, z;
   vector4f temp, temp2;
   float tx, td, tx2;

   pob = new polytype;

   pob->build(COUNTVERTEX, COUNTOBJECT, COUNTEDGE, (mctype > BW && ptob != (texpolygon **)NULL ));

   if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
      *ptob = new texpolygon;
      (*ptob)->countobject = COUNTOBJECT;
      (*ptob)->data = new texptrtype[COUNTOBJECT];

      for (i=0; i<COUNTSIDE; i++) {
         (*ptob)->data[i] = new tex;
         (*ptob)->data[i]->mapfunc = COORD;
         (*ptob)->data[i]->ob = tob;
         (*ptob)->data[i]->pt = new vector2f[5];
      }

      (*ptob)->data[COUNTSIDE] = (tex *)NULL;		// top/bottom dont have 2d texture
      (*ptob)->data[COUNTSIDE+1] = (tex *)NULL;		//	(lazy I know :) ) - alternative is to have a "pt" at the center of top/bottom, and triangulate the top/bottom
   }

   if (mcinfo & CI3DTEX && slist != (shaderlisttype *)NULL) {
      slist->scount = COUNTOBJECT;
      slist->shade = new shadertype[COUNTOBJECT];
   }

   for (i=0, j=COUNTSIDE, fi=-HALFPI, fd = TWOPI*ICOUNTSIDE; i < COUNTSIDE; i++, j++, fi += fd) {
      x = COS(fi);		// could precalc cos/sin in a table (only 10 entries each)
      z = -SIN(fi);

      pob->mvertex[i][0] = x*radius[0];
      pob->mvertex[i][1] = length;
      pob->mvertex[i][2] = z*radius[0];
      pob->mvertex[i][3] = 1;

      pob->mvertex[j][0] = x*radius[1];
      pob->mvertex[j][1] = -length;
      pob->mvertex[j][2] = z*radius[1];
      pob->mvertex[j][3] = 1;

      if (abs(radius[0] - radius[1]) < CORRECT) {
         pob->gnormal[i][0] = x;
         pob->gnormal[i][1] = 0;
         pob->gnormal[i][2] = z;
      }

      else {
         copyarray3(temp, pob->mvertex[j]);
         subarray3(temp, pob->mvertex[i]);	// "z" axis

         temp2[0] = temp[2];
         temp2[1] = 0;
         temp2[2] = -temp[0];		// "x" axis

//         if (radius[0] < radius[1])
            xproduct(pob->gnormal[i], temp, temp2);	// y = Z x X
//         else
//            xproduct(pob->gnormal[i], temp2, temp);	// y = X x Z
      }

      matvecmulto(world, pob->mvertex[i]);
      matvecmulto(world, pob->mvertex[j]);
      matvecmultv(world, pob->gnormal[i]);
      normalize3(pob->gnormal[i]);
      copyarray3(pob->gnormal[j], pob->gnormal[i]);
   }

   for (i=j=0, tx=0, tx2=td=texcoord[2][0]*ICOUNTSIDE, k=COUNTSIDE-1, l=COUNTSIDE; i<k; i++, l++, tx=tx2, tx2+=td) {
      pob->polynum[i] = 4;
      pob->edgeptr[i] = &pob->object[j];

      if (radius[1] > radius[0]) {

         pob->object[j] = l+1;
         j++;
         pob->object[j] = l;
         j++;
         pob->object[j] = i;
         j++;
         pob->object[j] = i+1;

         if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
            (*ptob)->data[i]->pt[0][0] = tx2;
            (*ptob)->data[i]->pt[0][1] = 0; 
            (*ptob)->data[i]->pt[1][0] = tx;
            (*ptob)->data[i]->pt[1][1] = 0;
            (*ptob)->data[i]->pt[2][0] = tx;
            (*ptob)->data[i]->pt[2][1] = texcoord[2][1]; 
            (*ptob)->data[i]->pt[3][0] = tx2;
            (*ptob)->data[i]->pt[3][1] = texcoord[2][1]; 
            (*ptob)->data[i]->pt[4][0] = tx2;
            (*ptob)->data[i]->pt[4][1] = 0;
         }

      }

      else {
         pob->object[j] = i;
         j++;
         pob->object[j] = i+1;
         j++;
         pob->object[j] = l+1;
         j++;
         pob->object[j] = l;

         if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
            (*ptob)->data[i]->pt[0][0] = tx;
            (*ptob)->data[i]->pt[0][1] = texcoord[2][1];
            (*ptob)->data[i]->pt[1][0] = tx2;
            (*ptob)->data[i]->pt[1][1] = texcoord[2][1];
            (*ptob)->data[i]->pt[2][0] = tx2;
            (*ptob)->data[i]->pt[2][1] = 0; 
            (*ptob)->data[i]->pt[3][0] = tx;
            (*ptob)->data[i]->pt[3][1] = 0; 
            (*ptob)->data[i]->pt[4][0] = tx;
            (*ptob)->data[i]->pt[4][1] = texcoord[2][1]; 
         }

      }

      j++;
   }

   pob->polynum[k] = 4;
   pob->edgeptr[k] = &pob->object[COUNTSIDEEDGE-4];
   l = COUNTSIDE-1;
   if (radius[1] > radius[0]) {
      pob->object[COUNTSIDEEDGE-4] = COUNTSIDE;
      pob->object[COUNTSIDEEDGE-3] = COUNTVERTEX-1;
      pob->object[COUNTSIDEEDGE-2] = k;
      pob->object[COUNTSIDEEDGE-1] = 0;

      if (ptob != (texpolygon **)NULL && mcinfo & CITEXTURE) {
         (*ptob)->data[l]->pt[0][0] = tx2;
         (*ptob)->data[l]->pt[0][1] = 0; 
         (*ptob)->data[l]->pt[1][0] = tx;
         (*ptob)->data[l]->pt[1][1] = 0;
         (*ptob)->data[l]->pt[2][0] = tx;
         (*ptob)->data[l]->pt[2][1] = texcoord[2][1]; 
         (*ptob)->data[l]->pt[3][0] = tx2;
         (*ptob)->data[l]->pt[3][1] = texcoord[2][1]; 
         (*ptob)->data[l]->pt[4][0] = tx2;
         (*ptob)->data[l]->pt[4][1] = 0;
      }

   }

   else {
      pob->object[COUNTSIDEEDGE-4] = COUNTSIDE-1;
      pob->object[COUNTSIDEEDGE-3] = 0;
      pob->object[COUNTSIDEEDGE-2] = COUNTSIDE;
      pob->object[COUNTSIDEEDGE-1] = COUNTVERTEX-1;

      if (ptob != (texpolygon **)NULL && mcinfo & CITEXTURE) {
         (*ptob)->data[l]->pt[0][0] = tx;
         (*ptob)->data[l]->pt[0][1] = texcoord[2][1];
         (*ptob)->data[l]->pt[1][0] = tx2;
         (*ptob)->data[l]->pt[1][1] = texcoord[2][1];
         (*ptob)->data[l]->pt[2][0] = tx2;
         (*ptob)->data[l]->pt[2][1] = 0; 
         (*ptob)->data[l]->pt[3][0] = tx;
         (*ptob)->data[l]->pt[3][1] = 0; 
         (*ptob)->data[l]->pt[4][0] = tx;
         (*ptob)->data[l]->pt[4][1] = texcoord[2][1]; 
      }

   }

   pob->polynum[COUNTSIDE] = COUNTSIDE;
   pob->edgeptr[COUNTSIDE] = &pob->object[COUNTSIDEEDGE];
   pob->polynum[COUNTSIDE+1] = COUNTSIDE;
   pob->edgeptr[COUNTSIDE+1] = &pob->object[COUNTSIDEEDGE+COUNTSIDE];

   for (i=COUNTSIDE-1, l=COUNTSIDEEDGE; i>-1; i--, l++)
      pob->object[l] = i;

   for (i=COUNTSIDE, j=COUNTSIDE<<1; i<j; i++, l++)
      pob->object[l] = i;

   if (ptob != (texpolygon **)NULL) {	// NULL signifies NO coloring
      if (mctype > BW)
         for (i=0; i<COUNTOBJECT; i++)
            pob->shade[i] = base_color;

      if (mcinfo & CI3DTEX) {
         for (i=0; i<COUNTOBJECT; i++)
            slist->shade[i] = shaderlist->shade[0];

         slist->scale = shaderlist->scale;
      }

   }

}
