
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */



// read in # of processors from io file


#if !defined __MASTER_H
#define __MASTER_H

#include "base.h"

#define CONTROLKEY 0xabcdef
#define MAXSEMAPHORE 2
#define PERMISSIONS 438		// rw permissions in shared mem

#include <sys/ipc.h>
#include <sys/sem.h>


typedef enum {
   LOCK_RESULT = 0,
   LOCK_QUEUE = 1
} locktype;

typedef struct ptrstruct {
   int    id;
   struct ptrstruct *next;
} ptrtype;

extern int MAXSUBPROC;
extern int MAXPROCESSOR;

class subslave {

   public:
      int          id;
      int          child_id;
      volatile int busy;
      volatile int active;
      slave        *control;
      volatile char bye;
      subslave     *next;

      grandchildtype *data;

           subslave() {

              busy   = active = bye = 0;
           }

      int  activate();
      int  spawn(grandchildtype *param);
      int  query_busy();
};


void create_subchild(void *owner);

class master;

class slave {

   public:
      camera       *ctr;
      light        *ltr;
      listtype     *ptr;
      FILE         *outfile;
      mapul        *mtr;
      int          doneflag;

      volatile int active;
      zbuffer      *zbuff;

      int       id;
      int       child_id;
      master    *control;
      volatile char bye;
      slave     *next;
      volatile int busy;
      volatile int done;

      subslave  *grandchild;

           slave();
           ~slave();
      int  activate();
      int  spawn(camera *cam, light *lit, FILE *io, listtype *mainlist, int frm, mapul *mapbuffer);

      int  query_busy();
};

void render_object(slave *sslave);
void create_child(void *owner);


class master {

   public:
      int     semid;
      slave   *circuit;
      slave   *freeslave;
      slave   *activeslave;
      lightbufftype *lbuff;

      ptrtype *result;

            master();
            ~master();
      void  done();
      void  activate();
      int   query_free();
      slave *get_slave();
      int   free_slave(int x);

      int   pop_result();
      int   push_result(int id);

      int   stack_lbuff(light *ltr);

      int   get_lbuff(light *ltr);

      void  kill_children();

      void set_lock(int semnum) {

         struct sembuf semstruct;

         semstruct.sem_op = -1;
         semstruct.sem_flg = 0;
         semstruct.sem_num = semnum;
         semop(semid, &semstruct, 1);
      }

      void free_lock(int semnum) {

         struct sembuf semstruct;

         semstruct.sem_op = 1;
         semstruct.sem_flg = 0;
         semstruct.sem_num = semnum;
         semop(semid, &semstruct, 1);
      }

};

#endif

