
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#include "cylinder.h"
#include <string.h>

#define COUNTVERTEX   20
		// COUNTVERTEX/2
#define COUNTSIDE     10
		// countside + 2
#define COUNTOBJECT   12
		// 1.0/countside
#define ICOUNTSIDE    0.1
		// countside*4
#define COUNTSIDEEDGE 40
		// countsideedge + 2*countside
#define COUNTEDGE     60


/* *************************************************************
************************************************************* */
int cylinder::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'l':
         if (!strcmp(token, "length")) {
            get_token(infile, token);
            length = atof(token) * 0.5;

            return 1;
         }

         break;

      case 'r':
         if (!strcmp(token, "radius")) {
            get_token(infile, token);
            radius = atof(token);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


/* *************************************************************
************************************************************* */
float cylinder::calc_area() {

   return cylinder_area(radius, length+length);
}


/* *************************************************************
************************************************************* */
void cylinder::pg(eye *parm, light *lparm, zbuffer *zbuff, int inc) {

   int  i, j;
   float surface[4], pzero[4];
   float temp[2];
   vector4f normal;
   float pt[4];
   float t;
   int   side;
   int   flag;
   float rcolor[3], acolor[3];
   light *lmain;
   char  flag2t = mcinfo & CITEXTURE;
   char  flag3t = mcinfo & CI3DTEX;
   char  flagl  = mcinfo & CILUM;
   char  flagt  = flag2t || flag3t;
   char  flagtr = mcinfo & CITRANSPARENT;

   texcolortype texcolor;
   shaderparamtype sparam;

   float piinverse = 1.0/PI;
   float ratio;
   float tx, ty;
   float di, dj, wi, wj, wt;

   float area;

   if (flag2t) {
      area = parm->query_scale();
      area = sqrt((tob->query_x()*tob->query_y())/(calc_area()*SQR(area)));
   }

   ratio = 1.0/(length + length);

   normal[3] = 0;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype != POINT && lmain->ltype != FAR; lmain=lmain->next);

   texcolor = *base_color;

   if (lmain != (light *)NULL && !flagt)
      lmain->set_ambient(texcolor.color.ka, texcolor.ambient);

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;

   matvecmulto(iworld, pzero);

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   parm->parascreen2map(surface);

   dj =  1.0/parm->query_scale();
   di = dj*MAXSUBPROC;

   wt = surface[0];
   wi = surface[1];

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {
         surface[0] = wj;
         surface[1] = wi;
         surface[2] = -1;

         copyarray3(pt, surface);
         matvecmultv(iworld, pt);

         flag = line_cylinder_intersect(pzero, pt, length, radius, &t, &side);

         if (flag && (-t>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front)) {
            surface[0] *= t;
            surface[1] *= t;
            surface[2] = -t;

            switch (side) {

               case CYL_TOP:
                  normal[0] = normal[2] = 0;
                  normal[1] = 1;

                  break;

               case CYL_BOTTOM:
                  normal[0] = normal[2] = 0;
                  normal[1] = -1;
              
                  break;

               default:			// CYL_SIDE
                  normal[0] = pzero[0] + t*pt[0];
                  normal[2] = pzero[2] + t*pt[2];
                  normal[1] = 0;
                  break;
            }

            matvecmultv(world, normal);
            normalize3(normal);

            if (flagt)
               texcolor = *base_color;

            if (flag2t) {
               smultarray3(pt, t);
               addarray3(pt, pzero);

               switch(side) {

                  case CYL_TOP:
                     ty = 1;
                     break;

                  case CYL_BOTTOM:
                     ty = 0;
                     break;

                  default:
                     ty = (pt[1] + length)*ratio;
                     break;
               }

               normalize3(pt);

               temp[0] = pt[0];
               temp[1] = pt[2];
               normalize2(temp);
               if (temp[0] >= 0.0)
                  tx = ACOS(temp[1])*piinverse*0.5;
               else
                  tx = 1-ACOS(temp[1])*piinverse*0.5;

               query_texel(tx, ty, &texcolor, area*(-surface[2]));

               if (lmain != (light *)NULL)
                  lmain->set_ambient(texcolor.color.ka, texcolor.ambient);
            }

            if (flag3t)
               calc_color(surface, shaderlist->scale, normal, lmain, lparm, parm, &texcolor, flagtr, 1, flagl, rcolor, acolor);
            else
               calc_color(surface, 1, normal, lmain, lparm, parm, &texcolor, flagtr, 0, flagl, rcolor, acolor);

            if (flagtr) {
               if (zbuff->zbuff[i][j] > front) {
                  zbuff->zbuff[i][j] = -999999999;
                  copyarray3(zbuff->ybuff[i][j], rcolor);
                  copyarray3(zbuff->ambientbuff[i][j], acolor);
               }

               else {
                  if (rcolor[0] > zbuff->ybuff[i][j][0])
                     zbuff->ybuff[i][j][0] = rcolor[0];
                  if (rcolor[1] > zbuff->ybuff[i][j][1])
                     zbuff->ybuff[i][j][1] = rcolor[1];
                  if (rcolor[2] > zbuff->ybuff[i][j][2])
                     zbuff->ybuff[i][j][2] = rcolor[2];

                  if (acolor[0] > zbuff->ambientbuff[i][j][0])
                     zbuff->ambientbuff[i][j][0] = acolor[0];
                  if (acolor[1] > zbuff->ambientbuff[i][j][1])
                     zbuff->ambientbuff[i][j][1] = acolor[1];
                  if (acolor[2] > zbuff->ambientbuff[i][j][2])
                     zbuff->ambientbuff[i][j][2] = acolor[2];
               }

            }

            else {
               zbuff->zbuff[i][j] = surface[2];
               copyarray3(zbuff->ybuff[i][j], rcolor);
               copyarray3(zbuff->ambientbuff[i][j], acolor);
            }

         }

      }

}


void cylinder::lt(light *lparm, int inc) {

   int i, j;
   float surface[4], pzero[4];
   float t;
   int   side;
   int   flag;
   float wi, wj, wt, di, dj;

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   copyarray3(pzero, surface);

   lparm->parascreen2map(surface);

   pzero[0] += 1;
   pzero[1] += MAXSUBPROC;
   pzero[3] = 0;
   lparm->parascreen2map(pzero);

   dj = pzero[0] - surface[0];
   di = pzero[1] - surface[1];
   wt = surface[0];
   wi = surface[1];

   pzero[0] = pzero[1] = 0;

   if (lparm->ltype == BEAM_POINT) {
      pzero[2] = 0;
      pzero[3]   = 1;
      matvecmulto(iworld, pzero);
      surface[3] = 0;
   }

   else {
      pzero[2] = -1;
      pzero[3]   = 0;
      matvecmultv(iworld, pzero);
      surface[3] = 1;
   }

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {

         surface[0] = wj;
         surface[1] = wi;

         if (lparm->ltype == BEAM_POINT) {
            surface[2] = -1;
            matvecmultv(iworld, surface);
            flag = line_cylinder_intersect(pzero, surface, length, radius, &t, &side);
         }

         else {
            surface[2] = 0;
            matvecmulto(iworld, surface);
            flag = line_cylinder_intersect(surface, pzero, length, radius, &t, &side);
         }

         if (flag && (-t>lparm->lbuff.lbuff->zbuff.pdata[i][j] || lparm->lbuff.lbuff->zbuff.pdata[i][j] > front)) {
            lparm->lbuff.lbuff->zbuff.pdata[i][j] = -t;
            lparm->lbuff.lbuff->idbuff.pdata[i][j] = id;
         }

      }

}


void cylinder::transform(eye *parm) {

   int i, j;

   copymx4x4(world, rotate);

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   matmatmulto(parm->transform, world);

   wcenter[0] = world[0][3];
   wcenter[1] = world[1][3];
   wcenter[2] = world[2][3];
   wcenter[3] = 1;

   length *= size;
   radius *= size;

   inversemx(world, iworld);

   bbox[0][0] = -radius;
   bbox[0][1] = -length;
   bbox[0][2] = -radius;
   bbox[0][3] = 1;

   bbox[1][0] = radius;
   bbox[1][1] = length;
   bbox[1][2] = radius;
   bbox[1][3] = 1;

   make_box(bbox, box);

   matvecmulto(world, box[0]);
   copyarray3(bbox[0], box[0]);
   copyarray3(bbox[1], box[0]);

   for (i=1; i<8; i++) {
      matvecmulto(world, box[i]);

      for (j=0; j<3; j++)
         if (box[i][j] < bbox[0][j])
            bbox[0][j] = box[i][j];
         else
            if (box[i][j] > bbox[1][j])
               bbox[1][j] = box[i][j];
   }


}


void cylinder::geo2poly(texpolygon **ptob, shaderlisttype *slist) {

   int i, j, k, l;
   float fi, fd;
   float x, z;
   float tx, td, tx2;

   pob = new polytype;

   pob->build(COUNTVERTEX, COUNTOBJECT, COUNTEDGE, (mctype > BW && ptob != (texpolygon **)NULL ));

   if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
      *ptob = new texpolygon;
      (*ptob)->countobject = COUNTOBJECT;
      (*ptob)->data = new texptrtype[COUNTOBJECT];

      for (i=0; i<COUNTSIDE; i++) {
         (*ptob)->data[i] = new tex;
         (*ptob)->data[i]->mapfunc = COORD;
         (*ptob)->data[i]->ob = tob;
         (*ptob)->data[i]->pt = new vector2f[5];
      }

      (*ptob)->data[COUNTSIDE] = (tex *)NULL;		// top/bottom dont have 2d texture
      (*ptob)->data[COUNTSIDE+1] = (tex *)NULL;		//	(lazy I know :) ) - alternative is to have a "pt" at the center of top/bottom, and triangulate the top/bottom
   }

   if (mcinfo & CI3DTEX && slist != (shaderlisttype *)NULL) {
      slist->scount = COUNTOBJECT;
      slist->shade = new shadertype[COUNTOBJECT];
   }

   for (i=0, j=COUNTSIDE, fi=-HALFPI, fd = TWOPI*ICOUNTSIDE; i < COUNTSIDE; i++, j++, fi += fd) {
      x = COS(fi);		// could precalc cos/sin in a table (only 10 entries each)
      z = -SIN(fi);

      pob->mvertex[j][0] = pob->mvertex[i][0] = x*radius;
      pob->mvertex[j][1] = -(pob->mvertex[i][1] = length);
      pob->mvertex[j][2] = pob->mvertex[i][2] = z*radius;
      pob->mvertex[j][3] = pob->mvertex[i][3] = 1;

      pob->gnormal[i][0] = x;
      pob->gnormal[i][1] = 0;
      pob->gnormal[i][2] = z;

      matvecmulto(world, pob->mvertex[i]);
      matvecmulto(world, pob->mvertex[j]);
      matvecmultv(world, pob->gnormal[i]);
      normalize3(pob->gnormal[i]);
      copyarray3(pob->gnormal[j], pob->gnormal[i]);
   }

   for (i=j=0, tx=0, tx2=td=texcoord[2][0]*ICOUNTSIDE, k=COUNTSIDE-1, l=COUNTSIDE; i<k; i++, l++, tx=tx2, tx2+=td) {
      pob->polynum[i] = 4;
      pob->edgeptr[i] = &pob->object[j];

      pob->object[j] = l+1;
      j++;
      pob->object[j] = l;
      j++;
      pob->object[j] = i;
      j++;
      pob->object[j] = i+1;
      j++;

      if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
         (*ptob)->data[i]->pt[0][0] = tx2;
         (*ptob)->data[i]->pt[0][1] = 0; 
         (*ptob)->data[i]->pt[1][0] = tx;
         (*ptob)->data[i]->pt[1][1] = 0;
         (*ptob)->data[i]->pt[2][0] = tx;
         (*ptob)->data[i]->pt[2][1] = texcoord[2][1]; 
         (*ptob)->data[i]->pt[3][0] = tx2;
         (*ptob)->data[i]->pt[3][1] = texcoord[2][1]; 
         (*ptob)->data[i]->pt[4][0] = tx2;
         (*ptob)->data[i]->pt[4][1] = 0;
      }

   }

   pob->polynum[k] = 4;
   pob->edgeptr[k] = &pob->object[COUNTSIDEEDGE-4];
   l = COUNTSIDE-1;

   pob->object[COUNTSIDEEDGE-4] = COUNTSIDE;
   pob->object[COUNTSIDEEDGE-3] = COUNTVERTEX-1;
   pob->object[COUNTSIDEEDGE-2] = k;
   pob->object[COUNTSIDEEDGE-1] = 0;

   if (ptob != (texpolygon **)NULL && mcinfo & CITEXTURE) {
      (*ptob)->data[l]->pt[0][0] = tx2;
      (*ptob)->data[l]->pt[0][1] = 0; 
      (*ptob)->data[l]->pt[1][0] = tx;
      (*ptob)->data[l]->pt[1][1] = 0;
      (*ptob)->data[l]->pt[2][0] = tx;
      (*ptob)->data[l]->pt[2][1] = texcoord[2][1]; 
      (*ptob)->data[l]->pt[3][0] = tx2;
      (*ptob)->data[l]->pt[3][1] = texcoord[2][1]; 
      (*ptob)->data[l]->pt[4][0] = tx2;
      (*ptob)->data[l]->pt[4][1] = 0;
   }

   pob->polynum[COUNTSIDE] = COUNTSIDE;
   pob->edgeptr[COUNTSIDE] = &pob->object[COUNTSIDEEDGE];
   pob->polynum[COUNTSIDE+1] = COUNTSIDE;
   pob->edgeptr[COUNTSIDE+1] = &pob->object[COUNTSIDEEDGE+COUNTSIDE];

   for (i=COUNTSIDE-1, l=COUNTSIDEEDGE; i>-1; i--, l++)
      pob->object[l] = i;

   for (i=COUNTSIDE, j=COUNTSIDE<<1; i<j; i++, l++)
      pob->object[l] = i;

   if (ptob != (texpolygon **)NULL) {	// NULL signifies NO coloring
      if (mctype > BW)
         for (i=0; i<COUNTOBJECT; i++)
            pob->shade[i] = base_color;

      if (mcinfo & CI3DTEX) {
         for (i=0; i<COUNTOBJECT; i++)
            slist->shade[i] = shaderlist->shade[0];
         slist->scale = shaderlist->scale;
      }

   }

}

