
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#include "geo.h"
#include <string.h>

/* *************************************************************
************************************************************* */
int geo::read_data(char filename[], datatype **dptr) {

   return 0;
}


/* *************************************************************
************************************************************* */
int geo::bound_box(eye *parm) {

   int    i, j;
   float  extreme[2][2];
   float  d[2];

   if (bbox[0][2] > front)
      return 0;

   if (bbox[1][2] > front)
      return 1;

   for (i=0; i<8; i++)
      parm->map2screen(box[i]);

   copyarray2(extreme[0], box[0]);      // bounding box
   copyarray2(extreme[1], box[0]);

   for (i=1; i<8; i++)
      for (j=0; j<2; j++)
         if (extreme[0][j] > box[i][j])
            extreme[0][j] = box[i][j];
         else
            if (extreme[1][j] < box[i][j])
               extreme[1][j] = box[i][j];

   d[0] = extreme[1][0] - extreme[0][0];
   d[1] = extreme[1][1] - extreme[0][1];

   if (abs(d[0]) < MINBBOX  || abs(d[1]) < MINBBOX)
      return 0;

   if (abs(d[0]) < MINDIM  || abs(d[1]) < MINDIM) {
      if (mctype > BW)
         mctype = CONSTANT;

      mcinfo &= CIRAYCAST;

      setup_transfer(4,1);

      pob->countvertex = 4;
      pob->countobject = 1;
      pob->countedge   = 4;

      pob->mvertex[0][0] = extreme[0][0];
      pob->mvertex[0][1] = extreme[0][1];
      pob->mvertex[0][2] = bbox[1][2];
      pob->mvertex[0][3] = 1;
      parm->screen2map(pob->mvertex[0]);

      pob->mvertex[1][0] = extreme[0][0];
      pob->mvertex[1][1] = extreme[1][1];
      pob->mvertex[1][2] = bbox[1][2];
      pob->mvertex[1][3] = 1;
      parm->screen2map(pob->mvertex[1]);

      pob->mvertex[2][0] = extreme[1][0];
      pob->mvertex[2][1] = extreme[1][1];
      pob->mvertex[2][2] = bbox[1][2];
      pob->mvertex[2][3] = 1;
      parm->screen2map(pob->mvertex[2]);

      pob->mvertex[3][0] = extreme[1][0];
      pob->mvertex[3][1] = extreme[0][1];
      pob->mvertex[3][2] = bbox[1][2];
      pob->mvertex[3][3] = 1;
      parm->screen2map(pob->mvertex[3]);

      pob->object[0] = 0;
      pob->object[1] = 1;
      pob->object[2] = 2;
      pob->object[3] = 3;

      if (mctype == PALEX || (mctype > WFBW && mctype <= PHONG))
         for (i=0; i<4; i++) {
            pob->gnormal[i][0] = 0;
            pob->gnormal[i][1] = 0;
            pob->gnormal[i][2] = 1;
         }

      pob->polynum[0] = 4;

      pob->edgeptr[0] = pob->object;

      if (mctype == CONSTANT)
         for (i=0; i<pob->countobject; i++)
            pob->shade[i].color = base_color;

      return 2;
   }

   return 1;}


/* *************************************************************
************************************************************* */
int geo::read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) {

   for (tob=*tptr; tob!=(texbase *)NULL; tob=tob->next)
      if (!strcmp(filename, tob->filename))
         return 1;

   if (query_video(filename)) {
      tob = new texvid;
      tob->read_texture(filename, tptr);
   }

   else {
      tob = new texture;
      tob->read_texture(filename);
   }

   tob->query_coord(texcoord);

   tob->next = *tptr;
   *tptr = tob;
   return 1;
}


/* *************************************************************
************************************************************* */
int geo::read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) {

   int i;
   char buffer[MAXSTRLEN];
   FILE *infile;
   int count;

   for (shaderlist=*slptr; shaderlist!=(shaderlisttype *)NULL; shaderlist=shaderlist->next)
      if (!strcmp(filename, shaderlist->shname) && !strcmp(dfilename, shaderlist->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CI3DTEX;
      return 0;
   }

   shaderlist = new shaderlisttype;
   strcpy(shaderlist->shname, filename);
   strcpy(shaderlist->dataname, dfilename);
   shaderlist->init(1);
   shaderlist->next   = *slptr;
   *slptr       = shaderlist;

   fscanf(infile, "%s %f", buffer, &shaderlist->scale);
   lower_case(buffer);

   if (strcmp(buffer, "mono"))                 //ALL
      shaderlist->shade->read_data(infile, sptr);

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {
      fscanf(infile, "%d", &i);

      if (!(i-1)) 
         shaderlist->shade->read_data(infile, sptr);
   }

   fclose(infile);
   return 1;
}



unsigned int geo::query_texel(float tx, float ty, texcolortype *color, float d) {

   float a, b;
   unsigned int x;
   vector3f c;
   a = (tx - (int)tx)*texcoord[2][0];
   b = (ty - (int)ty)*texcoord[2][1];

   x = tob->query_texel(a, b, c, frame, 1, d);
   multarray3(color->color.ka, c);
   multarray3(color->color.kp, c);
   multarray3(color->color.ks, c);
   multarray3(color->color.lum, c);

   return x;
}


int geo::clip(eye *parm, int maxx, int maxy) {

   float corner[4];

   if (bbox[0][2] > front ||
       (bbox[1][2] <= front &&
        (-bbox[0][0]/bbox[1][2] > parm->vrc[1] ||     // outside boundaries
         -bbox[1][0]/bbox[1][2] < parm->vrc[0] ||
         -bbox[0][1]/bbox[1][2] > parm->vrc[3] ||
         -bbox[1][1]/bbox[1][2] < parm->vrc[2]))) {
      return 0;
   }

   copyarray3(corner, bbox[0]);
   corner[3] = 1;

   parm->map2screen(corner);

   row[0] = round(corner[1]);
   col[0] = round(corner[0]);

   corner[0] = bbox[1][0];
   corner[1] = bbox[1][1];
   corner[2] = bbox[0][2];

   parm->map2screen(corner);

   row[1] = round(corner[1]);
   col[1] = round(corner[0]);

   corner[0] = bbox[0][0];
   corner[1] = bbox[0][1];
   corner[2] = bbox[1][2];

   parm->map2screen(corner);

   if (row[0] > corner[1])
      row[0] = round(corner[1]);
   if (col[0] > corner[0])
      col[0] = round(corner[0]);

   copyarray3(corner, bbox[1]);

   parm->map2screen(corner);

   if (row[1] < corner[1])
      row[1] = round(corner[1]);
   if (col[1] < corner[0])
      col[1] = round(corner[0]);

   if (col[0] < 0)
      col[0] = 0;

   if (col[1] >= maxx)
      col[1] = (int)maxx-1;

   if (row[0] < 0)
      row[0] = 0;

   if (row[1] >= maxy)
      row[1] = (int)maxy-1;

   return 1;
}


int geo::beamclip(light *parm, int maxx, int maxy) {

   float corner[4];

   if (bbox[0][2] > front ||
       (parm->ltype == BEAM_POINT &&
        (-bbox[0][0]/bbox[1][2] > parm->vrc[1] ||     // outside boundaries
         -bbox[1][0]/bbox[1][2] < parm->vrc[0] ||
         -bbox[0][1]/bbox[1][2] > parm->vrc[3] ||
         -bbox[1][1]/bbox[1][2] < parm->vrc[2])) ||
       (parm->ltype == BEAM_FAR &&
        (bbox[0][0] > parm->vrc[1] ||     // outside boundaries
         bbox[1][0] < parm->vrc[0] ||
         bbox[0][1] > parm->vrc[3] ||
         bbox[1][1] < parm->vrc[2])))
      return 0;

   copyarray3(corner, bbox[0]);
   corner[3] = 1;

   if (parm->ltype == BEAM_POINT)
      parm->map2screen(corner);
   else
      parm->paramap2screen(corner);

   row[0] = round(corner[1]);
   col[0] = round(corner[0]);

   corner[0] = bbox[1][0];
   corner[1] = bbox[1][1];
   corner[2] = bbox[0][2];

   if (parm->ltype == BEAM_POINT)
      parm->map2screen(corner);
   else
      parm->paramap2screen(corner);

   row[1] = round(corner[1]);
   col[1] = round(corner[0]);

   corner[0] = bbox[0][0];
   corner[1] = bbox[0][1];
   corner[2] = bbox[1][2];

   if (parm->ltype == BEAM_POINT)
      parm->map2screen(corner);
   else
      parm->paramap2screen(corner);

   if (row[0] > corner[1])
      row[0] = round(corner[1]);
   if (col[0] > corner[0])
      col[0] = round(corner[0]);

   copyarray3(corner, bbox[1]);

   if (parm->ltype == BEAM_POINT)
      parm->map2screen(corner);
   else
      parm->paramap2screen(corner);

   if (row[1] < corner[1])
      row[1] = round(corner[1]);
   if (col[1] < corner[0])
      col[1] = round(corner[0]);

   if (col[0] < 0)
      col[0] = 0;

   if (col[1] >= maxx)
      col[1] = (int)maxx-1;

   if (row[0] < 0)
      row[0] = 0;

   if (row[1] >= maxy)
      row[1] = (int)maxy-1;

   return 1;
}


/* **************************************************
************************************************** */
void geo::colorize(grandchildtype *child) {

   if ((eye *)child->cam == (eye *)child->lit) {
      lt(child->lit, child->id);
      child->done = 1;
      return;
   }

   pg(child->cam, child->lit, child->zbuff, child->id);

   child->done = 1;
}


/* **************************************************
************************************************** */
void geo::datacopy() {

   if (lob != (shadelist *)NULL)
      base_color = &lob->base;
};


/* **************************************************
************************************************** */
void geo::beamscan(light *lparm, subslave grandchild[]) {

   int i, j;
   grandchildtype *offspring;
   polygon polyob;
   colortype  nctype;
   unsigned int ncinfo;

   if (mctype < BW)
      return;

   nctype = mctype;
   ncinfo = mcinfo;

   mcinfo &= ~CITEXTURE;

   if (mctype <= PHONG)
      mctype = BW;
   else
      mctype = PBW;

   transform(lparm);

   if (!(i=bound_box(lparm))) {
      mctype = nctype;
      mcinfo = ncinfo;

      return;
   }

   if (i==2) {
      polyob.setup_patch(pob, BW, mcinfo, id, 0, splane, bbox, base_color);
      polyob.beamscan(lparm, grandchild);
      polyob.end_scan();

      mctype = nctype;
      mcinfo = ncinfo;
      return;

   }

   if (!beamclip(lparm, lparm->lbuff.maxx, lparm->lbuff.maxy)) {
      mctype = nctype;
      mcinfo = ncinfo;

      return;
   }

   if (mctype < PALEX) {
      geo2poly((texpolygon **)NULL, (shaderlisttype *)NULL);
      polyob.setup_patch(pob, BW, mcinfo, id, 0, splane, bbox, base_color);
      polyob.beamscan(lparm, grandchild);
      polyob.end_scan();

      mctype = nctype;
      mcinfo = ncinfo;
      return;
   }

   j = MAXSUBPROC-1;

   offspring = new grandchildtype[MAXSUBPROC];

   for (i=0; i<j; i++) {                  // interlace lines
      offspring[i].cam = offspring[i].lit = lparm;
      offspring[i].source = this;
      offspring[i].id     = i;
      offspring[i].done   = 0;
      grandchild[i].spawn(&offspring[i]);
   }

   offspring[j].cam = offspring[j].lit = lparm;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;

   colorize(&offspring[j]);

   for (i=0; i<MAXSUBPROC; i++)
      while (!offspring[i].done);

   delete offspring;

   nctype = mctype;
   ncinfo = mcinfo;
}


void geo::scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) {

   int i, j;
   light *lmain;
   shadow    *str;
   grandchildtype *offspring;
   texpolygon *ptob;
   shaderlisttype slist;
   char texflag = mcinfo & CITEXTURE;
   char flag3d = mcinfo & CI3DTEX;

   polygon polyob;

   transform(cparm);

   if (!(i=bound_box(cparm)))
      return;

   if (i == 2) {
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);
      polyob.scan(cparm, lparm, zbuff, grandchild, shadptr);
      polyob.end_scan();

      return;
   }

   if (!clip(cparm, WINX, WINY))
      return;

   if (mctype < PALEX) {
      geo2poly(&ptob, &slist);
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);

      if (texflag)
         polyob.setup_texture(ptob);

      if (flag3d)
         polyob.setup_texture3d(&slist, world, iworld);

      polyob.scan(cparm, lparm, zbuff, grandchild, shadptr);

      polyob.end_scan();

      if (texflag)
         delete ptob;

      if (flag3d) { 			// causes seg fault w/ linux (pixcon -f a -x 320 -y 240, marble texture)
         delete slist.shade;
         slist.shade = (shadertype *)NULL;
      }

      return;
   }

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);
   if (sflag && lmain!=(light *)NULL && !NOSHADOW && !(mcinfo & CITRANSPARENT)) {
      geo2poly((texpolygon **)NULL, (shaderlisttype *)NULL);
      polyob.setup_patch(pob, ALEX, 0, id, 1, splane, bbox, base_color);

      str = new shadow;
      str->begin_scan();

      if (str->datacopy2(&polyob, cparm, lmain, 1)) {
         str->next = *shadptr;
         *shadptr = str;
      }

      else {
         str->end_scan();
         delete str;
      }

      polyob.end_scan();
   }

   j = MAXSUBPROC-1;

   offspring = new grandchildtype[MAXSUBPROC];

   for (i=0; i<j; i++) {                  // interlace lines
      offspring[i].cam    = cparm;
      offspring[i].lit    = lparm;
      offspring[i].zbuff  = zbuff;
      offspring[i].source = this;
      offspring[i].id     = i;
      offspring[i].done   = 0;
      grandchild[i].spawn(&offspring[i]);
   }

   offspring[j].cam    = cparm;
   offspring[j].lit    = lparm;
   offspring[j].zbuff  = zbuff;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;

   colorize(&offspring[j]);

   for (i=0; i<MAXSUBPROC; i++)
      while (!offspring[i].done);

   delete offspring;
}


void geo::scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

   polygon polyob;

   mctype = WFBW;
   mcinfo = 0;

   transform(cparm);

   if (!bound_box(cparm)) {
      polyob.setup_patch(pob, WFBW, 0, id, sflag, splane, bbox, base_color);
      polyob.scan(cparm, lparm, grandchild, shadptr);
      polyob.end_scan();

      return;
   }

   if (!clip(cparm, WINX, WINY))
      return;

   geo2poly((texpolygon **)NULL, (shaderlisttype *)NULL);
   polyob.setup_patch(pob, WFBW, 0, id, sflag, splane, bbox, base_color);
   polyob.scan(cparm, lparm, grandchild, shadptr);
   polyob.end_scan();
}

