
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */


#include <string.h>
#include "global.h"


/* *************************************************************
   This file contains all the global data vars for my programs
************************************************************* */


illtype illtype::operator = (shadetype *x) {

   color = x;
   ambient[0] = ambient[1] = ambient[2] = 0;

   return *this;
}


texcolortype texcolortype::operator = (illtype &x) {

   color = *x.color;
   copyarray3(ambient, x.ambient);

   return *this;
}


texcolortype texcolortype::operator = (shadetype &x) {

   color = x;
   ambient[0] = ambient[1] = ambient[2] = 0;

   return *this;
}


shadetype::shadetype() {

   ka[0] = ka[1] = ka[2] = 1.0;
   kp[0] = kp[1] = kp[2] = 0.7;
   ks[0] = ks[1] = ks[2] = 0.3;
   specn = 1;
   lum[0] = lum[1] = lum[2] = 0.0;
}


void shadetype::read_data(FILE *infile) {

   char junk[MAXSTRLEN];

   fscanf(infile, "%s %f %f %f", junk, &ka[0], &ka[1], &ka[2]);
   fscanf(infile, "%s %f %f %f", junk, &kp[0], &kp[1], &kp[2]);
   fscanf(infile, "%s %f %f %f", junk, &ks[0], &ks[1], &ks[2]);
   fscanf(infile, "%s %f", junk, &specn);
   fscanf(infile, "%s %f %f %f", junk, &lum[0], &lum[1], &lum[2]);
}


void shadetype::write_data(FILE *outfile) {

   fprintf(outfile, "KA %f %f %f\n", ka[0], ka[1], ka[2]);
   fprintf(outfile, "KP %f %f %f\n", kp[0], kp[1], kp[2]);
   fprintf(outfile, "KS %f %f %f\n", ks[0], ks[1], ks[2]);
   fprintf(outfile, "SPECN %f\n", specn);
   fprintf(outfile, "LUM %f %f %f\n\n", lum[0], lum[1], lum[2]);
}


int shadelist::read_data(char *filename) {

   FILE *infile;                        // file pointer
   char junk[MAXSTRLEN];
   int i;

   strcpy(shadename, filename);

   if (!(infile = fopen(filename, "r"))) {
      printf("Couldn't open %s\n", filename);
      return 0;
   }

   base.read_data(infile);

   fscanf(infile, "%s %d", junk, &count);

   if (list != (shadetype *)NULL)
      delete list;

   list = new shadetype[count];

   for (i=0; i<count; i++) {
      fscanf(infile, "%s %d", junk, &list[i].index);
      list[i].index--;
      list[i].read_data(infile);
   }

   fclose(infile);
   return 1;
}


int shadelist::write_data(char *filename) {

   FILE *outfile;                        // file pointer
   int i;

   if (!(outfile = fopen(filename, "w"))) {
      printf("Couldn't open %s for output\n", filename);
      return 0;
   }

   base.write_data(outfile);

   fprintf(outfile, "no_sides %d\n\n", count);

   for (i=0; i<count; i++) {
      fprintf(outfile, "# %d\n", i+1);
      list[i].write_data(outfile);
   }

   fclose(outfile);
   return 1;
}


bindex *bindex::find(int x) {

   if (x == id)
      return this;

   if (x < id) {
      if (back == (bindex *)NULL)
         return (bindex *)NULL;

      return back->find(x);
   }

   if (next == (bindex *)NULL)
      return (bindex *)NULL;

   return next->find(x);
}


void bindex::insert(bindex *pbtr) {

   if (pbtr->id < id) {
      if (back == (bindex *)NULL) {
         back = pbtr;
         return;
      }

      back->insert(pbtr);
      return;
   }

   if (next == (bindex *)NULL) {
      next = pbtr;
      return;
   }

   next->insert(pbtr);
}


void bindex::build(edgetype **head) {

   ebucket->next = *head;
   *head = bucket;

   if (back != (bindex *)NULL)
      back->build(head);
   if (next != (bindex *)NULL)
      next->build(head);
}


void bindex::free_index(bindex **freelist) {

   if (next != (bindex *)NULL)
      next->free_index(freelist);

   if (back != (bindex *)NULL) {
      back->free_index(freelist);
      back = (bindex *)NULL;
   }

   bucket = ebucket = (edgetype *)NULL;

   next = *freelist;
   *freelist = this;
}


void polytype::init() {

   normal = mvertex = gnormal = (vector4f *)NULL;
   object = polynum = (int *)NULL;
   edgeptr = (int **)NULL;
   shade = (illtype *)NULL;
}


void polytype::dest() {

   if (mvertex != (vector4f *)NULL) {
      delete mvertex;
      delete normal;
      delete gnormal;
      delete polynum;
      delete edgeptr;
      delete object;
   }

   if (shade != (illtype *)NULL)
      delete shade;

   init();
}


int polytype::build(int cvertex, int cobject, int cedge, int sflag) {

   if (!cvertex || !cobject || !cedge)
      return 0;

   dest();

   countvertex = cvertex;
   countobject = cobject;
   countedge   = cedge;

   mvertex = new vector4f[countvertex];
   polynum = new int[countobject];
   edgeptr = new pint[countobject];
   object = new int[countedge];
   normal  = new vector4f[countobject];
   gnormal = new vector4f[countvertex];

   if (sflag)
      shade = new illtype[countobject];

   return 1;
}


void spatchtype::init() {

   ndata = data = (vector4f *)NULL;
   gnormal = mvertex = (pvector4f *)NULL;
   shade = (illtype *)NULL;
   mshade = (illtype **)NULL;
}

void spatchtype::dest() {

   if (mvertex != (pvector4f *)NULL) {
      delete data;
      delete mvertex;
   }

   if (gnormal != (pvector4f *)NULL) {
      delete ndata;
      delete gnormal;
   }

   if (mshade != (illtype **)NULL) {
      delete shade;
      delete mshade;
   }

   init();
}

int spatchtype::build(int crow, int ccol, int nflag, int sflag) {

   int i, j;

   if (!crow || !ccol)
      return 0;

   dest();

   vrow = crow;
   vcol = ccol;

   srow = vrow - 2;
   scol = vcol - 2;

   prow = vrow - 3;
   pcol = vcol - 3;

   vsize = vrow*vcol;
   psize = prow*pcol;

   data = new vector4f[vsize];
   mvertex = new pvector4f[vrow];

   for (i=j=0; i<vrow; i++, j+=vcol)
      mvertex[i] = &data[j];

   if (nflag) {
      ndata = new vector4f[vsize];
      gnormal = new pvector4f[vrow];

      for (i=j=0; i<vrow; i++, j+=vcol)
         gnormal[i] = &ndata[j];
   }

   if (sflag) {
      shade  = new illtype[psize];
      mshade = new pilltype[prow];

      for (i=j=0; i<prow; i++, j+=pcol)
         mshade[i] = &shade[j];
   }

   return 1;
}


static char char_diff = 'A' - 'a';


/* **************************************************
************************************************** */
void lower_case(char *x) {

   int i;

   for (i=0; x[i]; i++)
      if (x[i] >= 'A' && x[i] <= 'Z')
         x[i] -= char_diff;
}


/* **************************************************
************************************************** */
void upper_case(char *x) {

   int i;

   for (i=0; x[i]; i++)
      if (x[i] >= 'a' && x[i] <= 'z')
         x[i] += char_diff;
}


/* **************************************************
************************************************** */
int get_token(FILE *infile, char *x) {

   int errorflag;

   do {
      if (errorflag = (fscanf(infile, "%255s", x) == EOF))
         return 0;

      if (!strcmp("//", x))
         while ((errorflag=fgetc(infile)) != '\n')
            if (errorflag == EOF)
               return 0;
   } while (!strcmp("//", x));

   return 1;
}


float      front = -1;                          // front clipping plane

AOGLwindow gwindow;

unsigned int WINX = DEFWINX;
unsigned int WINY = DEFWINY;

int PENCILTEST = 0;
int NOSHADOW = 0;
int NOSCREEN = 0;
int SILENT = 0;
int NIXMIP = 0;

int GAMMA = 0;
float FGAMMA;

int RENDER = 0;
int RENDER_START;
int RENDER_END;

int ANTIALIAS = 0;
int DEFLIGHT = 0;
int DEFLST = 0;
int DEFCAMERA = 0;
int SAVE = 0;
int RGBFILE = 0;

char SANTIALIAS[MAXSTRLEN];
char SDEFLIGHT[MAXSTRLEN];
char SDEFCAMERA[MAXSTRLEN];
char SDEFLST[MAXSTRLEN];
char SSAVE[MAXSTRLEN];
char SRGBFILE[MAXSTRLEN];

shadetype gshade;
