
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file contains all the global data types for my programs
************************************************************* */

#if !defined (__GLOBAL_H)
#define __GLOBAL_H

#include <stdlib.h>
#include "matrix.h"
#include "alex.h"


typedef enum renderenum {
   CAMERA, LIGHT, POLYGON, SPATCH, PARTICLE, LINE,
   RADIANT, SPHERE, CYLINDER, CONE, SBFAST
} rendertype;


typedef struct edgestruct {		// struct of edge table data
   vector3f         pt1;                // point data
   vector3f         pt2;                // point data			// only need y after clipping
   float            deltax, dx;		// dx/dy, multiscanline dx/dy
   float            deltaz, dz;		// dz/dy, multiscanline dz/dy

   float            t[2][2];						// dont need t[1][*] after clipping
   float            deltat[2], dt[2];	// dtx/dty, multiscan dtx/dty

   vector3f         colorstuff1;        // gouraud/phong shade data
   vector3f         delta, d;           // gouraud/phong delta/dy

   int             color;              // polygon #
   struct edgestruct *next;             // next edge
} edgetype;

typedef edgetype *pedgetype;
typedef edgetype **ppedgetype;


class shadetype {

   public:
      int index;
      vector3f ka, kp, ks;
      float specn;                        // power of cosine
      vector3f lum;

      shadetype();

      void read_data(FILE *infile);
      void write_data(FILE *outfile);
};

typedef shadetype * pshadetype;
typedef shadetype ** ppshadetype;


class illtype {

   public:
      shadetype *color;
      vector3f ambient;

      illtype operator = (shadetype *x);
};

typedef illtype *pilltype;


class texcolortype {

   public:

      shadetype color;
      vector3f  ambient;

      texcolortype operator = (illtype &x);
      texcolortype operator = (shadetype &x);
};


class shadelist {

   public:
      char shadename[MAXSTRLEN];
      
      shadetype base;

      int count;
      shadetype *list;

      shadelist *next;

      shadelist() {

         list = (shadetype *)NULL;
         next = (shadelist *)NULL;
      }

      ~shadelist() {

         if (list != (shadetype *)NULL)
            delete list;

         if (next != (shadelist *)NULL)
            delete next;
      }

      int read_data(char *filename);
      int write_data(char *filename);
};


class polytype {                     // composite data

   public:
      int		countvertex;           // vertex count
      int		countedge;
      int		countobject;           // # of polygons

      vector4f		*mvertex;
      vector4f		*gnormal;              // normals for vertex

      int		*polynum;              // # of verticies in a polygon
      int		**edgeptr;
      vector4f		*normal;               // normals for polygons

      int		*object;

      illtype		*shade;

      void init();
      void dest();
      int build(int cvertex, int cobject, int cedge, int sflag);

      polytype() {
         init();
      }

      ~polytype() {
         dest();
      }

};


class bindex {

   public:
      bindex *next, *back;
      int     id;
      edgetype  *bucket, *ebucket;

      bindex() {
         next = back = (bindex *)NULL;
         bucket = ebucket = (edgetype *)NULL;
      }

      ~bindex() {
         if (next != (bindex *)NULL)
            delete next;
         if (back != (bindex *)NULL)
            delete back;
      }

      bindex *find(int x);
      void insert(bindex *pbtr);
      void build(edgetype **head);
      void free_index(bindex **freelist);
};


class spatchtype {

   public:
      vector4f		*data;
      pvector4f		*mvertex;
      vector4f		*ndata;               // normals for vertex
      pvector4f		*gnormal;
      int		vsize, psize;           // vertex/patch count
      int		prow, pcol;
      int		vrow, vcol;             // x, y count
      int		srow, scol;

      illtype		*shade;
      illtype		**mshade;

      void init();
      void dest();
      int build(int crow, int ccol, int nflag, int sflag);

      spatchtype() {
         init();
      }

      ~spatchtype() {
         dest();
      }

};

void lower_case(char *x);
void upper_case(char *x);
int get_token(FILE *infile, char *x);

extern float     front;                         // front clipping plane
extern AOGLwindow gwindow;
extern unsigned int WINX, WINY;

extern int PENCILTEST;
extern int NOSHADOW;
extern int NOSCREEN;
extern int SILENT;
extern int NIXMIP;

extern int GAMMA;
extern float FGAMMA;

extern int RENDER;
extern int RENDER_START;
extern int RENDER_END;

extern int ANTIALIAS;
extern int DEFLIGHT;
extern int DEFLST;
extern int DEFCAMERA;
extern int SAVE;
extern int RGBFILE;

extern char SANTIALIAS[MAXSTRLEN];
extern char SDEFLIGHT[MAXSTRLEN];
extern char SDEFCAMERA[MAXSTRLEN];
extern char SDEFLST[MAXSTRLEN];
extern char SSAVE[MAXSTRLEN];
extern char SRGBFILE[MAXSTRLEN];

extern shadetype gshade;

#endif
