
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <termio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "laser.h"

static int  video_outfile = -1;
static char laser_portname[256];

#define LQ4000_STX          0x02 /* Start transmission of Command sequence        */
#define LQ4000_ETX          0x03 /* End Transmission of command sequence          */
#define LQ4000_EOT          0x04 /* Same as ETX					  */

#define LQ4000_ACK          0x06 /* Acknowledge receipt of valid command          */
#define LQ4000_NAK          0x15 /* Reject receipt of command                     */


static int LQ4000_error_check(char *response) {

	int error_num;
	
	if (response[0] == 'E')
		{
		printf("%s",response);
		sscanf(&response[1],"%d",&error_num);
		switch(error_num)
			{
			case 1:
				printf("LQ4000, Invalid Key or Overflow\n");
				break;
			case 2:
				printf("LQ4000, Disc Error\n");
				break;
			case 3:
				printf("LQ4000, Disc not loaded\n");
				break;
			case 4:
				printf("LQ4000, search timeout error\n");
				break;
			case 5:
				printf("LQ4000, Laser Error\n");
				break;
			case 6:
				printf("LQ4000, Focus Error\n");
				break;
			case 7:
				printf("LQ4000, Disc Motor speed or Sync Error\n");
				break;
			case 8:
				printf("LQ4000, Head Locked Error\n");
				break;
			case 9:
				printf("LQ4000, Insufficient Recording Space Error\n");
				break;
			case 10:
				printf("LQ4000, Video Signal Sync Error\n");
				break;
			case 11:
				printf("LQ4000, Write Protect Error\n");
				break;
			case 12:
				printf("LQ4000, Dew Error\n");
				break;
			case 15:
				printf("LQ4000, Thermic Error\n");
				break;
			case 16:
				printf("LQ4000, Off Track Error\n");
				break;
			case 17:
				printf("LQ4000, Loader Stop Error\n");
				break;
			case 18:
				printf("LQ4000, Disc Tilt Error\n");
				break;
			case 19:
				printf("LQ4000, Magnet Error\n");
				break;
			case 20:
				printf("LQ4000, Buffer Overflow\n");
				break;
			case 21:
				printf("LQ4000, Transmission Error\n");
				break;
			case 25:
				printf("LQ4000, Servo Drive Error\n");
				break;
			case 26:
				printf("LQ4000, Laser Power Setup Error\n");
				break;
			case 30:
				printf("LQ4000, Change Battery Request\n");
				break;
			case 32:
				printf("LQ4000, Nesting Error\n");
				break;
			case 33:
				printf("LQ4000, Program Error\n");
				break;
			case 34:
				printf("LQ4000, Program Too Large Error\n");
				break;
			case 36:
				printf("LQ4000, Memory Backup Error\n");
				break;
			case 41:
				printf("LQ4000, Disc ID Write Error\n");
				break;
			case 43:
				printf("LQ4000, Duplicate Disc ID Error\n");
				break;
			case 44:
				printf("LQ4000, Bad Disc ID Number Error\n");
				break;
			case 45:
				printf("LQ4000, Disc ID Read Error\n");
				break;
			default:
				printf("LQ4000, Unknown Error number\n");
				break;
			}
		return(-1);
		}

   return(0);
}

int LQ4000_read_cr_lf()
	{
	char	io_char;
	
	if (read(video_outfile, &io_char, 1) != 1)
		{
		printf("error in LQ4000_read: read error on port %s\n",laser_portname);
		return(-1);
		}
	if (io_char != 0xD)
		{
		printf("Error, expected (CR) character 0xD got %X\n",(int) io_char);
		}
	if (read(video_outfile, &io_char, 1) != 1)
		{
		printf("error in LQ4000_read: read error on port %s\n",laser_portname);
		return(-1);
		}
	if (io_char != 0xA)
		{
		printf("Error, expected (NL) character 0xA got %X\n",(int) io_char);
		}

   return 0;
}


/**********************************************************************************

 LQ4000_read - An internal procedure for reading from the laser disc player

 **********************************************************************************/

static int LQ4000_read()
	{
	int i;
	char	io_char;
	char	response[256];
	

	if (read(video_outfile, &io_char, 1) != 1)
		{
		printf("error in LQ4000_read: read error on port %s\n",laser_portname);
		return(-1);
		}

	switch(io_char)
		{
		case LQ4000_ACK:
			/* printf("ACK "); */
			LQ4000_read_cr_lf();
			LQ4000_read();
			return(0);
		case LQ4000_NAK:
			/* printf("[NAK] "); */
			return(LQ4000_read());
		case LQ4000_STX:
			io_char = (char) NULL;
			i = 0;
			/* printf("[STX] ");  */
			while ((io_char != (char) LQ4000_ETX) && (io_char != (char) LQ4000_EOT))
				{
				if (read(video_outfile, &io_char, 1) != 1)
					{
					printf("error in LQ4000_read: read error on port %s\n",laser_portname);
					return(-1);
					}
				response[i] = io_char;
				i++;
				}
			response[i-1] = (char) NULL;
			/* printf("%s [ETX] ",response); */
			LQ4000_read_cr_lf();
			/* printf("\n"); */
			return(LQ4000_error_check(response));
		default:	
			printf("Unknown character read hex = %x %c\n",(int) io_char,io_char);
		}

   return 0;
}

/**********************************************************************************

 LQ4000_write - send a command to the laser disc

 **********************************************************************************/

int LQ4000_write(char *command)
	{
	register int size;
	register int i;
	char         io_char;
	int error;

	io_char = (char) LQ4000_STX;
	if ((error = write(video_outfile, &io_char, 1)) != 1)
		{
		printf("error %d in LQ4000_write: Error sending STX byte\n",error);
		return(-1);
		}
	
	size = strlen(command);
	if ((i=write(video_outfile, command, size)) != size)
		{
		printf("error in LQ4000_write: write error on port %s on byte %d\n",laser_portname, i);
		return(-1);
		}
		
	io_char = (char) LQ4000_ETX;
	if (write(video_outfile, &io_char, 1) != 1)
		{
		printf("error in LQ4000_write: Error sending ETX byte\n");
		return(-1);
		}
   LQ4000_read();
   return(0);
}

/**********************************************************************************

 LQ4000_init - Called once to open a communication port and initializes the laser 
                disc player. The char string is the name of the port. An example 
                would be /dev/ttyd2.  Port speed is 9600.

 **********************************************************************************/

int LQ4000_init(char *port_name)
	{
	struct termio term;
	char   command[8];

	video_outfile = open(port_name, O_RDWR);
	if (video_outfile == -1)
		{
		printf("error in LQ4000_init: could not open port %s for reading and writing\n",port_name);
		return(-1);
		}
	memset(&term, 0, sizeof(term));
	term.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
	term.c_iflag = 0;
	term.c_oflag = 0;
	term.c_lflag = 0;
	term.c_cc[VMIN] = 1;
	term.c_cc[VTIME] = 0;

	if (ioctl(video_outfile, TCSETA, &term) < 0)
		{
		video_outfile = -1;
		printf("error in LQ4000_init: could not configure port %s\n",port_name);
		return(-1);
		}
	/* Dis-able screen blanking */
	strcpy(laser_portname, port_name);

	/* Online command, mode 8 */
	strcpy(command,"ON 0:");

	if (LQ4000_write(command))
		{
		printf("error in LQ4000_init: command failed\n");
		return(-1);
		}

   return(0);
}

/**********************************************************************************

 LQ4000_clear - Clears or reset the laser disc player

 **********************************************************************************/

int LQ4000_clear(void)
	{
	char command[8];

	/* clear command */
	strcpy(command,"AC");

	if (LQ4000_write(command))
		{
		printf("error in LQ4000_clear: command failed\n");
		return(-1);
		}

   return(0);
}

/**********************************************************************************

 LQ4000_close - used to close the communication port.

 **********************************************************************************/

int LQ4000_close(void) 
	{
	char command[32];
	
	if (video_outfile == -1)
		{
		printf("error in LQ4000_close: port already closed\n");
		return(-1);
		}

	/* clear command */
	strcpy(command,"AC");

	if (LQ4000_write(command))
		{
		printf("error in LQ4000_close: all clear command failed\n");
		return(-1);
		}

	strcpy(command,"OF");

	if (LQ4000_write(command))
		{
		printf("error in LQ4000_close: off line command failed\n");
		return(-1);
		}

   return(close(video_outfile));
}



/**********************************************************************************

 LQ4000_rec_standby - Puts the laser dics player in record standby mode. The player
                       is prepared to record from start (in frame number) to end
                       (in frame number). The actual recording occurs when 
                       LQ4000_rec is called. 

 **********************************************************************************/

int LQ4000_rec_standby(int start, int stop)
	{
	char command[32];

	if (start < 1)
		{
		printf("error in LQ4000_rec_standby: start point must be a positive number\n");
		return(-1);
		}

	if (stop <= start)
		{
		printf("error in LQ4000_rec_standby: stop point must be larger then start\n");
		return(-1);
		}

	sprintf(command,"SR %d:",start);
	if (LQ4000_write(command))
		{
		printf("error in LQ4000_rec_standby: command failed while searching start point\n");
		return(-1);
		}
	
	sprintf(command,"RM %d:",stop - start);
	if (LQ4000_write(command))
		{
		printf("error in LQ4000_rec_standby: command failed while sending write area\n");
		return(-1);
		}

   return(0);
}

/**********************************************************************************

 LQ4000_rec - Causes the laser disc player to record.  If the player is in frame by
               frame recording then this procedure is called after each new frame is
               displayed.  

 **********************************************************************************/

int LQ4000_rec(void) {

   char command[32];

   strcpy(command,"GS 1:1:");

   if (LQ4000_write(command)) {
      printf("error in LQ4000_rec: command failed\n");
      return -1;
   }

   return 0;
}

