
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
************************************************************* */

#include <bstring.h>
#include <string.h>
#include "light.h"

/* *************************************************************
************************************************************* */
lightbuff lightbuff::operator = (lightbuff x) {

   int i, j;

   if (lbuff != (lightbufftype *)NULL && (maxx != x.maxx || maxy != x.maxy)) {
      delete lbuff;
      lbuff = (lightbufftype *)NULL;
   }

   maxx = x.maxx;
   maxy = x.maxy;

   j = maxx*maxy;

   if (lbuff == (lightbufftype *)NULL) {
      lbuff = new lightbufftype;
      lbuff->idbuff.init_map(MAXLDIM, MAXLDIM);
      lbuff->zbuff.init_map(MAXLDIM, MAXLDIM);
   }

   for (i=0; i<j; i++) {
      lbuff->idbuff.data[i] = x.lbuff->idbuff.data[i];
      lbuff->zbuff.data[i] = x.lbuff->zbuff.data[i];
   }

   return *this;
}


/* *************************************************************
************************************************************* */
int lightbuff::init_buff() {

   unsigned int i;

   if (lbuff == (lightbufftype *)NULL) {
      lbuff = new lightbufftype;
      lbuff->idbuff.init_map(MAXLDIM, MAXLDIM);
      lbuff->zbuff.init_map(MAXLDIM, MAXLDIM);
   }

   else
      if (lbuff->idbuff.data == (unsigned int *)NULL) {
         lbuff->idbuff.init_map(MAXLDIM, MAXLDIM);
         lbuff->zbuff.init_map(MAXLDIM, MAXLDIM);
      }

   i = SQR(MAXLDIM) << 2;

   bzero(lbuff->zbuff.data, i);		// float - 4 bytes
   bzero(lbuff->idbuff.data, i);	// int - 4 bytes

   return 1;
}


/* *************************************************************


id - 0 		always (transparent)
id - "+"	id of object
id - "-"	ignore (should never call... ergo never happen)
************************************************************* */
float lightbuff::calc_intensity(eye *parm, float *c, float *surface, int id, int frame) {

   vector4f pt;
   int  i, j, k, l;
   float mag;
   int  frac;
   int  pt2[2][2];
   float temp;

   copyarray3(pt, surface);
   pt[3] = 1;

   matvecmulto(lparm->transform, pt);

   if (lparm->ltype == BEAM_POINT) {
      mag = -pt[2];
      lparm->map2screen(pt);
   }

   else {
      mag = magnitude3(pt);
      lparm->paramap2screen(pt);
   }

   j = (int)pt[0];
   i = (int)pt[1];

   if (i < 0 || i >= maxy || j < 0 || j >= maxx)
      return 0;

   if (id == lbuff->idbuff.pdata[i][j]) {

      i = (int)(i*ztscale);
      j = (int)(j*ztscale);

      copyarray3(pt, surface);
      subarray3(pt, parm->location);
      temp = magnitude3(pt)*lparm->area;

      if (lparm->ltype == BEAM_FAR)
         lparm->tob->query_texel(j, i, c, frame,1,temp);
      else {
         copyarray3(pt, surface);
         subarray3(pt, lparm->location);
         lparm->tob->query_texel(j, i, c, frame,1,temp/magnitude3(pt));		// alt - could use zbuff for z values instead of calc distance....
      }

      return mag;
   }

   pt2[0][0] = j-1;
   pt2[1][0] = j+1;
   pt2[0][1] = i-1;
   pt2[1][1] = i+1;

   i = (int)(i*ztscale);
   j = (int)(j*ztscale);

   if (pt2[0][0] < 0)
      pt2[0][0] = 0;
   if (pt2[1][0] >= maxx)
      pt2[1][0] = maxx-1;
   if (pt2[0][1] < 0)
      pt2[0][1] = 0;
   if (pt2[1][1] >= maxy)
      pt2[1][1] = maxy-1;

   frac = 0;

   for (l=pt2[0][1]; l<pt2[1][1]; l++)
      for (k=pt2[0][0]; k<pt2[1][0]; k++)
         if (id == lbuff->idbuff.pdata[l][k])
            frac += 25;

   if (!frac)
      return 0;
   else
      frac += 50;

   copyarray3(pt, surface);
   subarray3(pt, parm->location);
   temp = magnitude3(pt)*lparm->area;

   if (lparm->ltype == BEAM_FAR)
      lparm->tob->query_texel(j, i, c, frame,1,temp);
   else {
      copyarray3(pt, surface);
      subarray3(pt, lparm->location);
      lparm->tob->query_texel(j, i, c, frame,1,temp/magnitude3(pt));		// alt - could use zbuff for z values instead of calc distance....
   }

   temp = frac*0.000039215686;
   smultarray3(c, temp);

   return mag;
}


/* *************************************************************
************************************************************* */
void light::alter_beam_mx(eye *parm) {

   float temp[4][4];

   copymx4x4(temp, transform);
   copymx4x4(transform, parm->Tinverse);
   matmatmulto(temp, transform);
   matmatmulto(parm->transform, Tinverse);
}


/* *************************************************************
************************************************************* */
int light::parse(FILE *infile, char *token) {

   switch (token[0]) {
      case '2':
         if (!strcmp(token, "2d_texture")) {
            get_token(infile, texname);

            return 1;
         }

         return 0;

      case 'a':
         if (!strcmp(token, "ambient")) {
            get_token(infile, token);
            Ia[0] = atof(token);
            get_token(infile, token);
            Ia[1] = atof(token);
            get_token(infile, token);
            Ia[2] = atof(token);

            return 1;
         }

         return 0;

      case 'd':
         if (strlen(token) < 3)
            return 0;

         switch (token[2]) {
            case 'f':
               if (!strcmp(token, "diffuse")) {
                  get_token(infile, token);
                  Ip[0] = atof(token);
                  get_token(infile, token);
                  Ip[1] = atof(token);
                  get_token(infile, token);
                  Ip[2] = atof(token);

                  return 1;
               }

               return 0;

            case 's':
               if (!strcmp(token, "distance2plane")) {
                  get_token(infile, token);
                  distance2plane = atof(token);

                  return 1;
               }

               return 0;

            case 'r':
               if (!strcmp(token, "direction")) {
                  get_token(infile, token);
                  vpn[0] = -atof(token);
                  get_token(infile, token);
                  vpn[1] = -atof(token);
                  get_token(infile, token);
                  vpn[2] = -atof(token);

                  return 1;
               }

               return 0;

            default:
               break;
         }

         return 0;

      case 'f':
         if (!strcmp(token, "fatt")) { 
            get_token(infile, token);
            fatt[0] = atof(token);
            get_token(infile, token);
            fatt[1] = atof(token);
            get_token(infile, token);
            fatt[2] = atof(token);

            return 1;
         }

         if (!strcmp(token, "frame")) { 
            get_token(infile, token);
            frame = atoi(token) - 1;

            return 1;
         }

         return 0;

      case 'l':
         if (strlen(token) < 2)
            return 0;

         switch (token[1]) {
            case 'o':
               if (!strcmp(token, "location")) {
                  get_token(infile, token);
                  location[0] = atof(token);
                  get_token(infile, token);
                  location[1] = atof(token);
                  get_token(infile, token);
                  location[2] = atof(token);

                  return 1;
               }

               return 0;

            case 'e':
               if (!strcmp(token, "length")) {
                  get_token(infile, token);
                  length = atof(token);

                  return 1;
               }

               return 0;

            default:
               break;
         }

         return 0;

      case 't':
         if (!strcmp(token, "type")) {
            get_token(infile, token);
            lower_case(token);

            if (!strcmp(token, "beam_point"))
               ltype = BEAM_POINT;
            else if (!strcmp(token, "beam_far"))
               ltype = BEAM_FAR;
            else if (!strcmp(token, "point"))
               ltype = POINT;
            else
               ltype = FAR;

            return 1;
         }

         return 0;

      case 'u':
         if (!strcmp(token, "up")) {
            get_token(infile, token);
            vup[0] = atof(token);
            get_token(infile, token);
            vup[1] = atof(token);
            get_token(infile, token);
            vup[2] = atof(token);

            return 1;
         }

         return 0;

      case 'w':
         if (!strcmp(token, "width")) {
            get_token(infile, token);
            width = atof(token);

            return 1;
         }

         return 0;

      default:
         break;
   }

   return 0;
}


/* *************************************************************
************************************************************* */
void light::preprocess(texbase **tptr) {

   float temp[2];
   float i, j;

   if (ltype == POINT)
      return;

   normalize3(vpn);

   if (ltype == FAR)
      return;

   normalize3(vup);

   width /= distance2plane;
   length /= distance2plane;

   for (tob=*tptr; tob!=(texbase *)NULL; tob=tob->next)
      if (!strcmp(texname, tob->filename))
         break;

   if (tob == (texbase *)NULL) {

      if (query_video(texname)) {
         tob = new texvid;
         tob->read_texture(texname, tptr);
      }

      else {
         tob = new texture;
         tob->read_texture(texname);
      }

      tob->next = *tptr;
      *tptr = tob;
   }

   i = tob->query_x();
   j = tob->query_y();

   temp[0] = length/width;
   temp[1] = i/j;

   if (temp[0] > temp[1])            // reshape viewport to texture ratios;
      length = temp[1]*width;
   else
      width = length/temp[1];

   temp[0] = 0.5*length;
   temp[1] = 0.5*width;

   vrc[0] = -temp[0];
   vrc[1] =  temp[0];
   vrc[2] = -temp[1];
   vrc[3] =  temp[1];

   if (width > length)
      mscale = MAXLDIM/width;
   else
      mscale = MAXLDIM/length;

   imscale = 1.0/mscale;
 
   lbuff.maxx = (int)(length * mscale);
   lbuff.maxy = (int)(width * mscale);

   if (i > j)
      lbuff.ztscale = i/(float)MAXLDIM;      // scale from lbuff to tmap
   else
      lbuff.ztscale = j/(float)MAXLDIM;

   nper();
}


/* *************************************************************
************************************************************* */
int light::read_data(FILE *infile, texbase **tptr) {

   char tempstr[MAXSTRLEN];

   fscanf(infile, "%s", tempstr);

   if (!strcmp(tempstr, "BEAM_POINT"))
      ltype = BEAM_POINT;
   else if (!strcmp(tempstr, "BEAM_FAR"))
      ltype = BEAM_FAR;
   else if (!strcmp(tempstr, "POINT"))
      ltype = POINT;
   else
      ltype = FAR;

   if (ltype == FAR) {
      fscanf(infile, "%f %f %f", &vpn[0], &vpn[1], &vpn[2]);
      fscanf(infile, "%f %f %f", &Ia[0], &Ia[1], &Ia[2]);
      fscanf(infile, "%f %f %f", &Ip[0], &Ip[1], &Ip[2]);
      smultarray3(vpn, -1);
      preprocess(tptr);
   }

   else {
      fscanf(infile, "%f %f %f", &location[0], &location[1], &location[2]);

      if (ltype == POINT) {
         fscanf(infile, "%f %f %f", &Ia[0], &Ia[1], &Ia[2]);
         fscanf(infile, "%f %f %f", &Ip[0], &Ip[1], &Ip[2]);
      }

      else {
         fscanf(infile, "%f %f %f", &vpn[0], &vpn[1], &vpn[2]);
         fscanf(infile, "%f %f %f", &vup[0], &vup[1], &vup[2]);
         fscanf(infile, "%f", &distance2plane);
         fscanf(infile, "%f %f", &width, &length);
         fscanf(infile, "%s", texname);               // texture name
         smultarray3(vpn, -1);
         preprocess(tptr);
      }

   }

   fscanf(infile, "%f %f %f", &fatt[0], &fatt[1], &fatt[2]);

   return 1;
}


void light::write_lst(FILE *outfile, int frameno) {

   fprintf(outfile, "%d\nLIGHT\n", frameno);

   switch (ltype) {

      case BEAM_FAR:
         fprintf(outfile, "BEAM_FAR\n");
         break;

      case BEAM_POINT:
         fprintf(outfile, "BEAM_POINT\n");
         break;

      case FAR:
         fprintf(outfile, "FAR\n");
         break;

      default:
         fprintf(outfile, "POINT\n");
         break;
   }

   fprintf(outfile, "%f %f %f\n", location[0], location[1], location[2]);

   if (ltype == FAR || ltype == POINT) {
      fprintf(outfile, "%f %f %f\n", Ia[0], Ia[1], Ia[2]);
      fprintf(outfile, "%f %f %f\n", Ip[0], Ip[1], Ip[2]);
   }

   else {
      fprintf(outfile, "%f %f %f\n", -vpn[0], -vpn[1], -vpn[2]);
      fprintf(outfile, "%f %f %f\n", vup[0], vup[1], vup[2]);
      fprintf(outfile, "1\n");
      fprintf(outfile, "%f %f\n", vrc[3]+vrc[3], vrc[1]+vrc[1]);
      fprintf(outfile, "%s\n", tob->filename);               // texture name
   }

   fprintf(outfile, "%f %f %f\n", fatt[0], fatt[1], fatt[2]);
}


/* *************************************************************
   This procedure calculates the intensity of light at a certain point
   * this function also takes into account the position of the light
     source, and shadows the appropriate sides :)
************************************************************* */
void light::intensity(float surface[], float gnormal[], float rcolor[],
                      eye *parm, shadetype *shade, int id) {

   int    i;                                                /* looping var  */
   float vlight[3];                                        /* light vector */
   float V[3];                                             /* view vector  */
   float R[4];                                             /* R vector     */
   float NL;                                               /* normal.light */
   float fatten;                                           /* light atten  */
   float beam_distance;
   float mIp[3];
   float temp;

   if ( (ltype == BEAM_POINT || ltype == BEAM_FAR) && id) {
      if ((beam_distance = lbuff.calc_intensity(parm, mIp, surface, id, frame)) < CORRECT)
         return;
   }

   else
      copyarray3(mIp, Ip);		// BEAM_FAR/POINT use Ip defaults for CONSTANT

   if (ltype == FAR || ltype == BEAM_FAR) {
      copyarray3(vlight, transvec);
      temp = 1;
   }

   else {
      for (i=0; i<3; i++)               // calculate light
         vlight[i] = transloc[i] - surface[i];

      temp = normalize3(vlight);
   }

   NL = dotproduct3(gnormal, vlight);	// normal . light
   if (-NL >= 0) 			// calculate if not in light
      return;

   for (i=0; i<3; i++)                                    // V, view vector
      V[i] = parm->location[i] - surface[i];

   fatten = normalize3(V);

   if (ltype != FAR)
      if (ltype == BEAM_FAR)		// vlight = normalize(vpn)
         fatten += beam_distance;
      else
         fatten += temp;

   fatten = fatt[0] + (fatt[1] + fatt[2]*fatten)*fatten;    // calculate fatt

   for (i=0; i<3; i++) {                              // calculate R vector
      R[i] = NL*gnormal[i];
      R[i] += R[i] - vlight[i];
   }

   R[0] = dotproduct3(R, V);

   if (R[0] > 0.0)
      if (shade->specn > CORRECT)
         R[3] = pow(R[0], shade->specn);
      else
         R[3] = R[0];
   else
      R[3] = 0;

   if (fatten > 1) {
      R[3] /= fatten;
      NL   /= fatten;
   }
                                                      // assign pixel color
   rcolor[0] += mIp[0]*(shade->kp[0]*NL + shade->ks[0]*R[3]);
   rcolor[1] += mIp[1]*(shade->kp[1]*NL + shade->ks[1]*R[3]);
   rcolor[2] += mIp[2]*(shade->kp[2]*NL + shade->ks[2]*R[3]);
}


/* *************************************************************
************************************************************* */
void light::xform(camera *cparm) {

   if (ltype != POINT) {
      copyarray3(transvec, vpn);
      transvec[3] = 0;
      matvecmultv(cparm->transform, transvec);
   }

   if (ltype != FAR) {
      copyarray3(transloc, location);
      transloc[3] = 1;
      matvecmulto(cparm->transform, transloc);
   }

   area = 1.0/(mscale*cparm->query_scale());
}


/* *************************************************************
************************************************************* */
light light::operator = (light x) {

   ltype = x.ltype;
   copyarray3(location, x.location);
   copyarray4(transloc, x.transloc);
   copyarray4(transvec, x.transvec);
   copyarray3(Ia, x.Ia);
   copyarray3(Ip, x.Ip);
   copyarray3(fatt, x.fatt);
   frame = x.frame;
   next = x.next;

   if (ltype == BEAM_FAR || ltype == BEAM_POINT) {
      tob = x.tob;
      mscale = x.mscale;
      imscale = x.imscale;
      copyarray3(vpn, x.vpn);
      copyarray3(vup, x.vup);
      copyarray4(vrc, x.vrc);
      copymx4x4(transform, x.transform);
      copymx4x4(Tinverse, x.Tinverse);
      area = x.area;
      lbuff = x.lbuff;
   }

   return *this;
}


/* *************************************************************
************************************************************* */
void light::add_ambient(float *ka, float *ambient) {

   if (ltype == FAR || ltype == POINT) {
      ambient[0] += Ia[0]*ka[0];
      ambient[1] += Ia[1]*ka[1];
      ambient[2] += Ia[2]*ka[2];
   }

}


/* *************************************************************
************************************************************* */
void light::set_ambient(float *ka, float *ambient) {

   if (ltype == FAR || ltype == POINT) {
      ambient[0] = Ia[0]*ka[0];
      ambient[1] = Ia[1]*ka[1];
      ambient[2] = Ia[2]*ka[2];
   }

   else
      ambient[0] = ambient[1] = ambient[2] = 0;
}
