
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */



#if !defined __LIGHT_H
#define __LIGHT_H

#include "texture.h"
#include "camera.h"
#include "global.h"

#define MAXLDIM 200


typedef enum lightenum {
   POINT = 0,
   FAR   = 1,
   BEAM_POINT = 2,
   BEAM_FAR = 3
} lighttype;


typedef struct lightbuffstruct {

   mapul idbuff;
   mapf  zbuff;
   struct lightbuffstruct *next;
} lightbufftype;


class light;

class lightbuff {

   public:
      light     *lparm;

      int      maxx, maxy;
      float     ztscale;

      lightbufftype *lbuff;

      lightbuff *next;

      lightbuff operator = (lightbuff x);
      int       init_buff();
      float     calc_intensity(eye *parm, float *c, float *surface, int id, int frame);

      lightbuff() {
         lbuff   = (lightbufftype *)NULL;
         next    = (lightbuff *)NULL;
      }

      void set_light(light *lit) {
         lparm   = lit;
      }

      ~lightbuff() {
         if (lbuff != (lightbufftype *)NULL)
            delete lbuff;
      }

};


/* *************************************************************
************************************************************* */


class light : public eye {

   public:
      lighttype ltype;

      vector4f  transloc, transvec;

      float     Ia[3], Ip[3];              // light intensities
      float     fatt[3];

      texbase   *tob;
      float     area;
      lightbuff lbuff;

      float length, width, distance2plane;
      char texname[MAXSTRLEN];

      light     *next;

      void      preprocess(texbase **tptr);
      int	parse(FILE *infile, char *token);
      int       read_data(FILE *infile, texbase **tptr);
      void      write_lst(FILE *outfile, int frameno);
      void      xform(camera *cparm);
      void      intensity(float surface[], float gnormal[], float rcolor[],
                          eye *parm, shadetype *shade, int id);
      void      add_ambient(float *ka, float *ambient);
      void      set_ambient(float *ka, float *ambient);
      light     operator = (light x);

      void      alter_beam_mx(eye *parm);

      light() {
         ltype = FAR;
         vpn[0] = vpn[1] = 0; vpn[2] = 1;

         Ia[0] = Ia[1] = Ia[2] = 0.3;
         Ip[0] = Ip[1] = Ip[2] = 0.7;
         fatt[0] = fatt[1] = fatt[2] = 0.0;
         tob = (texbase *)NULL;
         area   = 1;
         lbuff.set_light(this);

         length = width = distance2plane = 1.0;
         texname[0] = 0;

         next   = (light *)NULL;
      }

};

typedef light * lightptrtype;

#endif
