
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */


#include "control.h"

int MAXSUBPROC = 1;
int MAXPROCESSOR = 1;

master::master() {

   freeslave   = (slave *)NULL;
   activeslave = (slave *)NULL;

   result      = (ptrtype *)NULL;
   lbuff       = (lightbufftype *)NULL;

   circuit     = (slave *)NULL;
}


master::~master() {

   if (circuit != (slave *)NULL)
      delete circuit;
}


void master::done() {

   lightbufftype *ltr;

   for (ltr=lbuff; ltr != (lightbufftype *)NULL; lbuff=lbuff->next, delete ltr, ltr=lbuff);
}


void master::activate() {

   int i;

   circuit = new slave[MAXPROCESSOR];

   for (i=0; i<MAXPROCESSOR-1; i++) {
      circuit[i].next = &circuit[i+1];
      circuit[i].control = this;
      circuit[i].id = i;
   }

   circuit[MAXPROCESSOR-1].next = (slave *)NULL;
   circuit[MAXPROCESSOR-1].control = this;
   circuit[MAXPROCESSOR-1].id = MAXPROCESSOR-1;

   freeslave = &circuit[0];

   for (i=0; i<MAXPROCESSOR; i++)
      circuit[i].activate();

   semid = semget(CONTROLKEY, MAXSEMAPHORE, IPC_CREAT | PERMISSIONS);

   for (i=0; i<MAXSEMAPHORE; i++)
      free_lock(i);
}


int master::query_free() {

   return (freeslave != (slave *)NULL);
}


slave *master::get_slave() {

   slave *ptr;

   set_lock(LOCK_QUEUE);

   if (freeslave == (slave *)NULL) {
      free_lock(LOCK_QUEUE);
      return (slave *)NULL;
   }

   ptr = freeslave;
   freeslave = freeslave->next;

   ptr->next = activeslave;
   activeslave = ptr;

   free_lock(LOCK_QUEUE);
   return ptr;
}


int master::free_slave(int x) {

   slave *ptr, *qtr;

   set_lock(LOCK_QUEUE);

   if (activeslave == (slave *)NULL) {
      free_lock(LOCK_QUEUE);
      return -1;
   }

   if (activeslave->id == x) {
      ptr = activeslave;
      activeslave = activeslave->next;
      ptr->next = freeslave;
      freeslave = ptr;
      free_lock(LOCK_QUEUE);
      return 1;
   }

   for (ptr=activeslave; ptr->next != (slave *)NULL; ptr = ptr->next)
      if (ptr->next->id == x) {
         qtr = ptr->next;
         ptr->next = qtr->next;
         qtr->next = freeslave;
         freeslave = qtr;
         free_lock(LOCK_QUEUE);
         return 1;
      }

   free_lock(LOCK_QUEUE);
   return -1;
}


int master::get_lbuff(light *ltr) {

   if (lbuff != (lightbufftype *)NULL) {
      ltr->lbuff.lbuff = lbuff;
      lbuff = lbuff->next;
   }

   else
      ltr->lbuff.lbuff = new lightbufftype;

   return 1;
}


int master::stack_lbuff(light *ltr) {

   if (ltr->lbuff.lbuff != (lightbufftype *)NULL) {
      ltr->lbuff.lbuff->next = lbuff;
      lbuff = ltr->lbuff.lbuff;
      ltr->lbuff.lbuff = (lightbufftype *)NULL;
      return 1;
   }

   return 0;
}


int master::pop_result() {

   ptrtype *ptr;

   if (result == (ptrtype *)NULL)
      return 0;

   ptr       = result;
   result    = result->next;
   delete ptr;

   return 1;
}


int master::push_result(int id) {

   ptrtype *ptr;

   ptr = new ptrtype;
   ptr->id = id;
   ptr->next = result;
   result    = ptr;

   return 1;
}


/* **************************************************
************************************************** */
void master::kill_children() {

   int i, j;

#ifdef SGI
   for (i=0; i<MAXPROCESSOR; i++) {

      while(circuit[i].query_busy());

      for (j=0; j<MAXSUBPROC-1; j++) {           // comment out scanline
         circuit[i].grandchild[j].bye = 1;
//         if (!kill(circuit[i].grandchild[j].child_id, 9))
            printf("Killed subchild %d %d\n", i, j);
      }

      circuit[i].bye = 1;
//      if (!kill(circuit[i].child_id, 9))
         printf("Killed child %d\n", i);
   }

   semctl(semid, 1, IPC_RMID);                  // kill semaphores
#else
   semun x;
   semctl(semid, 0, IPC_RMID, x);
#endif
}
