
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file handles all the matrix/vector/math operations

   Reminder: Spherical Linear intERPolation (SLERP) can be
             done by splining the quaternions

************************************************************* */

#include <string.h>
#include "matrix.h"

vector4f MCR[4]  = { -0.5,      1.5,     -1.5,      0.5,
                      1.0,     -2.5,      2.0,     -0.5,
                     -0.5,      0.0,      0.5,      0.0,
                      0.0,      1.0,      0.0,      0.0 };     // catmull-rom

vector4f MCRT[4] = { -0.5,      1.0,     -0.5,      0.0,       // Tcatmull-rom
                      1.5,     -2.5,      0.0,      1.0,
                     -1.5,      2.0,      0.5,      0.0,
                      0.5,     -0.5,      0.0,      0.0 };

vector4f MB[4]   = { -1.0/6.0,  0.5,     -0.5,      1.0/6.0,   // B-spline
                      0.5,     -1.0,      0.5,      0.0,
                     -0.5,      0.0,      0.5,      0.0,
                      1.0/6.0,  2.0/3.0,  1.0/6.0,  0.0 };

vector4f MBT[4]  = { -1.0/6.0,   0.5,    -0.5,       1.0/6.0,   // TB-spline
                     0.5,      -1.0,     0.0,       2.0/3.0,
                    -0.5,       0.5,     0.5,       1.0/6.0,
                     1.0/6.0,   0.0,     0.0,       0.0 };


/* *************************************************************
   this function extends a vector to a specified magnitude
************************************************************* */
void extend_vector3(float x[], float dist) {

   register float mag = magnitude3(x);
   register float imag;

   if (mag > 0) {
      imag = dist/mag;
      smultarray3(x, imag);
   }

}


/* *************************************************************
   This procedure normalizes a vector and returns its magnitude
************************************************************* */
float normalize2(float x[]) {

   register float mag = magnitude2(x);
   register float imag;

   if (mag > 0) {
      imag = 1.0/mag;
      x[0] *= imag;
      x[1] *= imag;
   }

   return mag;
}


/* *************************************************************
   This procedure normalizes a vector and returns its magnitude
************************************************************* */
float normalize3(float x[]) {

   register float mag = magnitude3(x);
   register float imag;

   if (mag > 0) {
      imag = 1.0/mag;
      smultarray3(x, imag);
   }

   return mag;
}


/* *************************************************************
   This procedure normalizes a 4D vector and returns its magnitude
************************************************************* */
float normalize4(float x[]) {

   register float mag = magnitude4(x);
   register float imag;

   if (mag > 0) {
      imag = 1.0/mag;
      smultarray4(x, imag);
   }

   return mag;
}


/* *************************************************************
   This procedure normalizes a vector and returns its magnitude
************************************************************* */
void xproduct(float *a, float *x, float *y) {

   a[0] = x[1]*y[2] - x[2]*y[1];
   a[1] = x[2]*y[0] - x[0]*y[2];
   a[2] = x[0]*y[1] - x[1]*y[0];
}


/* *************************************************************
   this function calculates the transpose of a matrix
************************************************************* */
void transpose(float r[][4], float s[][4]) {

   register int    i, j;

   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
         r[j][i] = s[i][j];
}


/* *************************************************************
   this function calculates the transpose of a matrix
************************************************************* */
void transpose(float r[][4]) {

   register int    i, j;
   register float  s;

   for (i=0; i<4; i++)
      for (j=0; j<i; j++) {
         s = r[j][i];
         r[j][i] = r[i][j];
         r[i][j] = s;
      }

}


/* *************************************************************
   This procedure multiplies 2 4X4 matricies
************************************************************* */
void matmatmult(float y[][4], float x[][4]) {

   vector4f z[4];
   register int    i, j, k;

   for (i=0; i<4; i++)
      copyarray4(z[i], x[i]);

   for (i=0; i<4; i++)
      for (j=0; j<4; j++) {
         x[i][j] = 0.0;
         for (k=0; k<4; k++)
            x[i][j] += y[i][k]*z[k][j];
      }

}
//
/* *************************************************************
   This procedure multiplies a 4X4 matrix with a vector
************************************************************* */
void matvecmult(float y[][4], float x[]) {

   vector4f z;
   register int i, j;

   copyarray4(z, x);

   for (i=0; i<4; i++) {
       x[i] = 0.0;
       for (j=0; j<4; j++)
          x[i] += y[i][j]*z[j];
   }

}


/* *************************************************************
   This function multiplies a vector . matrix
************************************************************* */
void vecmatmult(float y[][4], float x[]) {

   vector4f z;
   register int i, j;

   copyarray4(z, x);

   for (i=0; i<4; i++) {
      x[i] = 0.0;
      for (j=0; j<4; j++)
         x[i] += z[j]*y[j][i];
   }

}


/* *************************************************************
   This procedure multiplies 2 4X4 matricies
************************************************************* */
void matmatmulto(float y[][4], float x[][4]) {

   vector4f z[3];
   register int i;

   copyarray4(z[0], x[0]);
   copyarray4(z[1], x[1]);
   copyarray4(z[2], x[2]);

   for (i=0; i<3; i++) {
      x[i][0] = y[i][0]*z[0][0] + y[i][1]*z[1][0] + y[i][2]*z[2][0];
      x[i][1] = y[i][0]*z[0][1] + y[i][1]*z[1][1] + y[i][2]*z[2][1];
      x[i][2] = y[i][0]*z[0][2] + y[i][1]*z[1][2] + y[i][2]*z[2][2];
      x[i][3] = y[i][0]*z[0][3] + y[i][1]*z[1][3] + y[i][2]*z[2][3] + y[i][3];
   }

}


/* *************************************************************
   This procedure multiplies a 4X4 matrix with a vector
************************************************************* */
void matvecmultv(float y[][4], float x[]) {

   vector3f z;

   copyarray3(z, x);

   x[0] = dotproduct3(y[0], z);
   x[1] = dotproduct3(y[1], z);
   x[2] = dotproduct3(y[2], z);
}


/* *************************************************************
   This procedure multiplies a 4X4 matrix with a vector
************************************************************* */
void matvecmulto(float y[][4], float x[]) {

   vector3f z;

   copyarray3(z, x);

   x[0] = dotproduct3(y[0], z) + y[0][3];
   x[1] = dotproduct3(y[1], z) + y[1][3];
   x[2] = dotproduct3(y[2], z) + y[2][3];
}


/* *************************************************************
   This function multiplies a vector . matrix
************************************************************* */
void vecmatmultv(float y[][4], float x[]) {

   vector3f z;

   copyarray3(z, x);

   x[0] = y[0][0]*z[0] + y[1][0]*z[1] + y[2][0]*z[2];
   x[1] = y[0][1]*z[0] + y[1][1]*z[1] + y[2][1]*z[2];
   x[2] = y[0][2]*z[0] + y[1][2]*z[1] + y[2][2]*z[2];
   x[3] = 1;
//   x[3] = y[0][3]*z[0] + y[1][3]*z[1] + y[2][3]*z[2];
}


/* *************************************************************
   This function multiplies a vector . matrix
************************************************************* */
void vecmatmulto(float y[][4], float x[]) {

   vector3f z;

   copyarray3(z, x);

   x[0] = y[0][0]*z[0] + y[1][0]*z[1] + y[2][0]*z[2];
   x[1] = y[0][1]*z[0] + y[1][1]*z[1] + y[2][1]*z[2];
   x[2] = y[0][2]*z[0] + y[1][2]*z[1] + y[2][2]*z[2];
   x[3] = y[0][3]*z[0] + y[1][3]*z[1] + y[2][3]*z[2] + y[3][3];
}


/* *************************************************************
   this procedure swaps two vectors
************************************************************* */
void swap4(float mx1[], float mx2[]) {

   float temp;
   register int i;

   for (i=0; i<4; i++) {
      temp   = mx1[i];
      mx1[i] = mx2[i];
      mx2[i] = temp;
   }

}
// 
/* *************************************************************
   this procedure calculates the inverse of a matrix
************************************************************* */
void inversemxrt(float mx[][4], float mxinverse[][4]) {

   vector4f temp;

   transpose(mxinverse, mx);

   copyarray4(temp, mxinverse[3]);

   mxinverse[3][0] = mxinverse[3][1] = mxinverse[3][2] = 0;

   matvecmulto(mxinverse, temp);

   mxinverse[0][3] = -temp[0];
   mxinverse[1][3] = -temp[1];
   mxinverse[2][3] = -temp[2];
}



/* *************************************************************
   this procedure calculates the inverse of a matrix
************************************************************* */
int inversemx(float a[][4],float b[][4]) {

   register double det, idet;

   b[0][0] = a[1][1]*a[2][2] - a[2][1]*a[1][2];
   b[0][1] = a[0][1]*a[2][2] - a[2][1]*a[0][2];
   b[0][2] = a[0][1]*a[1][2] - a[1][1]*a[0][2];

   b[1][0] = a[1][0]*a[2][2] - a[2][0]*a[1][2];
   b[1][1] = a[0][0]*a[2][2] - a[2][0]*a[0][2];
   b[1][2] = a[0][0]*a[1][2] - a[1][0]*a[0][2];

   b[2][0] = a[1][0]*a[2][1] - a[2][0]*a[1][1];
   b[2][1] = a[0][0]*a[2][1] - a[2][0]*a[0][1];
   b[2][2] = a[0][0]*a[1][1] - a[1][0]*a[0][1];

   det = a[0][0]*b[0][0] - a[1][0]*b[0][1] + a[2][0]*b[0][2];

   if (abs(det) < CORRECT)
      return 0;

   det = 1.0/det;
   idet = -det;

   b[0][0] *= det;
   b[0][1] *= idet;
   b[0][2] *= det;

   b[1][0] *= idet;
   b[1][1] *= det;
   b[1][2] *= idet;

   b[2][0] *= det;
   b[2][1] *= idet;
   b[2][2] *= det;

   b[0][3] = -(b[0][0]*a[0][3] + b[0][1]*a[1][3] + b[0][2]*a[2][3]);
   b[1][3] = -(b[1][0]*a[0][3] + b[1][1]*a[1][3] + b[1][2]*a[2][3]);
   b[2][3] = -(b[2][0]*a[0][3] + b[2][1]*a[1][3] + b[2][2]*a[2][3]);

   b[3][0] = 0;
   b[3][1] = 0;
   b[3][2] = 0;
   b[3][3] = 1;

   return 1;
}


/* *************************************************************
   this procedure calculates the inverse of a matrix
************************************************************* */
int inversemx2(float mx[][4], float mxinverse[][4]) {

   vector4f tempmx[4];
   register int i, j, k;
   register float temp;

   init_mx(i, j, mxinverse);

   copymx4x4(tempmx, mx);

   for (i=0; i<4; i++) {
      if (!not_zero(tempmx[i][i]))
         for (j=i+1; (j<4 && !not_zero(tempmx[i][i])); j++)
            if (tempmx[j][i]) {
               swap4(tempmx[i], tempmx[j]);
               swap4(mxinverse[i], mxinverse[j]);
            }

      if (!not_zero(tempmx[i][i]))
         return 0;

      temp = 1.0/tempmx[i][i];

      for (j=i+1; j<4; j++)
         tempmx[i][j] *= temp;

      mxinverse[i][0] *= temp;
      mxinverse[i][1] *= temp;
      mxinverse[i][2] *= temp;
      mxinverse[i][3] *= temp;

      for (j=i+1; j<4; j++) {
         for (k=i+1; k<4; k++)
            tempmx[j][k] -= tempmx[j][i]*tempmx[i][k];

         mxinverse[j][0] -= tempmx[j][i]*mxinverse[i][0];
         mxinverse[j][1] -= tempmx[j][i]*mxinverse[i][1];
         mxinverse[j][2] -= tempmx[j][i]*mxinverse[i][2];
         mxinverse[j][3] -= tempmx[j][i]*mxinverse[i][3];
      }

   }

   for (i=3; i>-1; i--)
      for (j=i-1; j>-1; j--) {
         mxinverse[j][0] -= tempmx[j][i]*mxinverse[i][0];
         mxinverse[j][1] -= tempmx[j][i]*mxinverse[i][1];
         mxinverse[j][2] -= tempmx[j][i]*mxinverse[i][2];
         mxinverse[j][3] -= tempmx[j][i]*mxinverse[i][3];
      }

   return 1;
}


/* *************************************************************
   this function calculates the current value of a surface patch
   spline - s is dy, t is dx
************************************************************* */
void sppatch(float s, float t, float in[][4], float out[], int ptr[], int stype) {

   vector4f T, S, temp;
   vector4f Q[4][4];
   int  i, j, k, l;

   for (j=0; j<4; j++)
      for (k=0; k<4; k++)
         for (l=0, i=ptr[k]; l<4; l++, i+=4)
            Q[j][k][l] = in[i][j];

   T[3] = 1;
   T[2] = t;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = S[2]*S[2];
   S[0] = S[1]*S[2];

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);
   }

   for (j=0; j<4; j++) {
      for (k=0; k<4; k++)
         temp[k] = dotproduct4(Q[j][k], T);
      out[j] = dotproduct4(S, temp);
   }

}


/* *************************************************************
   this function calculates the current value of a surface patch
   spline - s is dy, t is dx
************************************************************* */
void sppatch(float s, float t, vector4f **in, float *out, int stype) {

   vector4f T, S, temp;
   vector4f Q[4][4];
   int  j, k, l;

   for (j=0; j<4; j++)
      for (k=0; k<4; k++)
         for (l=0; l<4; l++)
            Q[j][k][l] = in[k][l][j];

   T[3] = 1;
   T[2] = t;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = S[2]*S[2];
   S[0] = S[1]*S[2];

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);
   }

   for (j=0; j<4; j++) {
      for (k=0; k<4; k++)
         temp[k] = dotproduct4(Q[j][k], T);

      out[j] = dotproduct4(S, temp);
   }

}


/* *************************************************************
   this function calculates the current value of a surface patch
   spline - s is dy, t is dx
************************************************************* */
void sppatch(float s, float t, float Q[][4][4], float *out, int stype) {

   vector4f T, S, temp;
   int  j, k;

   T[3] = 1;
   T[2] = t;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = S[2]*S[2];
   S[0] = S[1]*S[2];

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);
   }

   for (j=0; j<4; j++) {
      for (k=0; k<4; k++)
         temp[k] = dotproduct4(Q[j][k], T);

      out[j] = dotproduct4(S, temp);
   }

}
//


/* *************************************************************
   This function calculates the patch normals
************************************************************* */
void spnormal(float s, float t, float in[][4], float out[], int ptr[], int stype) {

   vector4f ds, dt;
   vector4f S, T;
   vector4f Q[4][4];
   vector4f temp1, temp2;
   vector3f pt1, pt2;
   int   i, j, k, l;

   for (j=0; j<4; j++)
      for (k=0; k<4; k++)
         for (l=0, i=ptr[k]; l<4; l++, i+=4)
            Q[j][k][l] = in[i][j];

   T[3] = 1;
   T[2] = t;
   T[1] = SQR(T[2]);
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = SQR(S[2]);
   S[0] = S[1]*S[2];

   ds[3] = 0;
   ds[2] = 1;
   ds[1] = double(s);
   ds[0] = 3*SQR(s);

   dt[3] = 0;
   dt[2] = 1;
   dt[1] = double(t);
   dt[0] = 3*SQR(t);

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);

      matvecmult(MCRT, dt);
      vecmatmult(MCR, ds);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);

      matvecmult(MBT, dt);
      vecmatmult(MB, ds);
   }

   for (i=0; i<3; i++) {
      for (j=0; j<4; j++) {
         temp1[j] = dotproduct4(Q[i][j], T);
         temp2[j] = dotproduct4(Q[i][j], dt);
      }

      pt1[i] = dotproduct4(ds, temp1);
      pt2[i] = dotproduct4(S, temp2);
   }

   xproduct(out, pt1, pt2);
   normalize3(out);
}


/* *************************************************************
   This function calculates the patch normals
************************************************************* */
void spnormal(float s, float t, vector4f **in, float *out, int stype) {

   vector4f ds, dt;
   vector4f S, T;
   vector4f Q[4][4];
   vector4f temp1, temp2;
   vector3f pt1, pt2;
   int j, k, l;

   for (j=0; j<4; j++)
      for (k=0; k<4; k++)
         for (l=0; l<4; l++)
            Q[j][k][l] = in[k][l][j];

   T[3] = 1;
   T[2] = t;
   T[1] = SQR(t);
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = SQR(s);
   S[0] = S[1]*S[2];

   ds[3] = 0;
   ds[2] = 1;
   ds[1] = double(s);
   ds[0] = 3*S[1];

   dt[3] = 0;
   dt[2] = 1;
   dt[1] = double(t);
   dt[0] = 3*T[1];

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);

      matvecmult(MCRT, dt);
      vecmatmult(MCR, ds);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);

      matvecmult(MBT, dt);
      vecmatmult(MB, ds);
   }

   for (k=0; k<3; k++) {
      for (j=0; j<4; j++) {
         temp1[j] = dotproduct4(Q[k][j], T);
         temp2[j] = dotproduct4(Q[k][j], dt);
      }

      pt1[k] = dotproduct4(ds, temp1);
      pt2[k] = dotproduct4(S, temp2);
   }

   xproduct(out, pt1, pt2);
   normalize3(out);
}


/* *************************************************************
   This function calculates the patch normals
************************************************************* */
void spnormal(float s, float t, float Q[][4][4], float *out, int stype) {

   vector4f ds, dt;
   vector4f S, T;
   vector4f temp1, temp2;
   vector3f pt1, pt2;
   int j, k;

   T[3] = 1;
   T[2] = t;
   T[1] = SQR(t);
   T[0] = T[1]*T[2];

   S[3] = 1;
   S[2] = s;
   S[1] = SQR(s);
   S[0] = S[1]*S[2];

   ds[3] = 0;
   ds[2] = 1;
   ds[1] = double(s);
   ds[0] = 3*S[1];

   dt[3] = 0;
   dt[2] = 1;
   dt[1] = double(t);
   dt[0] = 3*T[1];

   if (stype == M_CR) {
      matvecmult(MCRT, T);
      vecmatmult(MCR, S);

      matvecmult(MCRT, dt);
      vecmatmult(MCR, ds);
   }

   else {
      matvecmult(MBT, T);
      vecmatmult(MB, S);

      matvecmult(MBT, dt);
      vecmatmult(MB, ds);
   }

   for (k=0; k<3; k++) {
      for (j=0; j<4; j++) {
         temp1[j] = dotproduct4(Q[k][j], T);
         temp2[j] = dotproduct4(Q[k][j], dt);
      }

      pt1[k] = dotproduct4(ds, temp1);
      pt2[k] = dotproduct4(S, temp2);
   }

   xproduct(out, pt1, pt2);
   normalize3(out);
}


/* *************************************************************
   This function calculates the normal at s = t = 0
************************************************************* */
void spnormal_base(vector4f **in, float out[], int stype) {

   vector4f S, T;
   vector4f Q[4][4];
   vector4f temp1, temp2;
   vector3f pt1, pt2;
   int   i, j, k, l;

   if (stype == M_CR)
      for (i=0; i<3; i++) {
         pt1[i] = -in[0][1][i] + in[2][1][i];
         pt2[i] = -in[1][0][i] + in[1][2][i];
      }

   else {
      for (j=0; j<4; j++)
         for (k=0; k<4; k++)
            for (l=0; l<4; l++)
               Q[j][k][l] = in[k][l][j];

      S[0] = T[0] = MB[3][0];
      S[0] = T[1] = MB[3][1];
      S[0] = T[2] = MB[3][2];

      for (i=0; i<3; i++) {
         for (j=0; j<4; j++) {
            temp1[j] = dotproduct3(Q[i][j], T);
            temp2[j] = 0.5*(-Q[i][j][0] + Q[i][j][2]);
         }

         pt1[i] = 0.5*(-temp1[0] + temp1[2]);
         pt2[i] = dotproduct3(S, temp2);
      }

   }

   xproduct(out, pt1, pt2);
   normalize3(out);
}


/* *************************************************************
   This procedure calculates the current placement of objects
   using Catmull-Rom or B splines
************************************************************* */
void spline(float s, float in[][4], float out[], int i, int stype) {

   vector4f Q[4];
   vector4f T;
   int j, k;

   for (j=0; j<4; j++)
      for (k=0; k<4; k++)
         Q[j][k] = in[i+k][j];

   T[3] = 1;
   T[2] = s;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   if (stype == M_CR)				// Catmull-Rom spline
      for (j=0; j<4; j++) {
         matvecmult(MCR, Q[j]);
         out[j] = dotproduct4(T, Q[j]);
      }

   else						// B spline
      for (j=0; j<4; j++) {
         matvecmult(MB, Q[j]);
         out[j] = dotproduct4(T, Q[j]);
      }

}


/* *************************************************************
   This procedure calculates the current placement of objects
   using Catmull-Rom or B splines
************************************************************* */
void splinet(float s, vector4f *in, float *out, int stype) {

   vector4f Q[4];
   vector4f T;
   int j;

   transpose(Q, in);

   T[3] = 1;
   T[2] = s;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   if (stype == M_CR)				// Catmull-Rom spline
      for (j=0; j<4; j++) {
         matvecmult(MCR, Q[j]);
         out[j] = dotproduct4(T, Q[j]);
      }

   else						// B spline
      for (j=0; j<4; j++) {
         matvecmult(MB, Q[j]);
         out[j] = dotproduct4(T, Q[j]);
      }

}


/* *************************************************************
   This procedure calculates the current placement of objects
   using Catmull-Rom or B splines
************************************************************* */
void spline(float s, float Q[][4], float *out, int stype) {

   vector4f T, temp;
   int j;

   T[3] = 1;
   T[2] = s;
   T[1] = T[2]*T[2];
   T[0] = T[1]*T[2];

   if (stype == M_CR)
      for (j=0; j<4; j++) {
         copyarray4(temp, Q[j]);
         matvecmult(MCR, temp);
         out[j] = dotproduct4(T, temp);
      }

   else
      for (j=0; j<4; j++) {
         copyarray4(temp, Q[j]);
         matvecmult(MB, temp);
         out[j] = dotproduct4(T, temp);
      }

}


/* *************************************************************
   Quaternion multiplication
************************************************************* */
void quatquatmult(float y[], float x[]) {

   float x1, y1;                       // real part of quaternion
   vector3f x24, y24;               // complex part of quaternion
   int    i;                            // looping var

   x1 = x[0];
   y1 = y[0];

   for (i=0; i<3; i++) {
      x24[i] = x[i+1];
      y24[i] = y[i+1];
   }

   x[0] = y1*x1 - dotproduct3(y24, x24);         // calc real part

   xproduct(&x[1], y24, x24);                   // calc complex part

   for (i=0; i<3; i++)
      x[i+1] += y1*x24[i] + x1*y24[i];
}
//
/* *************************************************************
   convert euler to quaternion
************************************************************* */
void euler2quat(float angle[], float quat[]) {

   vector3f s, c;
   int i;
   float temp;

   for (i=0; i<3; i++) {
      temp = -angle[i]*0.5;
      s[i] = SIN(temp);
      c[i] = COS(temp);
   }

   quat[0] = c[0]*c[1]*c[2] + s[0]*s[1]*s[2];
   quat[1] = s[0]*c[1]*c[2] - c[0]*s[1]*s[2];
   quat[2] = c[0]*s[1]*c[2] + s[0]*c[1]*s[2];
   quat[3] = c[0]*c[1]*s[2] - s[0]*s[1]*c[2];
}


/* *************************************************************
   calculate euler rotation
************************************************************* */
void calc_euler(float mx[][4], float current[]) {

   vector4f temp[4];
   int    i, j;

   init_mx(i, j, mx);
   mx[2][2] = mx[1][1] = COS(current[0]);          // x
   mx[1][2] = -(mx[2][1] = SIN(current[0]));

   init_mx(i, j, temp);
   temp[2][2] = temp[0][0] = COS(current[1]);        // y
   temp[2][0] = -(temp[0][2] = SIN(current[1]));
   matmatmulto(temp, mx);

   init_mx(i, j, temp);
   temp[1][1] = temp[0][0] = COS(current[2]);        // z
   temp[0][1] = -(temp[1][0] = SIN(current[2]));
   matmatmulto(temp, mx);
}
//
/* *************************************************************
   calculate quaternion mx
************************************************************* */
void calc_quat(float mx[][4], float current[]) {

   vector4f vect, sqr;
   register float temp;

   copyarray4(vect, current);
   normalize4(vect);

//   sqr[0] = SQR(vect[0]);
   sqr[1] = SQR(vect[1]);
   sqr[2] = SQR(vect[2]);
   sqr[3] = SQR(vect[3]);

   temp = sqr[2] + sqr[3];
   mx[0][0] = 1 - (temp + temp);
   temp = vect[1]*vect[2] + vect[0]*vect[3];
   mx[0][1] = temp + temp;
   temp = vect[1]*vect[3] - vect[0]*vect[2];
   mx[0][2] = temp + temp;
   mx[0][3] = 0;

   temp = vect[1]*vect[2] - vect[0]*vect[3];
   mx[1][0] = temp + temp;
   temp = sqr[1] + sqr[3];
   mx[1][1] = 1 - (temp + temp);
   temp = vect[2]*vect[3] + vect[0]*vect[1];
   mx[1][2] = temp + temp;
   mx[1][3] = 0;

   temp = vect[1]*vect[3] + vect[0]*vect[2];
   mx[2][0] = temp + temp;
   temp = vect[2]*vect[3] - vect[0]*vect[1];
   mx[2][1] = temp + temp;
   temp = sqr[1] + sqr[2];
   mx[2][2] = 1 - (temp + temp);
   mx[2][3] = 0;

   mx[3][0] = mx[3][1] = mx[3][2] = 0;
   mx[3][3] = 1;
}

/* *************************************************************
	roll -	x	axis from back -> front
	pitch -	y	axis from 9 oclock -> 3 oclock (looking from top)
	yaw -	z	axis represents height of object

	mx[	 cpcy,		-cpsy,		 sp,
		 srspcy+cpsy	-srspsy+cpcy	-srcp	
		-crspcy+spsy	 crspsy+spcy	 crcp
	   ]

	Assumption: matrix is a pure rotation matrix
************************************************************* */
void decompose_mx(vector4f mx[4], float *roll, float *pitch, float *yaw) {

   double temp, tempc, temps;

   if (abs(mx[0][2]) > 1-CORRECT) {
      if (mx[0][2] > 0)				// pitch = PI/2
         *pitch = HALFPI;
      else
         *pitch = -HALFPI;			// pitch = -PI/2

      *yaw = 0;
      temp = 1.0/COS(*pitch);

      if ((tempc = mx[0][1]*temp) > 1-CORRECT)
         *roll = 0;
      else
         if (tempc < CORRECT-1)
            *roll = PI;
         else
            if ((temps = mx[0][0]*temp) > 1-CORRECT)
               *roll = HALFPI;
            else
               if (temps < CORRECT-1)
                  *roll = PI + HALFPI;
               else
                  *roll = atan2(temps, tempc);

      return;
   }

   *pitch = ASIN(mx[0][2]);
   temp = 1.0/COS(*pitch);

   if ((tempc = mx[0][0]*temp) > 1-CORRECT)
      *yaw = 0;
   else
      if  (tempc < CORRECT-1)
         *yaw = PI;
      else
         if ((temps = -mx[0][1]*temp) > 1-CORRECT)
            *yaw = HALFPI;
         else
            if (temps < CORRECT-1)
               *yaw = PI + HALFPI;
            else
               *yaw = atan2(temps, tempc);

   if ((tempc = mx[2][2]*temp) > 1-CORRECT)
      *roll = 0;
   else
      if (tempc < CORRECT-1)
         *roll = PI;
      else
         if ((temps = -mx[1][2]*temp) > 1-CORRECT)
            *roll = HALFPI;
         else
            if (temps < CORRECT-1)
               *roll = PI + HALFPI;
            else
               *roll = atan2(temps, tempc);
}

/* *************************************************************
************************************************************* */
void rotate_mx_x(vector4f mx[4], float x) {

   vector4f rot[4];
   int i, j;

   init_mx(i, j, rot);

   rot[1][1] = rot[2][2] = COS(x);
   rot[1][2] = -(rot[2][1] = SIN(x));

   matmatmult(rot, mx);
}


/* *************************************************************
************************************************************* */
void rotate_mx_y(vector4f mx[4], float x) {

   vector4f rot[4];
   int i, j;

   init_mx(i, j, rot);

   rot[0][0] = rot[2][2] = COS(x);
   rot[2][0] = -(rot[0][2] = SIN(x));

   matmatmult(rot, mx);
}


/* *************************************************************
************************************************************* */
void rotate_mx_z(vector4f mx[4], float x) {

   vector4f rot[4];
   int i, j;

   init_mx(i, j, rot);

   rot[0][0] = rot[1][1] = COS(x);
   rot[0][1] = -(rot[1][0] = SIN(x));

   matmatmult(rot, mx);
}


/* *************************************************************
************************************************************* */
float cylinder_area(float radius, float length) {

   register float temp;

   temp = PI*radius*length;
   return temp + temp;
}


/* *************************************************************
************************************************************* */
void make_box(vector4f *in, vector4f *out) {

   out[0][0] = in[0][0];
   out[0][1] = in[0][1];
   out[0][2] = in[1][2];
   out[0][3] = 1;

   out[1][0] = in[0][0];
   out[1][1] = in[1][1];
   out[1][2] = in[1][2];
   out[1][3] = 1;

   out[2][0] = in[1][0];
   out[2][1] = in[1][1];
   out[2][2] = in[1][2];
   out[2][3] = 1;

   out[3][0] = in[1][0];
   out[3][1] = in[0][1];
   out[3][2] = in[1][2];
   out[3][3] = 1;

   out[4][0] = in[0][0];
   out[4][1] = in[0][1];
   out[4][2] = in[0][2];
   out[4][3] = 1;

   out[5][0] = in[0][0];
   out[5][1] = in[1][1];
   out[5][2] = in[0][2];
   out[5][3] = 1;

   out[6][0] = in[1][0];
   out[6][1] = in[1][1];
   out[6][2] = in[0][2];
   out[6][3] = 1;

   out[7][0] = in[1][0];
   out[7][1] = in[0][1];
   out[7][2] = in[0][2];
   out[7][3] = 1;
}


float *cos_table = (float *)NULL, *sin_table = (float *)NULL, *acos_table = (float *)NULL, *asin_table = (float *)NULL;
double trigfactor = MAX_TRIG_SIZE/TWOPI;
double atrigfactor = MAX_TRIG_SIZE*0.5;

void init_trig(int fmask) {

   int i;
   float temp;

   if (fmask & MASK_COS) {
      if (cos_table != (float *)NULL)
         delete cos_table;

      cos_table = new float[MAX_TRIG_SIZEPLUS];

      for (i=0 ; i <MAX_TRIG_SIZEPLUS; i++) {
         temp = (i*TWOPI)/0xffff;
         cos_table[i] = cos(temp);
      }

   }

   if (fmask & MASK_SIN) {
      if (sin_table != (float *)NULL)
         delete sin_table;

      sin_table = new float[MAX_TRIG_SIZEPLUS];

      for (i=0 ; i <MAX_TRIG_SIZEPLUS; i++) {
         temp = (i*TWOPI)/0xffff;
         sin_table[i] = sin(temp);
      }

   }

   if (fmask & MASK_ACOS) {
      if (acos_table != (float *)NULL)
         delete acos_table;

      acos_table = new float[MAX_TRIG_SIZEPLUS];

      for (i=0 ; i <MAX_TRIG_SIZEPLUS; i++) {
         temp = (((float)(i+i))/0xffff) - 1;
         acos_table[i] = acos(temp);
      }

   }

   if (fmask & MASK_ASIN) {
      if (asin_table != (float *)NULL)
         delete asin_table;

      asin_table = new float[MAX_TRIG_SIZEPLUS];

      for (i=0 ; i <MAX_TRIG_SIZEPLUS; i++) {
         temp = (((float)(i+i))/0xffff) - 1;
         asin_table[i] = asin(temp);
      }

   }

}
