
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file handles all the matrix/vector/math operations
************************************************************* */

#if !defined (__MATH_H)
#define __MATH_H

#include <stdlib.h>
#include <math.h>

#define TWOPI2D    6.28

#ifndef PI
#define PI         3.14159265358979323846
#endif

#define TWOPI      6.28318530717958647692
#define HALFPI     1.57079632679489661923
#define ANGLE2RAD  0.01745329251994329577
#define LOG2E      1.4426950408889634074
#define LOG2(x)    (LOG2E * log(x))	// nat. log to log 2
#define rad2deg(x) ((x)*57.29577951)	// times 180.0/PI
#define deg2rad(x) ((x)*ANGLE2RAD)	// times 180.0/PI
#define CORRECT    0.00001              // accuracy limit
#define CORRECT2   0.01                 // accuracy limit
#define CORRECT3   0.001                // accuracy limit
#define CORRECT4   0.0001               // accuracy limit
#define X          0
#define Y          1
#define Z          2

#define CYL_TOP		0
#define CYL_BOTTOM	1
#define CYL_SIDE	3

#define MASK_COS	0x0001
#define MASK_SIN	0x0010
#define MASK_ACOS	0x0100
#define MASK_ASIN	0x1000

#define SQR(x) ((x)*(x))
#define rnd(x) ((x)*(rand() / (float)((1<<15)-1)))
#define randomizer(x, y) ((x) + (((y)-(x))*(rand() / (float)((1<<15)-1)))
#define round(x) ((int)((x) + 0.5))
#define round0(x) (((x)<0) ? (-(int)(-x)) : ((int)(x)))

#ifndef sgn
#define sgn(x)   ( ((x)<0) ? -1 : ((x) ? 1:0) )
#endif

#define abs(x) (((x)<0) ? -(x):(x))
#define not_zero(x) (((x)>0) ? (int)((x)+1-CORRECT):(int)((x)-1+CORRECT))
#define not_zero3(x) (((x)>0) ? (int)((x)+1-CORRECT2):(int)((x)-1+CORRECT2))
#define double(x) ((x) + (x))

#define dotproduct3(x, y) (x[0]*y[0] + x[1]*y[1] + x[2]*y[2])
#define dotproduct4(x, y) (x[0]*y[0] + x[1]*y[1] + x[2]*y[2] + x[3]*y[3])
#define magnitude2(x) sqrt(SQR(x[0]) + SQR(x[1]))
#define magnitude3(x) sqrt(SQR(x[0]) + SQR(x[1]) + SQR(x[2]))
#define magnitude4(x) sqrt(SQR(x[0]) + SQR(x[1]) + SQR(x[2]) + SQR(x[3]))
#define distance3(x,y) sqrt(SQR(x[0]-y[0]) +SQR(x[1]-y[1]) + SQR(x[2]-y[2]))

#define copyarray2(x, y) {x[0] = y[0]; x[1] = y[1];}
#define copyarray3(x, y) {copyarray2(x, y); x[2] = y[2];}
#define copyarray4(x, y) {copyarray3(x, y); x[3] = y[3];}

#define addarray2(x, y) {x[0] += y[0]; x[1] += y[1];}
#define addarray3(x, y) {x[0] += y[0]; x[1] += y[1]; x[2] += y[2];}
#define addarray4(x, y) {x[0] += y[0]; x[1] += y[1]; x[2] += y[2]; x[3] += y[3];}

#define subarray2(x, y) {x[0] -= y[0]; x[1] -= y[1]; }
#define subarray3(x, y) {x[0] -= y[0]; x[1] -= y[1]; x[2] -= y[2];}

#define multarray3(x, y) {x[0] *= y[0]; x[1] *= y[1]; x[2] *= y[2];}

#define divarray3(x, y) {x[0] /= y[0]; x[1] /= y[1]; x[2] /= y[2];}

#define saddarray2(x,y)  {x[0] += y; x[1] += y;}
#define saddarray3(x,y)  {x[0] += y; x[1] += y; x[2] += y;}

#define ssubarray3(x,y)  {x[0] -= y; x[1] -= y; x[2] -= y;}

#define smultarray3(x,y) {x[0] *= y; x[1] *= y; x[2] *= y;}
#define smultarray4(x,y) {x[0] *= y; x[1] *= y; x[2] *= y; x[3] *= y;}

#define sdivarray3(x,y)  {x[0] /= y; x[1] /= y; x[2] /= y;}
#define sdivarray4(x,y)  {x[0] /= y; x[1] /= y; x[2] /= y; x[3] /= y;}

#define nte_vector3(x, y) (x[0]!=y[0] || x[1]!=y[1] || x[2]!=y[2])
#define nre_vector3(x, y) (abs(x[0]-y[0])<=CORRECT && abs(x[1]-y[1])<=CORRECT && abs(x[2]-y[2])<=CORRECT)
#define nnre_vector3(x, y) (abs(x[0]-y[0]) > CORRECT || abs(x[1]-y[1]) > CORRECT || abs(x[2]-y[2]) > CORRECT)

#define copymx4x4(x, y) { copyarray4(x[0], y[0]); copyarray4(x[1], y[1]); copyarray4(x[2], y[2]);  copyarray4(x[3], y[3]); }
#define copymx3x3(x, y) { copyarray3(x[0], y[0]); copyarray3(x[1], y[1]); copyarray3(x[2], y[2]); }
#define init_mx(i, j, x) for (i=0; i<4; i++) for (j=0; j<4; j++) x[i][j] = (i==j)
#define init_mx3x3(x)    { x[0][0] = x[1][1] = x[2][2] = 1; x[0][1] = x[0][2] = x[1][0] = x[1][2] = x[2][0] = x[2][1] = 0; }
#define print_mx(i, x) for (i=0; i<4; i++) printf("%f %f %f %f\n", x[i][0], x[i][1], x[i][2], x[i][3])

#define step(a, x) ((x) >= (a))
#define clamp(x, a, b) ( (x) < (a) ? (a) : ((x) > (b) ? (b) : (x) )
#define getgamma(gamma) (1.0/(gamma))
#define gammacorrect(gamma, x) ((x) < 0 ? 0 : powf(x, gamma))

typedef float vector2f[2];
typedef vector2f *pvector2f;
typedef float vector3f[3];
typedef vector3f *pvector3f;
typedef float vector4f[4];
typedef vector4f *pvector4f;

void   extend_vector(float x[], float dist);
float  normalize2(float x[]);
float  normalize3(float x[]);
float  normalize4(float x[]);
void   swap4(float mx1[], float mx2[]);
void   xproduct(float *a, float *x, float *y);
#define xproduct2(x, y) (x[0]*y[1] - x[1]*y[0])
void   transpose(float r[][4], float s[][4]);
void   transpose(float r[][4]);

void   matmatmult(float y[][4], float x[][4]);
void   matvecmult(float y[][4], float x[]);
void   vecmatmult(float y[][4], float x[]);

void   matmatmulto(float y[][4], float x[][4]);
void   matvecmultv(float y[][4], float x[]);
void   vecmatmultv(float y[][4], float x[]);
void   matvecmulto(float y[][4], float x[]);
void   vecmatmulto(float y[][4], float x[]);

int    inversemx(float a[][4],float b[][4]);
int    inversemx2(float mx[][4], float mxinverse[][4]);
void   inversemxrt(float mx[][4], float mxinverse[][4]);

void calc_quat(float mx[][4], float current[]);
void calc_euler(float mx[][4], float current[]);
void euler2quat(float angle[], float quat[]);
void quatquatmult(float y[], float x[]);

#define M_CR	0
#define M_B	1

void spline(float s, float in[][4], float out[], int i, int stype);
void spline(float s, vector4f *in, float *out, int stype);
void splinet(float s, float Q[][4], float *out, int stype);

void spnormal(float s, float t, float in[][4], float out[], int ptr[], int stype);
void spnormal(float s, float t, vector4f **in, float *out, int stype);
void spnormal(float s, float t, float Q[][4][4], float *out, int stype);

void spnormal_base(vector4f **in, float out[], int stype);

void sppatch(float s, float t, float **in, float out[], int stype);
void sppatch(float s, float t, vector4f **in, float *out, int stype);
void sppatch(float s, float t, float Q[][4][4], float *out, int stype);

void decompose_mx(vector4f mx[4], float *roll, float *pitch, float *yaw);
void rotate_mx_x(vector4f mx[4], float x);
void rotate_mx_y(vector4f mx[4], float x);
void rotate_mx_z(vector4f mx[4], float x);

#define cone_area(radius, length) (PI*(radius)*sqrt(SQR(radius) + SQR(length)))
float cylinder_area(float radius, float length);
#define sphere_area(radius) (4*PI*SQR(radius))

// misc
#define dist_pt2plane(pt, plane) (dotproduct3(pt, plane) + plane[3])

void make_box(vector4f *in, vector4f *out);

#define MAX_TRIG_SIZE		0xffff
#define MAX_TRIG_SIZEPLUS	0x10000

extern float *cos_table, *sin_table, *acos_table, *asin_table;
extern double trigfactor, atrigfactor;

#define COS(x) (cos_table[((int)(abs(x)*trigfactor)) & MAX_TRIG_SIZE])
#define SIN(x) (((x)<0) ? (-sin_table[((int)(-x*trigfactor)) & MAX_TRIG_SIZE]) : \
			sin_table[((int)(x*trigfactor)) & MAX_TRIG_SIZE])
#define ACOS(x) (acos_table[((int)((x+1)*atrigfactor)) & MAX_TRIG_SIZE])
#define ASIN(x) (asin_table[((int)((x+1)*atrigfactor)) & MAX_TRIG_SIZE])

void init_trig(int fmask = 0x1111);


// collision stuff

#define query_point_point_intersect(pt1, pt2) (abs(pt1[0]-pt2[0])<CORRECT && abs(pt1[1]-pt2[1])<CORRECT && abs(pt1[2]-pt2[2])<CORRECT)
int query_point_sphere_intersect(float *pt, float *center, float r);
int query_point_poly_intersect(float *pt, float *normal, int count, vector4f *vertex);
int query_point_plane_intersect(float *pt, float *n);
int query_point_rect_intersect(float *pt, vector4f *n);

int query_sphere_sphere_intersect(float *c1, float r1, float *c2, float r2);
int query_sphere_plane_intersect(float *c, float r, float *n);
int query_sphere_rect_intersect(float *c, float r, vector4f *n);

#define query_plane_plane_intersect(n1, n2) (abs(n1[0]-n2[0]) < CORRECT && abs(n1[1]-n2[1]) < CORRECT && abs(n1[2]-n2[2]) < CORRECT)
int query_plane_rect(float *n, float *pt1, float *pt7);

int query_rect_rect(vector4f *n1, vector4f *p1, vector4f *n2, vector4f *p2);

int line_rect_intersect(float *pt, float *v, vector4f *p, vector4f *n, float *flag);
int line_plane_intersect(float *plane, float *pt, float *vector, float *t);
int line_sphere_intersect(float radius, float *sphere, float *pt, float *vector, float *t);
int line_ellipse_intersect(float *scale, float *pt, float *vector, float *t);
int line_poly_intersect(float *pt, float *vector, float *plane, vector4f *vertex, int count, float *point);
int line2d_line2d_intersect(float *v1, float *v2, float *pt);

void   reflect(float normal[], float vector[]);

int point_sphere_reflectn3vf(float *pt, float *center, float *normal);
int sphere_sphere_reflectn3vf(float *c1, float *c2, float *normal);
int sphere_plane_reflectn3vf(float *n, float *normal);
int sphere_rect_reflectn3vf(float *c, vector4f *p, vector4f *n, float *normal);

int line_cylinder_intersect(vector4f pt, vector4f v, float length, float radius, float *t, int *side);
int line_cone_intersect(vector4f pt, vector4f v, float length, float *radius, float *t, int *side);


// perlin stuff

#define bias(a, b) (pow((a), log(b) / (-0.693147181)))

void  init_noise();
float gain(float a, float b);
float noise1(float arg);
float noise2(float vec[2]);
float noise3(float vec[3]);
float noise(float vec[], int len);
float turbulence(float *v, float freq);
float fBm( float *point, float H, float lacunarity, float octaves, int rebuild, float **freq );
float fBm2( float *point, float H, float lacunarity, float octaves, int rebuild, float **freq );
float multifractal( float *point, float H, float lacunarity,
                    float octaves, float offset, int rebuild, float **freq);
float Hetero_Terrain( float *point, float H, float lacunarity, float octaves,
                float offset, int rebuild, float **freq );
float HybridMultifractal( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain, int rebuild, float **freq);
float RidgedMultifractal( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain,  int rebuild, float **freq);
float RidgedMultifractal2( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain,  int rebuild, float **freq);

#endif
