
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "matrix.h"

#define B  0x100
#define BSTRING "perlin.4"
#define BM 0xff

#define N  0x1000
#define NP 12   /* 2^N */
#define NM 0xfff

#ifdef REAL_FAST

//	linear interpolation from 0..1
#define s_curve(t) (t)

#else

//	slerp interpolation from 0..1
#define s_curve(t) ( SQR(t) * (3. - ((t) + (t))) )

#endif

#define lerp(t, a, b) ( (a) + (t) * ((b) - (a)) )

#define setup(i,b0,b1,r0,r1)\
        t = vec[i] + N;\
        b0 = ((int)t) & BM;\
        b1 = (b0+1) & BM;\
        r0 = t - (int)t;\
        r1 = r0 - 1.;


static int     p[B + B + 2];
static vector3f g3[B + B + 2];		// need g1, g2, g3 cause they are normalized!!!!
static vector2f g2[B + B + 2];
static float    g1[B + B + 2];
static int      start = 1;

float gain(float a, float b) {

   if (a < 0.5) {
      if (a < .001)
         return 0.;

      return 0.5*bias(a + a, 1-b);
   }

   if (a > .999)
      return 1.;

   return 1. - 0.5*bias(2 - (a + a), 1-b);
}


void init_noise(void) {

   int i, j, k;
   FILE *iofile;
   int db = B+B;
   float b;

   start = 0;

   if ((iofile = fopen(BSTRING, "r")) != NULL) {

      k = db + 2;
      for (i=0; i<k; i++)
         fscanf(iofile, "%f %f %f", &g3[i][0], &g3[i][1], &g3[i][2]);
      for (i=0; i<k; i++)
         fscanf(iofile, "%f %f", &g2[i][0], &g2[i][1]);
      for (i=0; i<k; i++)
         fscanf(iofile, "%f", &g1[i]);
      for (i=0; i<k; i++)
         fscanf(iofile, "%d", &p[i]);

      fclose(iofile);
      return;
   }

   b = 1.0/B;

   for (i=0; i<B; i++) {
      p[i] = i;

      g1[i] = g2[i][0] = g3[i][0] = (float)((random() % db) - B);
      g2[i][1] = g3[i][1] = (float)((random() % db) - B);
      g3[i][2] = (float)((random() % db) - B);
      normalize3(g3[i]);
      normalize2(g2[i]);
      g1[i] *= b;
   }

   while (--i) {
      k = p[i];
      p[i] = p[j = random() % B];
      p[j] = k;
   }

   for (i=0, k=B+2, j=B; i < k; i++, j++) {
      p[j] = p[i];
      copyarray3(g3[j], g3[i]);
      copyarray2(g2[j], g2[i]);
      g1[j] = g1[i];
   }

   iofile = fopen(BSTRING, "w");        // save lattice

   k = db + 2;
   for (i=0; i<k; i++)
      fprintf(iofile, "%f %f %f\n", g3[i][0], g3[i][1], g3[i][2]);
   for (i=0; i<k; i++)
      fprintf(iofile, "%f %f\n", g2[i][0], g2[i][1]);
   for (i=0; i<k; i++)
      fprintf(iofile, "%f\n", g1[i]);
   for (i=0; i<k; i++)
      fprintf(iofile, "%d\n", p[i]);

   fclose(iofile);
}


float noise1(float arg) {

   int bx0, bx1;
   float rx0, rx1, sx, t, u, v, vec[1];

   vec[0] = arg;

   if (start)
      init_noise();

   setup(0, bx0,bx1, rx0,rx1);

   sx = s_curve(rx0);

   u = rx0 * g1[ p[ bx0 ] ];
   v = rx1 * g1[ p[ bx1 ] ];

   return lerp(sx, u, v);
}


float noise2(float vec[2]) {

   int bx0, bx1, by0, by1, b00, b10, b01, b11;
   float rx0, rx1, ry0, ry1, *q, sx, sy, a, b, t, u, v;
   register int i, j;

   if (start)
      init_noise();

   setup(0, bx0,bx1, rx0,rx1);
   setup(1, by0,by1, ry0,ry1);

   i = p[ bx0 ];
   j = p[ bx1 ];

   b00 = p[ i + by0 ];
   b10 = p[ j + by0 ];
   b01 = p[ i + by1 ];
   b11 = p[ j + by1 ];

   sx = s_curve(rx0);
   sy = s_curve(ry0);

#define at2(rx,ry) ( rx * q[0] + ry * q[1] )

   q = g2[ b00 ] ; u = at2(rx0,ry0);
   q = g2[ b10 ] ; v = at2(rx1,ry0);
   a = lerp(sx, u, v);

   q = g2[ b01 ] ; u = at2(rx0,ry1);
   q = g2[ b11 ] ; v = at2(rx1,ry1);
   b = lerp(sx, u, v);

   return lerp(sy, a, b);
}

float noise3(float vec[3]) {

   int bx0, bx1, by0, by1, bz0, bz1, b00, b10, b01, b11;
   float rx0, rx1, ry0, ry1, rz0, rz1, *q, sy, sz, a, b, c, d, t, u, v;
   register int i, j;

   if (start)
      init_noise();

   setup(0, bx0,bx1, rx0,rx1);
   setup(1, by0,by1, ry0,ry1);
   setup(2, bz0,bz1, rz0,rz1);

   i = p[ bx0 ];
   j = p[ bx1 ];

   b00 = p[ i + by0 ];
   b10 = p[ j + by0 ];
   b01 = p[ i + by1 ];
   b11 = p[ j + by1 ];

   t  = s_curve(rx0);
   sy = s_curve(ry0);
   sz = s_curve(rz0);

#define at3(rx,ry,rz) ( rx * q[0] + ry * q[1] + rz * q[2] )

   q = g3[ b00 + bz0 ] ; u = at3(rx0,ry0,rz0);
   q = g3[ b10 + bz0 ] ; v = at3(rx1,ry0,rz0);
   a = lerp(t, u, v);

   q = g3[ b01 + bz0 ] ; u = at3(rx0,ry1,rz0);
   q = g3[ b11 + bz0 ] ; v = at3(rx1,ry1,rz0);
   b = lerp(t, u, v);

   c = lerp(sy, a, b);

   q = g3[ b00 + bz1 ] ; u = at3(rx0,ry0,rz1);
   q = g3[ b10 + bz1 ] ; v = at3(rx1,ry0,rz1);
   a = lerp(t, u, v);

   q = g3[ b01 + bz1 ] ; u = at3(rx0,ry1,rz1);
   q = g3[ b11 + bz1 ] ; v = at3(rx1,ry1,rz1);
   b = lerp(t, u, v);

   d = lerp(sy, a, b);

   return lerp(sz, c, d);
}


float noise(float vec[], int len) {

   switch (len) {
      case 3:
         return noise3(vec);
      case 2:
         return noise2(vec);
      case 1:
         return noise1(vec[0]);
      default:
         return 0;
   }

}

// freq = high frequency, low freq defaulted to 1
float turbulence(float *v, float freq) {

   float t, vec[3];

   for (t=0; freq>=1. ; freq*=0.5) {
      vec[0] = freq * v[0];
      vec[1] = freq * v[1];
      vec[2] = freq * v[2];
      t += abs(noise3(vec)) / freq;
   }

   return t;
}


/*
 * Procedural fBm evaluated at "point"; returns value stored in "value".
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Parameters:
 *    ``H''  is the fractal increment parameter
 *    ``lacunarity''  is the gap between successive frequencies
 *    ``octaves''  is the number of frequencies in the fBm
 */
float fBm( float *point, float H, float lacunarity, float octaves, int rebuild, float **freq ) {

   float             value, frequency, remainder;
   int               i;
   float             p[3];
   int               j = (int)octaves;

   copyarray3(p, point);

   /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   value = 0.0;            /* initialize vars to proper values */

   /* inner loop of spectral construction */
   for (i=0; i<=j; i++) {
      value += noise3( p ) * (*freq)[i];
      smultarray3(p, lacunarity);
   } /* for */

   if ( ((remainder = octaves - j) > CORRECT) )      /* add in ``octaves''  remainder */
         /* ``i''  and spatial freq. are preset in loop above */
      value += remainder * noise3( p ) * (*freq)[i];

   return value;

} /* fBm() */


/*
 * Procedural fBm evaluated at "point"; returns value stored in "value".
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Parameters:
 *    ``H''  is the fractal increment parameter
 *    ``lacunarity''  is the gap between successive frequencies
 *    ``octaves''  is the number of frequencies in the fBm
 */
float fBm2( float *point, float H, float lacunarity, float octaves, int rebuild, float **freq ) {

   float             value, frequency, remainder;
   int               i;
   float             p[3];
   int               j = (int)octaves;

   copyarray3(p, point);

   /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   value = 0.0;            /* initialize vars to proper values */

   /* inner loop of spectral construction */
   for (i=0; i<=j; i++) {
      value += noise2( p ) * (*freq)[i];
      smultarray3(p, lacunarity);
   } /* for */

   if ( ((remainder = octaves - j) > CORRECT) )      /* add in ``octaves''  remainder */
         /* ``i''  and spatial freq. are preset in loop above */
      value += remainder * noise2( p ) * (*freq)[i];

   return value;

} /* fBm() */


/*
 * Procedural multifractal evaluated at "point";
 * returns value stored in "value".
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Parameters:
 *    ``H''  determines the highest fractal dimension
 *    ``lacunarity''  is gap between successive frequencies
 *    ``octaves''  is the number of frequencies in the fBm
 *    ``offset''  is the zero offset, which determines multifractality
 *      (ken sets offset to 1)
 */
float multifractal( float *point, float H, float lacunarity,
                     float octaves, float offset, int rebuild, float **freq) {

   float             value, frequency, remainder;
   int               i;
   float             p[3];
   int               j = (int)octaves;

   copyarray3(p, point);

      /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   value = 1.0;            /* initialize vars to proper values */

   /* inner loop of multifractal construction */
   for (i=0; i<=j; i++) {
//      value *= (*freq)[i] * noise3( p ) + offset;
      value += (*freq)[i] * noise3( p ) + offset;
      smultarray3(p, lacunarity);
   } /* for */

   remainder = octaves - j;
   if ( ((remainder = octaves - j) > CORRECT) )      /* add in ``octaves''  remainder */
         /* ``i''  and spatial freq. are preset in loop above */
	 // not comfortable, above mult, here add
         value += remainder * noise3( p ) * (*freq)[i] + offset;

   return value;

} /* multifractal() */


/*
 * Heterogeneous procedural terrain function: stats by altitude method.
 * Evaluated at "point"; returns value stored in "value".
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Parameters:
 *       ``H''  determines the fractal dimension of the roughest areas
 *       ``lacunarity''  is the gap between successive frequencies
 *       ``octaves''  is the number of frequencies in the fBm
 *       ``offset''  raises the terrain from `sea level'
 */
float Hetero_Terrain( float *point, float H, float lacunarity, float octaves,
                float offset, int rebuild, float **freq ) {

   float          value, increment, frequency, remainder;
   int            i;
   float          p[3];
   int            j = (int)octaves;

   copyarray3(p, point);

   /* precompute and store spectral weights, for efficiency */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   /* first unscaled octave of function; later octaves are scaled */
   value = offset + noise3( p );

   /* spectral construction inner loop, where the fractal is built */
   for (i=1; i<=j; i++) {
      /* raise spatial frequency */
      smultarray3(p, lacunarity);
      /* obtain displaced noise value */
      increment = noise3( p ) + offset;
      /* scale amplitude appropriately for this frequency */
      increment *= (*freq)[i];
      /* scale increment by current `altitude' of function */
      increment *= value;
      /* add increment to ``value''  */
      value += increment;
   } /* for */

   /* take care of remainder in ``octaves''  */
   remainder = octaves - j;
   if ( ((remainder = octaves - j) > CORRECT) ) {
      /* ``i''  and spatial freq. are preset in loop above */
      /* note that the main loop code is made shorter here */
      /* you may want to that loop more like this */
      /* raise spatial frequency */
      smultarray3(p, lacunarity);
      increment = (noise3( p ) + offset) * (*freq)[i];
      value += remainder * increment * value;
   }

   return value;

} /* Hetero_Terrain() */


/* Hybrid additive/multiplicative multifractal terrain model.
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Some good parameter values to start with:
 *
 *      H:           0.25
 *      offset:      0.7
 */
float HybridMultifractal( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain, int rebuild, float **freq) {

   float          frequency, result, signal, weight, remainder;
   int            i;
   float          p[3];
   int            j = (int)octaves;

   copyarray3(p, point);

      /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   /* get first octave of function */
   result = ( noise3( p ) + offset ) * (*freq)[0];
   weight = gain * result;

   /* spectral construction inner loop, where the fractal is built */
   for (i=1; i<=j; i++) {
      /* increase frequency */
      smultarray3(p, lacunarity);

      /* prevent divergence */
      if ( weight > 1.0 )  weight = 1.0;

      /* get next higher frequency */
      signal = ( noise3( p ) + offset ) * (*freq)[i];
      /* add it in, weighted by previous freq's local value */
      result += weight * signal;
      /* update the (monotonically decreasing) weighting value */
      /* (this is why H must specify a high fractal dimension) */
      weight *= gain * signal;
   } /* for */

   /* take care of remainder in ``octaves''  */
   remainder = octaves - j;
   if ( ((remainder = octaves - j) > CORRECT) ) {
      /* ``i''  and spatial freq. are preset in loop above */
      smultarray3(p, lacunarity);

      if ( weight > 1.0 )  weight = 1.0;
      signal = ( noise3( p ) + offset ) * (*freq)[i];
      result += remainder*weight * signal;
   }

   return result;

} /* HybridMultifractal() */


/* Ridged multifractal terrain model.
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Some good parameter values to start with:
 *
 *      H:           1.0 | 0.9
 *      offset:      1.0 | 1
 *      gain:        2.0 | 2
 */
float RidgedMultifractal( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain,  int rebuild, float **freq) {

   float          result, frequency, signal, weight;
   int            i;
   float          p[3];
   float          remainder;
   int            j = (int)octaves;

   copyarray3(p, point);

   /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   /* get first octave */
   /* get absolute value of signal (this creates the ridges) */
   /* invert and translate (note that "offset" should be ~= 1.0) */
   signal = offset - abs(noise3( p ));

   /* square the signal, to increase "sharpness" of ridges */
   signal *= signal;

   /* assign initial values */
   result = signal;
   weight = 1.0;

   for( i=1; i<=j; i++ ) {
      /* increase the frequency */
      smultarray3(p, lacunarity);

      /* weight successive contributions by previous signal */
      weight = signal * gain;
      if ( weight > 1.0 ) weight = 1.0;
      else if ( weight < 0.0 ) weight = 0.0;
      signal = offset - abs(noise3(p));
      signal *= signal;
      /* weight the contribution */
      signal *= weight;
      result += signal * (*freq)[i];
   }

   /* take care of remainder in ``octaves''  */
   remainder = octaves - j;
   if ( ((remainder = octaves - j) > CORRECT) ) {
      /* ``i''  and spatial freq. are preset in loop above */
      smultarray3(p, lacunarity);

      weight = signal * gain;
      if ( weight > 1.0 ) weight = 1.0;
      else if ( weight < 0.0 ) weight = 0.0;
      signal = offset - abs(noise3(p));
      signal *= signal;
      /* weight the contribution */
      signal *= weight;
      result += remainder*signal * (*freq)[i];
   }

   return result;

} /* RidgedMultifractal() */


/* Ridged multifractal terrain model.
 *
 * Copyright 1994 F. Kenton Musgrave
 *
 * Some good parameter values to start with:
 *
 *      H:           1.0 | 0.9
 *      offset:      1.0 | 1
 *      gain:        2.0 | 2
 */
float RidgedMultifractal2( float *point, float H, float lacunarity,
                    float octaves, float offset, float gain,  int rebuild, float **freq) {

   float          result, frequency, signal, weight;
   int            i;
   float          p[3];
   float          remainder;
   int            j = (int)octaves;

   copyarray2(p, point);

   /* precompute and store spectral weights */
   if ( rebuild ) {
      if (*freq != (float *)NULL)
         delete (*freq);

      *freq = new float[(int)octaves+1];

      frequency = 1.0;
      for (i=0; i<=octaves; i++) {
         /* compute weight for each frequency */
         (*freq)[i] = pow( frequency, -H );
         frequency *= lacunarity;
      }

   }

   /* get first octave */
   /* get absolute value of signal (this creates the ridges) */
   /* invert and translate (note that "offset" should be ~= 1.0) */
   signal = offset - abs(noise2( p ));

   /* square the signal, to increase "sharpness" of ridges */
   signal *= signal;

   /* assign initial values */
   result = signal;
   weight = 1.0;

   for( i=1; i<=j; i++ ) {
      /* increase the frequency */
      p[0] *= lacunarity;
      p[1] *= lacunarity;

      /* weight successive contributions by previous signal */
      weight = signal * gain;
      if ( weight > 1.0 ) weight = 1.0;
      else if ( weight < 0.0 ) weight = 0.0;
      signal = offset - abs(noise2(p));
      signal *= signal;
      /* weight the contribution */
      signal *= weight;
      result += signal * (*freq)[i];
   }

   /* take care of remainder in ``octaves''  */
   remainder = octaves - j;
   if ( ((remainder = octaves - j) > CORRECT) ) {
      /* ``i''  and spatial freq. are preset in loop above */
      p[0] *= lacunarity;
      p[1] *= lacunarity;

      weight = signal * gain;
      if ( weight > 1.0 ) weight = 1.0;
      else if ( weight < 0.0 ) weight = 0.0;
      signal = offset - abs(noise2(p));
      signal *= signal;
      /* weight the contribution */
      signal *= weight;
      result += remainder*signal * (*freq)[i];
   }

   return result;

} /* RidgedMultifractal2() */



