
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#include "polygon.h"
#include <string.h>


/* *************************************************************
************************************************************* */
int polygon::parse(FILE *infile, char *token) {

   int temp[3];

   switch (token[0]) {

      case 'f':
         if (!strcmp(token, "file")) {
            get_token(infile, filename);

            return 1;
         }

         break;

      case 'i':
         if (!strcmp(token, "invis")) {
            get_token(infile, token);
            temp[0] = atoi(token);
            get_token(infile, token);
            temp[1] = atoi(token);
            get_token(infile, token);
            temp[2] = atoi(token);
            inviso = temp[0] | (((temp[2] << 8) | temp[1]) << 8);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


/* *************************************************************
************************************************************* */
void polygon::preprocess(listtype *mainlist, shader *sparm) {

   read_data(filename, &mainlist->polydtr);
   pc::preprocess(mainlist, sparm);
}


/* *************************************************************
************************************************************* */
int polygon::read_data(FILE *infile) {

   return 0;
}


/* *************************************************************
   This procedure reads in the object data.  File format is:

   M                                    - # of vertices   m
   N                                    - # of polygons   n
   X1 Y1 Z1                             - x y z coordinates for vertex 1
    - - -
   XM YM ZM                             - x y z coordinates for vertex N
   C1 V1 ... Vc1                        - # of vertices in this polygon, and
    - - -                                   pointers to vertex list...
   CN VN ... VcN
************************************************************* */
int polygon::read_data(char filename[], datatype **dptr) {

   FILE *infile;                        // file pointer
   int i, j;                           // looping var

   for (dob=*dptr; dob != (datatype *)NULL; dob=dob->next)
      if (!strcmp(filename, dob->dataname))
         return dob->pdata.countobject;

   if (!(infile = fopen(filename, "r"))) {
      printf("Could not access %s....\n", filename);
      return 0;
   }

   dob       = new datatype;
   dob->next = *dptr;
   *dptr     = dob;

   strcpy(dob->dataname, filename);

   fscanf(infile, "%d %d %d", &dob->pdata.countvertex, &dob->pdata.countobject, &dob->pdata.countedge);

   dob->pdata.mvertex = new vector4f[dob->pdata.countvertex];
   dob->pdata.object  = new int[dob->pdata.countedge];
   dob->pdata.polynum = new int[dob->pdata.countobject];
   dob->pdata.edgeptr = new pint[dob->pdata.countobject];

   fscanf(infile, "%f %f %f", &dob->pdata.mvertex[0][0], &dob->pdata.mvertex[0][1],
                              &dob->pdata.mvertex[0][2]);
   dob->pdata.mvertex[0][3] = 1;

   copyarray3(dob->bbox[0], dob->pdata.mvertex[0]);
   copyarray3(dob->bbox[1], dob->pdata.mvertex[0]);

   for (i=1; i<dob->pdata.countvertex; i++) {
      fscanf(infile, "%f %f %f", &dob->pdata.mvertex[i][0], &dob->pdata.mvertex[i][1],
                                 &dob->pdata.mvertex[i][2]);
      dob->pdata.mvertex[i][3] = 1;

      for (j=0; j<3; j++)
         if (dob->pdata.mvertex[i][j] < dob->bbox[0][j])
            dob->bbox[0][j] = dob->pdata.mvertex[i][j];
         else
            if (dob->pdata.mvertex[i][j] > dob->bbox[1][j])
               dob->bbox[1][j] = dob->pdata.mvertex[i][j];
   }

   dob->pdata.countedge = 0;
   for (i=0; i<dob->pdata.countobject; i++) {
      fscanf(infile, "%d", &dob->pdata.polynum[i]);

      dob->pdata.edgeptr[i] = &dob->pdata.object[dob->pdata.countedge];

      dob->pdata.countedge += dob->pdata.polynum[i];

      for (j=0; j<dob->pdata.polynum[i]; j++) {
         fscanf(infile, "%d", &dob->pdata.edgeptr[i][j]);
         dob->pdata.edgeptr[i][j]--;
      }

   }

   fclose(infile);
   return dob->pdata.countobject;
}


/* *************************************************************
************************************************************* */
int polygon::read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) {

   FILE *infile;                        // file pointer
   int i, j, k;
   int count;
   char tempstr[MAXSTRLEN];             // generic string

   for (tob=*tpptr; tob!=(texpolygon *)NULL; tob=tob->next)
      if (!strcmp(filename, tob->filename) && !strcmp(dfilename, tob->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CITEXTURE;
      return 0;
   }

   tob = new texpolygon;
   strcpy(tob->filename, filename);
   strcpy(tob->dataname, dfilename);
   tob->countobject = dob->pdata.countobject;

   tob->next = *tpptr;
   *tpptr    = tob;

   tob->setup_data(tob->countobject);

   fscanf(infile, "%s", tempstr);
   lower_case(tempstr);
   if (!strcmp(tempstr, "mono"))
      for (i=0; i<tob->countobject; i++)
         tob->data[i] = (tex *)NULL;
   else {
      tob->data[0] = new tex;

      fscanf(infile, "%s", tempstr);
      lower_case(tempstr);

      if (!strcmp(tempstr, "morph")) {
         tob->data[0]->mapfunc = MORPH;
         fscanf(infile, "%d", &tob->data[0]->vertex[0]);
         tob->data[0]->vertex[0]--;
      }

      else {
         tob->data[0]->mapfunc = CORNER;
         fscanf(infile, "%d %d %d %d", &tob->data[0]->vertex[0], &tob->data[0]->vertex[1],
					&tob->data[0]->vertex[2], &tob->data[0]->vertex[3]);
         tob->data[0]->vertex[0]--;
         tob->data[0]->vertex[1]--;
         tob->data[0]->vertex[2]--;
         tob->data[0]->vertex[3]--;
      }

      fscanf(infile, "%s", tempstr);               // texture name
      for (tob->data[0]->ob=*tptr; tob->data[0]->ob!=(texbase *)NULL; tob->data[0]->ob=tob->data[0]->ob->next)
         if (!strcmp(tempstr, tob->data[0]->ob->filename))
            break;

      if (tob->data[0]->ob == (texbase *)NULL) {

         if (query_video(tempstr)) {
            tob->data[0]->ob = new texvid;
            tob->data[0]->ob->read_texture(tempstr, tptr);
         }

         else {
            tob->data[0]->ob = new texture;
            tob->data[0]->ob->read_texture(tempstr);
         }

         tob->data[0]->ob->next = *tptr;
         *tptr = tob->data[0]->ob;
      }

      tob->data[0]->ob->query_coord(tob->data[0]->coord);

      for (i=1; i<tob->countobject; i++) {
         tob->data[i] = new tex;

         tob->data[i]->mapfunc = tob->data[0]->mapfunc;
         if (tob->data[i]->mapfunc == MORPH)
            tob->data[i]->vertex[0] = tob->data[0]->vertex[0];
         else
            copyarray4(tob->data[i]->vertex, tob->data[0]->vertex);

         tob->data[i]->ob = tob->data[0]->ob;

         copyarray2(tob->data[i]->coord[0], tob->data[0]->coord[0]);
         copyarray2(tob->data[i]->coord[1], tob->data[0]->coord[1]);
         copyarray2(tob->data[i]->coord[2], tob->data[0]->coord[2]);
         copyarray2(tob->data[i]->coord[3], tob->data[0]->coord[3]);
      }

   }

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {
      fscanf(infile, "%d", &j);
      j--;

      if (tob->data[j] == (tex *)NULL)
         tob->data[j] = new tex;
      else
         if (tob->data[j]->pt != (vector2f *)NULL) {
            delete tob->data[j]->pt;
            tob->data[j]->pt = (vector2f *)NULL;
         }

      fscanf(infile, "%s", tempstr);
      lower_case(tempstr);

      if (!strcmp(tempstr, "morph")) {
         tob->data[j]->mapfunc = MORPH;
         fscanf(infile, "%d", &tob->data[j]->vertex[0]);
         tob->data[j]->vertex[0]--;
      }

      else
         if (!strcmp(tempstr, "coord")) {
            fscanf(infile, "%d", &i);

            if (i == dob->pdata.polynum[j]) {
               tob->data[j]->mapfunc = COORD;

               tob->data[j]->pt = new vector2f[i+1];
               for (k=0; k<i; k++) {
                  fscanf(infile, "%f %f", &tob->data[j]->pt[k][0], &tob->data[j]->pt[k][1]);
                  tob->data[j]->pt[k][0]--;
                  tob->data[j]->pt[k][1]--;
               }

               copyarray2(tob->data[j]->pt[i], tob->data[j]->pt[0]);
            }

         }

         else {
            tob->data[j]->mapfunc = CORNER;
            fscanf(infile, "%d %d %d %d", &tob->data[j]->vertex[0], &tob->data[j]->vertex[1],
						&tob->data[j]->vertex[2], &tob->data[j]->vertex[3]);
            tob->data[j]->vertex[0]--;
            tob->data[j]->vertex[1]--;
            tob->data[j]->vertex[2]--;
            tob->data[j]->vertex[3]--;
         }

      fscanf(infile, "%s", tempstr);               // texture name
      for (tob->data[j]->ob=*tptr; tob->data[j]->ob!=(texbase *)NULL; tob->data[j]->ob=tob->data[j]->ob->next)
         if (!strcmp(tempstr, tob->data[j]->ob->filename))
            break;

      if (tob->data[j]->ob == (texbase *)NULL) {
         if (query_video(tempstr)) {
            tob->data[j]->ob = new texvid;
            tob->data[j]->ob->read_texture(tempstr, tptr);
         }

         else {
            tob->data[j]->ob = new texture;
            tob->data[j]->ob->read_texture(tempstr);
         }

         tob->data[j]->ob->next = *tptr;
         *tptr = tob->data[j]->ob;
      }

      tob->data[j]->ob->query_coord(tob->data[j]->coord);

//      tob->next    = *tpptr;
//      *tpptr = tob;
   }

   fclose(infile);
   return 1;
}


int polygon::read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) {

   int i, count;
   char buffer[MAXSTRLEN];
   FILE *infile;
   int objectcount = dob->pdata.countobject;

   for (shaderlist=*slptr; shaderlist!=(shaderlisttype *)NULL; shaderlist=shaderlist->next)
      if (!strcmp(filename, shaderlist->shname) && !strcmp(dfilename, shaderlist->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CI3DTEX;
      return 0;
   }

   shaderlist = new shaderlisttype;
   strcpy(shaderlist->shname, filename);
   strcpy(shaderlist->dataname, dfilename);
   shaderlist->init(objectcount);
   shaderlist->next   = *slptr;
   *slptr       = shaderlist;

   fscanf(infile, "%s %f", buffer, &shaderlist->scale);
      lower_case(buffer);

   if (strcmp(buffer, "mono")) {                        // ALL
      shaderlist->shade[0].read_data(infile, sptr);

      for (i=1; i<objectcount; i++) 
         shaderlist->shade[i] = shaderlist->shade[0];
   }

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {
      fscanf(infile, "%d", &i);
      shaderlist->shade[i-1].read_data(infile, sptr);
   }

   fclose(infile);
   return 1;
}

