
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




/* ***************************************************************
   this file handles the routines to manipulate polygons
**************************************************************** */

#include "polygon.h"
#include <bstring.h>


/* *************************************************************
   this function calculates the surface normals and vertex normals
************************************************************* */
void polygon::calc_normal() {

   vector3f v1, v2;
   int   i, j;

   for (i=0; i<ob->countobject; i++) {
      for (j=0; j<3; j++) {
         v1[j]      = ob->mvertex[ob->edgeptr[i][1]][j] -
                      ob->mvertex[ob->edgeptr[i][0]][j];
         v2[j]      = ob->mvertex[ob->edgeptr[i][2]][j] -
                      ob->mvertex[ob->edgeptr[i][1]][j];
      }

      xproduct(ob->normal[i], v2, v1);
      normalize3(ob->normal[i]);
                                        // used for phong2  ax+by+cz+d
      ob->normal[i][3] = -dotproduct3(ob->normal[i], ob->mvertex[ob->edgeptr[i][0]]);
   }

   if (mctype < GOURAUD || !bbflag)
      return;

   bzero(ob->gnormal, ob->countvertex << 4);	// 4 floats x 4 bytes

   for (i=0; i<ob->countobject; i++)
      for (j=0; j<ob->polynum[i]; j++)
         addarray3(ob->gnormal[ob->edgeptr[i][j]], ob->normal[i]);

   for (i=0; i<ob->countvertex; i++)
      normalize3(ob->gnormal[i]);
}


/* *************************************************************
   this function removes back face surfaces
************************************************************* */
void polygon::removeside() {

   float     *vp;
   int      i, k;

   for (i=0; i<ob->countobject; i++) {
      for (k=0; k<ob->polynum[i]; k++)
         if (ob->mvertex[ob->edgeptr[i][k]][2] < 0) {
            vp = ob->mvertex[ob->edgeptr[i][0]];
            if (dotproduct3(ob->normal[i], vp) > 0)
               ob->polynum[i] = 0;

            break;
         }

      if (k == ob->polynum[i])
         ob->polynum[i] = 0;
   }


//   float  v1[2], v2[2];
//   int   i, j;

//   for (i=0; i<ob->countobject; i++) {

//      for (j=0; j<2; j++) {
//         v1[j]      = ob->mvertex[ob->object[ob->edgeptr[i]+1]][j] -
//                      ob->mvertex[ob->object[ob->edgeptr[i]]][j];
//         v2[j]      = ob->mvertex[ob->object[ob->edgeptr[i]+2]][j] -
//                      ob->mvertex[ob->object[ob->edgeptr[i]+1]][j];
//      }

//      if (v2[0]*v1[1] <= v2[1]*v1[0])
//         ob->polynum[i] = 0;
//   }

}


/* *************************************************************
   This procedure puts the object in the edge table
************************************************************* */
void polygon::transform(eye *parm) {

   int i, j;			// looping vars
   vector4f temp;

   copymx4x4(world, rotate);

   for (i=0; i<3; i++)
      for (j=0; j<3; j++)
         world[i][j] *= size;

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   if (!(mcinfo & CIBITMAP))		// remove rotation for bitmaps :)
      matmatmulto(parm->transform, world);
   else {
      temp[0] = world[0][3];
      temp[1] = world[1][3];
      temp[2] = world[2][3];
      matvecmulto(parm->transform, temp);
      world[0][3] = temp[0];
      world[1][3] = temp[1];
      world[2][3] = temp[2];
   }

/*
   if (mcinfo & CIBITMAP) {		// remove rotation for bitmaps :)
      world[0][0] = world[1][1] = world[2][2] = size;
      world[0][1] = world[0][2] = 
      world[1][0] = world[1][2] =
      world[2][0] = world[2][1] = 0;
   }
*/

   inversemx(world, iworld);

   make_box(bbox, box);

   matvecmulto(world, box[0]);
   copyarray3(bbox[0], box[0]);
   copyarray3(bbox[1], box[0]);

   for (i=1; i<8; i++) {
      matvecmulto(world, box[i]);

      for (j=0; j<3; j++)
         if (box[i][j] < bbox[0][j])
            bbox[0][j] = box[i][j];
         else
            if (box[i][j] > bbox[1][j])
               bbox[1][j] = box[i][j];
   }

   if (bbox[0][2] <= front)
      for (i=0; i<ob->countvertex; i++)
         matvecmulto(world, ob->mvertex[i]);
}
