
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

#if !defined __POLYGON_H
#define __POLYGON_H


#include "base.h"

class shadow;

/* *************************************************************
   object to render poligonal data.
************************************************************* */
class polygon : public pc {

   friend class shadow;

   protected:
      polytype     *ob;                         // pointer to composite data
      datatype     *dob;                        // pointer to data
      texpolygon   *tob;

      vector3f *colortable;
      float    *area;				// texture compression factor
      char     *ptflag;

      int bflag, nflag, tflag, bbflag;

      void   clip_greater(edgetype *ptr, int i, float  plane, float *normal, eye *parm);
      void   clip_lesser(edgetype *ptr, int i, float  plane, float *normal, eye *parm);
      void   bfclip_greater(edgetype *ptr, int i, float  plane, float *normal, eye *parm);
      void   bfclip_lesser(edgetype *ptr, int i, float  plane, float *normal, eye *parm);
      void   post_clip(edgetype *ptr);
      int    prepare_edge(int polynum, edgetype *ptr, eye *parm, light *lparm);
      void   insert_edge(edgetype *ptr, bindex **btr, bindex **freelist);
      void   add_edge_table(edgetype ***et, eye *parm, light *lparm, int maxy);
      void   add_edge_table(edgetype ***et, edgetype *ptr, int i, eye *parm, light *lparm, int maxy);
      edgetype *edgelist(int i);
      void   maptexture(edgetype **ptr, int i);
      int    polyclet(edgetype ***et, eye *parm, light *lparm, int maxx, int maxy);
      int    beamclet(edgetype ***et, light *lparm, int maxx, int maxy);

      void   calc_normal();
      void   removeside();
      void   transform(eye *parm);

      void   constcolor(eye *parm, light *lparm);
      void   constcolor(eye *parm, light *lparm, int i, edgetype *ptr);
      float  engine(float  normal[], float  middle[], eye *parm);
      float  bfengine(float  normal[], float  middle[], eye *parm);

      void polywfbw(zbuffer *zbuff, edgetype **et, int inc);
      void polybw(eye *parm, zbuffer *zbuff, edgetype **et, int inc);
      void polyct(eye *parm, light *lparm, zbuffer *zbuff, edgetype **et, int inc);
      void polyal(eye *parm, light *lparm, zbuffer *zbuff, edgetype **et, int inc);
      void polygd(eye *parm, light *lparm, zbuffer *zbuff, edgetype **et, int inc);
      void polypg(eye *parm, light *lparm, zbuffer *zbuff, edgetype **et, int inc);
      void polylt(light *lparm, edgetype **et, int inc);

      int bound_box(eye *parm);

   public:

      ALEXcolor inviso;

      polygon() {
         tflag = bflag = nflag = bbflag = 1;
         colortable = (vector3f *)NULL;
         area = (float *)NULL;
         ptflag = (char *)NULL;
         ob = (polytype *)NULL;
      }

      ~polygon() {
         if (colortable != (vector3f *)NULL)
            delete colortable;
         if (area != (float *)NULL)
            delete area;
         if (ptflag != (char *)NULL)
            delete ptflag;
      }

      rendertype query_whatami() { return POLYGON; }

      void begin_scan() {
         ob = new polytype;
      }

      void end_scan() {

         if (ob != (polytype *)NULL) {
            delete ob;
            ob = (polytype *)NULL;
         }

         if (colortable != (vector3f *)NULL) {
            delete colortable;
            colortable = (vector3f *)NULL;
         }

      }

      void setup_patch(polytype *sob, colortype sctype, unsigned int scinfo, int sid, int ssflag,
                       float  ssplane[], float  box[][4], shadetype *c);
      void setup_texture(texpolygon *stob);
      void setup_texture3d(shaderlisttype *stob, float sworld[][4], float siworld[][4]);

      void datacopy();
      void scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr);
      void scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr);
      void beamscan(light *lparm, subslave grandchild[]);

      int parse(FILE *infile, char *token);
      void preprocess(listtype *mainlist, shader *sparm);

      int read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr);
      int read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr);

      int read_data(char filename[], datatype **dptr);
      int read_data(FILE *infile);

      void colorize(grandchildtype *child);
};


class make_poly : public pc {

   protected:
      polytype     *pob;
      texpolygon   *tpob;

      void setup_transfer(int vcount, int pcount);
};


#include "shadow.h"


#endif

