
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */






/* *************************************************************
   This function scans poligonal objects into a zbuffer
************************************************************* */


#include "polygon.h"
#include <string.h>

typedef edgetype *pedgetype;


/* *************************************************************
************************************************************* */
void make_poly::setup_transfer(int vcount, int pcount) {

   int i, j;

   i = pcount << 1;

   pob = new polytype;

   pob->build(vcount, i, i<<2, (mctype > PBW || (mctype > BW && mctype <=PHONG)));

   pob->countvertex = 0;
   pob->countobject = 0;
   pob->countedge   = 0;

   if (mcinfo & CITEXTURE) {
      tpob = new texpolygon;
      tpob->countobject = 0;

      tpob->data = new texptrtype[i];

      for (j=0; j<i; j++)
         tpob->data[j] = (tex *)NULL;
   }

}


/* *************************************************************
************************************************************* */
int polygon::bound_box(eye *parm) {

   int    i, j;
   float  extreme[2][2];
   float  d[2];

   if (bbox[0][2] > front)
      return 0;

   if (bbox[1][2] > front)
      return 1;

   for (i=0; i<8; i++)
      parm->map2screen(box[i]);

   copyarray2(extreme[0], box[0]);      // bounding box
   copyarray2(extreme[1], box[0]);

   for (i=1; i<8; i++)
      for (j=0; j<2; j++)
         if (extreme[0][j] > box[i][j])
            extreme[0][j] = box[i][j];
         else
            if (extreme[1][j] < box[i][j])
               extreme[1][j] = box[i][j];

   d[0] = extreme[1][0] - extreme[0][0];
   d[1] = extreme[1][1] - extreme[0][1];

   if (abs(d[0]) < MINBBOX  || abs(d[1]) < MINBBOX)
      return 0;

   if (abs(d[0]) < MINDIM  || abs(d[1]) < MINDIM) {
      mcinfo &= CIRAYCAST;

      if (mctype > CONSTANT)
         mctype = CONSTANT;

      if (bflag) {
         if (ob->countvertex > 3 && ob->countobject > 0 && ob->countedge > 3) {

            ob->countvertex = 4;
            ob->countobject = 1;
            ob->countedge   = 4;

            ob->mvertex[0][0] = extreme[0][0];
            ob->mvertex[0][1] = extreme[0][1];
            ob->mvertex[0][2] = bbox[1][2];
            ob->mvertex[0][3] = 1;
            parm->screen2map(ob->mvertex[0]);

            ob->mvertex[1][0] = extreme[0][0];
            ob->mvertex[1][1] = extreme[1][1];
            ob->mvertex[1][2] = bbox[1][2];
            ob->mvertex[1][3] = 1;
            parm->screen2map(ob->mvertex[1]);

            ob->mvertex[2][0] = extreme[1][0];
            ob->mvertex[2][1] = extreme[1][1];
            ob->mvertex[2][2] = bbox[1][2];
            ob->mvertex[2][3] = 1;
            parm->screen2map(ob->mvertex[2]);

            ob->mvertex[3][0] = extreme[1][0];
            ob->mvertex[3][1] = extreme[0][1];
            ob->mvertex[3][2] = bbox[1][2];
            ob->mvertex[3][3] = 1;
            parm->screen2map(ob->mvertex[3]);

            ob->object[0] = 0;
            ob->object[1] = 1;
            ob->object[2] = 2;
            ob->object[3] = 3;

            if (nflag) {
               ob->normal[0][0] = 0;
               ob->normal[0][1] = 0;
               ob->normal[0][2] = 1;
               ob->normal[0][3] = -bbox[1][2];
            }

            ob->polynum[0] = 4;

            ob->edgeptr[0] = ob->object;

            nflag = 0;
            bflag = 0;
         }

         if (mctype == CONSTANT)
            for (i=0; i<ob->countobject; i++)
               ob->shade[i] = base_color;
      }

   }

   return 1;
}


/* *************************************************************
************************************************************* */
void polygon::setup_texture(texpolygon *stob) {

   tob     = stob;
}


/* *************************************************************
************************************************************* */
void polygon::setup_texture3d(shaderlisttype *stob, float sworld[][4], float siworld[][4])  {

   shaderlist = stob;
   copymx4x4(world, sworld);
   copymx4x4(iworld, siworld);
}


/* *************************************************************
   This function converts surface patch data into polygonal data
************************************************************* */
void polygon::setup_patch(polytype *sob, colortype sctype, unsigned scinfo,
                          int sid, int ssflag, float  ssplane[],
                          float  box[][4], shadetype *c) {

   int i, j;

   base_color = c;
   copyarray4(bbox[0], box[0]);
   copyarray4(bbox[1], box[1]);

   ob     = sob;

   if (sctype >=PDOT)
      sctype = (colortype)(sctype - PDOT + DOT);

   mctype = sctype;
   mcinfo = scinfo;
   id     = sid;
   size   = 1;
   center[0] = center[1] = center[2] = 0;
   init_mx(i, j, rotate);

   bbflag = tflag  = 0;

   if (ssflag)
      sflag = 2;
   else
      sflag  = 0;

   bflag = (mctype > WFBW);

   copyarray4(splane, ssplane);
}


/* *************************************************************
   This procedure takes the basic object data, and creates a composite
   structure to be worked on.
************************************************************* */
void polygon::datacopy() {

   int i, j;               // looping variables

   ob->build(dob->pdata.countvertex, dob->pdata.countobject, dob->pdata.countedge, (mctype > BW));

   j = sizeof(int);
   i = ob->countobject*j;
   memcpy(ob->mvertex, dob->pdata.mvertex, ob->countvertex*sizeof(vector4f));
   memcpy(ob->polynum, dob->pdata.polynum, i);
   memcpy(ob->object, dob->pdata.object, ob->countedge*j);

   ob->edgeptr[0] = ob->object;
   for (i=1; i<ob->countobject; i++)
      ob->edgeptr[i] = ob->edgeptr[i-1] + ob->polynum[i-1];

   if (mctype > BW) {
      base_color = &lob->base;

      for (i=0; i<ob->countobject; i++)
         ob->shade[i] = base_color;

      for (i=0; i<lob->count; i++)
         ob->shade[lob->list[i].index] = &lob->list[i];
   }

   bflag = (mctype > WFBW);

   copyarray4(bbox[0], dob->bbox[0]);
   copyarray4(bbox[1], dob->bbox[1]);
}


/* **************************************************
   This function scans polygonal data and puts it in a zbuffer.
************************************************** */
void polygon::scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) {

   edgetype ***et;
   grandchildtype *offspring;
   int       i, j;                        // loop counter
   edgetype  *ptr, *qtr;
   shadow    *str;
   light     *lmain;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);

   if (sflag && lmain!=(light *)NULL && !NOSHADOW && !((mcinfo & CITRANSPARENT) | (mcinfo & CIBITMAP))) {
      str = new shadow;
      str->begin_scan();

      if (str->datacopy2(this, cparm, lmain, sflag-1)) {
         str->next = *shadptr;
         *shadptr = str;
      }

      else {
         str->end_scan();
         delete str;
      }

   }

   if (tflag)
      transform(cparm);                  // transform objects

   if (!ob->countobject || (bbflag && !bound_box(cparm)))
      return;

   if (nflag)
      calc_normal();

   if (bflag)
      removeside();

   et = new ppedgetype[MAXSUBPROC];
   offspring = new grandchildtype[MAXSUBPROC];

   for (i=0; i<MAXSUBPROC; i++) {
      et[i] = new pedgetype[WINY+1];

      for (j=0; j <= WINY; j++)           // initialize the edge table to null
         et[i][j] = (edgetype *)NULL;
   }

   if (lmain != (light *)NULL) {
      if (mctype > BW)
            for (i=0; i<ob->countobject; i++)
               if (ob->polynum[i])
                  lmain->set_ambient(ob->shade[i].color->ka, ob->shade[i].ambient);
   }

   else
      if (mctype > BW)
         for (i=0; i<ob->countobject; i++)
            if (ob->polynum[i])
               ob->shade[i].ambient[0] = ob->shade[i].ambient[1] = ob->shade[i].ambient[2] = 0;

   if (!polyclet(et, cparm, lmain, WINX, WINY)) {
      for (i=0; i<MAXSUBPROC; i++)
         delete et[i];

      delete et;
      delete offspring;
      return;
   }

   if (MAXSUBPROC > 1) {
      for (i=j=0; i<WINY && et[j][i]==(edgetype *)NULL; i++) {
         j++;
         if (j == MAXSUBPROC)
            j = 0;
      }

      while (i < WINY-1) {                 // scan all the lines

         ptr=et[j][i];
         i++;
         j++;
         if (j == MAXSUBPROC)
            j = 0;

         for (; ptr!=(edgetype *)NULL; ptr=ptr->next) {

            if (!(i >= round(ptr->pt2[1]) || i-round(ptr->pt1[1]) >= MAXSUBPROC)) {
               qtr = new edgetype;
               *qtr = *ptr;

               qtr->pt1[0] += qtr->deltax;
               qtr->pt1[2] += qtr->deltaz;

               if (mctype == PHONG)
                  addarray3(qtr->colorstuff1, qtr->delta);

               if (ptflag[qtr->color])
                  addarray2(qtr->t[0], qtr->deltat);

               qtr->next = et[j][i];
               et[j][i] = qtr;
            }

         }

         while (i<WINY && et[j][i]==(edgetype *)NULL) {
            i++;
            j++;
            if (j == MAXSUBPROC)
               j = 0;
         }

      }

      j = MAXSUBPROC-1;

      for (i=0; i<j; i++) {                  // interlace lines
         offspring[i].cam    = cparm;
         offspring[i].lit    = lparm;
         offspring[i].zbuff  = zbuff;
         offspring[i].e      = et[i];
         offspring[i].source = this;
         offspring[i].id     = i;
         offspring[i].done   = 0;
         grandchild[i].spawn(&offspring[i]);
      }

   }

   else
      j = MAXSUBPROC-1;

   offspring[j].cam    = cparm;
   offspring[j].lit    = lparm;
   offspring[j].zbuff  = zbuff;
   offspring[j].e      = et[j];
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;

   colorize(&offspring[j]);

   for (i=0; i<MAXSUBPROC; i++)
      while (!offspring[i].done);

   for (i=0; i<MAXSUBPROC; i++)
      delete et[i];

   delete et;
   delete offspring;
}


/* **************************************************
   This function scans polygonal data and puts it in a zbuffer.
************************************************** */
void polygon::scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

   int      i;                           // loop counter
   int       pt[2][2];
   shadow    *str;
   colortype x;
   edgetype  **et;
   edgetype  *ptr;
   light     *lmain;

   x       = mctype;
   mctype  = WFBW;

   mcinfo = CINULL;

   for (lmain=lparm; lmain!=(light *)NULL && lparm->ltype!=POINT && lparm->ltype!=FAR; lmain=lmain->next);

   if (sflag && lmain!=(light *)NULL && !NOSHADOW && !((mcinfo & CITRANSPARENT) | (mcinfo & CIBITMAP))) {

      str = new shadow;
      str->begin_scan();

      if (str->datacopy2(this, cparm, lmain, sflag-1)) {
         str->next = *shadptr;
         *shadptr = str;
      }

      else {
         str->end_scan();
         delete str;
      }

   }

   if (tflag)
      transform(cparm);                  // transform objects

   if (!ob->countobject || (bbflag && !bound_box(cparm))) {
      mctype = x;
      return;
   }

   et = new pedgetype[WINY+1];

   for (i=0; i <= WINY; i++)            // initialize the edge table to null
      et[i] = (edgetype *)NULL;

   if (!polyclet(&et, cparm, lmain, WINX, WINY)) {
      mctype = x;

      delete et;
      return;
   }

   glColor3f(1.0, 1.0, 1.0);

   for (ptr=et[WINY]; et[WINY]!=(edgetype *)NULL; et[WINY]=ptr->next, delete ptr, ptr=et[WINY]) {
      i = (int)(round(ptr->pt1[1]));
      pt[0][0] = round(ptr->pt1[0]);
      pt[0][1] = i;
      pt[1][0] = round(ptr->pt2[0]);
      pt[1][1] = i;

      glBegin(GL_LINES);
         glVertex2iv(pt[0]);
         glVertex2iv(pt[1]);
      glEnd();
   }

   for (i=0; i<WINY && et[i]==(edgetype *)NULL; i++);

   while (i < WINY) {                 // scan all the lines

      for (ptr=et[i]; et[i]!=(edgetype *)NULL; et[i]=ptr->next, delete ptr, ptr=et[i]) {

         pt[0][0] = round(ptr->pt1[0]);
         pt[0][1] = round(ptr->pt1[1]);
         pt[1][0] = round(ptr->pt2[0]);
         pt[1][1] = round(ptr->pt2[1]);

         glBegin(GL_LINES);
            glVertex2iv(pt[0]);
            glVertex2iv(pt[1]);
         glEnd();
      }

      for(i++; i<WINY && et[i]==(edgetype *)NULL; i++);
   }

   mctype = x;
   delete et;
}


/* **************************************************
   This function scans polygonal data and puts it in a zbuffer.
************************************************** */
void polygon::beamscan(light *lparm, subslave grandchild[]) {

   edgetype  ***et;
   grandchildtype *offspring;
   int       i, j;                        // loop counter
   edgetype  *ptr, *qtr;
   colortype tempcolor;

   if (mctype < BW)
      return;

   tempcolor = mctype;
   mctype = BW;

   if (tflag)
      transform(lparm);                  // transform objects

   if (!ob->countobject || (bbflag && !bound_box(lparm))) {
      mctype = tempcolor;
      return;
   }

   if (nflag)
      calc_normal();

   if (bflag)
      removeside();

   et = new ppedgetype[MAXSUBPROC];
   offspring = new grandchildtype[MAXSUBPROC];

   for (i=0; i<MAXSUBPROC; i++) {
      et[i] = new pedgetype[lparm->lbuff.maxy+1];

      for (j=0; j<=lparm->lbuff.maxy; j++)           // initialize the edge table to null
         et[i][j] = (edgetype *)NULL;
   }

   if (!beamclet(et, lparm, lparm->lbuff.maxx, lparm->lbuff.maxy)) {
      for (i=0; i<MAXSUBPROC; i++)
          delete et[i];

      delete et;
      delete offspring;
      mctype = tempcolor;
      return;
   }

   if (MAXSUBPROC > 1) {

      for (i=j=0; i<lparm->lbuff.maxy && et[j][i]==(edgetype *)NULL; i++) {
        j++;
         if (j == MAXSUBPROC)
            j = 0;
      }

      while (i < lparm->lbuff.maxy-1) {                 // scan all the lines

         ptr=et[j][i];
         i++;
         j++;
         if (j == MAXSUBPROC)
            j = 0;

         for (; ptr!=(edgetype *)NULL; ptr=ptr->next) {

            if (!(i >= round(ptr->pt2[1]) || i-round(ptr->pt1[1]) >= MAXSUBPROC)) {
               qtr = new edgetype;
               *qtr = *ptr;

               qtr->pt1[0] += qtr->deltax;
               qtr->pt1[2] += qtr->deltaz;

               qtr->next = et[j][i];
               et[j][i] = qtr;
            }

         }

         while (i<lparm->lbuff.maxy && et[j][i]==(edgetype *)NULL) {
            i++;
            j++;
            if (j == MAXSUBPROC)
               j = 0;
         }

      }

      for (i=0; i < MAXSUBPROC; i++)
         while (et[i][lparm->lbuff.maxy] != (edgetype *)NULL) {    // delete horizontal lines
            ptr = et[i][lparm->lbuff.maxy];
            et[i][lparm->lbuff.maxy] = ptr->next;
            delete ptr;
         }

      j = MAXSUBPROC-1;

      for (i=0; i<j; i++) {                  // interlace lines
         offspring[i].cam    = offspring[i].lit = lparm;
         offspring[i].source = this;
         offspring[i].id     = i;
         offspring[i].e      = et[i];
         offspring[i].done   = 0;
         grandchild[i].spawn(&offspring[i]);
      }

   }

   else {
      while (et[0][lparm->lbuff.maxy] != (edgetype *)NULL) {    // delete horizontal lines
         ptr = et[0][lparm->lbuff.maxy];
         et[0][lparm->lbuff.maxy] = ptr->next;
         delete ptr;
      }

      j = MAXSUBPROC-1;
   }

   offspring[j].cam    = offspring[j].lit = lparm;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].e      = et[j];
   offspring[j].done   = 0;

   colorize(&offspring[j]);

   for (i=0; i<MAXSUBPROC; i++)
      while (!offspring[i].done);

   mctype = tempcolor;

   for (i=0; i<MAXSUBPROC; i++)
      delete et[i];

   delete et;
   delete offspring;
}

