
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This function scans poligonal objects into a zbuffer
************************************************************* */


#include "polygon.h"


/* *************************************************************
   this function calculates the constant coloring by interpolating
   a point from which to calculate the polygon's color
************************************************************* */
void polygon::constcolor(eye *parm, light *lparm) {

   int  i, j;                           // looping variables
   float surface[3];                     // surface point

   if (lparm == (light *)NULL)
      return;

   for (i=0; i<ob->countobject; i++)
      if (ob->polynum[i]) {
         copyarray3(surface, ob->mvertex[ob->edgeptr[i][0]]);

         for (j=1; j<ob->polynum[i]; j++)
            addarray3(surface, ob->mvertex[ob->edgeptr[i][j]]);

         sdivarray3(surface, ob->polynum[i]);
         lparm->intensity(surface, ob->normal[i], colortable[i], parm, ob->shade[i].color, 0);
      }

}


/* *************************************************************
   this function calculates the constant coloring by interpolating
   a point from which to calculate the polygon's color
************************************************************* */
void polygon::constcolor(eye *parm, light *lparm, int i, edgetype *ptr) {

   int        j;                           // looping variables
   float       surface[3];                  // surface point
   edgetype *qtr;

   if (lparm == (light *)NULL || ptr == (edgetype *)NULL)
      return;

   copyarray3(surface, ptr->pt1);

   for (qtr=ptr->next, j=1; qtr!=(edgetype *)NULL; qtr=qtr->next, j++)
      addarray3(surface, qtr->pt1);

   sdivarray3(surface, (float)j);

   lparm->intensity(surface, ob->normal[i], colortable[i], parm, ob->shade[i].color, 0);
}


/* *************************************************************
   This procedure calculates linear interpolation in object space
************************************************************* */
float polygon::engine(float normal[], float middle[], eye *parm) {

   float n;
   vector3f v;

   copyarray2(v, middle);
   parm->parascreen2map(v);

   n = v[0]*normal[0] + v[1]*normal[1] - normal[2];	// pt = (0,0,0)
							// ray = (x, y, -1)
   if (not_zero(n))
      return normal[3]/n;

   return -1000000;
}


/* *************************************************************
************************************************************* */
float polygon::bfengine(float normal[], float middle[], eye *parm) {

   float n, num;
   vector3f pt;

   copyarray2(pt, middle);
   parm->parascreen2map(pt);

   num = pt[0]*normal[0] + pt[1]*normal[1];		// pt  = (x, y, 0)

   n = - normal[2];					// ray = (0, 0, -1)

   if (not_zero(n))
      return (normal[3] + num)/n;

   return -1000000;
}


/* **************************************************
************************************************** */
void polygon::polywfbw(zbuffer *zbuff, edgetype **et, int inc) {
   edgetype      *aet;
   edgetype      *ptr, *qtr;            // temp pointer
   int           scany;                 // current line being scanned
   int          i, j;                  // loop counter
   int           mtempx[2];
   float         mtempz, mdeltaz;

   while (et[WINY] != (edgetype *)NULL) {               // horizontal lines
      ptr = et[WINY];
      et[WINY] = ptr->next;

      i = round(ptr->pt1[1]);
      mtempx[0] = round(ptr->pt1[0]);
      mtempx[1] = round(ptr->pt2[0]);

      for (j=mtempx[0]; j<mtempx[1]; j++) {

         if (ptr->pt1[2]>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front) {
            zbuff->zbuff[i][j]       = ptr->pt1[2];
            zbuff->ambientbuff[i][j][0] = zbuff->ambientbuff[i][j][1] = zbuff->ambientbuff[i][j][2] = 1;
            zbuff->ybuff[i][j][0] = zbuff->ybuff[i][j][1] = zbuff->ybuff[i][j][2] = 0;
         }

         ptr->pt1[2] += ptr->deltaz;
      }

      delete ptr;
   }

   aet = (edgetype *)NULL;

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY)
      aet = et[scany];

   while (scany < WINY) {               // scan all the lines
      for (ptr=aet; ptr!=(edgetype *)NULL; ptr=ptr->next) {
         mtempx[0] = round(ptr->pt1[0]);

#include "polywfbw.c"

      }

      scany += MAXSUBPROC;

      while (aet != (edgetype *)NULL)      // update aet
         if (scany >= round(aet->pt2[1])) {
            qtr=aet;
            aet = aet->next;
            delete qtr;
         }

         else {
            aet->pt1[0] += aet->dx;
            break;
         }

      if (aet != (edgetype *)NULL) {
         ptr = aet;
         while (ptr->next != (edgetype *)NULL)        // update aet
            if (scany >= round(ptr->next->pt2[1])) {
               qtr=ptr->next;
               ptr->next = qtr->next;
               delete qtr;
            }

            else {
               ptr = ptr->next;
               ptr->pt1[0] += ptr->dx;
            }

      }

      if (aet == (edgetype *)NULL) {
         while (scany < WINY && et[scany] == (edgetype *)NULL)
            scany += MAXSUBPROC;
         if (scany<WINY)
            aet = et[scany];
      }

      else
         if (scany<WINY && et[scany]!=(edgetype *)NULL) {
            for (ptr=et[scany]; ptr->next!=(edgetype *)NULL; ptr=ptr->next);

            ptr->next = aet;
            aet       = et[scany];
         }

   }
                                                        // clean garbage
   for (ptr=aet; ptr!=(edgetype *)NULL; ptr=aet) {
      aet = aet->next;
      delete ptr;
   }

}


/* **************************************************
************************************************** */
void polygon::polybw(eye *parm, zbuffer *zbuff, edgetype **et, int inc) {

   edgetype  *aet;
   edgetype  *ptr, *qtr;                     // temp pointer
   int       scany;                       // current line being scanned
   int      i, j;                        // loop counter
   int       mtempx[2];
   float     mtempz[2];
   int       mtemp[3];
   float     mdeltaz;                     // delta z for a scan line
   float     surface[4];
   char      flagray = mcinfo & CIRAYCAST;
   char      nflagray = !flagray;
   int       bucketflag;
   bindex    *btr, *freelist;

   surface[3] = 1;

   while (et[WINY] != (edgetype *)NULL) {            // horizontal lines
      ptr = et[WINY];
      et[WINY] = ptr->next;

      surface[1] = ptr->pt1[1];
      i = round(surface[1]);
      mtemp[0] = round(ptr->pt1[0]);
      mtemp[1] = round(ptr->pt2[0]);

      for (j=mtemp[0]; j<mtemp[1]; j++) {
         surface[0] = j;

         if ((mcinfo & CIRAYCAST))
            surface[2] = engine(ob->normal[ptr->color], surface, parm);
         else {
            surface[2] = ptr->pt1[2];
            ptr->pt1[2] += ptr->deltaz;
         }

         if (surface[2]>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front) {
            zbuff->zbuff[i][j]    = surface[2];
            zbuff->ambientbuff[i][j][0] = zbuff->ambientbuff[i][j][1] = zbuff->ambientbuff[i][j][2] = 1;
            zbuff->ybuff[i][j][0] = zbuff->ybuff[i][j][1] = zbuff->ybuff[i][j][2] = 0;
         }

      }

      delete ptr;
   }

   btr = freelist = (bindex *)NULL;

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < WINY) {                 // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            if (mtempx[0] != mtempx[1])
#include "polybw.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<WINY && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::polyct(eye *parm, light *lparm, zbuffer *zbuff,
                     edgetype **et, int inc) {

   edgetype   *aet;
   edgetype   *ptr, *qtr;               // temp pointer
   int        scany;                 // current line being scanned
   int       i;                     // loop counter
   int        mtempx[2];
   float      mtempz, mtempt[2];     // current x, z value for a pixel
   float      mdeltaz, mdeltat[2];   // delta z for a scan line
   float      mtemp;
   texcolortype  texcolor;
   ALEXcolor  tempcolor;
   float      surface[4];
   float      rcolor[3], acolor[3];
   light      *lmain;
   char       flagray = mcinfo & CIRAYCAST;
   char       nflagray = !flagray;
   char       flag2t, flagtr, flag2d, flag3t, flag3d, flagl, flagb;
   shaderparamtype sparam;
   int       bucketflag;
   bindex     *btr, *freelist;

   surface[3] = 1;

   btr = freelist = (bindex *)NULL;

   flagtr = mcinfo & CITRANSPARENT;
   flagl = mcinfo & CILUM;
   flag2d = mcinfo & CITEXTURE;
   flag3d = mcinfo & CI3DTEX;
   flagb = mcinfo & CIBITMAP;
   scany = inc;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < WINY) {               // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            flag2t = flag2d && ptflag[ptr->color];
            flag3t = flag3d && shaderlist->shade[ptr->color].scount;
            if (mtempx[0] != mtempx[1])
#include "polyct.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            if (ptflag[ptr->color])
               addarray2(ptr->t[0], ptr->dt);

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<WINY && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::polyal(eye *parm, light *lparm, zbuffer *zbuff,
                     edgetype **et, int inc) {

   edgetype  *aet;
   edgetype  *ptr, *qtr;               // temp pointer
   int       scany;                 // current line being scanned
   float     surface[4];            // current point on surface
   int      i;                  // loop counter
   int       mtempx[2];
   float     rcolor[3], acolor[3];  // current color
   texcolortype texcolor;
   ALEXcolor  tempcolor;
   float     mtemp;
   float     mtempz, mtempt[2];     // current x, z value for a pixel
   float     mdeltaz, mdeltat[2];   // delta z for a scan line
   light     *lmain;
   char      flagray = mcinfo & CIRAYCAST;
   char      nflagray = !flagray;
   char      flag2t, flag2d, flag3t, flag3d, flagl, flagtr, flagb;
   shaderparamtype sparam;
   int       bucketflag;
   bindex    *btr, *freelist;
   float     normal[3];

   surface[3] = 1;

   btr = freelist = (bindex *)NULL;

   flagtr = mcinfo & CITRANSPARENT;
   flagl = mcinfo & CILUM;
   flag2d = mcinfo & CITEXTURE;
   flag3d = mcinfo & CI3DTEX;
   flagb = mcinfo & CIBITMAP;

   scany = inc;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < WINY) {               // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            flag2t = flag2d && ptflag[ptr->color];
            flag3t = flag3d && shaderlist->shade[ptr->color].scount;
            if (mtempx[0] != mtempx[1])
#include "polyal.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            if (ptflag[ptr->color])
               addarray2(ptr->t[0], ptr->dt);

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<WINY && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::polygd(eye *parm, light *lparm, zbuffer *zbuff,
                     edgetype **et, int inc) {

   edgetype  *aet;
   edgetype  *ptr, *qtr;               // temp pointer
   int       scany;                 // current line being scanned
   int      i;                     // loop counter
   int       mtempx[2];
   float     mtempz, mtempt[2];     // current x, z value for a pixel
   float     rcolor[3], acolor[3], tcolor[3];             // current color
   float     mdeltaz, mdeltat[2];   // delta z for a scan line
   float     mtemp;
   float     mdeltac[3];
   float     surface[4];
   texcolortype texcolor;
   ALEXcolor  tempcolor;
   light     *lmain;
   char      flagray = mcinfo & CIRAYCAST;
   char      nflagray = !flagray;
   char      flag2t, flagtr, flag2d, flag3t, flag3d, flagl, flagb;
   shaderparamtype sparam;
   int       bucketflag;
   bindex    *btr, *freelist;

   surface[3] = 1;

   btr = freelist = (bindex *)NULL;

   flagtr = mcinfo & CITRANSPARENT;
   flagl = mcinfo & CILUM;
   flag2d = mcinfo & CITEXTURE;
   flag3d = mcinfo & CI3DTEX;
   flagb = mcinfo & CIBITMAP;

   scany = inc;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < WINY) {               // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            flag2t = flag2d && ptflag[ptr->color];
            flag3t = flag3d && shaderlist->shade[ptr->color].scount;
            if (mtempx[0] != mtempx[1])

#include "polygd.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            for (i=0; i<3; i++)
               ptr->colorstuff1[i] += ptr->d[i];

            if (ptflag[ptr->color])
               addarray2(ptr->t[0], ptr->dt);

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<WINY && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::polypg(eye *parm, light *lparm, zbuffer *zbuff,
                     edgetype **et, int inc) {

   edgetype  *aet;
   edgetype  *ptr, *qtr;               // temp pointer
   int       scany;                 // current line being scanned
   float     surface[4];            // current point on surface
   int      i;                     // loop counter
   int       mtempx[2];
   float     mdeltac[3];            // delta color stuff
   float     mtempn[3], mtemp[3];   // current surface normal
   float     rcolor[3], acolor[3];             // current color
   texcolortype texcolor;
   ALEXcolor  tempcolor;
   float     mtempz, mtempt[2];     // current x, z value for a pixel
   float     mdeltaz, mdeltat[2];   // delta z for a scan line
   light     *lmain;
   char      flagray = mcinfo & CIRAYCAST;
   char      nflagray = !flagray;
   char      flag2t, flagtr, flag2d, flag3t, flag3d, flagl, flagb;
   shaderparamtype sparam;
   int       bucketflag;
   bindex    *btr, *freelist;
   float     normal[3];

   surface[3] = 1;

   btr = freelist = (bindex *)NULL;

   flagtr = mcinfo & CITRANSPARENT;
   flagl = mcinfo & CILUM;
   flag2d = mcinfo & CITEXTURE;
   flag3d = mcinfo & CI3DTEX;
   flagb = mcinfo & CIBITMAP;

   scany = inc;

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype!=POINT && lmain->ltype!=FAR; lmain=lmain->next);

   for (scany=inc; scany<WINY && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < WINY) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < WINY) {               // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            flag2t = flag2d && ptflag[ptr->color];
            flag3t = flag3d && shaderlist->shade[ptr->color].scount;
            if (mtempx[0] != mtempx[1])
#include "polypg.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            if (ptflag[ptr->color])
               addarray2(ptr->t[0], ptr->dt);

            for (i=0; i<3; i++)
               ptr->colorstuff1[i] += ptr->d[i];

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<WINY && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < WINY) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::polylt(light *lparm, edgetype **et, int inc) {

   edgetype *aet;
   edgetype *ptr, *qtr;               // temp pointer
   int      scany;                 // current line being scanned
   int     i;                     // loop counter
   int      mtempx[2];
   float    mtempz;                // current x, z value for a pixel
   float    mdeltaz;               // delta z for a scan line
   float    surface[4];
   char     flagray = mcinfo & CIRAYCAST;
   char     nflagray = !flagray;
   int       bucketflag;
   bindex    *btr, *freelist;

   surface[3] = 1;

   btr = freelist = (bindex *)NULL;

   scany = inc;

   for (scany=inc; scany<lparm->lbuff.maxy && et[scany]==(edgetype *)NULL; scany+=MAXSUBPROC);

   if (scany < lparm->lbuff.maxy) {
      for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
         et[scany] = ptr->next;
         insert_edge(ptr, &btr, &freelist);
      }

      aet = (edgetype *)NULL;
      btr->build(&aet);
      btr->free_index(&freelist);
      btr = (bindex *)NULL;
   }

   while (scany < lparm->lbuff.maxy) {               // scan all the lines

      surface[1] = scany;

      for (ptr=aet; ptr!=(edgetype *)NULL; ) {

         if (ptr->next != (edgetype *)NULL &&
             ptr->color == ptr->next->color) {

            qtr    = ptr->next;
            mtempx[0] = round(ptr->pt1[0]);
            mtempx[1] = round(qtr->pt1[0]);

            if (mtempx[0] != mtempx[1])
#include "polylt.c"

            ptr = qtr->next;
         }

         else {
            printf("%d SHIT BEAM ", scany);
            for (qtr=ptr; qtr!=(edgetype *)NULL; qtr=qtr->next)
               printf(" %f %d",qtr->pt2[1], qtr->color);

            printf("\n");
            ptr = ptr->next;
         }

      }

      scany += MAXSUBPROC;

      bucketflag = 0;

      qtr = aet;
      aet = (edgetype *)NULL;

      while (qtr != (edgetype *)NULL) {      // update aet
         ptr = qtr;
         qtr = qtr->next;

         if (scany >= round(ptr->pt2[1]))
            delete ptr;
         else {
            ptr->pt1[0] += ptr->dx;
            ptr->pt1[2] += ptr->dz;

            insert_edge(ptr, &btr, &freelist);
            bucketflag = 1;
         }

      }

      if (!bucketflag) {
         while (scany<lparm->lbuff.maxy && et[scany]==(edgetype *)NULL)
            scany += MAXSUBPROC;

         if (scany < lparm->lbuff.maxy) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

      }

      else
         if (scany < lparm->lbuff.maxy) {
            for (ptr=et[scany]; ptr!=(edgetype *)NULL; ptr=et[scany]) {
               et[scany] = ptr->next;
               insert_edge(ptr, &btr, &freelist);
            }

            btr->build(&aet);
            btr->free_index(&freelist);
            btr = (bindex *)NULL;
         }

         else {
            btr->build(&aet);
            btr->free_index(&freelist);

            for (ptr=aet; ptr!=(edgetype *)NULL; aet=ptr->next, delete ptr, ptr=aet);
         }

   }

   delete freelist;
}


/* **************************************************
************************************************** */
void polygon::colorize(grandchildtype *child) {

   if ((eye *)child->cam == (eye *)child->lit) {
      if (!(mcinfo & CIBITMAP))
         polylt(child->lit, child->e, child->id);

      child->done = 1;
      return;
   }

   switch (mctype) {

      case PHONG:
         polypg(child->cam, child->lit, child->zbuff, child->e, child->id);
         break;

      case GOURAUD:
        polygd(child->cam, child->lit, child->zbuff, child->e, child->id);
         break;

      case ALEX:
         polyal(child->cam, child->lit, child->zbuff, child->e, child->id);
         break;

      case CONSTANT:
         polyct(child->cam, child->lit, child->zbuff, child->e, child->id);
         break;

      case BW:
         polybw(child->cam, child->zbuff, child->e, child->id);
         break;

      default:
         polywfbw(child->zbuff, child->e, child->id);
         break;
   }

   child->done = 1;
}

