
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */


#include "radiant.h"
#include "sphere.h"


/* *************************************************************
************************************************************* */
int radiant::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'h':
         if (!strcmp(token, "hollow")) {
            flagh = 1;
            return 1;
         }

         break;

      case 'i':
         if (!strcmp(token, "inner_radius")) {
            get_token(infile, token);
            inner_circle = atof(token);

            return 1;
         }

         break;

      case 'o':
         if (!strcmp(token, "outer_radius")) {
            get_token(infile, token);
            radius = atof(token);

            return 1;
         }

         break;

      case 'l':
         if (!strcmp(token, "location")) {
            get_token(infile, token);
            center[0] = atof(token);
            get_token(infile, token);
            center[1] = atof(token);
            get_token(infile, token);
            center[2] = atof(token);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


void radiant::preprocess(listtype *mainlist, shader *sparm) {

   wratio = 1.0 - inner_circle/radius;
   pc::preprocess(mainlist, sparm);
}


int radiant::read_data(char filename[], datatype **dptr) {

   return 0;
}


int radiant::read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) {

   return 0;
}


int radiant::read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) {

   return 0;
}


int radiant::read_data(FILE *infile) {

   char junk[MAXSTRLEN], junk1[MAXSTRLEN];

   fscanf(infile, "%s %f %f %f", junk, &center[0], &center[1], &center[2]);
   fscanf(infile, "%s %f %s %f", junk, &radius, junk1, &inner_circle);

   return 0;
}


float radiant::radiance(float pt[]) {

   float i, k;
   float vector[2], sv[2];

   vector[0] = pt[0]-wcenter[0];
   vector[1] = pt[1]-wcenter[1];

   sv[0] = SQR(vector[0]);
   sv[1] = SQR(vector[1]);

   k = sv[0] + sv[1];
   i = (sqrwradius-k)*scale;				// point outside perimeter

   if (i < 0)
      return -1.0;

   pt[2] = wcenter[2] + sqrt(i);

   if (pt[2] > front)
      return -1.0;

   if (k <= winner_circle)
      return (!flagh);

   i = wratio*sqrt(k);		// calc intensity

   if (tradius < i)
      return -1.0;

   return SIN((1.0-i/tradius)*HALFPI) + 0.1;
}


int radiant::bound_box(eye *parm) {

   bbox[0][0] = wcenter[0] - wradius;
   bbox[0][1] = wcenter[1] - wradius;
   bbox[0][2] = wcenter[2] - wradius;

   bbox[1][0] = wcenter[0] + wradius;
   bbox[1][1] = wcenter[1] + wradius;
   bbox[1][2] = wcenter[2] + wradius;

   return 1;
}


int radiant::rad_clip(eye *parm, int maxx, int maxy) {

   if (bbox[0][2] > front || wcenter[2] > -CORRECT ||
       (bbox[1][2] <= front &&
        (-bbox[0][0]/bbox[1][2] > parm->vrc[1] ||     // outside boundaries
         -bbox[1][0]/bbox[1][2] < parm->vrc[0] ||
         -bbox[0][1]/bbox[1][2] > parm->vrc[3] ||
         -bbox[1][1]/bbox[1][2] < parm->vrc[2]))) {
      return 0;
   }

   parm->map2screen(wcenter);                       // map to screen

   scale = parm->query_scale()/(-wcenter[2]);

   winner_circle *= scale;
   wradius       *= scale;
   tradius        = wradius - winner_circle;
   sqrwradius	  = SQR(wradius);
   winner_circle *= winner_circle;
   scale = 1.0/SQR(scale);

   row[0] = round(wcenter[1] - wradius);
   row[1] = round(wcenter[1] + wradius);
   col[0] = round(wcenter[0] - wradius);
   col[1] = round(wcenter[0] + wradius);

   if (col[0] < 0)
      col[0] = 0;

   if (col[1] > maxx-1)
      col[1] = (int)maxx-1;

   if (row[0] < 0)
      row[0] = 0;

   if (row[1] > maxy-1)
      row[1] = (int)maxy-1;

   return 1;
}


void radiant::transform(eye *parm) {

   matvecmulto(parm->transform, wcenter);
}


void radiant::datacopy() {

   copyarray3(wcenter, center);
   wcenter[3] = 1;
   wradius = radius;
   winner_circle = inner_circle;

   if (mctype > BW)
      base_color = &lob->base;
}


void radiant::colorize(grandchildtype *child) {

   eye     *parm;
   light   *lparm, *lmain;
   zbuffer *zbuff;
   int     inc;

   int   i, j;
   float intens;
   float pt[4];
   texcolortype texcolor;
   float normal[4];
   float ycenter[4];
   float ca[3], cd[3];

   parm  = child->cam;
   lparm = child->lit;
   zbuff = child->zbuff;
   inc   = child->id;

   pt[3] = 1;

   switch (mctype) {
      case IRRADIANT:

         for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC) {
            pt[1] = (float)i;
            for (j=col[0]; j<=col[1]; j++) {

               pt[0]  = (float)j;
               intens = radiance(pt);

               if (intens > 0 && (zbuff->zbuff[i][j] > front || pt[2]>zbuff->zbuff[i][j])) {
                  ca[0] = base_color->lum[0]*intens;
                  ca[1] = base_color->lum[1]*intens;
                  ca[2] = base_color->lum[2]*intens;

                  if (zbuff->zbuff[i][j] > front) {
                     zbuff->zbuff[i][j] = -999999999;
                     copyarray3(zbuff->ambientbuff[i][j], ca);
                  }

                  else {
                     if (ca[0] > zbuff->ambientbuff[i][j][0])
                        zbuff->ambientbuff[i][j][0] = ca[0];
                     if (ca[1] > zbuff->ambientbuff[i][j][1])
                        zbuff->ambientbuff[i][j][1] = ca[1];
                     if (ca[2] > zbuff->ambientbuff[i][j][2])
                        zbuff->ambientbuff[i][j][2] = ca[2];
                  }

               }

            }

         }

         break;

      case RSOLID:              // pseudo 3D - needs work

         for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype != POINT && lmain->ltype != FAR; lmain=lmain->next);

         pt[3] = 1;
         copyarray4(ycenter, wcenter);
         parm->screen2map(ycenter);

         for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC)
            for (j=col[0]; j<=col[1]; j++) {

               pt[1]  = (float)i;
               pt[0]  = (float)j;
               intens = radiance(pt);

               if (intens > 0 && (pt[2]>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front)) {
                  parm->screen2map(pt);

                  normal[0] = pt[0] - ycenter[0];
                  normal[1] = pt[1] - ycenter[1];
                  normal[2] = pt[2] - ycenter[2];
                  normalize3(normal);

                  texcolor = *base_color;
                  smultarray3(texcolor.color.ka, intens);
                  smultarray3(texcolor.color.kp, intens);
                  smultarray3(texcolor.color.ks, intens);
                  smultarray3(texcolor.color.lum, intens);

                  ca[0] = ca[1] = ca[2] = 0;
                  cd[0] = cd[1] = cd[2] = 0;

                  if (lmain != (light *)NULL) {
                     lmain->set_ambient(texcolor.color.ka, texcolor.ambient);
                     copyarray3(ca, texcolor.ambient);
                     lmain->intensity(pt, normal, cd, parm, &texcolor.color, id);
                  }

                  addarray3(ca, texcolor.color.lum);

                  if (zbuff->zbuff[i][j] > front) {
                     zbuff->zbuff[i][j] = -999999999;
                     copyarray3(zbuff->ambientbuff[i][j], ca);
                     copyarray3(zbuff->ybuff[i][j], cd);
                  }

                  else {
                     if (ca[0] > zbuff->ambientbuff[i][j][0])
                        zbuff->ambientbuff[i][j][0] = ca[0];
                     if (ca[1] > zbuff->ambientbuff[i][j][1])
                        zbuff->ambientbuff[i][j][1] = ca[1];
                     if (ca[2] > zbuff->ambientbuff[i][j][2])
                        zbuff->ambientbuff[i][j][2] = ca[2];

                     if (cd[0] > zbuff->ybuff[i][j][0])
                        zbuff->ybuff[i][j][0] = cd[0];
                     if (cd[1] > zbuff->ybuff[i][j][1])
                        zbuff->ybuff[i][j][1] = cd[1];
                     if (cd[2] > zbuff->ybuff[i][j][2])
                        zbuff->ybuff[i][j][2] = cd[2];
                  }

               }

            }

         break;

      default:
         break;
   }

   child->done = 1;
}


void radiant::scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) {

   int i, j;
   grandchildtype *offspring;

   sphere r;

   if (mctype == WFBW) {
      r.frame = frame;
      r.id = id;

      r.sflag = 0;

      copyarray3(r.center, center);
      r.size = 1;
      init_mx(i, j, r.rotate);
      r.mctype = WFBW;
      r.mcinfo = CINULL;
      r.wscale[0] = r.wscale[1] = r.wscale[2] = radius;
   
      r.scan(cparm, lparm, zbuff, grandchild, shadptr);

      return;
   }

   transform(cparm);

   if (!bound_box((eye *)NULL) || !rad_clip(cparm, WINX, WINY))
      return;

   j = MAXSUBPROC-1;

   offspring = new grandchildtype[MAXSUBPROC];

   for (i=0; i<j; i++) {                  // interlace lines
      offspring[i].cam    = cparm;
      offspring[i].lit    = lparm;
      offspring[i].zbuff  = zbuff;
      offspring[i].source = this;
      offspring[i].id     = i;
      offspring[i].done   = 0;
      grandchild[i].spawn(&offspring[i]);
   }

   offspring[j].cam    = cparm;
   offspring[j].lit    = lparm;
   offspring[j].zbuff  = zbuff;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;

   colorize(&offspring[j]);

   for (i=0; i<MAXSUBPROC; i++)
      while (!offspring[i].done);

   delete offspring;
}


void radiant::scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

   sphere r;
   int i, j;

   r.frame = frame;
   r.id = id;

   r.sflag = 0;

   copyarray3(r.center, center);
   r.size = 1;
   init_mx(i, j, r.rotate);
   r.mctype = WFBW;
   r.mcinfo = CINULL;
   r.wscale[0] = r.wscale[1] = r.wscale[2] = radius;
   
   r.scan(cparm, lparm, grandchild, shadptr);
}


void radiant::beamscan(light *lparm, subslave grandchild[]) {

}
