
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */



#include "sbfast.h"



void sbfast::sbfastbm_llurh(light *lparm, int inc, float dy) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float row;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pray[3] = 0;
   surface[3] = 1;
   pzero[3] = 1;

   if (lparm->ltype == BEAM_POINT) {
      pzero[0] = pzero[1] = pzero[2] = 0;
      matvecmulto(iworld, pzero);
   }

   else {
      pray[0] = pray[1] = 0;
      pray[2] = -1;
      matvecmultv(iworld, pray);
   }

   for ((!inc) ? h=MAXSUBPROC : h=inc; h<lparm->lbuff.maxy; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      row = h;

      if (!SILENT)
         printf("y %d\n", h);

      for (j=0; j<lparm->lbuff.maxx && (i=(int)row)<lparm->lbuff.maxy; j++, row+=dy) {
#include "sblt.c"
      }

      for (; j<lparm->lbuff.maxx && (i=(int)row)<lparm->lbuff.maxy; j++, row+=dy)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_llurv(light *lparm, int inc, float dx) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float col;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (h=inc; h<lparm->lbuff.maxx; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      col = h;

      if (!SILENT)
         printf("x %d\n", h);

      for (i=0; i<lparm->lbuff.maxy && (j=(int)col)<lparm->lbuff.maxx; i++, col+=dx) {
#include "sblt.c"
      }

      for (; i<lparm->lbuff.maxy && (j=(int)col)<lparm->lbuff.maxx; i++, col+=dx)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_lrulh(light *lparm, int inc, float dy) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float row;
   float increment;

                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for ((!inc) ? h=MAXSUBPROC : h=inc; h<lparm->lbuff.maxy; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      row = h;

      if (!SILENT)
         printf("y %d\n", h);

      for (j=lparm->lbuff.maxx-1; j>-1 && (i=(int)row)<lparm->lbuff.maxy; j--, row -= dy) {  // dy is negative - go up
#include "sblt.c"
      }

      for (; j>-1 && (i=(int)row)<lparm->lbuff.maxy; j--, row -= dy)    // dy is negative - go up
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_lrulv(light *lparm, int inc, float dx) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float col;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (h=inc; h<lparm->lbuff.maxx; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      col = h;

      if (!SILENT)
         printf("x %d\n", h);

      for (i=0; i<lparm->lbuff.maxy && (j=(int)col)>-1; i++, col += dx) {  // dx is negative, go left
#include "sblt.c"
      }

      for (; i<lparm->lbuff.maxy && (j=(int)col)>-1; i++, col += dx)    // dx is negative, go left
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_negh(light *lparm, int inc, float dy, float *up, float *horizon) {

   int  g, h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float intersect[2];
   float row;
   float column[3];
   float incup;
   float delta[2], curr[2];

   float yaxis[3];
   float xzero[3] = {1,0,0};

   float op[4];
   float increment;

   yaxis[0] = 0;
   yaxis[1] = 1;
   yaxis[2] = 1.0-lparm->lbuff.maxy;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   copyarray2(column, up);
   incup = -up[1]*MAXSUBPROC;

   if (inc)
      h = inc;
   else
      h = MAXSUBPROC;

   column[2] = -(up[0]*(WINX-1) + up[1]*h);

   line2d_line2d_intersect(column, horizon, curr);
   column[2] += incup;
   line2d_line2d_intersect(column, horizon, delta);
   subarray2(delta, curr);

   for (column[2] = -(up[0]*(WINX-1)+up[1]*h); h<lparm->lbuff.maxy; h+=MAXSUBPROC, column[2]+=incup) {
      if (!SILENT)
         printf("y %d\n", h);

      start_flag[inc] = 1;

      copyarray2(intersect, curr);

      if (intersect[0] >= lparm->lbuff.maxx-1) {
         intersect[0] = lparm->lbuff.maxx-1;
         intersect[1] = h;
      }

      else {
         if (intersect[1] > lparm->lbuff.maxy-1)
            line2d_line2d_intersect(column, yaxis, intersect);

         if (intersect[0] < 0)
            line2d_line2d_intersect(column, xzero, intersect);
      }

      g = j = (int)intersect[0];
      i = (int)(intersect[1] = h - (lparm->lbuff.maxx-1-j)*dy);

      if (i > lparm->lbuff.maxy-1) {
         j++;
         i = (int)(intersect[1] += dy);
      }

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (j++, row=intersect[1]+dy; j<lparm->lbuff.maxx && (i=(int)row)>-1; j++, row+=dy) {
#include "sblt.c"
      }

      for (; j<lparm->lbuff.maxx && (i=(int)row)>-1; j++, row+=dy)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (j=g-1, row=intersect[1]-dy; j>-1 && (i=(int)row)<lparm->lbuff.maxy; j--, row -= dy) {
#include "sblt.c"
      }

      for (; j>-1 && (i=(int)row)<lparm->lbuff.maxy; j--, row -= dy)
         sbbuffer->pdata[i][j] = 0;

      addarray2(curr, delta);
   }

}


void sbfast::sbfastbm_negv(light *lparm,int inc, float dx, float *up, float *horizon) {

   int  g, h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float intersect[2];
   float col;
   float column[3];
   float incup;
   float delta[2], curr[2];

   float yaxis[3];
   float xzero[3] = {1,0,0};

   float op[4];
   float increment;

   yaxis[0] = 0;
   yaxis[1] = 1;
   yaxis[2] = 1.0-lparm->lbuff.maxy;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   copyarray2(column, up);
   incup = -up[0]*MAXSUBPROC;

   h = inc;

   column[2] = -h*up[0];

   line2d_line2d_intersect(column, horizon, curr);
   column[2] += incup;
   line2d_line2d_intersect(column, horizon, delta);
   subarray2(delta, curr);

   for (column[2]=h*up[0]; h<lparm->lbuff.maxx; h+=MAXSUBPROC, column[2]+=incup) {
      if (!SILENT)
         printf("x %d\n", h);

      start_flag[inc] = 1;

      copyarray2(intersect, curr);

      if (intersect[1] <= 0) {
         intersect[0] = h;
         intersect[1] = 0;
      }

      else {
         if (intersect[1] > lparm->lbuff.maxy-1)
            line2d_line2d_intersect(column, yaxis, intersect);

         if (intersect[0] < 0)
            line2d_line2d_intersect(column, xzero, intersect);
      }

      g = i = (int)intersect[1];
      j = (int)(intersect[0] = h + i*dx);

      if (j > lparm->lbuff.maxx-1) {
         i++;
         j = (int)(intersect[0] += dx);
      }

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (i++, col=intersect[0]+dx; i<lparm->lbuff.maxy && (j=(int)col)>-1; i++, col+=dx) {
#include "sblt.c"
      }

      for (; i<lparm->lbuff.maxy && (j=(int)col)>-1; i++, col+=dx)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (i=g-1, col=intersect[0]-dx; i>-1 && (j=(int)col)<lparm->lbuff.maxx; i--, col-=dx) {
#include "sblt.c"
      }

      for (; i>-1 && (j=(int)col)<lparm->lbuff.maxx; i--, col-=dx)
         sbbuffer->pdata[i][j] = 0;

      addarray2(curr, delta);
   }

}


void sbfast::sbfastbm_posh(light *lparm, int inc, float dy, float *up, float *horizon) {

   int  g, h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float intersect[2];
   float row;
   float column[3];
   float incup;
   float delta[2], curr[2];

   float yaxis[3];
   float xaxis[3];

   float op[4];
   float increment;

   yaxis[0] = 0;
   yaxis[1] = 1;
   yaxis[2] = 1.0-lparm->lbuff.maxy;

   xaxis[0] = 1;
   xaxis[1] = 0;
   xaxis[2] = 1.0-lparm->lbuff.maxx;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   copyarray2(column, up);
   incup = -up[1]*MAXSUBPROC;

   if (inc)
      h = inc;
   else
      h = MAXSUBPROC;

   column[2] = -h*up[1];

   line2d_line2d_intersect(column, horizon, curr);
   column[2] += incup;
   line2d_line2d_intersect(column, horizon, delta);
   subarray2(delta, curr);

   for (column[2] = -h*up[1]; h<lparm->lbuff.maxy; h+=MAXSUBPROC, column[2]+=incup) {
      if (!SILENT)
         printf("y %d\n", h);

      start_flag[inc] = 1;

      copyarray2(intersect, curr);

      if (intersect[0] <= 0) {
         intersect[0] = 0;
         intersect[1] = h;
      }

      else {
         if (intersect[1] > lparm->lbuff.maxy-1)
            line2d_line2d_intersect(column, yaxis, intersect);

         if (intersect[0] > lparm->lbuff.maxx-1)
            line2d_line2d_intersect(column, xaxis, intersect);
      }

      g = j = (int)intersect[0];
      i = (int)(intersect[1] = h+j*dy);

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (j++, row=intersect[1]+dy; j<lparm->lbuff.maxx && (i=(int)row)<lparm->lbuff.maxy; j++, row+=dy) {
#include "sblt.c"
      }

      for (; j<lparm->lbuff.maxx && (i=(int)row)<lparm->lbuff.maxy; j++, row+=dy)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (j=g-1, row=intersect[1]-dy; j>-1 && (i=(int)row)>-1; j--, row -= dy) {
#include "sblt.c"
      }

      for (; j<lparm->lbuff.maxx && (i=(int)row)<lparm->lbuff.maxy; j++, row+=dy)
         sbbuffer->pdata[i][j] = 0;

      addarray2(curr, delta);
   }

}


void sbfast::sbfastbm_posv(light *lparm, int inc, float dx, float *up, float *horizon) {

   int  g, h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float intersect[2];
   float col;
   float column[3];
   float incup;
   float delta[2], curr[2];

   float yaxis[3];
   float xaxis[3];

   float op[4];
   float increment;

   yaxis[0] = 0;
   yaxis[1] = 1;
   yaxis[2] = 1.0-lparm->lbuff.maxy;

   xaxis[0] = 1;
   xaxis[1] = 0;
   xaxis[2] = 1.0-lparm->lbuff.maxx;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   copyarray2(column, up);
   incup = -up[0]*MAXSUBPROC;

   h = inc;

   column[2] = -h*up[0];

   line2d_line2d_intersect(column, horizon, curr);
   column[2] += incup;
   line2d_line2d_intersect(column, horizon, delta);
   subarray2(delta, curr);

   for (column[2] = h*up[0]; h<lparm->lbuff.maxx; h+=MAXSUBPROC, column[2]+=incup) {
      if (!SILENT)
         printf("x %d\n", h);

      start_flag[inc] = 1;

      copyarray2(intersect, curr);

      if (intersect[1] <=0) {
         intersect[0] = h;
         intersect[1] = 0;
      }

      else {
         if (intersect[1] > lparm->lbuff.maxy-1)
            line2d_line2d_intersect(column, yaxis, intersect);

         if (intersect[0] > lparm->lbuff.maxx-1)
            line2d_line2d_intersect(column, xaxis, intersect);
      }

      g = i = (int)intersect[1];
      j = (int)(intersect[0] = h + i*dx);

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (i++, col=intersect[0]+dx; i<lparm->lbuff.maxy && (j=(int)col)<lparm->lbuff.maxx; i++, col+=dx) {
#include "sblt.c"
      }

      for (; i<lparm->lbuff.maxy && (j=(int)col)<lparm->lbuff.maxx; i++, col+=dx)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (i=g-1, col=intersect[0]-dx; i>-1 && (j=(int)col)>-1; i--, col -= dx) {
#include "sblt.c"
      }

      for (; i>-1 && (j=(int)col)>-1; i--, col -= dx)
         sbbuffer->pdata[i][j] = 0;

      addarray2(curr, delta);
   }

}


void sbfast::sbfastbm_horizontal(light *lparm, int inc, int dflag) {

   int  i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   int up, down, vinc;
   float increment;

   if (dflag) {
      up   = lparm->lbuff.maxx;
      down = 0;
      vinc = 1;
   }

   else {
      up = -1;
      down = lparm->lbuff.maxx-1;
      vinc = -1;
   }
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (i=inc; i<lparm->lbuff.maxy; i+=MAXSUBPROC) {

      start_flag[inc] = 1;

      if (!SILENT)
         printf("y %d\n", i);
      for (j=down; j != up; j+=vinc)
#include "sblt.c"

      for (; j != up; j+=vinc)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_vertical(light *lparm, int inc, int dflag) {

   int  i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   int up, down, vinc;
   float increment;

   if (dflag) {
      up   = lparm->lbuff.maxy;
      down = 0;
      vinc = 1;
   }

   else {
      up = -1;
//      down = (lparm->lbuff.maxy-1) - ((lparm->lbuff.maxy-1-inc) % MAXSUBPROC);
      down = lparm->lbuff.maxy-1;
      vinc = -1;
   }
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (j=inc; j<lparm->lbuff.maxx; j+=MAXSUBPROC) {

      start_flag[inc] = 1;

      if (!SILENT)
         printf("x %d\n", j);
      for (i=down; i != up; i+=vinc)
#include "sblt.c"

      for (; i != up; i+=vinc)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_horizontald(light *lparm, int inc, int origin) {

   int  i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float op[4];
   float increment;

   if (origin < 0)              // shouldnt do this, but in case of fp error...
      origin = 0;
   else
      if (origin >= lparm->lbuff.maxx)
         origin = lparm->lbuff.maxx-1;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (i=inc; i<lparm->lbuff.maxy; i+=MAXSUBPROC) {

      if (!SILENT)
         printf("y %d\n", i);

      start_flag[inc] = 1;

      j = origin;

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (j--; j > -1; j--)
#include "sblt.c"

      for (; j > -1; j--)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (j=origin+1; j<lparm->lbuff.maxx; j++)
#include "sblt.c"

      for (; j<lparm->lbuff.maxx; j++)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_verticald(light *lparm, int inc, int origin) {

   int  i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];
   float op[4];
   float increment;

   if (origin < 0)              // shouldnt do this, but in case of fp error...
      origin = 0;
   else
      if (origin >= lparm->lbuff.maxy)
         origin = lparm->lbuff.maxy-1;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (j=inc; j<lparm->lbuff.maxx; j+=MAXSUBPROC) {
      if (!SILENT)
         printf("x %d\n", j);

      start_flag[inc] = 1;

      i = origin;

      do {
#include "sblt.c"
      } while (0);

      copyarray4(op, old_pos[inc]);

      for (i--; i > -1; i--)
#include "sblt.c"

      for (; i > -1; i--)
         sbbuffer->pdata[i][j] = 0;

      copyarray4(old_pos[inc], op);

      for (i=origin+1; i<lparm->lbuff.maxy; i++)
#include "sblt.c"

      for (; i<lparm->lbuff.maxy; i++)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_ullrh(light *lparm, int inc, float dy) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float row;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for ((!inc) ? h=MAXSUBPROC : h=inc; h<lparm->lbuff.maxy; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      row = h;

      if (!SILENT)
         printf("y %d\n", h);

      for (j=0; j<lparm->lbuff.maxx && (i=(int)row)>-1; j++, row += dy) {  // dy is negative - go down
#include "sblt.c"
      }

      for (; j<lparm->lbuff.maxx && (i=(int)row)>-1; j++, row += dy)   // dy is negative - go down
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_ullrv(light *lparm, int inc, float dx) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float col;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (h=inc; h<lparm->lbuff.maxx; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      col = h;

      if (!SILENT)
         printf("x %d\n", h);

      for (i=lparm->lbuff.maxy-1; i>-1 && (j=(int)col)<lparm->lbuff.maxx; i--, col -= dx) {  // dx is negative, go right
#include "sblt.c"
      }

      for (; i>-1 && (j=(int)col)<lparm->lbuff.maxx; i--, col -= dx)    // dx is negative, go right
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_urllh(light *lparm, int inc, float dy) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float row;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for ((!inc) ? h=MAXSUBPROC : h=inc; h<lparm->lbuff.maxy; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      row = h;

      if (!SILENT)
         printf("y %d\n", h);

      for (j=lparm->lbuff.maxx-1; j>-1 && (i=(int)row)>-1; j--, row -= dy) {
#include "sblt.c"
      }

      for (; j>-1 && (i=(int)row)>-1; j--, row -= dy)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_urllv(light *lparm, int inc, float dx) {

   int  h, i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float col;
   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (h=inc; h < lparm->lbuff.maxx; h+=MAXSUBPROC) {

      start_flag[inc] = 1;
      col = h;

      if (!SILENT)
         printf("x %d\n", h);

      for (i=lparm->lbuff.maxy-1; i>-1 && (j=(int)col)>-1; i--, col -= dx) {
#include "sblt.c"
      }

      for (; i>-1 && (j=(int)col)>-1; i--, col -= dx)
         sbbuffer->pdata[i][j] = 0;
   }

}


void sbfast::sbfastbm_postscan(light *lparm, int inc) {

   int  i, j, k;
   vector4f surface, normal;
   float pzero[4], pray[4];

   float increment;
                                           // calculate the ray-creep increment
   increment = (isize*(lparm->vrc[1]+lparm->vrc[1]))/(double)lparm->lbuff.maxx;

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;
   matvecmulto(iworld, pzero);

   pray[3] = 0;
   surface[3] = 1;

   for (i=inc; i < lparm->lbuff.maxy; i+=MAXSUBPROC)
      for (j=0; j < lparm->lbuff.maxx; j++)
         if (sbbuffer->pdata[i][j]) {
            start_flag[inc] = 1;
#include "sblt.c"
         }

}

