
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#include <string.h>
#include "sbfast.h"

/* *************************************************************
************************************************************* */
int sbfast::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'b':
         if (!strcmp(token, "back")) {
            get_token(infile, token);
            back = atof(token);

            return 1;
         }

         break;

      case 'd':
         if (!strcmp(token, "displacement")) {
            get_token(infile, shadername);

            return 1;
         }

         break;

      case 'h':
         if (!token[1]) {
            get_token(infile, token);
            H = atof(token);

            return 1;
         }

         break;

      case 'l':
         if (!strcmp(token, "lacunarity")) {
            get_token(infile, token);
            lacunarity = atof(token);

            return 1;
         }

         break;

      case 'o':
         if (!strcmp(token, "octaves")) {
            get_token(infile, token);
            octaves = atof(token);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


/* *************************************************************
************************************************************* */
void sbfast::preprocess(listtype *mainlist, shader *sparm) {

   displacement = sparm->find(shadername);

   pc::preprocess(mainlist, sparm);
}


/* *************************************************************
************************************************************* */
int sbfast::read_data(char filename[], datatype **dptr) {

  return 0;
}


/* *************************************************************
************************************************************* */
int sbfast::read_data(FILE *infile) {

   char  junk[MAXSTRLEN];

   fscanf(infile, "%s %f %s %f %s %f %s %f %s %s", junk, &back, junk, &octaves,
                     junk, &H, junk, &lacunarity, junk, shadername);
   return 1;
}


int sbfast::read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) {

   FILE *infile;
   int  i, j;
   char tempstr[MAXSTRLEN];
   vector4f pt[4];

   for (tob=*(texsbfast **)tpptr; tob!=(texsbfast *)NULL; tob=tob->next)
      if (!strcmp(filename, tob->filename) && !strcmp(dfilename, tob->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CITEXTURE;
      return 0;
   }

   tob = new texsbfast;
   strcpy(tob->filename, filename);
   strcpy(tob->dataname, dfilename);
   tob->next = *(texsbfast **)tpptr;
   *(texsbfast **)tpptr    = tob;

   fscanf(infile, "%d", &tob->countobject);

   tob->setup(tob->countobject);

   for (i=0; i<tob->countobject; i++) {
      fscanf(infile, "%s", tempstr);

      if (!strcmp(tempstr, "COORD"))
         tob->data[i].mapfunc = COORD;
      else
         tob->data[i].mapfunc = REPEAT;

      fscanf(infile, "%s", tempstr);               // texture name
      for (tob->data[i].ob=*tptr; tob->data[i].ob!=(texbase *)NULL; tob->data[i].ob=tob->data[i].ob->next)
         if (!strcmp(tempstr, tob->data[i].ob->filename))
            break;

      if (tob->data[i].ob == (texbase *)NULL) {
         if (query_video(tempstr)) {
            tob->data[i].ob = new texvid;
            tob->data[i].ob->read_texture(tempstr, tptr);
         }

         else {
            tob->data[i].ob = new texture;
            tob->data[i].ob->read_texture(tempstr);
         }

         tob->data[i].ob->next = *tptr;
         *tptr = tob->data[i].ob;
      }

      for (j=0; j<4; j++)
         fscanf(infile, "%f %f", &pt[j][0], &pt[j][1]);
      tob->data[i].setup(pt);
   }

   fclose(infile);
   return 1;
}


int sbfast::read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) {

   int i;
   char buffer[MAXSTRLEN];
   FILE *infile;
   int count;

   for (shaderlist=*slptr; shaderlist!=(shaderlisttype *)NULL; shaderlist=shaderlist->next)
      if (!strcmp(filename, shaderlist->shname) && !strcmp(dfilename, shaderlist->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CI3DTEX;
      return 0;
   }

   shaderlist = new shaderlisttype;
   strcpy(shaderlist->shname, filename);
   strcpy(shaderlist->dataname, dfilename);
   shaderlist->init(1);
   shaderlist->next   = *slptr;
   *slptr       = shaderlist;

   fscanf(infile, "%s %f", buffer, &shaderlist->scale);
   lower_case(buffer);

   if (strcmp(buffer, "mono"))                 //ALL
      shaderlist->shade->read_data(infile, sptr);

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {
      fscanf(infile, "%d", &i);

      if (!(i-1))
         shaderlist->shade->read_data(infile, sptr);
   }

   fclose(infile);
   return 1;
}


/* **************************************************
************************************************** */
int sbfast::bound_box(eye *parm) {

//octaves = 5.5 - log2(dist);
//SCALESB*max(rmf2 with pos, h, lac, octaves, 1, 2, 0)-1

   return 1;
}


/* **************************************************
 * Finds the intersection between line segment [a,b] and ray;
 * determines surface normal by constructing a plane.
 *
 * Note that this could be simplified by using a line/line intersection
 * test between line segment [a,b] and the ray.
************************************************** */
float sbfast::Intersect_Surface(float distance, float increment, float *cam, 
                   float *ray, float *surface, float *normal, int inc) {

   float   vf[3],     /* vector from p to pf */
           vr[3],     /* vector from p to pr */
           d;         /* ax + by + cz + d = 0 */
   float t;
   float p[3];

   copyarray3(p, old_pos[inc]);
   // get two vectors in the surface plane; cross for surface normal

   copyarray2(dis[inc].in, p);
   dis[inc].in[1] += increment;   // one pixel-width to front
   dis[inc].pt[0] = distance;
   displacement->query_data((void *)&dis[inc]);
   vf[2] = dis[inc].out[0] - p[2];
   vf[1] = increment;
   vf[0] = 0;

   copyarray2(dis[inc].in, p);
   dis[inc].in[0] += increment;   // one pixel-width to right
   dis[inc].pt[0] = distance;
   displacement->query_data((void *)&dis[inc]);
   vr[2] = dis[inc].out[0] - p[2];
   vr[1] = 0;
   vr[0] = increment;

   xproduct(normal, vr, vf);
   normalize3(normal);

   d = dotproduct3(normal, p);  /* this is the "-D" distance from the origin to the plane */

   // assign the various hit data 

   if (abs(t=dotproduct3(normal, ray)) < CORRECT)
      t = increment;
   else
      t = ( d - dotproduct3(normal, cam) ) / t;

   surface[0] = cam[0] + t*ray[0];
   surface[1] = cam[1] + t*ray[1];
   surface[2] = cam[2] + t*ray[2];

   return t;
}


/* **************************************************
 * Procedurally ray-trace a terrain given by function Displacement().
 *
 * Algorithm and code by F. Kenton Musgrave, 3/95
 *
 * Brute-force method steps across terrain at pixel-sized steps
 * until an intersection is found.
 *
 * Assumptions:
 *  - bottom-to-top scan
 *  - error inherent in pixel-sized horizontal steps is acceptable
 *  - parameter camera.near_clip is close enough to ray origin
 *  - downward-sloping rays: currently no vertical bound test
 *  - only for eye rays -- shadow rays (for directional light sources)
 *      should use a constant step size: the pixel size at ray origin
 *
 * Possible mods:
 *  - optimize based on prev col calc iff no horizontal jittering
 *  - use derivatives of terrain func to get greater step stride
 *  - add vertical bound test (only for upward-sloping rays)
 *  - modify for shadow rays with constant step size
************************************************** */
int sbfast::Intersect_Terrain(float increment, float *cam, float *ray, float *surface,
                      float *normal, int inc) {

   float   t, dt;              // ray parameter, equal to distance travelled
   float   s;
   float   alt = 0;           // alt at current step
   float   position[3];       // current posiiton along ray

   copyarray3(position, ray);

   t = normalize3(ray);         // get magnitude, then normalize
   s = increment/t;

//printf("t %f s %f sinc %f ", t, s, increment);

   if (start_flag[inc]) {
      start_flag[inc] = 0;
      smultarray3(position, isize);
      addarray3(position, cam);
   }

   else {
      dt = old_pos[inc][3];

      alt = dt-t;

      if (alt > 0) {
         increment = dt*s;
         t = dt;

         position[0] = cam[0] + t*ray[0];
         position[1] = cam[1] + t*ray[1];
         position[2] = cam[2] + t*ray[2];
      }

      else {
         smultarray3(position, isize);
         addarray3(position, cam);
      }

   }

   copyarray2(old_pos[inc], position);
   copyarray2(dis[inc].in, old_pos[inc]);
   dis[inc].pt[0] = t;
   displacement->query_data((void *)&dis[inc]);
   old_pos[inc][2] = dis[inc].out[0];

   if (position[2] < old_pos[inc][2]) {  // if here, then we have "cave/cliff" artifacts
      if (alt > 0) {				// go back only 1 pixel? should go back til no intersect...
         old_pos[inc][0] -= increment*ray[0];
         old_pos[inc][1] -= increment*ray[1];
         copyarray2(dis[inc].in, old_pos[inc]);
         dis[inc].pt[0] = t;
         displacement->query_data((void *)&dis[inc]);
         old_pos[inc][2] = dis[inc].out[0];
      }

      Intersect_Surface(t, increment, cam, ray, surface, normal, inc);

      old_pos[inc][3] = distance3(cam, surface);

      if (alt > 0) {
         position[0] = old_pos[inc][0];
         position[1] = old_pos[inc][1];
         position[2] -= increment*ray[2];

         alt = distance3(cam, position);
         if (alt < old_pos[inc][3])
            old_pos[inc][3] = alt;
      }

      return 1;
   }

   while ( t < back ) { // (add test for top of BB here)
      t += increment;           // the all-important ray-creep increment
      increment *= (1.0+s);

      position[0] += increment*ray[0];
      position[1] += increment*ray[1];
      position[2] += increment*ray[2];

      copyarray2(dis[inc].in, position);
      dis[inc].pt[0] = t;
      displacement->query_data((void *)&dis[inc]);
      alt = dis[inc].out[0];

      if ( position[2] < alt ) {       // surface penetrated
         position[0] = old_pos[inc][0];
         position[1] = old_pos[inc][1];
         position[2] -= increment*ray[2];

         old_pos[inc][3] = distance3(cam, position);

         Intersect_Surface(t, increment, cam, ray, surface, normal, inc);
         return 1;
      }

      old_pos[inc][2] = alt;
      copyarray2(old_pos[inc], position);
   }

   return -1;       // Exceeded far clip distance; update "prev_dist" appropriately  & exit.
}


/* **************************************************
************************************************** */
void sbfast::transform(eye *parm) {

   int i, j;

   copymx4x4(world, rotate);

   isize = 1.0/size;
   back *= isize;

   for (i=0; i<3; i++)
      for (j=0; j<3; j++)
         world[i][j] *= size;

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   matmatmulto(parm->transform, world);

   inversemx(world, iworld);
}


/* **************************************************
************************************************** */
void sbfast::colorize(grandchildtype *child) {

   vector4f pt = {0,0,1,0};             // vertical column
   float one = 1.0-CORRECT;
   vector4f pzero = {0,0,0,1};          // camera origin
   vector4f vnormal = {0,0,1,1};       // vpn
   vector4f vector;                     // vertical vector
   vector4f temp;
   float dx, dy;
   vector4f vertex[4];

   if (mctype != PALEX) {
      child->done = 1;
      return;
   }

   if (child->done == -1) {		// cleanup
      if (child->lit == child->cam)
         sbfastbm_postscan(child->lit, child->id);
      else
         sbfastpg_postscan(child->cam, child->lit, child->zbuff, child->id);

      child->done = 1;
      return;
   }

   matvecmultv(world, pt);            // up vector
   copyarray4(vector, pt);
   normalize3(pt);
   normalize2(pt);

   vertex[0][0] = child->cam->vrc[0];       // corners of viewplane "polygon"
   vertex[0][1] = child->cam->vrc[2];
   vertex[0][2] = -1;
   vertex[0][3] = 1;

   vertex[1][0] = child->cam->vrc[0];
   vertex[1][1] = child->cam->vrc[3];
   vertex[1][2] = -1;
   vertex[1][3] = 1;

   vertex[2][0] = child->cam->vrc[1];
   vertex[2][1] = child->cam->vrc[3];
   vertex[2][2] = -1;
   vertex[2][3] = 1;

   vertex[3][0] = child->cam->vrc[1];
   vertex[3][1] = child->cam->vrc[2];
   vertex[3][2] = -1;
   vertex[3][3] = 1;

   if (child->lit == child->cam) {
      if (line_poly_intersect( pzero, vector, vnormal, vertex, 4, temp)) {

         child->lit->map2screen(temp);

         // if here then we are in the "special case"
         if (abs(pt[2]) > one || abs(pt[1]) > one) // perpendicular to viewplane || vertical
            sbfastbm_verticald(child->lit, child->id, (int)temp[1]);
        else
            if (abs(pt[0]) > one)
               sbfastbm_horizontald(child->lit, child->id, (int)temp[0]);
            else {       // if here then we can do diagonal :)

               dx = pt[0]/pt[1];
               dy = pt[1]/pt[0];

               vector[1] = pt[0];
               vector[0] = -pt[1];
               vector[2] = -(vector[0]*temp[0] + vector[1]*temp[1]);    // column

               pt[2] = -(pt[0]*temp[0] + pt[1]*temp[1]);        // perpendicular line

               if (dx > 0) {
                  sbfastbm_posh(child->lit, child->id, dy, vector, pt);
                  sbfastbm_posv(child->lit, child->id, dx, vector, pt);
               }

               else {
                  sbfastbm_negh(child->lit, child->id, dy, vector, pt);
                  sbfastbm_negv(child->lit, child->id, dx, vector, pt);
               }

            }

      }

      else
         if (abs(pt[1]) > one)
            sbfastbm_vertical(child->lit, child->id, pt[1] >= 0);
         else
            if (abs(pt[0]) > one)
               sbfastbm_horizontal(child->lit, child->id, pt[0] >= 0);
            else        // if here then we can do diagonal :)
               if (pt[0] > 0)
                  if (pt[1] > 0) { // ll to ur
                     sbfastbm_llurh(child->lit, child->id, pt[1]/pt[0]);
                     sbfastbm_llurv(child->lit, child->id, pt[0]/pt[1]);
                  }

                  else { // ul to lr
                     sbfastbm_ullrh(child->lit, child->id, pt[1]/pt[0]);
                     sbfastbm_ullrv(child->lit, child->id, pt[0]/pt[1]);
                  }

               else
                  if (pt[1] > 0) { // lr to ul
                     sbfastbm_lrulh(child->lit, child->id, pt[1]/pt[0]);
                     sbfastbm_lrulv(child->lit, child->id, pt[0]/pt[1]);
                  }

                  else { // urll
                     sbfastbm_urllh(child->lit, child->id, pt[1]/pt[0]);
                     sbfastbm_urllv(child->lit, child->id, pt[0]/pt[1]);
                  }

   }

   else {
      if (line_poly_intersect( pzero, vector, vnormal, vertex, 4, temp)) {

         child->cam->map2screen(temp);

         // if here then we are in the "special case"
         if (abs(pt[2]) > one || abs(pt[1]) > one) // perpendicular to viewplane || vertical
            sbfastpg_verticald(child->cam, child->lit, child->zbuff, child->id, (int)temp[1]);
         else
            if (abs(pt[0]) > one)
               sbfastpg_horizontald(child->cam, child->lit, child->zbuff, child->id, (int)temp[0]);
            else {       // if here then we can do diagonal :)
               dx = pt[0]/pt[1];
               dy = pt[1]/pt[0];

               vector[1] = pt[0];
               vector[0] = -pt[1];
               vector[2] = -(vector[0]*temp[0] + vector[1]*temp[1]);    // column

               pt[2] = -(pt[0]*temp[0] + pt[1]*temp[1]);        // perpendicular line

               if (dx > 0) {
                  sbfastpg_posh(child->cam, child->lit, child->zbuff, child->id, dy, vector, pt);
                  sbfastpg_posv(child->cam, child->lit, child->zbuff, child->id, dx, vector, pt);
               }

               else {
                  sbfastpg_negh(child->cam, child->lit, child->zbuff, child->id, dy, vector, pt);
                  sbfastpg_negv(child->cam, child->lit, child->zbuff, child->id, dx, vector, pt);
               }

            }

      }

      else
         if (abs(pt[1]) > one)
            sbfastpg_vertical(child->cam, child->lit, child->zbuff, child->id, pt[1] >= 0);
         else
            if (abs(pt[0]) > one)
               sbfastpg_horizontal(child->cam, child->lit, child->zbuff, child->id, pt[0] >= 0);
            else        // if here then we can do diagonal :)
               if (pt[0] > 0)
                  if (pt[1] > 0) { // ll to ur
                     sbfastpg_llurh(child->cam, child->lit, child->zbuff, child->id, pt[1]/pt[0]);
                     sbfastpg_llurv(child->cam, child->lit, child->zbuff, child->id, pt[0]/pt[1]);
                  }

                  else { // ul to lr
                     sbfastpg_ullrh(child->cam, child->lit, child->zbuff, child->id, pt[1]/pt[0]);
                     sbfastpg_ullrv(child->cam, child->lit, child->zbuff, child->id, pt[0]/pt[1]);
                  }

               else
                  if (pt[1] > 0) { // lr to ul
                     sbfastpg_lrulh(child->cam, child->lit, child->zbuff, child->id, pt[1]/pt[0]);
                     sbfastpg_lrulv(child->cam, child->lit, child->zbuff, child->id, pt[0]/pt[1]);
                  }

                  else { // urll
                     sbfastpg_urllh(child->cam, child->lit, child->zbuff, child->id, pt[1]/pt[0]);
                     sbfastpg_urllv(child->cam, child->lit, child->zbuff, child->id, pt[0]/pt[1]);
                  }

   }

   child->done = 1;
}


/* **************************************************
************************************************** */
void sbfast::scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr) {

   unsigned int i, j;
   grandchildtype *offspring;
   float deg;

   deg = ACOS(1.0/sqrt(1 + SQR(cparm->vrc[1])));    // fov wrt x axis
   deg += deg;
   deg = LOG2(tan(deg/WINX)) + 3.92;          // usually around ~ -5, relates Noise() to Nyquist limit
// LOG2(tan(HALFPI/ WINX)) + 3.5

   transform(cparm);

   sbbuffer = new mapuc;
   sbbuffer->init_map(WINX, WINY);
   for (i=0, j=WINX*WINY; i<j; i++)
      sbbuffer->data[i] = 1;

   j = MAXSUBPROC-1;

   offspring = new grandchildtype[MAXSUBPROC];
   dis = new shaderparamtype[MAXSUBPROC];
         
   for (i=0; i<j; i++) {                  // interlace lines
      dis[i].octaves = deg;
      dis[i].H = H;
      dis[i].lacunarity = lacunarity;
      dis[i].freq = freq;

      offspring[i].cam    = cparm;
      offspring[i].lit    = lparm;
      offspring[i].zbuff  = zbuff;
      offspring[i].source = this;
      offspring[i].id     = i;
      offspring[i].done   = 0;
      grandchild[i].spawn(&offspring[i]);
   }

   dis[j].octaves = deg;
   dis[j].H = H;
   dis[j].lacunarity = lacunarity;
   dis[j].freq = freq;

   offspring[j].cam    = cparm;
   offspring[j].lit    = lparm;
   offspring[j].zbuff  = zbuff;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;
   colorize(&offspring[j]);

   for (i=0; i<j; i++)
      while (!offspring[i].done);

   for (i=0; i<j; i++) {                  // interlace lines
      offspring[i].done   = -1;
      grandchild[i].spawn(&offspring[i]);
   }

   offspring[j].done   = -1;
   colorize(&offspring[j]);

   for (i=0; i<j; i++)
      while (offspring[i].done < 1);

   for (i=0; i<MAXSUBPROC; i++)
      dis[i].freq = (float *)NULL;

   delete offspring;
   delete sbbuffer;

//	something causes a seg vio w/ "fclose" in write_rle when this is called
#ifdef SGI
   delete dis;
#endif
}


/* **************************************************
************************************************** */
void sbfast::scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

}


/* **************************************************
************************************************** */
void sbfast::beamscan(light *lparm, subslave grandchild[]) {

   int i, j;
   grandchildtype *offspring;
   float deg;

   deg = ACOS(1.0/sqrt(1 + SQR(lparm->vrc[1])));    // fov wrt x axis
   deg += deg;
   deg = LOG2(tan(deg/WINX)) + 3.92;          // usually around ~ -5, relates Noise() to Nyquist limit
// LOG2(tan(HALFPI/ WINX)) + 3.5

   transform(lparm);

   sbbuffer = new mapuc;
   sbbuffer->init_map(lparm->lbuff.maxx, lparm->lbuff.maxy);
   for (i=0, j=lparm->lbuff.maxx, lparm->lbuff.maxy; i<j; i++)
      sbbuffer->data[i] = 1;

   j = MAXSUBPROC-1;

   offspring = new grandchildtype[MAXSUBPROC];
   dis = new shaderparamtype[MAXSUBPROC];

   for (i=0; i<j; i++) {                  // interlace lines
      dis[i].octaves = deg;
      dis[i].H = H;
      dis[i].lacunarity = lacunarity;
      dis[i].freq = freq;

      offspring[i].cam    = offspring[i].lit = lparm;
      offspring[i].source = this;
      offspring[i].id     = i;
      offspring[i].done   = 0;
      grandchild[i].spawn(&offspring[i]);
   }

   dis[j].octaves = deg;
   dis[j].H = H;
   dis[j].lacunarity = lacunarity;
   dis[j].freq = freq;

   offspring[j].cam    = offspring[j].lit = lparm;
   offspring[j].source = this;
   offspring[j].id     = j;
   offspring[j].done   = 0;
   colorize(&offspring[j]);

   for (i=0; i<j; i++)
      while (!offspring[i].done);

   for (i=0; i<j; i++) {                  // interlace lines
      offspring[i].done   = -1;
      grandchild[i].spawn(&offspring[i]);
   }

   offspring[j].done   = -1;
   colorize(&offspring[j]);

   for (i=0; i<j; i++)
      while (offspring[i].done < 1);

   for (i=0; i<MAXSUBPROC; i++)
      dis[i].freq = (float *)NULL;

   delete offspring;
   delete sbbuffer;

//	something causes a seg vio w/ "fclose" in write_rle when this is called
#ifdef SGI
   delete dis;
#endif
}


