
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





#if !defined (__SBFAST_H)
#define __SBFAST_H

#include "base.h"

/* *************************************************************
************************************************************* */
class sbfast : public pc {

   protected:

      float   back;                     // max distance from camera
      float   octaves;
      float   H;
      float   lacunarity;
      float   *freq;
      shaderparamtype *dis;
      shader  *displacement;
      float   isize;

      texsbfast *tob;
      mapuc *sbbuffer;

      char shadername[MAXSTRLEN];

      volatile vector4f *old_pos;
      volatile int   *start_flag;

      int bound_box(eye *parm);
      void  transform(eye *parm);

      float Intersect_Surface(float distance, float increment,
                      float *cam, float *ray, float *surface, float *normal, int inc);
      int Intersect_Terrain(float increment, float *cam, float *ray,
                      float *surface, float *normal, int inc);

      void sbfastpg_vertical(eye *parm, light *lparm, zbuffer *zbuff, int inc, int dflag);
      void sbfastpg_horizontal(eye *parm, light *lparm, zbuffer *zbuff, int inc, int dflag);

      void sbfastpg_llurh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy);
      void sbfastpg_llurv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx);

      void sbfastpg_urllh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy);
      void sbfastpg_urllv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx);

      void sbfastpg_ullrv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx);
      void sbfastpg_ullrh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy);

      void sbfastpg_lrulv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx);
      void sbfastpg_lrulh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy);

      void sbfastpg_verticald(eye *parm, light *lparm, zbuffer *zbuff, int inc, int midy);
      void sbfastpg_horizontald(eye *parm, light *lparm, zbuffer *zbuff, int inc, int midx);

      void sbfastpg_posv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx, float *up, float *horizon);
      void sbfastpg_posh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy, float *up, float *horizon);

      void sbfastpg_negv(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dx, float *up, float *horizon);
      void sbfastpg_negh(eye *parm, light *lparm, zbuffer *zbuff, int inc, float dy, float *up, float *horizon);

      void sbfastpg_postscan(eye *parm, light *lparm, zbuffer *zbuff, int inc);

      void sbfastbm_vertical(light *lparm, int inc, int dflag);
      void sbfastbm_horizontal(light *lparm, int inc, int dflag);

      void sbfastbm_llurh(light *lparm, int inc, float dy);
      void sbfastbm_llurv(light *lparm, int inc, float dx);

      void sbfastbm_urllh(light *lparm, int inc, float dy);
      void sbfastbm_urllv(light *lparm, int inc, float dx);

      void sbfastbm_ullrv(light *lparm, int inc, float dx);
      void sbfastbm_ullrh(light *lparm, int inc, float dy);

      void sbfastbm_lrulv(light *lparm, int inc, float dx);
      void sbfastbm_lrulh(light *lparm, int inc, float dy);

      void sbfastbm_verticald(light *lparm, int inc, int midy);
      void sbfastbm_horizontald(light *lparm, int inc, int midx);

      void sbfastbm_posv(light *lparm, int inc, float dx, float *up, float *horizon);
      void sbfastbm_posh(light *lparm, int inc, float dy, float *up, float *horizon);

      void sbfastbm_negv(light *lparm, int inc, float dx, float *up, float *horizon);
      void sbfastbm_negh(light *lparm, int inc, float dy, float *up, float *horizon);

      void sbfastbm_postscan(light *lparm, int inc);

   public:

      sbfast() {
         back = 1000;
         octaves = 100;
         H = 0.9;
         lacunarity = 2.1;
         displacement = (shader *)NULL;       
//         SCALESB = 157.241;
         strcpy(shadername, "default_alt");
      }

      rendertype query_whatami() { return SBFAST; }

      void begin_scan() {

         freq = (float *)NULL;
         start_flag = new int[MAXSUBPROC];
         old_pos = new vector4f[MAXSUBPROC];

         if (displacement == (shader *)NULL)
            displacement = shades->find("default_alt");
      }

      void end_scan() {

         if (freq != (float *)NULL) {
            delete freq;
            freq = (float *)NULL;
         }

         delete start_flag;
         delete old_pos;
      }

      void datacopy() {

         float p[3] = {0,0,0};

         fBm(p, H, lacunarity, octaves, 1, &freq);

         if (mctype == PALEX)
            base_color = &lob->base;
      }

      void beamscan(light *lparm, subslave grandchild[]);

      void colorize(grandchildtype *child);

      void scan(camera *cparm, light *lparm, zbuffer *zbuff, subslave grandchild[], pc **shadptr);
      void scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr);

      int  parse(FILE *infile, char *token);
      void preprocess(listtype *mainlist, shader *sparm);

      int read_data(char filename[], datatype **dptr);
      int read_data(FILE *infile);
      int  read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr);
      int  read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr);
};

#endif
