
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */


#include <math.h>
#include <string.h>
#include "shader.h"

float terrain_shades[3][3] = {1, 0, 0,
                              0.628, 0, 0.08,
                              1, 0, 0.236};

shader *shades = (shader *)NULL;

void shadertype::read_data(FILE *infile, shader *sptr) {

   char tmpstr[MAXSTRLEN], tmpstr2[MAXSTRLEN];
   int j;

   fscanf(infile, "%d", &scount);

   if (s != (pshader *)NULL)
      delete s;

   if (stype != (shader_type *)NULL)
      delete stype;

   s = new pshader[scount];
   stype = new shader_type[scount];

   for (j=0; j<scount; j++) {
      fscanf(infile, "%s %s", tmpstr, tmpstr2);

      if (tmpstr2[0] == 'A')
         stype[j] = ATMOSPHERE;
      else
         stype[j] = SURFACE;

      if ((s[j] = sptr->find(tmpstr)) == (shader *)NULL)
         printf("Can't find shader %s\n", tmpstr);
   }

}


/* *************************************************************************
************************************************************************* */

#ifdef SGI
int shader::set(char *tmpstr, void *func_call) {
#else
int shader::set(char *tmpstr, int func_call(void *)) {
#endif

   strcpy(name, tmpstr);
   shade_fcn = func_call;
   return 1;
}


/***************************************************************************
****************************************************************************/
shader *shader::find(char *target) {

   int x;

   if (!(x = strcmp(target, name)))
      return this;

   if (x < 0)
      if (left != (shader *)NULL)
         return left->find(target);
      else
         return (shader *)NULL;

   if (right != (shader *)NULL)
      return right->find(target);

   return (shader *)NULL;
}


/***************************************************************************
****************************************************************************/
void shader::insert(shader *x) {

   if (strcmp(x->name, name) < 0)
      if (left == (shader *)NULL)
         left = x;
      else
         left->insert(x);
   else
      if (right == (shader *)NULL)
         right = x;
      else
         right->insert(x);
}


/***************************************************************************
****************************************************************************/
//void insert_shader(shader *head, char *str, void *fcn) {
void insert_shader(shader *head, char *str, int fcn(void *)) {

   shader *ptr;

   ptr = new shader;
   ptr->set(str, fcn);
   head->insert(ptr);
}


/***************************************************************************
****************************************************************************/
float grey_marble_color(float value) {

   if (value < 0.4)
      return 1 - 0.571429*value;
   if (value < 0.8)
      return 1.125 - 0.75 * value;

   return 1.7866336 - 1.4851485 * value;
}


/***************************************************************************
****************************************************************************/
float sub_marble_color(float value) {

   if (0.8 <= value)
      return -value + 1.2;

   return -0.5*value + 0.8;
}


/***************************************************************************
****************************************************************************/
float main_marble_color(float value) {

   if (value < 0.3)
      return 1.0;
   if ( 0.3 <= value < 0.7)
      return -1.0*value + 1.3;
   if ( 0.7 <= value < 0.9)
      return -0.75*value + 0.315;

   return -0.015*value + 0.315;
}


/***************************************************************************
* marble - calculate marble color for given point.
*
* Input : float point[3]
*
* Output: float color[3]
*
* Fc call: turbulence, marble_color;
****************************************************************************/
int marble (void *x) {

//   float  size = 2;             // high freq
   float  isize = 0.5;
   float  ans;
   int    freq = 13;
   vector4f point;

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   copyarray4(point, data->in);
   smultarray3(point, isize);

//   out[0] = main_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));
//   out[1] = sub_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));

   ans = grey_marble_color(SIN((point[0]+turbulence(point, 300)) * freq * PI));

   smultarray3(data->out, ans);

   return 1;
}


/***************************************************************************
****************************************************************************/
int wood (void *x) {

   float  size = 2;             // high freq
   float  isize = 0.5;
   float  ans;
   int    freq = 13;
   float   point[4];

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   copyarray4(point, data->in);

   point[0] *= isize;
   point[1] *= isize;
   point[2] *= isize;

//   out[0] = main_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));
//   out[1] = sub_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));

   ans = main_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));

   data->out[0] *= ans*1.0;
   data->out[1] *= ans*0.6;
   data->out[2] *= ans*0.30196;

   return 1;
}


/***************************************************************************
****************************************************************************/
int wood_panel (void *x) {

   float  size = 2;             // high freq
   float  isize = 0.5;
   float  ans;
   int    freq = 13;
   float   point[4];

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   point[0] = data->in[0] + 0.005 - round0(data->in[0]/4.0)*4.0;	 // 4x1x8
   point[1] = data->in[1] + 0.005 - round0(data->in[1]);
   point[2] = data->in[2] + 0.005 - round0(data->in[2]/8.0)*8.0;

   if (point[0] < 0)
      point[0] = 4.0 + point[0];

   if (point[1] < 0)
      point[1] = 1.0 + point[1];

   if (point[2] < 0)
      point[2] = 8.0 + point[2];

   if ((point[0] > 0.01 && point[0] < 0.2) ||
       (point[1] > 0.01 && point[1] < 0.2) ||
       (point[2] > 0.01 && point[2] < 0.2)) { 
      data->out[0] *= 0.1;
      data->out[1] *= 0.1;
      data->out[2] *= 0.1;
      return 1;
   }

   point[0] *= isize;
   point[1] *= isize;
   point[2] *= isize;

   ans = main_marble_color(SIN((point[0]+turbulence(point, size)) * freq * PI));

   data->out[0] *= ans*1.0;
   data->out[1] *= ans*0.6;
   data->out[2] *= ans*0.30196;

   return 1;
}


/***************************************************************************
****************************************************************************/
int moon (void *x) {

   float high[3] = {0,0.7,0.8};
   float basehi[3] = {0.1,0.3,0.2};

   float low[3]  = {0.1,0.4,0.3};
   float baselow[3] = {0.1,0.4, 0.6};
   float value;

//   float maxchange = 0.05;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0.0;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*2;
   p[1] = data->in[1]*2;
   p[2] = data->in[2]*2;

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if ( value > threshold) {
      data->out[0] *= (basehi[0] + high[0]*value);
      data->out[1] *= (basehi[1] + high[1]*value);
      data->out[2] *= (basehi[2] + high[2]*value);
   }

   else {
      data->out[0] *= (baselow[0] + low[0]*value);
      data->out[1] *= (baselow[1] + low[1]*value);
      data->out[2] *= (baselow[2] + low[2]*value);
   }

/*
   value *= maxchange;

   data->pt[0] += data->pt[0]*value * data->normal[0];
   data->pt[1] += data->pt[1]*value * data->normal[1];
   data->pt[2] += data->pt[2]*value * data->normal[2];
*/

   return 1;
}


int planet2 (void *x) {

   float high[3] = {0,1.254,0.2};
   float basehi[3] = {0.429,0.373,0.135};

   float high2[3] = {0.196,0.765,0};

   float low[3]  = {0.1,0.4,0.3};
   float baselow[3] = {0.1,0.4, 0.6};

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = -0.134544;
   float threshold2 = 0.134544;
   float threshold3 = 0.764352;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*2.9987983;
   p[1] = data->in[1]*2.9987983;
   p[2] = data->in[2]*2.9987983;

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if (abs(data->in[1] + 0.3*value) > 0.7) {
      value = 0.25*(3+value);
      data->out[0] = value;
      data->out[1] = value;
      data->out[2] = value;
   }

   else
      if ( value > threshold) {
         if (value < threshold2) {
            data->out[0] *= basehi[0];
            data->out[1] *= basehi[1];
            data->out[2] *= basehi[2];
         }

         else if (value < threshold3) {
            data->out[0] *= (basehi[0] + high[0]*value);
            data->out[1] *= (basehi[1] + high[1]*value);
            data->out[2] *= (basehi[2] + high[2]*value);
         }

         else {
            data->out[0] *= (basehi[0] + high2[0]*value);
            data->out[1] *= (basehi[1] + high2[1]*value);
            data->out[2] *= (basehi[2] + high2[2]*value);
         }

      }

      else {
         data->out[0] *= (baselow[0] + low[0]*value);
         data->out[1] *= (baselow[1] + low[1]*value);
         data->out[2] *= (baselow[2] + low[2]*value);
      }

   return 1;
}


int planet (void *x) {

   float high[3] = {0,1.254,0.2};
   float basehi[3] = {0.429,0.373,0.135};

   float high2[3] = {0.196,0.765,0};

   float low[3]  = {0.1,0.4,0.3};
   float baselow[3] = {0.1,0.4, 0.6};

   float value;
//   float maxchange = 0.05;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0;
   float threshold2 = 0.134544;
   float threshold3 = 0.764352;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*9.9987983;
   p[1] = data->in[1]*9.9987983;
   p[2] = data->in[2]*9.9987983;

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if ( value > threshold) {
      if (value < threshold2) {
         data->out[0] *= basehi[0];
         data->out[1] *= basehi[1];
         data->out[2] *= basehi[2];
      }

      else if (value < threshold3) {
         data->out[0] *= (basehi[0] + high[0]*value);
         data->out[1] *= (basehi[1] + high[1]*value);
         data->out[2] *= (basehi[2] + high[2]*value);
      }

      else {
         data->out[0] *= (basehi[0] + high2[0]*value);
         data->out[1] *= (basehi[1] + high2[1]*value);
         data->out[2] *= (basehi[2] + high2[2]*value);
      }

   }

   else {
      data->out[0] *= (baselow[0] + low[0]*value);
      data->out[1] *= (baselow[1] + low[1]*value);
      data->out[2] *= (baselow[2] + low[2]*value);
   }

/*
   value *= maxchange;
   data->pt[0] += data->pt[0]*value * data->normal[0];
   data->pt[1] += data->pt[1]*value * data->normal[1];
   data->pt[2] += data->pt[2]*value * data->normal[2];
*/

   return 1;
}


/***************************************************************************
****************************************************************************/
int cloud (void *x) {

   float value;

   float high     = 0.4;
   float basehigh = 0.175;
   float low      = 0.175;
   float baselow  = 0.15;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0.0;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*0.09472388;
   p[1] = data->in[1]*0.09472388;
   p[2] = data->in[2]*0.09472388;

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if (value > threshold)
      value = basehigh + high*value;
   else
      value = baselow + low*value;

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] *= value;

   return 1;
}


/***************************************************************************
****************************************************************************/
int cloud2 (void *x) {

   float value;

   float high     = 0.4;
   float basehigh = 0.475;
   float low      = 0.275;
   float baselow  = 0.15;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0.1;

   float p[3];
   shaderparamtype *data;

   marble(x);

   data = (shaderparamtype *)x;

   p[0] = data->in[0]; ///10.557;
   p[1] = data->in[1]; ///10.557;
   p[2] = data->in[2]; ///10.557;

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if (value > threshold)
      value = basehigh + high*value;
   else
      value = baselow + low*value;

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] *= value;

   return 1;
}


/***************************************************************************
****************************************************************************/
int warp (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0.6;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   copyarray3(p, data->in);

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if (value < threshold) {
      data->out[0] = data->out[1] = data->out[2] = 0;
      return 1;
   }

   if ((p[2]=abs(data->in[2])) > 60)
      value *= 60/p[2];

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] *= value;

   return 1;
}


/***************************************************************************
****************************************************************************/
int stars (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float threshold = 0.6;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = abs(data->in[0]);
   p[1] = data->in[1];
   p[2] = data->in[2];

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      value = fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      value = fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   if (value < threshold) {
      data->out[0] = data->out[1] = data->out[2] = 0;
      return 1;
   }

   p[0] += value;
   p[1] -= value;
   p[2] += value;

   value = fBm(p, H, lacunarity, octaves, 0, &data->freq);

   if (value < threshold) {
      data->out[0] = data->out[1] = data->out[2] = 0;
      return 1;
   }

   value = abs(value);

   if (value > 1)
      value -= (int)value;

   if ((p[2]=-data->pt[2]) > 100)
      value *=100/p[2];
   else
      value += 0.5*(1-value);

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] *= value;

   return 1;
}


/***************************************************************************
****************************************************************************/
int wave(void *x) {

   float value;

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   value = turbulence(data->in, 5.745);

   data->normal[0] += value;

   data->normal[1] += value;

   data->normal[2] += value;

   normalize3(data->normal);

   return 1;
}


/***************************************************************************
****************************************************************************/
int terrain (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float p[3], p2[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*0.0002;
   p[1] = data->in[1]*0.0002;
   p[2] = abs(data->in[2])*0.0005;                      // remove negative to get bluemoon.rgb

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   p[2] *= 2000;

   if (p[2] < 100) {            // water
      p[2] *= 0.01;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.1;
      data->out[1] *= 0.3 - 0.2*value;
      data->out[2] *= 0.6 - 0.25*value;
      return 0;
   }

   copyarray3(p2, data->in);

   data->in[0] *= 0.0005;
   data->in[1] *= 0.0005;
   data->in[2] *= 0.0005;

   marble(data);

   copyarray3(data->in, p2);

   if (p[2] > 1600) {           // snow
      p[2] *= 0.0005;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      value = 0.5 * (1+ value);

      data->out[0] *= value;
      data->out[1] *= value;
      data->out[2] *= value;

      return 1;
   }

   if (p[2] > 1000) {           // red rock
      p[2] *= 0.000625;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.5*(0.6+value);
      data->out[1] *= 0.1*(0.6+value);
      data->out[2] *= 0.1;
      return 1;
   }

   if (p[2] > 200) {            // vegetation
      p[2] *= 0.001;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.5*(1-value);
      data->out[1] *= 0.5*(1+value);
      data->out[2] *= 0.1;
      return 1;
   }

   p[2] *= 0.005;               // sand

   value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

   value = 0.3 * (1+value);

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] = 0;

   return 1;
}


/***************************************************************************
****************************************************************************/
int terrain2 (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float p[3], p2[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*0.002;
   p[1] = data->in[1]*0.002;
   p[2] = data->in[2]*0.006666667;                      // remove negative to get bluemoon.rgb

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   p[2] *= 150;

   if (p[2] < -0.09) {            // water deep
      p[2] *= 0.006666667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.02;
      data->out[1] *= 0.14 - 0.09*value;
      data->out[2] *= 0.28 - 0.12*value;

      return 1;
   }

   if (p[2] < 0.01) {            // water
      p[2] *= 0.006666667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.1;
      data->out[1] *= 0.3 - 0.2*value;
      data->out[2] *= 0.6 - 0.25*value;

      copyarray3(p2, data->in);

      data->in[0] *= 3.5473;
      data->in[1] *= 3.5473;
      data->in[2] *= 3.5473;

      wave(data);

      copyarray3(data->in, p2);

      return 1;
   }

   copyarray3(p2, data->in);

   data->in[0] *= 0.0005;
   data->in[1] *= 0.0005;
   data->in[2] *= 0.001;

   marble(data);

   copyarray3(data->in, p2);

   if (p[2] > 110) {           // snow
      p[2] *= 0.006666667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));
      value = 0.5 * (1+ value);

      data->out[0] *= value;
      data->out[1] *= value;
      data->out[2] *= value;

      return 1;
   }

   if (p[2] > 70) {           // red rock
      p[2] *= 0.009090909;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[2] *= 0.1 * (1+ value);

      value = 0.08 * (1+ value);

      data->out[0] *= value;
      data->out[1] *= value;

      return 1;
   }

   if (p[2] > 10) {            // vegetation
      p[2] *= 0.014285714;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.5*(1-value);
      data->out[1] *= 0.5*(1+value);
      data->out[2] *= 0.1;
      return 1;
   }

   p[2] *= 0.1;               // sand

   value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));
   value = 0.3 * (1+ value);

   data->out[0] *= value;
   data->out[1] *= value;
   data->out[2] = 0;

   return 1;
}


/***************************************************************************
****************************************************************************/
int terrain3 (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float p[3], p2[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

   p[0] = data->in[0]*0.002;
   p[1] = data->in[1]*0.002;
   p[2] = data->in[2]*0.0066667;                      // remove negative to get bluemoon.rgb

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   p[2] *= 150;

   if (p[2] < -0.09) {            // water deep
      p[2] *= 0.0066667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.02;
      data->out[1] *= 0.14 - 0.09*value;
      data->out[2] *= 0.28 - 0.12*value;

      return 1;
   }

   if (p[2] < 0.01) {            // water
      p[2] *= 0.0066667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.1;
      data->out[1] *= 0.3 - 0.2*value;
      data->out[2] *= 0.6 - 0.25*value;

      return 1;
   }

   copyarray3(p2, data->in);

   data->in[0] *= 0.0005;
   data->in[1] *= 0.0005;
   data->in[2] *= 0.001;

   marble(data);

   copyarray3(data->in, p2);

   if (p[2] > 100) {           // snow
      p[2] *= 0.0066667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));
      value = 0.5 * (1+ value);

      data->out[0] *= value;
      data->out[1] *= value;
      data->out[2] *= value;

      return 1;
   }

   if (p[2] > 80) {
      p[2] *= 0.01;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.73 + 0.27*value;
      data->out[1] *= 0.75 + 0.25*value;
      data->out[2] *= 0.69 + 0.31*value;
      return 1;
   }

   if (p[2] > 60) {
      p[2] *= 0.0125;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.79 - 0.06*value;
      data->out[1] *= 0.75;
      data->out[2] *= 0.59 + 0.10*value;
      return 1;
   }

   if (p[2] > 40) {
      p[2] *= 0.0166667;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.70 + 0.09*value;
      data->out[1] *= 0.61 + 0.14*value;
      data->out[2] *= 0.36 + 0.23*value;
      return 1;
   }

   if (p[2] > 20) {
      p[2] *= 0.025;

      value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

      data->out[0] *= 0.34 + 0.36*value;
      data->out[1] *= 0.65 - 0.04*value;
      data->out[2] *= 0.04 + 0.32*value;
      return 1;
   }

   p[2] *= 0.05;

   value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

   data->out[0] *= 0.29 + 0.06*value;
   data->out[1] *= 0.36 + 0.29*value;
   data->out[2] = 0.04*value;

   return 1;
}


/***************************************************************************
****************************************************************************/
int checker (void *x) {

   int value;

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   value = abs((int)data->in[0]);
   value += abs((int)data->in[0]);
   value += abs((int)data->in[0]);

   if (value & 0x01)
      data->out[0] = data->out[1] = data->out[2] = 0;

   return 1;
}


/***************************************************************************
        note: this is an alpha blending shader - should be the last
                shader called...
****************************************************************************/
int fog_mus(void *x) {

   float ray_dir[3];
   float t_e;                           // distance to object (i think)
   float alpha[3];
   float beta;
   float bsqrt;
   float b = 1;                         // fallout w/ altitude
   float a = 0.5;                       // average fog density
   float near_val, far_val;
   float origin[3] = {0, 0, -100};
   float rc_square;
   float tau[3];

   shaderparamtype *data;

   data = (shaderparamtype *)x;

                // compute distance from fog origin to rays closest approach
   alpha[0] = -origin[0];
   alpha[1] = -origin[1];
   alpha[2] = -origin[2];

   copyarray3(ray_dir, data->pt);

   t_e = normalize3(ray_dir);

   beta = abs(dotproduct3(ray_dir, alpha));    // calc fallout w/ altitude

   bsqrt = sqrt(b);
   near_val = erf(bsqrt*beta);
   far_val = erf(bsqrt*(beta+t_e));

                // comput dist from fog origin to rays closest approach

   rc_square = dotproduct3(alpha, alpha) - SQR(beta);

   tau[0] = a/bsqrt*exp(-b*rc_square)*(far_val-near_val);
   tau[1] = tau[0]*2.25;
   tau[2] = tau[0]*21.0;

   tau[0] = exp(-tau[0]);
   tau[1] = exp(-tau[1]);
   tau[2] = exp(-tau[2]);

   data->out[0] = data->out[0]*tau[0] + (1-tau[0]);     // alpha blend color with white
   data->out[1] = data->out[1]*tau[1] + (1-tau[1]);     // alpha blend color with white
   data->out[2] = data->out[2]*tau[2] + (1-tau[2]);     // alpha blend color with white

   return 1;
}


/***************************************************************************
        note: this is an alpha blending shader - should be the last
                shader called...
		based on height
****************************************************************************/
int fog(void *x) {

   float b = 0.007;                      // fallout w/ altitude
   float a = 0.0004;                     // average fog density
   float tau, taui;
   float distance;
   float height;

   shaderparamtype *data;

   data = (shaderparamtype *)x;

   height = abs(data->pt[1]);

   distance = sqrt(dotproduct3(data->pt, data->pt));

   tau = a*distance*exp(-b*height);
   if (tau > 1.0)
      tau = 1.0;

   taui = 1.0 - tau;

   data->add[0] += tau;
   data->add[1] += tau;
   data->add[2] += tau;

   data->out[0] = data->out[0]*taui;     // alpha blend color with white
   data->out[1] = data->out[1]*taui;     // alpha blend color with white
   data->out[2] = data->out[2]*taui;     // alpha blend color with white

   return 1;
}


/***************************************************************************
        note: this is an alpha blending shader - should be the last
                shader called...

	based on distance
****************************************************************************/
int fog2(void *x) {

//   float b = 0.0007;                      // fallout w/ altitude
   float a = 0.0004;                     // average fog density
   float tau, taui;
   float distance;
//   float height;

   shaderparamtype *data;

   data = (shaderparamtype *)x;

//   height = abs(data->pt[1]);

   distance = sqrt(dotproduct3(data->pt, data->pt));

//   tau = b*height*exp(-a*distance);
   tau = exp(-a*distance);

   if (tau > 1.0)
      tau = 1.0;

   taui = 1.0 - tau;

   data->add[0] += taui;
   data->add[1] += taui;
   data->add[2] += taui;

   data->out[0] = data->out[0]*tau;     // alpha blend color with white
   data->out[1] = data->out[1]*tau;     // alpha blend color with white
   data->out[2] = data->out[2]*tau;     // alpha blend color with white

   return 1;
}


/***************************************************************************
****************************************************************************/
int cave_floor (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

			// calc general color
   p[0] = data->in[0]*0.002;
   p[1] = data->in[1]*0.002;
   p[2] = data->in[2]*0.006666667;                      // remove negative to get bluemoon.rgb

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

				// red bands
   switch (((int)(value*6.0)) % 6)	{
      case 0:
         data->out[0] *= value*terrain_shades[0][0];
         data->out[1] *= value*terrain_shades[0][1];
         data->out[2] *= value*terrain_shades[0][2];
         break;

      case 2:
         data->out[0] *= value*terrain_shades[1][0];
         data->out[1] *= value*terrain_shades[1][1];
         data->out[2] *= value*terrain_shades[1][2];
         break;

      default:
         data->out[0] *= value*terrain_shades[2][0];
         data->out[1] *= value*terrain_shades[2][1];
         data->out[2] *= value*terrain_shades[2][2];
         break;
   }

   return 1;
}


/***************************************************************************
****************************************************************************/
int cave_ceiling (void *x) {

   float value;

   float H = 0.632;
   float lacunarity = 2;
   float octaves = 7;

   float p[3];
   shaderparamtype *data;

   data = (shaderparamtype *)x;

			// calc general color
   p[0] = data->in[0]*0.002;
   p[1] = data->in[1]*0.002;
   p[2] = data->in[2]*0.006666667;                      // remove negative to get bluemoon.rgb

   if (data->H == H && data->lacunarity == lacunarity && data->octaves >= octaves)
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 0, &data->freq);
   else {
      p[2] *= 1 + 0.732743*fBm(p, H, lacunarity, octaves, 1, &data->freq);
      data->H = H;
      data->lacunarity = lacunarity;
      data->octaves = octaves;
   }

   value = abs(fBm(p, H, lacunarity, octaves, 0, &data->freq));

/*
   value += 1;

   data->out[2] *= 0.1*value;

   value = 0.08*value;

   data->out[0] *= value;
   data->out[1] *= value;
*/

   switch (((int)(value*6.0)) % 6)	{
      case 0:
         value *= 0.5;
         data->out[0] *= value;
         data->out[1] *= value;
         data->out[2] *= value*1.2;
         break;

      case 2:
         value *= 0.5;
         data->out[0] *= value*1.1;
         data->out[1] *= value;
         data->out[2] *= value*1.2;
         break;

      default:
         value *= 0.5;
         data->out[0] *= value;
         data->out[1] *= value;
         data->out[2] *= value;
         break;
   }

   return 1;
}


/* *************************************************************
  calculate level of detail, based on distance from eye
		run, chase, sbfast pics

	pt[0] - distance
        in - pos;
        octaves - octave_scalar
        out[0] - height

#define SBSCALE 0.0063596645
#define SCALESB 157.241

************************************************************* */
int default_alt(void *x) {

   shaderparamtype *data = (shaderparamtype *)x;
   float octaves;
   float dist;

   if (data->pt[0] < 1)
      dist = 1;
   else
      dist = data->pt[0];

   octaves = -(data->octaves + LOG2(dist))+0.5;
                                                // use any fBm function
   data->out[0] = RidgedMultifractal2(data->in, data->H, data->lacunarity, octaves, 1.0, 2.0, 0, &data->freq) -1;  // altitude - mountains

   return 1;

/*
   float p[2];

   p[0] = data->SCALESB*pos[0];
   p[1] = data->SCALESB*pos[1];

   return RidgedMultifractal2( p, H, lacunarity, octaves, 1.0, 2.0, 0, &freq);  // altitude between 0.2 - 2 - water
*/

//   return 10*RidgedMultifractal2( p, H, lacunarity, octaves, 1.0, 2.0, 0, &freq);  // altitude between 2 - 21
//   return 0.1*RidgedMultifractal2( pos, H, lacunarity, octaves, 1.0, 2.0, 0, &freq);  // altitude
}


/* *************************************************************
  calculate level of detail, based on distance from eye
		"dip" for road

#define SBSCALE 0.0063596645
#define SCALESB 157.241

************************************************************* */
int mnb_alt(void *x) {

   shaderparamtype *data = (shaderparamtype *)x;
   float octaves;
   float dist;

   float road; 
   float edge;

   road = 0.31798322;		// sbscale * 50
   edge = abs(data->in[0]);

   if (edge < road) {		// road 50+ units wide
      data->out[0] = 0;
      return 1;
   }

   if (data->pt[0] < 1)
      dist = 1;
   else
      dist = data->pt[0];

   octaves = -(data->octaves + LOG2(dist))+0.5;

   if (edge < road+road)
      dist = edge/road - 1.0;
   else
      dist = 1;
                                                // use any fBm function
   data->out[0] = dist*(RidgedMultifractal2(data->in, data->H, data->lacunarity, octaves, 1.0, 2.0, 0, &data->freq)-1);  // altitude - mountains

   return 1;
}


/* **************************************************
************************************************** */
void setup_shader() {

   init_noise();

   shades = new shader;
   shades->set("marble", marble);		// surface: marble texture

   insert_shader(shades, "moon", moon);		// surface: fractal moon
   insert_shader(shades, "planet", planet);	// surface: fractal planet - variation of moon
   insert_shader(shades, "wood", wood);		// surface: plywood texture
   insert_shader(shades, "cloud", cloud);	// surface: cloud texture
   insert_shader(shades, "stars", stars);	// surface: star texture
   insert_shader(shades, "warp", warp);		// surface: displaces xyz for color calculations
   insert_shader(shades, "terrain", terrain);	// surface: color base on altitude
   insert_shader(shades, "planet2", planet2);	// surface: another planet
   insert_shader(shades, "cloud2", cloud2);	// surface: another cloud texture
   insert_shader(shades, "terrain2", terrain2); // surface: another terrain texture
   insert_shader(shades, "checker", checker);	// surface: 3D checker pattern
   insert_shader(shades, "fog_mus", fog_mus);	// atmosphere: musgrave's fog shader
   insert_shader(shades, "fog", fog);		// atmosphere: another fog shader
   insert_shader(shades, "terrain3", terrain3); // surface: yet another terrain
   insert_shader(shades, "wood_panel", wood_panel); // surface: plywood panels
   insert_shader(shades, "wave", wave);		// surface: normal manipulator
   insert_shader(shades, "fog2", fog2);		// atmosphere: yet another fog shader
   insert_shader(shades, "cave_ceiling", cave_ceiling); // surface: (stalagmite?) shader
   insert_shader(shades, "cave_floor", cave_floor);	// surface: (stalagtite?) shader

   insert_shader(shades, "default_alt", default_alt);	// sbfast: a height field
   insert_shader(shades, "mnb_alt", mnb_alt);		// sbfast: another height field
}



