
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#if !defined __SHADER_H
#define __SHADER_H

#include <string.h>
#include <bstring.h>
#include "global.h"

typedef enum {

   SURFACE    = 0,
   ATMOSPHERE = 1,
   DISPLACEMENT = 2
} shader_type;


class shaderparamtype {

   public:
      float in[4];		// pt in local coordinates
      float out[3];		// out color
      float add[3];		// out additional color

      float normal[4];		// normal in world coordinates
      float pt[4];		// pt in world coordinates (out)

      float lacunarity, H;
      float octaves;
      float *freq;

      shaderparamtype() {
         H = lacunarity = octaves = 0;
         freq = (float *)NULL;
      }

      ~shaderparamtype () {
         if (freq != (float *)NULL)
            delete freq;
      }

};


class shader {

   protected:
      void *shade_fcn;
      shader *left, *right;

   public:
      char name[MAXSTRLEN];

      shader()  { name[0] = 0; left = right = (shader *)NULL; }
      ~shader() { if (left != (shader *)NULL) delete left; if (right != (shader *)NULL) delete right; }

      shader & operator = (shader &x) {

         strcpy(name, x.name);
         shade_fcn = x.shade_fcn;

         if (left != (shader *)NULL)
            delete left;

         if (x.left != (shader *)NULL) {
            left = new shader;
            *left = *(x.left);
         }

         else
            left = (shader *)NULL;

         if (right != (shader *)NULL)
            delete right;

         if (x.right != (shader *)NULL) {
            right = new shader;
            *right = *(x.right);
         }

         else
            right = (shader *)NULL;

         return x;
      }

      shader    *find(char *target);
      void      insert(shader *x);

#ifdef SGI
      int       set(char *tmpstr, void *func_call);
#else
      int       set(char *tmpstr, int func_call(void *));
#endif

      int       query_data(void *x) { return ((int (*)(void *))shade_fcn)(x); }
};


typedef shader *pshader;

class shadertype {

   public:
      shader **s;
      shader_type *stype;

      int scount;

      shadertype() { s = (shader **)NULL; stype = (shader_type *)NULL; scount = 0; }
      ~shadertype() { if (s != (shader **)NULL) delete s; if (stype != (shader_type)NULL) delete stype; }
      void read_data(FILE *infile, shader *sptr);

      shadertype &operator = (shadertype &x) {

         scount = x.scount;
         if (s != (shader **)NULL)
            delete s;
         if (stype != (shader_type *)NULL)
            delete stype;

         s = new pshader[scount];
         stype = new shader_type[scount];

         bcopy(x.stype, stype, sizeof(shader_type *)*scount);
         bcopy(x.s, s, sizeof(pshader)*scount);
         return x;
      }

};


class shaderlisttype {

   public:
      shadertype        *shade;
      float             scale;
      char              shname[MAXSTRLEN];            // illumination file
      char              dataname[MAXSTRLEN];          // object file
      int               scount;
      shaderlisttype    *next;

      shaderlisttype() { shade = (shadertype *)NULL; scale = 1; }
      ~shaderlisttype() { if (shade != (shadertype *)NULL) delete shade; }

      void init(int x) {

         if (shade != (shadertype *)NULL)
            delete shade;
         shade = new shadertype[x];
         scount = x;
      }

};


void setup_shader();

extern shader *shades;

#endif

