
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */



#include <string.h>
#include "control.h"


slave::slave() {

   int i, j;

   busy   = active = bye = 0;
   done   = 1;

   j = MAXSUBPROC-1;

   if (!j)
      grandchild = (subslave *)NULL;
   else {
      grandchild = new subslave[j];

      j--;

      for (i=0; i<j; i++) {
         grandchild[i].id = i;
         grandchild[i].control = this;
         grandchild[i].next = &grandchild[i+1];
      }

      grandchild[j].id = j;
      grandchild[j].control = this;
      grandchild[j].next = (subslave *)NULL;
   }

   zbuff = new zbuffer;
}


slave::~slave() {

   if (grandchild != (subslave *)NULL)
      delete grandchild;

   zbuff->pdata = (unsigned int **)NULL;
   zbuff->data = (unsigned int *)NULL;

   delete zbuff;
}


int slave::spawn(camera *cam, light *lit, FILE *io, listtype *mainlist, int frm, mapul *mapbuffer) {

   while (active);

//      printf("Firing off processor #%d %d\n", id, frm);

   control->set_lock(LOCK_RESULT);

   if (!PENCILTEST)
      for (ltr=lit; ltr!=(light *)NULL; ltr=ltr->next)
         if (ltr->ltype != FAR && ltr->ltype != POINT)
            control->get_lbuff(ltr);

   control->free_lock(LOCK_RESULT);

   ctr     = cam;
   ptr     = mainlist;
   outfile = io;
   mtr     = mapbuffer;
   ltr     = lit;

   doneflag = frm;

#ifdef SGI
   done     = 0;
   active   = 1;

   while (PENCILTEST && !done);
#else
   render_object(this);
#endif

   return 1;
}


int slave::activate() {

   int i;

   printf("activate slave %d\n", id);

#ifdef SGI
   child_id = sproc((void (*)(void *))create_child, PR_SALL, (void *)this);
#endif

   for (i=0; i<MAXSUBPROC-1; i++)               // comment out scanline
      grandchild[i].activate();

   return 1;
}


int slave::query_busy() {

   return busy;
}


void create_child(void *owner) {

   slave        *sslave;

   sslave = (slave *)owner;
   while (1) {
      while (!sslave->active)
         if (sslave->bye)
            return;

      sslave->busy = 1;
      sslave->active = 0;

      render_object(sslave);

      sslave->done = 1;
      sslave->busy = 0;
   }

}


void render_object(slave *sslave) {

   pc           *pctr, *ptr, *rtr;
   camera       *cctr, *ctr;
   light        *lctr, *ltr;
   int		mcanvas;
   FILE         *outfile;
   subslave     *grandchild;
   pc           *shadptr = (pc *)NULL;
   mapul        *mtr;

   mcanvas  = sslave->doneflag;
   ptr      = sslave->ptr->header[mcanvas];
   rtr      = sslave->ptr->raddtr[mcanvas];
   ctr      = sslave->ctr;
   ltr      = sslave->ltr;
   mtr      = sslave->mtr;
   outfile  = sslave->outfile;
   grandchild = sslave->grandchild;

   sslave->control->free_slave(sslave->id);

   for (lctr=ltr; lctr!=(light *)NULL; lctr=lctr->next)
      lctr->xform(ctr);

   if (PENCILTEST) {
      for (pctr=ptr; pctr!=(pc *)NULL; pctr=pctr->next)
         pctr->scan_children(ctr, ltr, grandchild, &shadptr);

      for (pctr=shadptr; pctr!=(pc *)NULL; pctr=pctr->next) {
          pctr->scan(ctr, (light *)NULL, grandchild, &shadptr);
          pctr->end_scan();
      }

      for (pctr=rtr; pctr!=(pc *)NULL; pctr=pctr->next)
         pctr->scan_children(ctr, ltr, grandchild, &shadptr);
   }

   else {
      if (NOSCREEN)
         sslave->zbuff->initbuff(WINX, WINY);
      else
         if (mtr != (mapul *)NULL)
            sslave->zbuff->initbuff(mtr->data, mtr->pdata, mtr->maxx, mtr->maxy);

      for (lctr=ltr; lctr!=(light *)NULL; lctr=lctr->next) // beamscan poly, spat, particle, not radiant
         if (lctr->ltype == BEAM_POINT || lctr->ltype == BEAM_FAR) {
            lctr->lbuff.init_buff();

            for (pctr=ptr; pctr!=(pc *)NULL; pctr=pctr->next)
               pctr->beamscan_children(lctr, grandchild);
            lctr->alter_beam_mx(ctr);
         }

      for (pctr=ptr; pctr!=(pc *)NULL; pctr=pctr->next)
         pctr->scan_children(ctr, ltr, sslave->zbuff, grandchild, &shadptr);

      for (pctr=shadptr; pctr!=(pc *)NULL; pctr=pctr->next) {
         pctr->scan(ctr, lctr, sslave->zbuff, grandchild, &shadptr);
         pctr->end_scan();
      }

      for (pctr=rtr; pctr!=(pc *)NULL; pctr=pctr->next)
         pctr->scan_children(ctr, ltr, sslave->zbuff, grandchild, &shadptr);

      if (ANTIALIAS)                           // anti aliasing
         if (GAMMA)
            sslave->zbuff->antialias(SANTIALIAS, FGAMMA);
         else
            sslave->zbuff->antialias(SANTIALIAS, -1);
      else
         if (GAMMA)
            sslave->zbuff->catbuff(FGAMMA);
         else
            sslave->zbuff->catbuff(-1);

      if (SAVE && strlen(SSAVE))
         sslave->zbuff->save_buffer(outfile, mcanvas);

      if (RGBFILE)
         sslave->zbuff->save_rgb(SRGBFILE, mcanvas);

      if (!NOSCREEN)
         sslave->zbuff->data = (unsigned int *)NULL;
   }

   sslave->control->set_lock(LOCK_RESULT);

   sslave->control->push_result(mcanvas);
   if (!PENCILTEST)
      for (lctr=ltr; lctr!=(light *)NULL; lctr=lctr->next)
         if (lctr->ltype != FAR && lctr->ltype != POINT)
            sslave->control->stack_lbuff(lctr);

   sslave->control->free_lock(LOCK_RESULT);

   for (pctr=ptr; pctr!=(pc *)NULL; pctr=ptr) {
      ptr = ptr->next;
      delete pctr;
   }

   for (pctr=rtr; pctr!=(pc *)NULL; pctr=rtr) {
      rtr = rtr->next;
      delete pctr;
   }

   for (pctr=shadptr; pctr!=(pc *)NULL; pctr=shadptr) {
      shadptr = shadptr->next;
      delete pctr;
   }

   for (cctr=ctr; cctr!=(camera *)NULL; cctr=ctr) {
      ctr = ctr->next;
      delete cctr;
   }

   for (lctr=ltr; lctr!=(light *)NULL; lctr=ltr) {
      ltr = ltr->next;
      delete lctr;
   }

}
