
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file scans surface patch data into the buffer
************************************************************* */


#include "spatch.h"
#include <string.h>


/* *************************************************************
************************************************************* */
int spatch::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'f':
         if (!strcmp(token, "file")) {
            get_token(infile, filename);

            return 1;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


/* *************************************************************
************************************************************* */
void spatch::preprocess(listtype *mainlist, shader *sparm) {

   read_data(filename, &mainlist->spatdtr);
   pc::preprocess(mainlist, sparm);
}


/* *************************************************************
************************************************************* */
int spatch::read_data(FILE *infile) {

   return 0;
}


int spatch::read_3dtex(char *filename, char *dfilename, shader *sptr, shaderlisttype **slptr) {

   int i, j;
   char buffer[MAXSTRLEN];
   FILE *infile;
   int count;
   int objectcount = (dob->sdata.vrow-3)*(dob->sdata.vcol-3);

   for (shaderlist=*slptr; shaderlist!=(shaderlisttype *)NULL; shaderlist=shaderlist->next)
      if (!strcmp(filename, shaderlist->shname) && !strcmp(dfilename, shaderlist->dataname))
         return 1;

   if (!(infile = fopen(filename, "r"))) {
      printf("Can't find file %s\n", filename);
      mcinfo &= ~CI3DTEX;
      return 0;
   }

   shaderlist = new shaderlisttype;
   strcpy(shaderlist->shname, filename);
   strcpy(shaderlist->dataname, dfilename);
   shaderlist->init(objectcount);
   shaderlist->next   = *slptr;
   *slptr       = shaderlist;

   fscanf(infile, "%s %f", buffer, &shaderlist->scale);
   lower_case(buffer);

   if (strcmp(buffer, "mono")) {                // ALL
      shaderlist->shade[0].read_data(infile, sptr);

      for (i=1; i<objectcount; i++)
         shaderlist->shade[i] = shaderlist->shade[0];
   }

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {
      fscanf(infile, "%d %d", &i, &j);
      i = (i-1)*(dob->sdata.vcol-2) + j-1;

      shaderlist->shade[i].read_data(infile, sptr);
   }

   fclose(infile);
   return 1;
}

/* *************************************************************
************************************************************* */
int spatch::read_tex(char *filename, char *dfilename, texbase **tptr, texpolygon **tpptr) {

   FILE *infile;                        // file pointer
   int i, j, k, l, m;
   int coord[2][2];
   float corner[4][2];
   int count;
   char tempstr[MAXSTRLEN];             // generic string
   int patch;
   texbase *ptr;
   vector2f *pointer;
   int   start[2], end[2];
   double pt[2], d[2], current[2];
   int   vtr;

   int prow, pcol, srow, scol;

   prow = dob->sdata.vrow - 3;
   pcol = dob->sdata.vcol - 3;
   srow = dob->sdata.vrow - 2;
   scol = dob->sdata.vcol - 2;

   for (tob=*tpptr; tob!=(texpolygon *)NULL; tob=tob->next)
      if (!strcmp(filename, tob->filename) && !strcmp(dfilename, tob->dataname))
         return 0;

   if (!(infile = fopen(filename, "r"))) {
      printf("Could not access %s.\n", filename);
      return 0;
   }

   tob = new texpolygon;
   strcpy(tob->filename, filename);
   strcpy(tob->dataname, dfilename);
   tob->countobject = prow*pcol;
   tob->next = *tpptr;
   *tpptr    = tob;

   tob->setup_data(tob->countobject);
   tob->setup_pdata(prow);

   pointer = new vector2f[scol*srow];

   for (i=0; i<tob->countobject; i++)
      tob->data[i] = (tex *)NULL;

   for (i=j=0; i<prow; i++, j+=pcol)
      tob->pdata[i] = &tob->data[j];

   fscanf(infile, "%d", &count);

   for (; count>0; count--) {

      fscanf(infile, "%s", tempstr);
      lower_case(tempstr);

      if (!strcmp(tempstr, "corner")) {
         fscanf(infile, "%d %d %d %d", &coord[0][0], &coord[0][1], &coord[1][0], &coord[1][1]);

         fscanf(infile, "%s", tempstr);               // texture name

         if (coord[0][0] != coord[1][0] && coord[0][1] != coord[1][1]) {
            for (ptr=*tptr; ptr!=(texbase *)NULL; ptr=ptr->next)

               if (!strcmp(tempstr, ptr->filename))
                  break;

            if (ptr == (texbase *)NULL) {
               if (query_video(tempstr)) {
                  ptr = new texvid;
                  ptr->read_texture(tempstr, tptr);
               }

               else {
                  ptr = new texture;
                  ptr->read_texture(tempstr);
               }

               ptr->next = *tptr;
               *tptr = ptr;
            }

            coord[0][0]--;
            coord[0][1]--;
            coord[1][0]--;
            coord[1][1]--;

            ptr->query_coord(corner);

            if (coord[0][0] > coord[1][0]) {
               start[0] = coord[1][0];
               end[0]   = coord[0][0];
            }

            else {
               start[0] = coord[0][0];
               end[0]   = coord[1][0];
            }

            if (coord[0][1] > coord[1][1]) {
               start[1] = coord[1][1];
               end[1]   = coord[0][1];
            }

            else {
               start[1] = coord[0][1];
               end[1]   = coord[1][1];
            }

            pt[0] = corner[0][0];
            pt[1] = corner[0][1];
            d[0]  = (corner[2][0]-corner[0][0]) / ((double)(coord[1][0]-coord[0][0]));
            d[1]  = (corner[2][1]-corner[0][1]) / ((double)(coord[1][1]-coord[0][1]));

            for (j=start[1], current[1]=pt[1], i=j*scol; j<=end[1]; j++, current[1]+=d[1], i+=scol)
               for (k=start[0], current[0]=pt[0], vtr=i+k; k<=end[0]; k++, current[0]+=d[0], vtr++) {
                   copyarray2(pointer[vtr], current);
               }

            l = end[1]+start[1]-1;
            for (j=start[1], i=j*scol; j<end[1]; j++, i+=scol)
               for (k=start[0], vtr=i+k, m=vtr+scol+1, patch=l-j; k<end[0]; k++, vtr++, m++) {
                  if (tob->pdata[patch][k] == (tex *)NULL) {
                     tob->pdata[patch][k]  = new tex;
                     tob->pdata[patch][k]->mapfunc = COORD;
                  }

                  tob->pdata[patch][k]->ob = ptr;

                  tob->pdata[patch][k]->coord[0][0] = pointer[vtr][0];
                  tob->pdata[patch][k]->coord[0][1] = pointer[m][1];

                  tob->pdata[patch][k]->coord[1][0] = pointer[m][0];
                  tob->pdata[patch][k]->coord[1][1] = pointer[vtr][1];
               }

         }

      }

      else
         if (!strcmp(tempstr, "coord")) {
            fscanf(infile, "%d %d %d %d", &coord[0][0], &coord[0][1], &coord[1][0], &coord[1][1]);
            fscanf(infile, "%f %f %f %f", &corner[0][0], &corner[0][1], &corner[2][0], &corner[2][1]);

            fscanf(infile, "%s", tempstr);               // texture name

            if (coord[0][0] != coord[1][0] && coord[0][1] != coord[1][1]) {
               for (ptr=*tptr; ptr!=(texbase *)NULL; ptr=ptr->next)
                  if (!strcmp(tempstr, ptr->filename))
                     break;

               if (ptr == (texbase *)NULL) {
                  if (query_video(tempstr)) {
                     ptr = new texvid;
                     ptr->read_texture(tempstr, tptr);
                  }

                  else {
                     ptr = new texture;
                     ptr->read_texture(tempstr);
                  }

                  ptr->next = *tptr;
                  *tptr = ptr;
               }

               coord[0][0]--;
               coord[0][1]--;
               coord[1][0]--;
               coord[1][1]--;

               corner[0][0]--;
               corner[0][1]--;
               corner[2][0]--;
               corner[2][1]--;

               if (coord[0][0] > coord[1][0]) {
                  start[0] = coord[1][0];
                  end[0]   = coord[0][0];
               }

               else {
                  start[0] = coord[0][0];
                  end[0]   = coord[1][0];
               }

               if (coord[0][1] > coord[1][1]) {
                  start[1] = coord[1][1];
                  end[1]   = coord[0][1];
               }

               else {
                  start[1] = coord[0][1];
                  end[1]   = coord[1][1];
               }

               pt[0] = corner[0][0];
               pt[1] = corner[0][1];
               d[0]  = (corner[2][0]-corner[0][0]) / ((double)(coord[1][0]-coord[0][0]));
               d[1]  = (corner[2][1]-corner[0][1]) / ((double)(coord[1][1]-coord[0][1]));

               for (j=start[1], current[1]=pt[1], i=j*scol; j<=end[1]; j++, current[1]+=d[1], i+=scol)
                  for (k=start[0], current[0]=pt[0], vtr=i+k; k<=end[0]; k++, current[0]+=d[0], vtr++) {
                      copyarray2(pointer[vtr], current);
                  }

               l = end[1]+start[1]-1;
               for (j=start[1], i=j*scol; j<end[1]; j++, i+=scol)
                  for (k=start[0], vtr=i+k, m=vtr+scol+1, patch=l-j; k<end[0]; k++, vtr++, m++) {
                     if (tob->pdata[patch][k] == (tex *)NULL) {
                        tob->pdata[patch][k]  = new tex;
                        tob->pdata[patch][k]->mapfunc = COORD;
                     }

                     tob->pdata[patch][k]->ob = ptr;

                     tob->pdata[patch][k]->coord[0][0] = pointer[vtr][0];
                     tob->pdata[patch][k]->coord[0][1] = pointer[m][1];

                     tob->pdata[patch][k]->coord[1][0] = pointer[m][0];
                     tob->pdata[patch][k]->coord[1][1] = pointer[vtr][1];
                  }

            }

         }

   }

   fclose(infile);
   delete pointer;
   return 1;
}


/* *************************************************************
   This function reads in the surface patch data
************************************************************* */
int spatch::read_data(char filename[], datatype **dptr) {

   FILE *infile;
   int i, j, size;

   for (dob=*dptr; dob!=(datatype *)NULL; dob=dob->next)
      if (!strcmp(filename, dob->dataname))
         return (dob->sdata.vrow-3)*(dob->sdata.vcol-3);

   if (!(infile = fopen(filename, "r"))) {
      printf("Could not access %s.\n", filename);
      return 0;
   }

   dob              = new datatype;
   dob->next        = *dptr;
   *dptr = dob;

   strcpy(dob->dataname, filename);

   fscanf(infile, "%d %d", &dob->sdata.vrow, &dob->sdata.vcol);

   size = dob->sdata.vrow * dob->sdata.vcol;
   dob->sdata.data = new vector4f[dob->sdata.vrow*dob->sdata.vcol];

   fscanf(infile, "%f %f %f", &dob->sdata.data[0][0], &dob->sdata.data[0][1], &dob->sdata.data[0][2]);
   dob->sdata.data[0][3] = 1;

   copyarray3(dob->bbox[0], dob->sdata.data[0]);
   copyarray3(dob->bbox[1], dob->sdata.data[0]);

   for (i=1; i<size; i++) {
      fscanf(infile, "%f %f %f", &dob->sdata.data[i][0], &dob->sdata.data[i][1], &dob->sdata.data[i][2]);
      dob->sdata.data[i][3] = 1;

      for (j=0; j<3; j++)
         if (dob->sdata.data[i][j] < dob->bbox[0][j])
            dob->bbox[0][j] = dob->sdata.data[i][j];
         else
            if (dob->sdata.data[i][j] > dob->bbox[1][j])
               dob->bbox[1][j] = dob->sdata.data[i][j];
   }

   fclose(infile);
   return (dob->sdata.vrow-3)*(dob->sdata.vcol-3);
}
