
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file scans surface patch data into the buffer
************************************************************* */


#include "spatch.h"
#include <string.h>


/* *************************************************************
************************************************************* */
int spatch::bound_box(eye *parm) {

   int    i, j;
   float  extreme[2][2];
   float  d[2];

   if (bbox[0][2] > front)
      return 0;

   if (bbox[1][2] > front)
      return 1;

   for (i=0; i<8; i++)
      parm->map2screen(box[i]);

   copyarray2(extreme[0], box[0]);      // bounding box
   copyarray2(extreme[1], box[0]);

   for (i=1; i<8; i++)
      for (j=0; j<2; j++)
         if (extreme[0][j] > box[i][j])
            extreme[0][j] = box[i][j];
         else
            if (extreme[1][j] < box[i][j])
               extreme[1][j] = box[i][j];

   d[0] = extreme[1][0] - extreme[0][0];
   d[1] = extreme[1][1] - extreme[0][1];

   if (abs(d[0]) < MINBBOX  || abs(d[1]) < MINBBOX)
      return 0;

   if (abs(d[0]) < MINDIM  || abs(d[1]) < MINDIM) {
      switch (mctype) {
         case DOT:
         case PDOT:
            mctype = DOT;

         case WFBW:
         case PWFBW:
            mctype = WFBW;
            break;

         case BW:
         case PBW:
            mctype = BW;
            break;

         default:
            mctype = CONSTANT;
      }

      mcinfo &= CIRAYCAST;

      setup_transfer(4,1);

      pob->countvertex = 4;
      pob->countobject = 1;
      pob->countedge   = 4;

      pob->mvertex[0][0] = extreme[0][0];
      pob->mvertex[0][1] = extreme[0][1];
      pob->mvertex[0][2] = bbox[1][2];
      pob->mvertex[0][3] = 1;
      parm->screen2map(pob->mvertex[0]);

      pob->mvertex[1][0] = extreme[0][0];
      pob->mvertex[1][1] = extreme[1][1];
      pob->mvertex[1][2] = bbox[1][2];
      pob->mvertex[1][3] = 1;
      parm->screen2map(pob->mvertex[1]);

      pob->mvertex[2][0] = extreme[1][0];
      pob->mvertex[2][1] = extreme[1][1];
      pob->mvertex[2][2] = bbox[1][2];
      pob->mvertex[2][3] = 1;
      parm->screen2map(pob->mvertex[2]);

      pob->mvertex[3][0] = extreme[1][0];
      pob->mvertex[3][1] = extreme[0][1];
      pob->mvertex[3][2] = bbox[1][2];
      pob->mvertex[3][3] = 1;
      parm->screen2map(pob->mvertex[3]);

      pob->object[0] = 0;
      pob->object[1] = 1;
      pob->object[2] = 2;
      pob->object[3] = 3;

      if (mctype > PWFBW || (mctype > WFBW && mctype <= PHONG))
         for (i=0; i<4; i++) {
            pob->gnormal[i][0] = 0;
            pob->gnormal[i][1] = 0;
            pob->gnormal[i][2] = 1;
         }

      pob->polynum[0] = 4;

      pob->edgeptr[0] = pob->object;

      if (mctype == CONSTANT)
         for (i=0; i<pob->countobject; i++)
            pob->shade[i] = base_color;

      return 2;
   }

   return 1;
}


/* *************************************************************
   This function converts surface patch data into polygonal data
************************************************************* */
void spatch::spline2poly(spatchtype *sob, texpolygon *stob, int xflag) {

   int   i, j, k, l, m;
   int   base;
   int   ptr[4];
   float  d[3];
   float  temp;
   int    rflag;                        // variable to keep track of splitting of patch
   pvector4f vtr[4];

   pob->countobject = pob->countvertex = pob->countedge = 0;

   base = l = 0;

   for (i=1; i<=sob->srow; i++)
      for (j=1, k=0; j<=sob->scol; k=j, j++, l++) {
         copyarray4(pob->mvertex[l], sob->mvertex[i][j]);

         if (mctype > PWFBW || (mctype > WFBW && mctype <= PHONG)) {
            if (mctype > PWFBW) {
               copyarray3(pob->gnormal[l], sob->gnormal[i][j]);
            }

            else {
               vtr[0] = &ob->mvertex[i-1][k];
               vtr[1] = &ob->mvertex[i][k];
               vtr[2] = &ob->mvertex[i+1][k];
               vtr[3] = &ob->mvertex[i+2][k];

               spnormal_base(vtr, (float *)pob->gnormal[l], M_CR);
            }

         }

      }

   pob->countvertex = l;

   for (i=0, m=base; i<sob->prow; i++, m++)
      for (j=0; j<sob->pcol; j++, m++) {
         ptr[0] = m;
         ptr[1] = ptr[0] + 1;
         ptr[3] = ptr[0] + sob->scol;
         ptr[2] = ptr[3] + 1;

         k = 0;
         for (l=0; l<4; l++)
            if (!k || nnre_vector3(pob->mvertex[pob->object[pob->countedge+k-1]], pob->mvertex[ptr[l]])) {
               pob->object[pob->countedge+k] = ptr[l];
               k++;
            }

         if (k == 3) {
            if (mctype > PBW || (mctype > BW && mctype <= PHONG))
               pob->shade[pob->countobject] = sob->mshade[i][j];

            pob->edgeptr[pob->countobject] = &pob->object[pob->countedge];

            if (xflag && stob->pdata[i][j] != (tex *)NULL) {
               pob->object[pob->countedge] = ptr[0];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[1];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[2];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[3];
               pob->countedge++;

               tpob->data[pob->countobject]          = new tex;
               tpob->data[pob->countobject]->mapfunc = COORD;
               tpob->data[pob->countobject]->ob      = stob->pdata[i][j]->ob;
               tpob->data[pob->countobject]->pt      = new vector2f[5];

               tpob->data[pob->countobject]->pt[4][0] =
               tpob->data[pob->countobject]->pt[3][0] =
               tpob->data[pob->countobject]->pt[0][0] = stob->pdata[i][j]->coord[0][0];

               tpob->data[pob->countobject]->pt[4][1] =
               tpob->data[pob->countobject]->pt[1][1] =
               tpob->data[pob->countobject]->pt[0][1] = stob->pdata[i][j]->coord[0][1];

               tpob->data[pob->countobject]->pt[2][0] =
               tpob->data[pob->countobject]->pt[1][0] = stob->pdata[i][j]->coord[1][0];

               tpob->data[pob->countobject]->pt[3][1] =
               tpob->data[pob->countobject]->pt[2][1] = stob->pdata[i][j]->coord[1][1];

               pob->polynum[pob->countobject] = 4;
            }

            else {
               pob->countedge += 3;
               pob->polynum[pob->countobject] = 3;
            }

            pob->countobject++;
         }

         else
            if (k == 4) {
               d[0] = pob->mvertex[ptr[0]][0] - pob->mvertex[ptr[2]][0];
               d[1] = pob->mvertex[ptr[0]][1] - pob->mvertex[ptr[2]][1];
               d[2] = pob->mvertex[ptr[0]][2] - pob->mvertex[ptr[2]][2];

               temp = magnitude3(d);

               d[0] = pob->mvertex[ptr[1]][0] - pob->mvertex[ptr[3]][0];
               d[1] = pob->mvertex[ptr[1]][1] - pob->mvertex[ptr[3]][1];
               d[2] = pob->mvertex[ptr[1]][2] - pob->mvertex[ptr[3]][2];

               if ((rflag = (magnitude3(d) < temp))) {
                  k = ptr[0];
                  ptr[0] = ptr[1];
                  ptr[1] = ptr[2];
                  ptr[2] = ptr[3];
                  ptr[3] = k;
               }

               pob->edgeptr[pob->countobject] = &pob->object[pob->countedge];

               if (mctype > PBW || (mctype > BW && mctype <= PHONG))
                  pob->shade[pob->countobject] = pob->shade[pob->countobject+1] = sob->mshade[i][j];

               pob->object[pob->countedge] = ptr[0];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[1];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[2];
               pob->countedge++;

               if (xflag && stob->pdata[i][j] != (tex *)NULL) {
                  tpob->data[pob->countobject]          = new tex;
                  tpob->data[pob->countobject]->mapfunc = COORD;
                  tpob->data[pob->countobject]->ob      = stob->pdata[i][j]->ob;
                  tpob->data[pob->countobject]->pt      = new vector2f[4];

                  if (rflag) {
                     tpob->data[pob->countobject]->pt[2][0] = stob->pdata[i][j]->coord[0][0];

                     tpob->data[pob->countobject]->pt[3][1] =
                     tpob->data[pob->countobject]->pt[0][1] = stob->pdata[i][j]->coord[0][1];

                     tpob->data[pob->countobject]->pt[3][0] =
                     tpob->data[pob->countobject]->pt[1][0] =
                     tpob->data[pob->countobject]->pt[0][0] = stob->pdata[i][j]->coord[1][0];

                     tpob->data[pob->countobject]->pt[2][1] =
                     tpob->data[pob->countobject]->pt[1][1] = stob->pdata[i][j]->coord[1][1];
                  }

                  else {
                     tpob->data[pob->countobject]->pt[3][0] =
                     tpob->data[pob->countobject]->pt[0][0] = stob->pdata[i][j]->coord[0][0];

                     tpob->data[pob->countobject]->pt[3][1] =
                     tpob->data[pob->countobject]->pt[1][1] =
                     tpob->data[pob->countobject]->pt[0][1] = stob->pdata[i][j]->coord[0][1];

                     tpob->data[pob->countobject]->pt[2][0] =
                     tpob->data[pob->countobject]->pt[1][0] = stob->pdata[i][j]->coord[1][0];

                     tpob->data[pob->countobject]->pt[2][1] = stob->pdata[i][j]->coord[1][1];
                  }

               }

               pob->polynum[pob->countobject] = 3;
               pob->countobject++;

               pob->edgeptr[pob->countobject] = &pob->object[pob->countedge];

               pob->object[pob->countedge] = ptr[2];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[3];
               pob->countedge++;
               pob->object[pob->countedge] = ptr[0];
               pob->countedge++;

               if (xflag && stob->pdata[i][j] != (tex *)NULL) {
                  tpob->data[pob->countobject]          = new tex;
                  tpob->data[pob->countobject]->mapfunc = COORD;
                  tpob->data[pob->countobject]->ob      = stob->pdata[i][j]->ob;
                  tpob->data[pob->countobject]->pt      = new vector2f[4];

                  if (rflag) {
                     tpob->data[pob->countobject]->pt[3][0] =
                     tpob->data[pob->countobject]->pt[1][0] =
                     tpob->data[pob->countobject]->pt[0][0] = stob->pdata[i][j]->coord[0][0];

                     tpob->data[pob->countobject]->pt[2][1] =
                     tpob->data[pob->countobject]->pt[1][1] = stob->pdata[i][j]->coord[0][1];

                     tpob->data[pob->countobject]->pt[3][1] =
                     tpob->data[pob->countobject]->pt[0][1] = stob->pdata[i][j]->coord[1][1];

                     tpob->data[pob->countobject]->pt[2][0] = stob->pdata[i][j]->coord[1][0];
                  }

                  else {
                     tpob->data[pob->countobject]->pt[2][0] =
                     tpob->data[pob->countobject]->pt[1][0] = stob->pdata[i][j]->coord[0][0];

                     tpob->data[pob->countobject]->pt[3][1] =
                     tpob->data[pob->countobject]->pt[1][1] =
                     tpob->data[pob->countobject]->pt[0][1] = stob->pdata[i][j]->coord[1][1];

                     tpob->data[pob->countobject]->pt[3][0] =
                     tpob->data[pob->countobject]->pt[0][0] = stob->pdata[i][j]->coord[1][0];

                     tpob->data[pob->countobject]->pt[2][1] = stob->pdata[i][j]->coord[0][1];
                  }

               }

               pob->polynum[pob->countobject] = 3;
               pob->countobject++;
            }

      }

   if (xflag)
      tpob->countobject = pob->countobject;
}


/* **************************************************
   This procedure scans 2D grids into surface patches
   or polygonal structures.
************************************************** */
void spatch::convert(int i, int j, spatchtype *sob, texpolygon *stob, shaderlisttype *ssob) {

   int h, k, l, m, n;
   float u, v, t;
   pvector4f ptr[4];
   float SPLQ[4][4], SPLQ2[4][4], SPPQ[4][4][4];
   float dt[2], pt[2][2], start[2];
   float delta = 1.0/SPDELTA;

   ptr[0] = &ob->mvertex[i][j];
   ptr[1] = &ob->mvertex[i+1][j];
   ptr[2] = &ob->mvertex[i+2][j];
   ptr[3] = &ob->mvertex[i+3][j];

   for (n=0; n<4; n++)
      for (m=0; m<4; m++)
         for (h=0; h<4; h++)
              SPPQ[n][m][h] = ptr[m][h][n];

						// top row
   for (n=0; n<4; n++)
      for (h=0; h<4; h++)
         SPLQ[n][h] = ptr[1][h][n];

   k=i+1;
   l=j+1;
   copyarray4(sob->mvertex[1][1], ob->mvertex[k][l]);

   if (mctype > PWFBW)
      spnormal(0, 0, SPPQ, (float *)&sob->gnormal[1][1], M_CR);

   for (n=1, m=2, t=delta; n<SPDELTA; n++, m++, t+=delta) {
      spline(t, SPLQ, (float *)&sob->mvertex[1][m], M_CR);
      sob->mvertex[1][m][3] = 1;

      if (mctype > PWFBW)
         spnormal(0, t, SPPQ, (float *)&sob->gnormal[1][m], M_CR);
   }

   l++;
   copyarray4(sob->mvertex[1][m], ob->mvertex[k][l]);
   if (mctype > PWFBW)
      spnormal(0, 1, SPPQ, (float *)&sob->gnormal[1][m], M_CR);

						// last row
   for (n=0; n<4; n++)
      for (h=0; h<4; h++)
         SPLQ[n][h] = ptr[2][h][n];

   k=i+2;
   l=j+1;
   h = 1+SPDELTA;
   copyarray4(sob->mvertex[h][1], ob->mvertex[k][l]);

   if (mctype > PWFBW)
      spnormal(1, 0, SPPQ, (float *)&sob->gnormal[h][1], M_CR);

   for (n=1, m=2, t=delta; n<SPDELTA; n++, m++, t+=delta) {
      spline(t, SPLQ, (float *)&sob->mvertex[h][m], M_CR);
      sob->mvertex[1][m][3] = 1;

      if (mctype > PWFBW)
         spnormal(1, t, SPPQ, (float *)&sob->gnormal[h][m], M_CR);
   }

   l++;
   copyarray4(sob->mvertex[h][m], ob->mvertex[k][l]);
   if (mctype > PWFBW)
      spnormal(1, 1, SPPQ, (float *)&sob->gnormal[h][m], M_CR);

					// rest
   for (n=0; n<4; n++)
      for (h=0; h<4; h++)
         SPLQ[n][h] = ptr[h][1][n];

   for (n=0; n<4; n++)
      for (h=0; h<4; h++)
         SPLQ2[n][h] = ptr[h][2][n];

   for (h=2, k=1, u=delta; k<SPDELTA; k++, u+=delta, h++) {
      spline(u, SPLQ, (float *)&sob->mvertex[h][1], M_CR);
      sob->mvertex[h][1][3] = 1;

      if (mctype > PWFBW)
         spnormal(u, 0, SPPQ, (float *)&sob->gnormal[h][1], M_CR);

      for (l=2,n=1, v=delta; n<SPDELTA; n++, v+=delta, l++) {
         sppatch(u, v, SPPQ, (float *)&sob->mvertex[h][l], M_CR);
         sob->mvertex[h][l][3] = 1;

         if (mctype > PWFBW)
            spnormal(u, v, SPPQ, (float *)&sob->gnormal[h][l], M_CR);
      }

      spline(u, SPLQ2, (float *)&sob->mvertex[h][l], M_CR);
      sob->mvertex[h][l][3] = 1;

      if (mctype > PWFBW)
         spnormal(u, 1, SPPQ, (float *)&sob->gnormal[h][l], M_CR);
   }

   if (mctype > PBW) {

      for (k=0; k<SPDELTA; k++)
         for (l=0; l<SPDELTA; l++)
            sob->mshade[k][l] = ob->mshade[i][j];

      if ((mcinfo & CITEXTURE) && tob->pdata[i][j]) {
         copyarray2(start, tob->pdata[i][j]->coord[0]);
         dt[0] = (tob->pdata[i][j]->coord[1][0]-start[0])*delta;
         dt[1] = (tob->pdata[i][j]->coord[1][1]-start[1])*delta;

         for (k=0, pt[0][1]=start[1], pt[1][1]=start[1]+dt[1]; k<SPDELTA;  k++, pt[0][1]=pt[1][1], pt[1][1]+=dt[1])
            for (l=0, pt[0][0]=start[0], pt[1][0]=start[0]+dt[0]; l<SPDELTA; l++, pt[0][0]=pt[1][0], pt[1][0]+=dt[0]) {
               stob->pdata[k][l] = new tex;
               stob->pdata[k][l]->mapfunc = COORD;
               stob->pdata[k][l]->ob = tob->pdata[i][j]->ob;
               copyarray2(stob->pdata[k][l]->coord[0], pt[0]);
               copyarray2(stob->pdata[k][l]->coord[1], pt[1]);
            }

      }

      if (mcinfo & CI3DTEX) {
         h = i*ob->pcol + j;
         m = sob->psize<<1;
         for (k=0; k<m; k++)
            ssob->shade[k] = shaderlist->shade[h];
      }

   }

   spline2poly(sob, stob, (mcinfo & CITEXTURE));
}


/* *************************************************************
   This procedure puts the object in the edge table
************************************************************* */
void spatch::transform(eye *parm) {

   int   i, j;                          // looping vars

   copymx4x4(world, rotate);

   for (i=0; i<3; i++)
      for (j=0; j<3; j++)
         world[i][j] *= size;

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   matmatmulto(parm->transform, world);
   inversemx(world, iworld);

   make_box(dob->bbox, box);

   matvecmulto(world, box[0]);
   copyarray3(bbox[0], box[0]);
   copyarray3(bbox[1], box[0]);

   for (i=1; i<8; i++) {
      matvecmulto(world, box[i]);

      for (j=0; j<3; j++)
         if (box[i][j] < bbox[0][j])
            bbox[0][j] = box[i][j];
         else
            if (box[i][j] > bbox[1][j])
               bbox[1][j] = box[i][j];
   }

   if (bbox[0][2] <= front)
      for (i=0; i<ob->vsize; i++)
         matvecmulto(world, ob->data[i]);
}


/* *************************************************************
   This function copies the basic object data into a composite
   structure for manipulation
************************************************************* */
void spatch::datacopy() {

   int i;                          // looping variables

   ob->build(dob->sdata.vrow, dob->sdata.vcol, 1, (i=(mctype > BW && mctype <= PHONG) || mctype > PBW));

   memcpy(ob->data, dob->sdata.data, ob->vsize*sizeof(vector4f));

   if (i) {
      base_color = &lob->base;

      for (i=0; i<ob->psize; i++)
         ob->shade[i] = base_color;

      for (i=0; i<lob->count; i++)
         ob->shade[lob->list[i].index] = &lob->list[i];
   }

   copyarray4(bbox[0], dob->bbox[0]);
   copyarray4(bbox[1], dob->bbox[1]);
}


/* *************************************************************
************************************************************* */
void spatch::colorize(grandchildtype *child) {

   return;
}


/* **************************************************
   This procedure scans 2D grids into surface patches
   or polygonal structures.
************************************************** */
void spatch::scan(camera *cparm, light *lparm, zbuffer *buff, subslave grandchild[], pc **shadptr) {

   polygon    polyob;                   // polygon structure
   int       i, j, k, l;
   spatchtype ptr;
   texpolygon ttr;
   shaderlisttype slist;
   int           texflag = mcinfo & CITEXTURE;
   int           flag3d = mcinfo & CI3DTEX;

   transform(cparm);

   if (!ob->psize || !(i=bound_box(cparm)))
      return;

   if (i == 2) {
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);
      polyob.scan(cparm, lparm, buff, grandchild, shadptr);
      polyob.end_scan();               // this deletes contents of x
      return;
   }

   if (mctype <= PHONG) {                // display data as polygons
      setup_transfer(ob->srow*ob->scol, ob->psize);
      spline2poly(ob, tob, texflag);
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);

      if (texflag)
         polyob.setup_texture(tpob);

      if (flag3d) {
         i = ob->psize<<1;
         slist.scount = i;
         slist.scale = shaderlist->scale;

         slist.shade = new shadertype[i];

         for (i=j=0; i<ob->psize; i++) {
            slist.shade[j] = shaderlist->shade[i];
            j++;
            slist.shade[j] = shaderlist->shade[i];
            j++;
         }

         polyob.setup_texture3d(&slist, world, iworld);
      }

      polyob.scan(cparm, lparm, buff, grandchild, shadptr);
      polyob.end_scan();               // this deletes contents of x
      return;
   }

   ptr.build(SPDELTA + 3, SPDELTA + 3, mctype > PWFBW, mctype > PBW);
   setup_transfer(ptr.vsize, ptr.psize);

   if (texflag) {
      ttr.countobject = ptr.psize;
      ttr.setup_data(ptr.psize);
      ttr.setup_pdata(ptr.prow);

      for (i=j=0; i<ptr.prow; i++, j+=ptr.pcol)
         ttr.pdata[i] = &ttr.data[j];

      for (i=0; i<ttr.countobject; i++)
         ttr.data[i] = (tex *)NULL;
 
      polyob.setup_texture(tpob);
   }

   l = ptr.psize<<1;

   if (flag3d) {
      slist.scount = 0;
      slist.shade = new shadertype[l];

      polyob.setup_texture3d(&slist, world, iworld);
    }

   polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);

   for (i=0; i<ob->prow; i++)
      for (j=0; j<ob->pcol; j++) {

         if (flag3d) {
            for (k=0; k<l; k++)
               slist.shade[k].s = (shader **)NULL;
         }

         if (texflag) {
            if (ttr.data[0] != (tex *)NULL)
               for (k=0; k<ttr.countobject; k++) {
                  delete ttr.data[k];
                  ttr.data[k] = (tex *)NULL;
               }

            if (tpob->data[0] != (tex *)NULL)
               for (k=0; k<tpob->countobject; k++) {
                  delete tpob->data[k];
                  tpob->data[k] = (tex *)NULL;
               }

         }

         convert(i, j, &ptr, &ttr, &slist);
         polyob.scan(cparm, lparm, buff, grandchild, shadptr);
      }

      polyob.end_scan();         // this deletes contents of x
}


/* **************************************************
   This procedure scans 2D grids into surface patches
   or polygonal structures.
************************************************** */
void spatch::scan(camera *cparm, light *lparm, subslave grandchild[], pc **shadptr) {

   polygon    polyob;                   // polygon structure
   int        i, j;
   spatchtype ptr;

   if (mctype <= PHONG)
      mctype = WFBW;
   else
      mctype = PWFBW;

   transform(cparm);

   if (!ob->psize || !(i=bound_box(cparm)))
      return;

   if (i == 2) {
      polyob.setup_patch(pob, WFBW, CINULL, id, sflag, splane, bbox, base_color);
      polyob.scan(cparm, lparm, grandchild, shadptr);
      polyob.end_scan();               // this deletes contents of x
      return;
   }

   if (mctype == BW) {                // display data as polygons
      setup_transfer(ob->srow*ob->scol, ob->psize);
      spline2poly(ob, (texpolygon *)NULL, 0);
      polyob.setup_patch(pob, WFBW, CINULL, id, sflag, splane, bbox, base_color);

      polyob.scan(cparm, lparm, grandchild, shadptr);
      polyob.end_scan();               // this deletes contents of x
      return;
   }

   ptr.build(SPDELTA + 3, SPDELTA + 3, 0, 0);
   setup_transfer(ptr.vsize, ptr.psize);

   for (i=0; i<ob->prow; i++)
      for (j=0; j<ob->pcol; j++) {
         convert(i, j, &ptr, (texpolygon *)NULL, (shaderlisttype *)NULL);
         polyob.setup_patch(pob, PWFBW, CINULL, id, sflag, splane, bbox, base_color);
         polyob.scan(cparm, lparm, grandchild, shadptr);
      }

   polyob.end_scan();               // this deletes contents of x
}


/* **************************************************
************************************************** */
void spatch::beamscan(light *lparm, subslave grandchild[]) {

   polygon    polyob;                   // polygon structure
   int        i, j;
   spatchtype ptr;
   colortype  nctype;
   unsigned int ncinfo;

   if (mctype < BW || (mctype > PHONG && mctype < PBW))
      return;

   nctype = mctype;
   ncinfo = mcinfo;

   mcinfo &= ~CITEXTURE;

   if (mctype <= PHONG)
      mctype = BW;
   else
      mctype = PBW;

   transform(lparm);

   if (!ob->psize || !(i=bound_box(lparm))) {
      mctype = nctype;
      mcinfo = ncinfo;
      return;
   }

   if (i == 2) {
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);
      polyob.beamscan(lparm, grandchild);
      polyob.end_scan();         // this deletes contents of x

      mctype = nctype;
      mcinfo = ncinfo;
      return;
   }

   if (mctype == BW) {                // display data as polygons
      setup_transfer(ob->srow*ob->scol, ob->psize);
      spline2poly(ob, (texpolygon *)NULL, 0);
      polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);

      polyob.beamscan(lparm, grandchild);
      polyob.end_scan();         // this deletes contents of x

      mctype = nctype;
      mcinfo = ncinfo;
      return;
   }

   ptr.build(SPDELTA + 3, SPDELTA + 3, mctype > PWFBW, 0);
   setup_transfer(ptr.vsize, ptr.psize);

   for (i=0; i<ob->prow; i++)
      for (j=0; j<ob->pcol; j++) {
         convert(i, j, &ptr, (texpolygon *)NULL, (shaderlisttype *)NULL);
         polyob.setup_patch(pob, mctype, mcinfo, id, sflag, splane, bbox, base_color);
         polyob.beamscan(lparm, grandchild);
      }

   polyob.end_scan();         // this deletes contents of x
   mctype = nctype;
   mcinfo = ncinfo;
}

