
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */




#include "sphere.h"
#include <string.h>


#define U 10
#define DU 0.1
#define V 16
#define DV 0.0625
				// (U+1-2) * V + 2
#define COUNTVERTEX   146
				// U * V
#define COUNTOBJECT   160
				// 4 * U * V
#define COUNTEDGE     640


/* *************************************************************
************************************************************* */
int sphere::parse(FILE *infile, char *token) {

   switch (token[0]) {

      case 'a':
         if (!strcmp(token, "axis")) {
            get_token(infile, token);
            wscale[0] = atof(token);
            get_token(infile, token);
            wscale[1] = atof(token);
            get_token(infile, token);
            wscale[2] = atof(token);

            return 1;
         }

         if (strlen(token) < 6)
            break;

         switch (token[5]) {

            case 'x':
               if (!strcmp(token, "axis_x")) {
                  get_token(infile, token);
                  wscale[0] = atof(token);

                  return 1;
               }

               break;

            case 'y':
               if (!strcmp(token, "axis_y")) {
                  get_token(infile, token);
                  wscale[1] = atof(token);

                  return 1;
               }

               break;

            case 'z':
               if (!strcmp(token, "axis_z")) {
                  get_token(infile, token);
                  wscale[2] = atof(token);

                  return 1;
               }

               break;

            default:
               break;
         }

         break;

      default:
         break;
   }

   return pc::parse(infile, token);
}


float sphere::calc_area() {

   return sphere_area((wscale[0]+wscale[1]+wscale[2])*0.33333333);
}


void sphere::pg(eye *parm, light *lparm, zbuffer *zbuff, int inc) {

   int  i, j;
   float surface[4], pzero[4];
   float temp[2];
   vector4f normal;
   float pt[4];
   float t;
   int   flag;
   float rcolor[3], acolor[3];
   light *lmain;
   int  flag2t = mcinfo & CITEXTURE;
   int  flag3t = mcinfo & CI3DTEX;
   int  flagl  = mcinfo & CILUM;
   int  flagt  = flag2t || flag3t;
   int  flagtr = mcinfo & CITRANSPARENT;

   texcolortype texcolor;
   shaderparamtype sparam;
   float fscale;

   float piinverse = 1.0/PI;
   float tx, ty;
   float di, dj, wi, wj, wt;

   float area;

   if (flag2t)
      area = sqrt((tob->query_x()*tob->query_y())/(calc_area()*SQR(parm->query_scale())));

   normal[3] = 0;

   fscale = wscale[0];

   if (wscale[1] < fscale)
      fscale = wscale[1];
   if (wscale[2] < fscale)
      fscale = wscale[2];

   fscale = 1.0/fscale;		// note: doesnt take into account scale
				// 	of rotation mx

   for (lmain=lparm; lmain!=(light *)NULL && lmain->ltype != POINT && lmain->ltype != FAR; lmain=lmain->next);

   texcolor = *base_color;

   if (lmain != (light *)NULL && !flagt)
      lmain->set_ambient(texcolor.color.ka, texcolor.ambient);

   pzero[0] = pzero[1] = pzero[2] = 0;
   pzero[3] = 1;

   matvecmulto(iworld, pzero);

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   parm->parascreen2map(surface);

   dj = 1.0/parm->query_scale();
   di = dj*MAXSUBPROC;

   wt = surface[0];
   wi = surface[1];

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {
         surface[0] = wj;
         surface[1] = wi;
         surface[2] = -1;

         copyarray3(pt, surface);
         matvecmultv(iworld, pt);

         flag = line_ellipse_intersect(wscale, pzero, pt, &t);

         if (flag && (-t>zbuff->zbuff[i][j] || zbuff->zbuff[i][j] > front)) {
            surface[0] *= t;
            surface[1] *= t;
            surface[2] = -t;

            if (flag > 0) {
               copyarray3(normal, surface);
               subarray3(normal, wcenter);
            }

            else {
               copyarray3(normal, wcenter);
               subarray3(normal, surface);
            }

            matvecmultv(normalcalc, normal);
            normalize3(normal);

            if (flagt)
               texcolor = *base_color;

            if (flag2t) {
               smultarray3(pt, t);
               addarray3(pt, pzero);

               normalize3(pt);
               ty = ACOS(-pt[1])*piinverse;

               temp[0] = pt[0];
               temp[1] = pt[2];
               normalize2(temp);
               if (temp[0] >= 0.0)
                  tx = ACOS(temp[1])*piinverse*0.5;
               else
                  tx = 1-ACOS(temp[1])*piinverse*0.5;

               query_texel(tx, ty, &texcolor, area*(-surface[2]));
               if (lmain != (light *)NULL)
                  lmain->set_ambient(texcolor.color.ka, texcolor.ambient);
            }

            if (flag3t)
               calc_color(surface, fscale*shaderlist->scale, normal, lmain, lparm, parm, &texcolor, flagtr, 1, flagl, rcolor, acolor);
            else
               calc_color(surface, 1, normal, lmain, lparm, parm, &texcolor, flagtr, 0, flagl, rcolor, acolor);

            if (flagtr) {
               if (zbuff->zbuff[i][j] > front) {
                  zbuff->zbuff[i][j] = -999999999;
                  copyarray3(zbuff->ybuff[i][j], rcolor);
                  copyarray3(zbuff->ambientbuff[i][j], acolor);
               }

               else {
                  if (rcolor[0] > zbuff->ybuff[i][j][0])
                     zbuff->ybuff[i][j][0] = rcolor[0];
                  if (rcolor[1] > zbuff->ybuff[i][j][1])
                     zbuff->ybuff[i][j][1] = rcolor[1];
                  if (rcolor[2] > zbuff->ybuff[i][j][2])
                     zbuff->ybuff[i][j][2] = rcolor[2];

                  if (acolor[0] > zbuff->ambientbuff[i][j][0])
                     zbuff->ambientbuff[i][j][0] = acolor[0];
                  if (acolor[1] > zbuff->ambientbuff[i][j][1])
                     zbuff->ambientbuff[i][j][1] = acolor[1];
                  if (acolor[2] > zbuff->ambientbuff[i][j][2])
                     zbuff->ambientbuff[i][j][2] = acolor[2];
               }

            }

            else {
               zbuff->zbuff[i][j] = surface[2];
               copyarray3(zbuff->ybuff[i][j], rcolor);
               copyarray3(zbuff->ambientbuff[i][j], acolor);
            }

         }

      }

}


void sphere::lt(light *lparm, int inc) {

   int i, j;
   float surface[4], pzero[4];
   float t;
   int   flag;
   float di, dj, wi, wj, wt;

   surface[0] = col[0];
   surface[1] = row[0]+inc;
   surface[2] = -1;
   surface[3] = 1;
   copyarray3(pzero, surface);

   lparm->parascreen2map(surface);

   pzero[0] += 1;
   pzero[1] += MAXSUBPROC;
   pzero[3] = 0;
   lparm->parascreen2map(pzero);

   dj = pzero[0] - surface[0];
   di = pzero[1] - surface[1];
   wt = surface[0];
   wi = surface[1];

   pzero[0] = pzero[1] = 0;

   if (lparm->ltype == BEAM_POINT) {
      pzero[2] = 0;
      pzero[3]   = 1;
      matvecmulto(iworld, pzero);
      surface[3] = 0;
   }

   else {
      pzero[2] = -1;
      pzero[3]   = 0;
      matvecmultv(iworld, pzero);
      surface[3] = 1;
   }

   for (i=row[0]+inc; i<=row[1]; i+=MAXSUBPROC, wi+=di)
      for (j=col[0], wj=wt; j<=col[1]; j++, wj+=dj) {

         surface[0] = wj;
         surface[1] = wi;

         if (lparm->ltype == BEAM_POINT) {
            surface[2] = -1;
            matvecmultv(iworld, surface);
            flag = line_ellipse_intersect(wscale, pzero, surface, &t);
         }

         else {
            surface[2] = 0;
            matvecmulto(iworld, surface);
            flag = line_ellipse_intersect(wscale, surface, pzero, &t);
         }

         if (flag && (-t>lparm->lbuff.lbuff->zbuff.pdata[i][j] || lparm->lbuff.lbuff->zbuff.pdata[i][j] > front)) {
            lparm->lbuff.lbuff->zbuff.pdata[i][j] = -t;
            lparm->lbuff.lbuff->idbuff.pdata[i][j] = id;
         }

      }

}


void sphere::transform(eye *parm) {

   int i, j;

   copymx4x4(world, rotate);

   world[0][3] += center[0];
   world[1][3] += center[1];
   world[2][3] += center[2];

   matmatmulto(parm->transform, world);

   wcenter[0] = world[0][3];
   wcenter[1] = world[1][3];
   wcenter[2] = world[2][3];
   wcenter[3] = 1;

   smultarray3(wscale, size);

   inversemx(world, iworld);

   copymx4x4(normalcalc, iworld);
   sdivarray4(normalcalc[0], wscale[0]);
   sdivarray4(normalcalc[1], wscale[1]);
   sdivarray4(normalcalc[2], wscale[2]);
   matmatmult(world, normalcalc);

/*
   inversemxrt(rotate, normalcalc);
   sdivarray4(normalcalc[0], wscale[0]);
   sdivarray4(normalcalc[1], wscale[1]);
   sdivarray4(normalcalc[2], wscale[2]);
   matmatmult(rotate, normalcalc);
*/

   bbox[0][0] = -wscale[0];
   bbox[0][1] = -wscale[1];
   bbox[0][2] = -wscale[2];
   bbox[0][3] = 1;

   bbox[1][0] = wscale[0];
   bbox[1][1] = wscale[1];
   bbox[1][2] = wscale[2];
   bbox[1][3] = 1;

   make_box(bbox, box);

   matvecmulto(world, box[0]);
   copyarray3(bbox[0], box[0]);
   copyarray3(bbox[1], box[0]);

   for (i=1; i<8; i++) {
      matvecmulto(world, box[i]);

      for (j=0; j<3; j++)
         if (box[i][j] < bbox[0][j])
            bbox[0][j] = box[i][j];
         else
            if (box[i][j] > bbox[1][j])
               bbox[1][j] = box[i][j];
   }

}


void sphere::geo2poly(texpolygon **ptob, shaderlisttype *slist) {

   unsigned int i, j, k, l;
   float fi, fdi, fj, fdj;
   float r;
   float tx, ty;
   vector4f sworld[4];

   pob = new polytype;

   pob->build(COUNTVERTEX, COUNTOBJECT, COUNTEDGE, (mctype > BW && ptob != (texpolygon **)NULL ));

   if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
      *ptob = new texpolygon;
      (*ptob)->countobject = COUNTOBJECT;
      (*ptob)->data = new texptrtype[COUNTOBJECT];

      for (i=0; i<COUNTOBJECT; i++) {
         (*ptob)->data[i] = new tex;
         (*ptob)->data[i]->mapfunc = COORD;
         (*ptob)->data[i]->ob = tob;
         (*ptob)->data[i]->pt = new vector2f[5];
      }

   }

   if (mcinfo & CI3DTEX && slist != (shaderlisttype *)NULL) {
      slist->scount = COUNTOBJECT;
      slist->shade = new shadertype[COUNTOBJECT];
   }

   init_mx(i, j, sworld);
   sworld[0][0] = wscale[0];
   sworld[1][1] = wscale[1];
   sworld[2][2] = wscale[2];

   matmatmulto(world, sworld);

   // optimization - do only 1/8 of below then calc the symetrical parts

   for (k=i=0, fdi = DU+DU, fi = -1 + fdi; i < U-1; i++, fi += fdi) {
      r = 1.0 - SQR(fi);
      if (r > CORRECT)
         r = sqrt(r);
      else
         r = 0;

      for (j=0, fj=-HALFPI, fdj = TWOPI*DV; j < V; j++, fj += fdj, k++) {
         pob->gnormal[k][0] = pob->mvertex[k][0] = COS(fj)*r;
         pob->gnormal[k][1] = pob->mvertex[k][1] = fi;
         pob->gnormal[k][2] = pob->mvertex[k][2] = -SIN(fj)*r;
         pob->gnormal[k][3] = pob->mvertex[k][3] = 1;

         matvecmulto(sworld, pob->mvertex[k]);
         matvecmultv(world, pob->gnormal[k]);
         normalize3(pob->gnormal[k]);
      }

   }

   pob->gnormal[k][0] = pob->mvertex[k][0] = 0;
   pob->gnormal[k][1] = pob->mvertex[k][1] = 1;
   pob->gnormal[k][2] = pob->mvertex[k][2] = 0;
   pob->gnormal[k][3] = pob->mvertex[k][3] = 1;
   matvecmulto(sworld, pob->mvertex[k]);
   matvecmultv(world, pob->gnormal[k]);
   normalize3(pob->gnormal[k]);
   k++;
   pob->gnormal[k][0] = pob->mvertex[k][0] = 0;
   pob->gnormal[k][1] = pob->mvertex[k][1] = -1;
   pob->gnormal[k][2] = pob->mvertex[k][2] = 0;
   pob->gnormal[k][3] = pob->mvertex[k][3] = 1;
   matvecmulto(sworld, pob->mvertex[k]);
   matvecmultv(world, pob->gnormal[k]);
   normalize3(pob->gnormal[k]);

   for (i=k=l=0, ty=fdi=DU*texcoord[2][1]; i < U-2; i++, ty+=fdi, k++) {
      for (j=0, tx=0, fdj=DV*texcoord[2][0]; j < V-1; j++, k++, tx+=fdj) {

         pob->polynum[k] = 4;
         pob->edgeptr[k] = &pob->object[l];

         pob->object[l] = k;
         l++;
         pob->object[l] = k+V;
         l++;
         pob->object[l] = k+V+1;
         l++;
         pob->object[l] = k+1;
         l++;

         if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
            (*ptob)->data[k]->pt[0][0] = tx;
            (*ptob)->data[k]->pt[0][1] = ty; 
            (*ptob)->data[k]->pt[1][0] = tx;
            (*ptob)->data[k]->pt[1][1] = ty+fdi;
            (*ptob)->data[k]->pt[2][0] = tx+fdj;
            (*ptob)->data[k]->pt[2][1] = ty+fdi; 
            (*ptob)->data[k]->pt[3][0] = tx+fdj;
            (*ptob)->data[k]->pt[3][1] = ty; 
            (*ptob)->data[k]->pt[4][0] = tx;
            (*ptob)->data[k]->pt[4][1] = ty;
         }

      }

      pob->polynum[k] = 4;
      pob->edgeptr[k] = &pob->object[l];

      pob->object[l] = k;
      l++;
      pob->object[l] = k+V;
      l++;
      pob->object[l] = k+1;
      l++;
      pob->object[l] = k+1-V;
      l++;

      if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
         (*ptob)->data[k]->pt[0][0] = tx;
         (*ptob)->data[k]->pt[0][1] = ty; 
         (*ptob)->data[k]->pt[1][0] = tx;
         (*ptob)->data[k]->pt[1][1] = ty+fdi;
         (*ptob)->data[k]->pt[2][0] = texcoord[2][0];
         (*ptob)->data[k]->pt[2][1] = ty+fdi; 
         (*ptob)->data[k]->pt[3][0] = texcoord[2][0];
         (*ptob)->data[k]->pt[3][1] = ty; 
         (*ptob)->data[k]->pt[4][0] = tx;
         (*ptob)->data[k]->pt[4][1] = ty;
      }

   }
				// top patches
   for (j=0, tx=0, fdj=DV*texcoord[2][0], ty=texcoord[2][1]-fdi; j < V-1; j++, k++, tx+=fdj) {

      pob->polynum[k] = 4;
      pob->edgeptr[k] = &pob->object[l];

      pob->object[l] = k+1;
      l++;
      pob->object[l] = k;
      l++;
      pob->object[l] = COUNTVERTEX-2;
      l++;
      pob->object[l] = COUNTVERTEX-2;
      l++;

      if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
         (*ptob)->data[k]->pt[0][0] = tx+fdj;
         (*ptob)->data[k]->pt[0][1] = ty; 
         (*ptob)->data[k]->pt[1][0] = tx;
         (*ptob)->data[k]->pt[1][1] = ty;
         (*ptob)->data[k]->pt[2][0] = tx;
         (*ptob)->data[k]->pt[2][1] = texcoord[2][1]; 
         (*ptob)->data[k]->pt[3][0] = tx+fdj;
         (*ptob)->data[k]->pt[3][1] = texcoord[2][1]; 
         (*ptob)->data[k]->pt[4][0] = tx+fdj;
         (*ptob)->data[k]->pt[4][1] = ty;
      }

   }

   pob->polynum[k] = 4;
   pob->edgeptr[k] = &pob->object[l];

   pob->object[l] = k+1-V;
   l++;
   pob->object[l] = k;
   l++;
   pob->object[l] = COUNTVERTEX-2;
   l++;
   pob->object[l] = COUNTVERTEX-2;
   l++;

   if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
      (*ptob)->data[k]->pt[0][0] = texcoord[2][0];
      (*ptob)->data[k]->pt[0][1] = ty; 
      (*ptob)->data[k]->pt[1][0] = tx;
      (*ptob)->data[k]->pt[1][1] = ty;
      (*ptob)->data[k]->pt[2][0] = tx;
      (*ptob)->data[k]->pt[2][1] = texcoord[2][1]; 
      (*ptob)->data[k]->pt[3][0] = texcoord[2][0];
      (*ptob)->data[k]->pt[3][1] = texcoord[2][1]; 
      (*ptob)->data[k]->pt[4][0] = texcoord[2][0];
      (*ptob)->data[k]->pt[4][1] = ty;
   }

   k++;
					// bottom
   for (j=0, tx=0, fdj=DV*texcoord[2][0], ty=texcoord[2][1]-fdi; j < V-1; j++, k++, tx+=fdj) {

      pob->polynum[k] = 4;
      pob->edgeptr[k] = &pob->object[l];

      pob->object[l] = j;
      l++;
      pob->object[l] = j+1;
      l++;
      pob->object[l] = COUNTVERTEX-1;
      l++;
      pob->object[l] = COUNTVERTEX-1;
      l++;

      if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
         (*ptob)->data[k]->pt[0][0] = tx;
         (*ptob)->data[k]->pt[0][1] = fdi; 
         (*ptob)->data[k]->pt[1][0] = tx+fdj;
         (*ptob)->data[k]->pt[1][1] = fdi;
         (*ptob)->data[k]->pt[2][0] = tx+fdj;
         (*ptob)->data[k]->pt[2][1] = 0.0; 
         (*ptob)->data[k]->pt[3][0] = tx;
         (*ptob)->data[k]->pt[3][1] = 0.0; 
         (*ptob)->data[k]->pt[4][0] = tx;
         (*ptob)->data[k]->pt[4][1] = fdi;
      }

   }

   pob->polynum[k] = 4;
   pob->edgeptr[k] = &pob->object[l];

   pob->object[l] = j;
   l++;
   pob->object[l] = 0;
   l++;
   pob->object[l] = COUNTVERTEX-1;
   l++;
   pob->object[l] = COUNTVERTEX-1;
   l++;

   if (mcinfo & CITEXTURE && ptob != (texpolygon **)NULL) {
      (*ptob)->data[k]->pt[0][0] = tx;
      (*ptob)->data[k]->pt[0][1] = fdi; 
      (*ptob)->data[k]->pt[1][0] = texcoord[2][0];
      (*ptob)->data[k]->pt[1][1] = fdi;
      (*ptob)->data[k]->pt[2][0] = texcoord[2][0];
      (*ptob)->data[k]->pt[2][1] = 0.0; 
      (*ptob)->data[k]->pt[3][0] = tx;
      (*ptob)->data[k]->pt[3][1] = 0.0; 
      (*ptob)->data[k]->pt[4][0] = tx;
      (*ptob)->data[k]->pt[4][1] = fdi;
   }

   if (ptob != (texpolygon **)NULL) {	// NULL signifies NO coloring
      if (mctype > BW)
         for (i=0; i<COUNTOBJECT; i++)
            pob->shade[i] = base_color;

      if (mcinfo & CI3DTEX) {
         slist->scale = shaderlist->scale;
         for (i=0; i<COUNTOBJECT; i++)
            slist->shade[i] = shaderlist->shade[0];
      }

   }

}

