
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   This file contains some of the file I/O routines
************************************************************** */

#include "pixcon.h"
#include <string.h>

master control;

/* **************************************************
************************************************** */
void killchild() {

   int i, j;
   semun x;

   for (i=0; i<MAXPROCESSOR; i++) {
      for (j=0; j<MAXSUBPROC-1; j++)            // comment out scanline
         if (!kill(control.circuit[i].grandchild[j].child_id, 9))
            printf("Killed subchild %d %d\n", i, j);
      if (!kill(control.circuit[i].child_id, 9))
         printf("Killed child %d\n", i);
   }

   semctl(control.semid, 1, IPC_RMID, x);			// kill semaphores

   exit(0);
}


/* **************************************************
************************************************** */
void kc(int x) {
   killchild();
}


/* **************************************************
   this function parses command line to find user defined files,
   or else use the default files.
************************************************** */
void parseinput(int argc, char *argv[]) {

   int i, j;

   for (i=1; i<argc; i++)
      if (*(argv[i]) == '-')
         switch (*(argv[i]+1)) {
            case 'a':                   // aliasing
               ANTIALIAS = 1;
               if (i+1 < argc && argv[i+1][0] != '-') {
                  i++;
                  strcpy(SANTIALIAS, argv[i]);
               }

               else
                  SANTIALIAS[0] = 0;

               break;

            case 'f':                   // image file
               i++;
               if (i < argc) {
                  RGBFILE = 1;
                  strcpy(SRGBFILE, argv[i]);
               }

               break;

            case 'g':
               i++;
               if (i < argc) {
                  GAMMA = 1;
                  FGAMMA = atof(argv[i]);
               }

               break;

            case 'i':                   // user default light
               i++;
               if (i < argc) {
                  DEFLIGHT = 1;
                  strcpy(SDEFLIGHT, argv[i]);
               }

               break;

            case 'l':                   // user defined .lst
               i++;
               if (i < argc) {
                  strcpy(SDEFLST, argv[i]);
                  DEFLST = 1;
               }

               break;

            case 'p':                   // pencil test
               PENCILTEST = 1;
               NOSCREEN = 0;
               MAXSUBPROC = 1;
               MAXPROCESSOR = 1;
               break;

            case 's':                   // save to file
               if (strlen(argv[i]) > 2) 
                  if (argv[i][2] == 'i')
                     SILENT = 1;
                  else
                     if (argv[i][2] == 'a') {
                        i++;
                        if (i < argc) {
                           strcpy(SSAVE, argv[i]);
                           SAVE = 1;
                        }

                     }

               break;

            case 'n':                   // no shadow
               if (strlen(argv[i]) > 2 && argv[i][2] == 'i')
                  NIXMIP = 1;
               else
                  if (strlen(argv[i]) > 4 && argv[i][4] == 'c') {
                     NOSCREEN = 1;
                     PENCILTEST = 0;
                     RENDER = 0;
                  }

                  else
                     NOSHADOW = 1;

               break;

            case 'r':                   // from frame to frame
               i++;
               if (i < argc) {
                  RENDER_START = atoi(argv[i]);
                  i++;
                  if (i < argc) {
                     RENDER_END = atoi(argv[i]);
                     RENDER = 1;
                     NOSCREEN = 0;
                  }

               }

               break;

            case 'x':
               i++;
               DEFWINX = atoi(argv[i]);
               break;

            case 'v':                   // user defined default camera
               i++;
               if (i < argc) {
                  strcpy(SDEFCAMERA, argv[i]);
                  DEFCAMERA = 1;
               }

               break;

            case 'y':
               i++;
               DEFWINY = atoi(argv[i]);
               break;

            case 'M':

#ifdef SGI
               if (strlen(argv[i]) > 4)
                  if (argv[i][4] == 'P') {
                     i++;
                     if (i < argc) {
                        j = atoi(argv[i]);
                        if (j > 0) {
                           PENCILTEST = 0;
                           MAXPROCESSOR = j;
                        }

                     }

                  }

                  else {
                     i++;
                     if (i < argc) {
                        j = atoi(argv[i]);
                        if (j > 0) {
                           PENCILTEST = 0;
                           MAXSUBPROC = j;
                        }

                     }

                  }
#endif

               break;

            default:
               break;
         }

}


/* *************************************************************
************************************************************* */
unsigned int parse_info(char *instring) {

   unsigned int mcinfo = 0;
   int  i, j;

   for (i=0, j=strlen(instring); i<j; i++) {
      switch (instring[i]) {
         case 'R':
            mcinfo |= CIRAYCAST;
            break;

         case '2':
            mcinfo |= CITEXTURE;
            break;

         case '3':
            mcinfo |= CI3DTEX;
            break;

         case 'L':
            mcinfo |= CILUM;
            break;

         case 'T':
            mcinfo |= CITRANSPARENT;
            break;

         case 'B':
            mcinfo |= (CIBITMAP | CITEXTURE);
            break;

         default:
            break;
      }

   }

   return mcinfo;
}


int calc_bucket(colortype ctype, unsigned int cinfo) {

   int k;

   switch (ctype) {
      case DOT:
      case PDOT:
      case WFBW:
         k = 0;
         break;

       case PWFBW:
       case BW:
       case CONSTANT:
       case GOURAUD:
          k = 1;
          break;

       case ALEX:
       case PHONG:
       case PBW:
       case PCONSTANT:
       case PGOURAUD:
          k = 2;
          break;

       default:
          k = 3;
    }

    if (cinfo & CITEXTURE)
       k++;
    if (cinfo & CIRAYCAST)
       k++;
    if (cinfo & CILUM)
       k++;
    if (cinfo & CI3DTEX)
       k+=2;

    return k;
}


void parse_color(char *ctype, colortype *mctype, unsigned int *mcinfo) {

   switch (ctype[0]) {
      case 'A':
         *mctype = ALEX;
         *mcinfo = parse_info(ctype+4);
         break;

      case 'B':
         *mctype = BW;
         *mcinfo = parse_info(ctype+2);
         break;

      case 'C':
         *mctype = CONSTANT;
         *mcinfo = parse_info(ctype+8);
         break;

      case 'D':
         *mctype = DOT;
         *mcinfo = parse_info(ctype+3);
         break;

      case 'G':
         *mctype = GOURAUD;
         *mcinfo = parse_info(ctype+7);
         break;

      case 'P':
         switch (ctype[1]) {
            case 'H':
               *mctype = PHONG;
               *mcinfo = parse_info(ctype+5);
               break;

            case 'A':
               *mctype = PALEX;
               *mcinfo = parse_info(ctype+5);
               break;

            case 'B':
               *mctype = PBW;
               *mcinfo = parse_info(ctype+3);
               break;

            case 'C':
               *mctype = PCONSTANT;
               *mcinfo = parse_info(ctype+9);
               break;

            case 'D':
               *mctype = PDOT;
               *mcinfo = parse_info(ctype+4);
               break;

            case 'G':
               *mctype = PGOURAUD;
               *mcinfo = parse_info(ctype+8);
               break;

            case 'P':
               *mctype = PPHONG;
               *mcinfo = parse_info(ctype+6);
               break;

            default:
               *mctype = PWFBW;
         }

         break;

      default:
         *mctype = WFBW;
   }

}


/* *************************************************************
   This procedure reads in a list of objects, their attributes,
   then reads in the object data.
************************************************************* */
pc *read_object(FILE *infile, char *otype, char *ctype, listtype *mainlist) {

   pc   *ptr = (pc *)NULL;
   char stype[MAXSTRLEN];
   int  j, k, l;

   fscanf(infile, "%s", ctype);        // object coloring

   switch (otype[0]) {

      case 'G':

         switch (otype[4]) {

            case 'P':			// SPHERE
               ptr = new sphere;
               break;

            case 'Y':			// CYLINDER
               ptr = new cylinder;
               break;

            case 'O':			// CONE
               ptr = new cone;
               break;

            default:
               break;
         }

         parse_color(ctype, &ptr->mctype, &ptr->mcinfo);

         if ((ptr->mctype > BW && ptr->mctype <=PHONG) || ptr->mctype > PBW) {
            fscanf(infile, "%s", ptr->colorname);                      // color file

            if (ptr->mcinfo & CITEXTURE)
               fscanf(infile, "%s", ptr->texname2);                   // texture filename

            if (ptr->mcinfo & CI3DTEX)
               fscanf(infile, "%s", ptr->texname3);                 // 3D texture file
         }

         fscanf(infile, "%s", stype);                      // shadow flag

         if (ptr->sflag = !strcmp(stype, "SHADOW"))
            fscanf(infile, "%f %f %f %f", &ptr->splane[0], &ptr->splane[1],
                                          &ptr->splane[2], &ptr->splane[3]);

         fscanf(infile, "%f %f %f", &ptr->center[0], &ptr->center[1],
					&ptr->center[2]);	// object location

         fscanf(infile, "%f", &ptr->size);          // object size

         for (j=0; j<4; j++)                         // object rotation
            fscanf(infile, "%f %f %f %f", &ptr->rotate[j][0],
                   &ptr->rotate[j][1], &ptr->rotate[j][2],
                   &ptr->rotate[j][3]);

         ptr->read_data(infile);

         break;

      case 'L':			// line
         ptr = new line;

         if (ctype[0] == 'G')
            ptr->mctype = GOURAUD;
         else
            ptr->mctype = WFBW;

         fscanf(infile, "%s", stype);                      // shadow flag

         if (ptr->sflag = !strcmp(stype, "SHADOW"))
            fscanf(infile, "%f %f %f %f", &ptr->splane[0], &ptr->splane[1],
                                          &ptr->splane[2], &ptr->splane[3]);
         ptr->read_data(infile);

         break;

      case 'P':
         if (!strcmp(otype, "PARTICLE")) {

            ptr = new particle;

            if (ctype[0] == 'G')
               ptr->mctype = GOURAUD;
            else
               ptr->mctype = WFBW;

            fscanf(infile, "%s", stype);                      // shadow flag

            if (ptr->sflag = !strcmp(stype, "SHADOW"))
               fscanf(infile, "%f %f %f %f", &ptr->splane[0], &ptr->splane[1],
                                             &ptr->splane[2], &ptr->splane[3]);

            ptr->read_data(infile);
         }

         else {
            ptr = new polygon;

            parse_color(ctype, &ptr->mctype, &ptr->mcinfo);

            fscanf(infile, "%s", ptr->filename);                      // object file

            if (ptr->mctype > BW) {
               fscanf(infile, "%s", ptr->colorname);                      // color file

               if (ptr->mcinfo & CITEXTURE) {
                  fscanf(infile, "%s", ptr->texname2);                   // texture filename
                  if ((ptr->mcinfo & CIBITMAP)) {
                     fscanf(infile, "%d %d %d", &j, &k, &l);
                     ((polygon *)ptr)->inviso = j | (((l<<8) | k)<<8);
                  }

               }

               if (ptr->mcinfo & CI3DTEX)
                  fscanf(infile, "%s", ptr->texname3);                 // 3D texture file
            }

            fscanf(infile, "%s", stype);                      // shadow flag

            if (ptr->sflag = !strcmp(stype, "SHADOW"))
               fscanf(infile, "%f %f %f %f", &ptr->splane[0], &ptr->splane[1],
                                             &ptr->splane[2], &ptr->splane[3]);

            fscanf(infile, "%f %f %f", &ptr->center[0], &ptr->center[1],
					&ptr->center[2]);

            fscanf(infile, "%f", &ptr->size);          // object size

            for (j=0; j<4; j++)                         // object rotation
               fscanf(infile, "%f %f %f %f", &ptr->rotate[j][0],
                      &ptr->rotate[j][1], &ptr->rotate[j][2], &ptr->rotate[j][3]);
         }

         break;

      case 'R':
         ptr = new radiant;

         if (ctype[1] == 'S') {
            ptr->mctype = RSOLID;
            for (j=6, k=strlen(ctype); j<k; j++)
               if (ctype[j] == 'L')
                  ptr->mcinfo |= CILUM;
               else
                  if (ctype[j] == 'H')			// flag for hollow - black hole
                     ((radiant *)ptr)->flagh = 1;
         }

         else
            if (ctype[0] == 'I') {
               ptr->mctype = IRRADIANT;
               ptr->mcinfo |= CILUM;
               for (j=6, k=strlen(ctype); j<k; j++)
                  if (ctype[j] == 'H')
                     ((radiant *)ptr)->flagh = 1;
            }

            else
               ptr->mctype = WFBW;

         if (ptr->mctype != WFBW)
            fscanf(infile, "%s", ptr->colorname);        // object coloring

         ptr->read_data(infile);

         break;

      case 'S':
         if (otype[1] == 'B') {                 // SBFAST
 
            ptr = new sbfast;

            ptr->mctype = PALEX;
            ptr->mcinfo = parse_info(&ctype[5]);

            fscanf(infile, "%s", ptr->colorname);                      // color file

            if (ptr->mcinfo & CITEXTURE)
               fscanf(infile, "%s", ptr->texname2);                   // texture filename

            if (ptr->mcinfo & CI3DTEX)
               fscanf(infile, "%s", ptr->texname3);                 // 3D texture file

            fscanf(infile, "%f %f %f", &ptr->center[0], // object location
                                          &ptr->center[1], &ptr->center[2]);
            fscanf(infile, "%f", &ptr->size);          // object size

            for (j=0; j<4; j++)                         // object rotation
               fscanf(infile, "%f %f %f %f", &ptr->rotate[j][0],
                      &ptr->rotate[j][1], &ptr->rotate[j][2],
                      &ptr->rotate[j][3]);

            ptr->read_data(infile);
         }

         else {						// SPATCH
            ptr = new spatch;

            parse_color(ctype, &ptr->mctype, &ptr->mcinfo);

            fscanf(infile, "%s", ptr->filename);                      // object file

            if ((ptr->mctype > BW && ptr->mctype <= PHONG) || ptr->mctype > PBW) {
               fscanf(infile, "%s", ptr->colorname);                      // color file

               if (ptr->mcinfo & CITEXTURE)
                  fscanf(infile, "%s", ptr->texname2);                   // texture filename

               if (ptr->mcinfo & CI3DTEX)
                  fscanf(infile, "%s", ptr->texname3);                 // 3D texture file
            }

            fscanf(infile, "%s", stype);                      // shadow flag

            if (ptr->sflag = !strcmp(stype, "SHADOW"))
               fscanf(infile, "%f %f %f %f", &ptr->splane[0], &ptr->splane[1],
                                             &ptr->splane[2], &ptr->splane[3]);

            fscanf(infile, "%f %f %f", &ptr->center[0], // object location
                                       &ptr->center[1], &ptr->center[2]);
            fscanf(infile, "%f", &ptr->size);          // object size

            for (j=0; j<4; j++)                         // object rotation
               fscanf(infile, "%f %f %f %f", &ptr->rotate[j][0],
                      &ptr->rotate[j][1], &ptr->rotate[j][2],
                      &ptr->rotate[j][3]);
         }

         break;

      default:
         break;
   }

   return ptr;
}



/* *************************************************************
   This procedure reads in a list of objects, their attributes,
   then reads in the object data.
************************************************************* */
void read_list(char *filename, listtype *mainlist, int *mapcount,
               camera *cparm, shader *sparm) {

   FILE    *infile;                // input file name
   pc      *ptr;                   // pointers to objects
   camera  *vtr;
   light   *ltr;
   char    otype[MAXSTRLEN], ctype[MAXSTRLEN];
   int     objectcount;            // # of objects
   int     i, j;
   pc      ***bucket, ***last;

   if (!(infile = fopen(filename, "r"))) {
      printf("Cannot access .lst file %s...\n", filename);
      return;
   }

   fscanf(infile, "%d %d", mapcount, &objectcount);	// # of bitmaps / objects

   mainlist->init(*mapcount);

   bucket = new pppc[*mapcount];
   last = new pppc[*mapcount];

   j = *mapcount * 9;
   bucket[0] = new pcptrtype[j];
   last[0] = new pcptrtype[j];

   for (i=0; i<j; i++)
      bucket[0][i] = (pc *)NULL;

   for (i=1, j=9; i<*mapcount; i++, j+=9) {
      bucket[i] = &bucket[0][j];
      last[i] = &last[0][j];
   }

   for (i=0; i<objectcount; i++) {
      fscanf(infile, "%d", &j);                 // map # to place object
      j--;

      fscanf(infile, "%s", otype);              // object type

      switch (otype[0]) {

         case 'C':                              // camera
            vtr = new camera(cparm);
            vtr->frame = j;

            vtr->read_data(infile);

            if (j < *mapcount) {
               vtr->next = mainlist->camdtr[j];
               mainlist->camdtr[j] = vtr;
            }

            else
               delete vtr;

            break;

         case 'L':                              // light

            if (otype[2] == 'G') {
               ltr = new light;
               ltr->frame = j;

               ltr->read_data(infile, &mainlist->tdtr);

               if (j < *mapcount) {
                  ltr->next = mainlist->lightdtr[j];
                  mainlist->lightdtr[j] = ltr;
               }

               else
                  delete ltr;

               break;

            }

         default:
            ptr = read_object(infile, otype, ctype, mainlist);

            if (ptr == (pc *)NULL)
               break;

            if (j >= *mapcount) {
               delete ptr;
               break;
            }

            ptr->preprocess(mainlist, shades);
            ptr->frame = j;
            ptr->id = i+1;

            if (ptr->mcinfo & CITRANSPARENT) {
               ptr->next = mainlist->raddtr[ptr->frame];
               mainlist->raddtr[ptr->frame] = ptr;
            }

            else {
               j = calc_bucket(ptr->mctype, ptr->mcinfo);

               if (bucket[ptr->frame][j] == (pc *)NULL)
                  last[ptr->frame][j] = ptr;

               ptr->next             = bucket[ptr->frame][j];
               bucket[ptr->frame][j] = ptr;
            }

            break;
      }

   }

   for (i=0; i<*mapcount; i++)
      for (j=8; j>-1; j--)
         if (bucket[i][j] != (pc *)NULL) {
            last[i][j]->next = mainlist->header[i];
            mainlist->header[i] = bucket[i][j];
         }

   delete bucket[0];
   delete bucket;
   delete last[0];
   delete last;

   fclose(infile);
}


/* *************************************************************
************************************************************* */
int parse_list(char *filename, listtype *mainlist, int *mapcount,
               camera *cparm, shader *sparm) {

   FILE    *infile;                // input file name
   char    otype[MAXSTRLEN], token[MAXSTRLEN];
   pc      *ptr;                   // pointers to objects
   camera  *vtr;
   light   *ltr;
   int     i, j;
   pc      ***bucket, ***last;
   int     errorflag;
   int     objectcount;            // # of objects

   if (!(infile = fopen(filename, "r"))) {
      printf("Cannot access .lst file %s...\n", filename);
      return 0;
   }

   get_token(infile, token);
   lower_case(token);

   if (strcmp(token, "list")) {
      printf("%s is an invalid .list file...\n", filename);
      fclose(infile);
      return 0;
   }

   get_token(infile, token);
   get_token(infile, token);
   *mapcount = atoi(token);

   get_token(infile, token);
   get_token(infile, token);
   objectcount = atoi(token);

   mainlist->init(*mapcount);

   bucket = new pppc[*mapcount];
   last = new pppc[*mapcount];

   j = *mapcount * 9;
   bucket[0] = new pcptrtype[j];
   last[0] = new pcptrtype[j];

   for (i=0; i<j; i++)
      bucket[0][i] = (pc *)NULL;

   for (i=1, j=9; i<*mapcount; i++, j+=9) {
      bucket[i] = &bucket[0][j];
      last[i] = &last[0][j];
   }

   for (i=0; i<objectcount; i++) {
      get_token(infile, token);
      lower_case(token);
      get_token(infile, otype);

      switch (token[0]) {

         case 'c':                              // camera
            vtr = new camera(cparm);

            do {
               if ((errorflag = !get_token(infile, token)) || token[0] == '}')
                  break;

               lower_case(token);

               if (!vtr->parse(infile, token)) {
                  printf("ERROR: Invalid data format... Last read token %s... Aborting...\n", token);
                  fclose(infile);
                  exit(1);
               }

            } while (1);

            if (errorflag) {
               printf("Error:Camera... EOF reached before '}'... Aborting...\n");
               fclose(infile);
               exit(1);
            }

            if (vtr->frame < *mapcount) {
               vtr->preprocess();
               vtr->next = mainlist->camdtr[vtr->frame];
               mainlist->camdtr[vtr->frame] = vtr;
            }

            else
               delete vtr;

            break;

         case 'l':                              // light

            if (token[2] == 'g') {
               ltr = new light;

               do {
                  if ((errorflag = !get_token(infile, token)) || token[0] == '}')
                     break;

                  lower_case(token);

                  if (!ltr->parse(infile, token)) {
                     printf("ERROR: Invalid data format... Last read token %s... Aborting...\n", token);
                     fclose(infile);
                     exit(1);
                  }

               } while (1);

               if (errorflag) {
                  printf("Error:Light... EOF reached before '}'... Aborting...\n");
                  fclose(infile);
                  exit(1);
               }

               if (ltr->frame < *mapcount) {
                  ltr->preprocess(&mainlist->tdtr);
                  ltr->next = mainlist->lightdtr[ltr->frame];
                  mainlist->lightdtr[ltr->frame] = ltr;
               }

               else
                  delete ltr;

               break;
            }

         default:

            switch (token[0]) {

               case 'g':

                  switch (token[4]) {

                     case 'p':			// SPHERE
                        ptr = new sphere;
                        break;

                     case 'y':			// CYLINDER
                        ptr = new cylinder;
                        break;

                     case 'o':			// CONE
                        ptr = new cone;
                        break;

                     default:
                        break;
                  }

                  break;

               case 'l':			// line
                  ptr = new line;
                  break;

               case 'p':
                  if (!strcmp(token, "particle"))
                     ptr = new particle;
                  else
                     ptr = new polygon;

                  break;

               case 'r':
                  ptr = new radiant;
                  break;

               case 's':
                  if (token[1] == 'b')                  // SBFAST
                     ptr = new sbfast;
                  else 					// SPATCH
                     ptr = new spatch;

                  break;

               default:
                  break;
            }

            if (ptr == (pc *)NULL) {
               printf("ERROR:%s... Invalid object type... Aborting...\n", token);
               fclose(infile);
               exit(1);
            }

            do {
               if ((errorflag = !get_token(infile, token)) || token[0] == '}')
                  break;

               lower_case(token);

               if (!ptr->parse(infile, token)) {
                  printf("ERROR: Invalid data format... Last read token %s... Aborting...\n", token);
                  fclose(infile);
                  exit(1);
               }

            } while (1);

            if (errorflag) {
               printf("Error:%s... EOF reached before '}'... Aborting...\n", token);
               fclose(infile);
               exit(1);
            }

            if (ptr->frame >= *mapcount) {
               delete ptr;
               break;
            }

            ptr->preprocess(mainlist, shades);
            ptr->id = i+1;

            if (ptr->mcinfo & CITRANSPARENT) {
               ptr->next = mainlist->raddtr[ptr->frame];
               mainlist->raddtr[ptr->frame] = ptr;
            }

            else {
               j = calc_bucket(ptr->mctype, ptr->mcinfo);

               if (bucket[ptr->frame][j] == (pc *)NULL)
                  last[ptr->frame][j] = ptr;

               ptr->next             = bucket[ptr->frame][j];
               bucket[ptr->frame][j] = ptr;
            }

            break;
      }

   }

   for (i=0; i<*mapcount; i++)
      for (j=8; j>-1; j--)
         if (bucket[i][j] != (pc *)NULL) {
            last[i][j]->next = mainlist->header[i];
            mainlist->header[i] = bucket[i][j];
         }

   delete bucket[0];
   delete bucket;
   delete last[0];
   delete last;

   fclose(infile);
   return 1;
}


/*
camera
	translate # # #
	up # # #
	coa # # #
        frame #

light
	type <type>

	direction # # #
	ambient # # #
	diffuse # # #
	location # # #
	up # # #

	distance2plane #
	width # 
	length # 
	2d_texture <filename>
	fatt # # #

        frame #

general primative
	rotate_x # deg
	rotate_y # deg
	rotate_z # deg
	translate # # #
	scale #
	tr_matrix [4x4]
	shadow # # # #
	2d_texture <filename>
	3d_texture <filename>
	transparent
	raycast
	luminate
	color <filename>
        frame #

polygon
	biplane
	file <filename>

spatch
	file <filename>

particle
        point # # #
	color # # #

line
	point # # # # # #
	color # # # # # #

radiant
	inner radius #
	outer radius #
	hollow

sbfast
	back
	octaves
	h
	lacunarity
	displacement <shadername>
	scalesb

sphere
	axis # # #
	axis_x #
	axis_y #
	axis_z #

cylinder
	length #
	radius #

cone	length #
	top_radius #
	bottom_radius #

}



}

*/
