
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */

#include <fcntl.h>
#include <gl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <termio.h>
#include <unistd.h>
#include "video.h"
#include <string.h>
#include <stdlib.h>
#include <CC/osfcn.h>

static int  video_outfile = -1;
static char filename[256];

#define LVR5000_ACK         0x0A /* Acknowledge receipt of valid command          */
#define LVR5000_NAK         0x0B /* Reject receipt of command                     */
#define LVR5000_ERROR       0x02 /* To notify of a communication error            */
#define LVR5000_CLEAR       0x56 /* Clears the mode currently selected            */
#define LVR5000_COMPLETION  0x01 /* The LVR5000 completed a command               */
#define LVR5000_REC_STANDBY 0xE0 /* Put the LVR5000 in recording standby mode     */
#define LVR5000_NO_FRAME    0x06 /* No currently selected frame number            */
#define LVR5000_REC_FRAME   0xE6 /* Sets the mode for recording 1 frame at a time */
#define LVR5000_REC         0xE9 /* Executes recording in the recording mode set  */
                                 /* from the specified IN point to OUT point.     */
#define LVR5000_ENTER       0x40 /* Enter command                                 */

/**********************************************************************************

 LVR5000_read - An internal procedure for reading from the laser disc player

 **********************************************************************************/

static int LVR5000_read()

{
  char         io_char;

  if (read(video_outfile, &io_char, 1) != 1)
   {
    printf("error in LVR5000_read: read error on port %s\n",filename);
    return(-1);
   }
  else
    return((int) io_char);
}

/**********************************************************************************

 LVR5000_write - An internal procedure for writing to the laser disc player

 **********************************************************************************/

static int LVR5000_write(char *command)

{
  register int size;
  register int i;
  char         io_char;
printf("3.1\n");
  size = strlen(command);
  for (i = 0; i < size; i++)
printf("%d\n", size);
   {
    if (write(video_outfile, &command[i], 1) != 1)
     {
      printf("error in LVR5000_write: write error on port %s on byte %d\n",filename, i);
      return(-1);
     }
printf(".\n");
    if (read(video_outfile, &io_char, 1) != 1)
     {
      printf("error in LVR5000_write: read error on port %s on byte %d\n",filename, i);
      return(-1);
     }
printf(".\n");

    if (io_char == (char) LVR5000_NAK)
     {
      printf("error in LVR5000_write: command not Acknowledge on byte %d\n",i);
      return(-1);
     }

    if (io_char == (char) LVR5000_ERROR)
     {
      printf("error in LVR5000_write: communication error\n");
      return(-1);
     }

    if (io_char != (char) LVR5000_ACK)
     {
      printf("error in LVR5000_write: internal error\n");
      return(-1);
     }
   }

  return(0);
}

/**********************************************************************************

 LVR5000_init - Called once to open a communication port and initializes the laser 
                disc player. The char string is the name of the port. An example 
                would be /dev/ttyd2.  Port speed is 9600.

 **********************************************************************************/

int LVR5000_init(char *port_name)

{
  struct termio term;
  char   command[8];

  video_outfile = open(port_name, O_RDWR);
printf("2\n");
  if (video_outfile == -1)
   {
    printf("error in LVR5000_init: could not open port %s for reading and writing\n",port_name);
    return(-1);
   }

  memset(&term, 0, sizeof(term));
  term.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
  term.c_iflag = 0;
  term.c_oflag = 0;
  term.c_lflag = 0;
  term.c_cc[VMIN] = 1;
  term.c_cc[VTIME] = 0;

  if (ioctl(video_outfile, TCSETA, &term) < 0)
   {
    video_outfile = -1;
    printf("error in LVR5000_init: could not configure port %s\n",port_name);
    return(-1);
   }

  /* Dis-able screen blanking */
  blanktime((long) 0);
  strcpy(filename, port_name);

  command[0] = (char) LVR5000_CLEAR;
  command[1] = (char) NULL;
printf("3\n");
  if (LVR5000_write(command))
   {
    printf("error in LVR5000_init: command failed\n");
    return(-1);
   }
  else
    return(0);
printf("4\n");
}

/**********************************************************************************

 LVR5000_clear - Clears or reset the laser disc player

 **********************************************************************************/

int LVR5000_clear(void)

{
  char command[8];

  command[0] = (char) LVR5000_CLEAR;
  command[1] = (char) NULL;

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_clear: command failed\n");
    return(-1);
   }
  else
    return(0);
}

/**********************************************************************************

 LVR5000_close - used to close the communication port.

 **********************************************************************************/

int LVR5000_close(void) 

{
  if (video_outfile == -1)
   {
    printf("error in LVR5000_close: port already closed\n");
    return(-1);
   }

  if (LVR5000_clear())
   {
    printf("error in LVR5000_close: could not clear laser disc player\n");
    return(-1);
   }

  return(close(video_outfile));
}

/**********************************************************************************

 LVR5000_rec_mode_frame - Puts the laser dics player in frame by frame recording mode

 **********************************************************************************/

int LVR5000_rec_mode_frame(void)

{
  char command[8];

  command[0] = (char) LVR5000_REC_FRAME;
  command[1] = (char) NULL;

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec_mode_frame: command failed\n");
    return(-1);
   }
  else
    return(0);
}

/**********************************************************************************

 LVR5000_rec_standby - Puts the laser dics player in record standby mode. The player
                       is prepared to record from start (in frame number) to end
                       (in frame number). The actual recording occurs when 
                       LVR5000_rec is called. 

 **********************************************************************************/

int LVR5000_rec_standby(int start, int stop)

{
  char command[8];

  if (start < 1)
   {
    printf("error in LVR5000_rec_standby: start point must be a positive number\n");
    return(-1);
   }

  if (stop <= start)
   {
    printf("error in LVR5000_rec_standby: stop point must be larger then start\n");
    return(-1);
   }

  command[0] = (char) LVR5000_REC_STANDBY;
  command[1] = (char) NULL;

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec_standby: command failed\n");
    return(-1);
   }
  if (LVR5000_read() != LVR5000_COMPLETION)
   {
    printf("error in LVR5000_rec_standby: command failed to complete\n");
    return(-1);
   }

  sprintf(command,"%d%c%c",start, (char) LVR5000_ENTER, (char) NULL);

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec_standby: command failed while sending start point\n");
    return(-1);
   }
 
  sprintf(command,"%d%c%c", stop, (char) LVR5000_ENTER, (char) NULL);

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec_standby: command failed while sending stop point\n");
    return(-1);
   }

  command[0] = (char) LVR5000_ENTER;
  command[1] = (char) NULL;

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec_standby: command failed while sending duration\n");
    return(-1);
   }

  if (LVR5000_read() != LVR5000_COMPLETION)
   {
    printf("error in LVR5000_rec_standby: command failed to complete after sending in and out points\n");
    return(-1);
   }

  return(0);
}

/**********************************************************************************

 LVR5000_rec - Causes the laser disc player to record.  If the player is in frame by
               frame recording then this procedure is called after each new frame is
               displayed.  

 **********************************************************************************/

int LVR5000_rec(void)

{
  char command[8];

  command[0] = (char) LVR5000_REC;
  command[1] = (char) NULL;

  if (LVR5000_write(command))
   {
    printf("error in LVR5000_rec: command failed\n");
    return(-1);
   }

  if (LVR5000_read() != LVR5000_COMPLETION)
   {
    printf("error in LVR5000_rec: command failed to complete\n");
    return(-1);
   }

  return(0);
}

