
/* ************************************************************************
		Copyright (c) 1994-1996 Alex Harper
************************************************************************ */





/* *************************************************************
   this file contains all the routines for the zbuffer class
************************************************************* */

#include <string.h>
#include <bstring.h>
#include "zbuffer.h"
#include "global.h"

#define MAXFILTER 7

static float  default_filter[MAXFILTER][MAXFILTER] = {
         0.001, 0.004, 0.008, 0.010, 0.008, 0.004, 0.001,
         0.004, 0.012, 0.025, 0.029, 0.025, 0.012, 0.002,
         0.008, 0.025, 0.049, 0.058, 0.049, 0.025, 0.008,
         0.010, 0.029, 0.058, 0.068, 0.058, 0.029, 0.010,
         0.008, 0.025, 0.049, 0.058, 0.049, 0.025, 0.008,
         0.004, 0.012, 0.025, 0.029, 0.025, 0.012, 0.002,
         0.001, 0.004, 0.008, 0.010, 0.008, 0.004, 0.001,
      };


/* *************************************************************
   this function reads in an anti-aliasing filter
************************************************************* */
int zbuffer::read_filter(char *fname, mapf *filter) {

   FILE   *infile;
   int    i, j;

   if (!(infile = fopen(fname, "r"))) {
      printf("Invalid Filter... using default\n");
      return 0;
   }

   fscanf(infile, "%d", &j);

   if (!j || !(j & 0x01)) {
      printf("Invalid Filter... using default\n");
      return 0;
   }
   
   filter->init_map(j, j);

   j *= j;

   for (i=0; i<j; i++)
      fscanf(infile, "%f", &filter->data[i]);

   fclose(infile);
   return 1;
}


/* *************************************************************
************************************************************* */
void zbuffer::save_rgb(char *fname, int mapno) {

   head.xsize = maxx;
   head.ysize = maxy;
   head.zsize = 3;

   sprintf(filename, "%s.%05d.rgb%c", fname, mapno, 0);

   write_rle(filename, 1);
}


/* *************************************************************
************************************************************* */
void zbuffer::save_buffer(FILE *outfile, int mapno) {

   int    i;

   fwrite(&mapno, sizeof(int), 1, outfile);

   for (i=0; i<maxy; i++)
      fwrite(pdata[i], sizeof(unsigned int)*maxx, 1, outfile);
}


/* *************************************************************
************************************************************* */
int zbuffer::read_buffer(FILE *infile) {

   int    i;
   int    mapno;

   fread(&mapno, sizeof(int), 1, infile);

   for (i=0; i<maxy; i++)
      fread(pdata[i], sizeof(unsigned int)*maxx, 1, infile);

   return mapno;
}


/* **************************************************
   this function anti-aliases the zbuffer
************************************************** */
void zbuffer::antialias(char *fname, float gamma) {

   mapf      filter;
   int       defflag = 1;
   int       current;				// center of filter
   int       i, j, k, l, m, n;                  // looping variables
   int       mk, nl;
   float     temp[3];                           // current pixel value
   float     g;
   int       cmaxx, cmaxy;

   if (gamma >= 0.0) 
      g = getgamma(gamma);

   if (gamma < 0.0) {
      for (i=0; i<maxy; i++)
         for (j=0; j<maxx; j++)
            if (zbuff[i][j] < front)
               addarray3(ybuff[i][j], ambientbuff[i][j]);
   }

   else
      for (i=0; i<maxy; i++)
         for (j=0; j<maxx; j++)
            if (zbuff[i][j] < front) {
               addarray3(ybuff[i][j], ambientbuff[i][j]);
               ybuff[i][j][0] = gammacorrect(g, ybuff[i][j][0]);
               ybuff[i][j][1] = gammacorrect(g, ybuff[i][j][1]);
               ybuff[i][j][2] = gammacorrect(g, ybuff[i][j][2]);
            }

   if (strlen(fname))                       // read filter from file
      defflag = read_filter(fname, &filter);

   if (defflag) {
      filter.init_map(MAXFILTER, MAXFILTER);

      for (i=0; i<MAXFILTER; i++)
         for (j=0; j<MAXFILTER; j++)
            filter.pdata[i][j] = default_filter[i][j];
   }

   current = (filter.maxx-1) >> 1;
   cmaxx = maxx - current;
   cmaxy = maxy - current;

   for (i=current, m=0; i<cmaxy; i++, m++)       // multiply filter X pixmap
      for (j=current, n=0; j<cmaxx; j++, n++)
         if (zbuff[i][j]<0) {
            temp[0] = temp[1] = temp[2] = 0;

            for (k=0, mk=m; k<filter.maxy; k++, mk++)
               for (l=0, nl=n; l<filter.maxx; l++, nl++)
                  if (zbuff[mk][n+l] < 0) {
                     temp[0] += (filter.pdata[k][l] * ybuff[mk][nl][0]);
                     temp[1] += (filter.pdata[k][l] * ybuff[mk][nl][1]);
                     temp[2] += (filter.pdata[k][l] * ybuff[mk][nl][2]);
                  }

            pdata[i][j] = gwindow.ALEXsetcolor(temp);
         }

         else
            pdata[i][j] = 0;

   for (i=0, k=cmaxy; i<current; i++, k++) {
      for (j=0; j<maxx; j++) {

         if (zbuff[i][j]<0)
            pdata[i][j] = gwindow.ALEXsetcolor(ybuff[i][j]);
         else
            pdata[i][j] = 0;

         if (zbuff[k][j]<0)
            pdata[k][j] = gwindow.ALEXsetcolor(ybuff[k][j]);
         else
            pdata[k][j] = 0;
      }

   }

   for (i=current; i<cmaxy; i++)
      for (j=0, k=cmaxx; j<current; j++, k++) {

         if (zbuff[i][j]<0)
            pdata[i][j] = gwindow.ALEXsetcolor(ybuff[i][j]);
         else
            pdata[i][j] = 0;

         if (zbuff[i][k]<0)
            pdata[i][k] = gwindow.ALEXsetcolor(ybuff[i][k]);
         else
            pdata[i][k] = 0;
      }

}


/* **************************************************
   this function initializes the buffer
************************************************** */
int zbuffer::initbuff(unsigned int x, unsigned int y) {

   unsigned int i;
   unsigned int a, b;

   if (data != (unsigned int *)NULL && (x != maxx || y != maxy)) {
      delete data;
      data = (unsigned int *)NULL;
      delete pdata;
      delete zdata;
      delete zbuff;
      delete adata;
      delete ambientbuff;
      delete ydata;
      delete ybuff;
   }

   a = x*y;

   if (data == (unsigned int *)NULL) {
      init_map(x, y);
      zdata = new float[a];
      zbuff = new pfloat[y];

      adata = new vector3f[a];
      ambientbuff = new pvector3f[y];

      ydata = new vector3f[a];
      ybuff = new pvector3f[y];

      for (i=b=0; i<y; i++, b+=x) {
         zbuff[i] = &zdata[b];
         ybuff[i] = &ydata[b];
         ambientbuff[i] = &adata[b];
      }

   }

   bzero((void *)zdata, a<<2);		// float - 4 bytes

/*
   a*=sizeof(vector3f);

   bzero((void *)ydata, a);
   bzero((void *)adata, a);
*/

   return 1;
}


/* **************************************************
   this function initializes the buffer
************************************************** */
int zbuffer::initbuff(unsigned int *d, puint *pd, unsigned int x, unsigned int y) {

   unsigned int i;
   unsigned int a, b;

   if (data != (unsigned int *)NULL) {
      delete data;
      delete pdata;
   }

   data = d;
   pdata = pd;

   if (zdata != (float *)NULL && (x != maxx || y != maxy)) {
      delete zdata;
      zdata = (float *)NULL;
      delete zbuff;
      delete adata;
      delete ambientbuff;
      delete ydata;
      delete ybuff;
   }

   a = x*y;

   if (zdata == (float *)NULL) {
      zdata = new float[a];
      zbuff = new pfloat[y];

      adata = new vector3f[a];
      ambientbuff = new pvector3f[y];

      ydata = new vector3f[a];
      ybuff = new pvector3f[y];

      for (i=b=0; i<y; i++, b+=x) {
         zbuff[i] = &zdata[b];
         ybuff[i] = &ydata[b];
         ambientbuff[i] = &adata[b];
      }

   }

   maxx = x;
   maxy = y;

   bzero((void *)zdata, a<<2);		// float - 4 bytes


/*
   a *= sizeof(vector3f);

   bzero((void *)ydata, a);
   bzero((void *)adata, a);
*/

   return 1;
}


/* **************************************************
************************************************** */
int zbuffer::get_buff(unsigned int **d, puint **pd, unsigned int *x, unsigned int *y) {

   if (data != (unsigned int *)NULL) {
      *d = data;
      data = (unsigned int *)NULL;
      *pd = pdata;
      pdata = (unsigned int **)NULL;
      *x = maxx;
      *y = maxy;
      return 1;
   }

   return 0;
}


/* **************************************************
   this function initializes the buffer
************************************************** */
void zbuffer::catbuff(float gamma) {

   int i, j;
   float g;

   if (gamma < 0.0) {
      for (i=0; i<maxy; i++)
         for (j=0; j<maxx; j++)
            if (zbuff[i][j] < front) {
               addarray3(ybuff[i][j], ambientbuff[i][j]);
               pdata[i][j] = gwindow.ALEXsetcolor(ybuff[i][j]);
            }

            else
               pdata[i][j] = 0;

      return;
   }

   g = getgamma(gamma);
   for (i=0; i<maxy; i++)
      for (j=0; j<maxx; j++)
         if (zbuff[i][j] < front) {
            addarray3(ybuff[i][j], ambientbuff[i][j]);
            ybuff[i][j][0] = gammacorrect(g, ybuff[i][j][0]);
            ybuff[i][j][1] = gammacorrect(g, ybuff[i][j][1]);
            ybuff[i][j][2] = gammacorrect(g, ybuff[i][j][2]);
            pdata[i][j] = gwindow.ALEXsetcolor(ybuff[i][j]);
         }

         else
            pdata[i][j] = 0;
}


/* *************************************************************
   this function puts a pixmap on the screen
************************************************************* */
void zbuffer::render(ALEXpixmap *mcanvas) {

   int           i, j;
   int           mpt[2][2];
   unsigned int temp;

   for (i=0; i<maxy; i++) {
      temp = pdata[i][0];
      mpt[1][0] = 0;
      mpt[0][1] = mpt[1][1] = i;

      for (j=0; j<maxx; j++)
         if (temp != pdata[i][j]) {
            gwindow.ALEXsetforeground(temp);
            mpt[0][0] = mpt[1][0];
            mpt[1][0] = j;
            gwindow.ALEXline(mpt[0], mpt[1], mcanvas);
            temp = pdata[i][j];
         }

      gwindow.ALEXsetforeground(temp);
      mpt[0][0] = mpt[1][0];
      mpt[1][0] = (int)(maxx-1);
      gwindow.ALEXline(mpt[0], mpt[1], mcanvas);
   }

}


/* *************************************************************
   this function puts an XImage on the screen
************************************************************* */
void zbuffer::render(ALEXimage mcanvas) {

   int  i, j;
   unsigned int *ptr;
   int diff;

   ptr  = ((unsigned int *)(mcanvas->data)) + mcanvas->width*mcanvas->height;
   diff = maxx << 1;

   for (i=0, ptr-=maxx; i<maxy; i++, ptr-=diff)
      for (j=0; j<maxx; j++, ptr++)
         *ptr = pdata[i][j];
}


/* *************************************************************
   this function puts an XImage on the screen
************************************************************* */
void zbuffer::render(mapul *mcanvas) {

   memcpy(mcanvas->data, data, maxx*maxy<<2);
}

